searchData={"items":[{"type":"module","title":"Scholar.Cluster.AffinityPropagation","doc":"Model representing affinity propagation clustering. The first dimension\nof `:clusters_centers` is set to the number of samples in the dataset.\nThe artificial centers are filled with `:infinity` values. To fillter\nthem out use `prune` function.\n\nThe algorithm has a time complexity of the order $O(N^2T)$, where $N$ is\nthe number of samples and $T$ is the number of iterations until convergence.\nFurther, the memory complexity is of the order $O(N^2)$.","ref":"Scholar.Cluster.AffinityPropagation.html"},{"type":"function","title":"Scholar.Cluster.AffinityPropagation.fit/2","doc":"Cluster the dataset using affinity propagation.","ref":"Scholar.Cluster.AffinityPropagation.html#fit/2"},{"type":"function","title":"Options - Scholar.Cluster.AffinityPropagation.fit/2","doc":"* `:iterations` (`t:pos_integer/0`) - Number of iterations of the algorithm. The default value is `300`.\n\n* `:damping_factor` (`t:float/0`) - Damping factor in the range [0.5, 1.0) is the extent to which the\n  current value is maintained relative to incoming values (weighted 1 - damping). The default value is `0.5`.\n\n* `:preference` - How to compute the preferences for each point - points with larger values\n  of preferences are more likely to be chosen as exemplars. The number of clusters is\n  influenced by this option.\n\n  The preferences is either an atom, each is a `Nx` reduction function to\n  apply on the input similarities (such as `:reduce_min`, `:median`, `:mean`,\n  etc) or a float.\n\n  The default value is `:reduce_min`.\n\n* `:key` - Determines random number generation for centroid initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.\n\n* `:learning_loop_unroll` (`t:boolean/0`) - If `true`, the learning loop is unrolled. The default value is `false`.\n\n* `:converge_after` (`t:pos_integer/0`) - Number of iterations with no change in the number of estimated clusters\n  that stops the convergence. The default value is `15`.","ref":"Scholar.Cluster.AffinityPropagation.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Cluster.AffinityPropagation.fit/2","doc":"The function returns a struct with the following parameters:\n\n  * `:clusters_centers` - Cluster centers from the initial data.\n\n  * `:cluster_centers_indices` - Indices of cluster centers.\n\n  * `:num_clusters` - Number of clusters.","ref":"Scholar.Cluster.AffinityPropagation.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.AffinityPropagation.fit/2","doc":"iex> key = Nx.Random.key(42)\n    iex> x = Nx.tensor([[12,5,78,2], [9,3,81,-2], [-1,3,6,1], [1,-2,5,2]])\n    iex> Scholar.Cluster.AffinityPropagation.fit(x, key: key)\n    %Scholar.Cluster.AffinityPropagation{\n      labels: Nx.tensor([0, 0, 2, 2]),\n      cluster_centers_indices: Nx.tensor([0, -1, 2, -1]),\n      cluster_centers: Nx.tensor(\n        [\n          [12.0, 5.0, 78.0, 2.0],\n          [:infinity, :infinity, :infinity, :infinity],\n          [-1.0, 3.0, 6.0, 1.0],\n          [:infinity, :infinity, :infinity, :infinity]\n        ]\n      ),\n      num_clusters: Nx.tensor(2, type: :u64),\n      iterations: Nx.tensor(22, type: :s64)\n    }","ref":"Scholar.Cluster.AffinityPropagation.html#fit/2-examples"},{"type":"function","title":"Scholar.Cluster.AffinityPropagation.predict/2","doc":"Predict the closest cluster each sample in `x` belongs to.","ref":"Scholar.Cluster.AffinityPropagation.html#predict/2"},{"type":"function","title":"Examples - Scholar.Cluster.AffinityPropagation.predict/2","doc":"iex> key = Nx.Random.key(42)\n    iex> x = Nx.tensor([[12,5,78,2], [9,3,81,-2], [-1,3,6,1], [1,-2,5,2]])\n    iex> model = Scholar.Cluster.AffinityPropagation.fit(x, key: key)\n    iex> model = Scholar.Cluster.AffinityPropagation.prune(model)\n    iex> Scholar.Cluster.AffinityPropagation.predict(model, Nx.tensor([[10,3,50,6], [8,3,8,2]]))\n    #Nx.Tensor","ref":"Scholar.Cluster.AffinityPropagation.html#predict/2-examples"},{"type":"function","title":"Scholar.Cluster.AffinityPropagation.prune/1","doc":"Optionally prune clusters, indices, and labels to only valid entries.\n\nIt returns an updated and pruned model.","ref":"Scholar.Cluster.AffinityPropagation.html#prune/1"},{"type":"function","title":"Examples - Scholar.Cluster.AffinityPropagation.prune/1","doc":"iex> key = Nx.Random.key(42)\n    iex> x = Nx.tensor([[12,5,78,2], [9,3,81,-2], [-1,3,6,1], [1,-2,5,2]])\n    iex> model = Scholar.Cluster.AffinityPropagation.fit(x, key: key)\n    iex> Scholar.Cluster.AffinityPropagation.prune(model)\n    %Scholar.Cluster.AffinityPropagation{\n      labels: Nx.tensor([0, 0, 1, 1]),\n      cluster_centers_indices: Nx.tensor([0, 2]),\n      cluster_centers: Nx.tensor(\n        [\n          [12.0, 5.0, 78.0, 2.0],\n          [-1.0, 3.0, 6.0, 1.0]\n        ]\n      ),\n      num_clusters: Nx.tensor(2, type: :u64),\n      iterations: Nx.tensor(22, type: :s64)\n    }","ref":"Scholar.Cluster.AffinityPropagation.html#prune/1-examples"},{"type":"module","title":"Scholar.Cluster.DBSCAN","doc":"Perform DBSCAN clustering from vector array or distance matrix.\n\nDBSCAN - Density-Based Spatial Clustering of Applications with Noise.\nFinds core samples of high density and expands clusters from them.\nGood for data which contains clusters of similar density.\n\nThe time complexity is $O(N^2)$ for $N$ samples.\nThe space complexity is $O(N^2)$.","ref":"Scholar.Cluster.DBSCAN.html"},{"type":"function","title":"Scholar.Cluster.DBSCAN.fit/2","doc":"Perform DBSCAN clustering from vector array or distance matrix.","ref":"Scholar.Cluster.DBSCAN.html#fit/2"},{"type":"function","title":"Options - Scholar.Cluster.DBSCAN.fit/2","doc":"* `:eps` - The maximum distance between two samples for them to be considered as in the same neighborhood. The default value is `0.5`.\n\n* `:min_samples` (`t:integer/0`) - The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.\n  This includes the point itself. The default value is `5`.\n\n* `:metric` - The function that measures the pairwise distance between two points. Possible values:\n\n  * `{:minkowski, p}` - Minkowski metric. By changing value of `p` parameter (a positive number or `:infinity`)\n  we can set Manhattan (`1`), Euclidean (`2`), Chebyshev (`:infinity`), or any arbitrary $L_p$ metric.\n\n  * `:cosine` - Cosine metric.\n\n  * Anonymous function of arity 2 that takes two rank-2 tensors.\n\n  The default value is `&Scholar.Metrics.Distance.pairwise_minkowski/2`.\n\n* `:weights` - The weights for each observation in `x`. If equals to `nil`,\n  all observations are assigned equal weight.","ref":"Scholar.Cluster.DBSCAN.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Cluster.DBSCAN.fit/2","doc":"The function returns a struct with the following parameters:\n\n* `:core_sample_indices` - Indices of core samples represented as a mask.\n  The mask is a boolean array of shape `{num_samples}` where `1` indicates\n  that the corresponding sample is a core sample and `0` otherwise.\n\n* `:labels` - Cluster labels for each point in the dataset given to fit().\n  Noisy samples are given the label `-1`.","ref":"Scholar.Cluster.DBSCAN.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.DBSCAN.fit/2","doc":"iex> x = Nx.tensor([[1, 2], [2, 2], [2, 3], [8, 7], [8, 8], [25, 80]])\n    iex> Scholar.Cluster.DBSCAN.fit(x, eps: 3, min_samples: 2)\n    %Scholar.Cluster.DBSCAN{\n      core_sample_indices: Nx.tensor(\n        [1, 1, 1, 1, 1, 0], type: :u8\n      ),\n      labels: Nx.tensor(\n        [0, 0, 0, 1, 1, -1]\n      )\n    }","ref":"Scholar.Cluster.DBSCAN.html#fit/2-examples"},{"type":"module","title":"Scholar.Cluster.GaussianMixture","doc":"Gaussian Mixture Model.\n\nGaussian Mixture Model is a probabilistic model that assumes every data point is generated\nby choosing one of several fixed Gaussian distributions and then sampling from it.\nIts parameters are estimated using the Expectation-Maximization (EM) algorithm, which is an\niterative algorithm alternating between the two steps: the E-step which computes the\nexpectation of the Gaussian assignment for each data point x and the M-step which updates the\nparameters to maximize the expectations found in E-step. While every iteration of the algorithm\nis guaranteed to improve the log-likelihood, the final result depends on the initial values of\nthe parameters. Thus the procedure consists of repeating the algorithm several times\nand taking the best obtained result.\n\nTime complexity is $O(NKD^3)$ for $N$ data points, $K$ Gaussian components and $D$ dimensions\n\nReferences:\n\n* [1] - Mixtures of Gaussians and the EM algorithm https://cs229.stanford.edu/notes2020spring/cs229-notes7b.pdf\n* [2] - Density Estimation with Gaussian Mixture Models https://mml-book.github.io/book/mml-book.pdf Chapter 11","ref":"Scholar.Cluster.GaussianMixture.html"},{"type":"function","title":"Scholar.Cluster.GaussianMixture.fit/2","doc":"Fits a Gaussian Mixture Model for sample inputs `x`.","ref":"Scholar.Cluster.GaussianMixture.html#fit/2"},{"type":"function","title":"Options - Scholar.Cluster.GaussianMixture.fit/2","doc":"* `:num_gaussians` (`t:pos_integer/0`) - Required. The number of Gaussian distributions in the mixture.\n\n* `:num_runs` (`t:pos_integer/0`) - The number of times to initialize parameters and run the entire EM algorithm. The default value is `1`.\n\n* `:max_iter` (`t:pos_integer/0`) - The number of EM iterations to perform. The default value is `100`.\n\n* `:tol` - The convergence threshold. The default value is `0.001`.\n\n* `:covariance_regularization_eps` - The non-negative number that is added to each element of the diagonal\n  of the covariance matrix to ensure it is positive. Usually a small number. The default value is `1.0e-6`.\n\n* `:key` - Used for random number generation in parameter initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.","ref":"Scholar.Cluster.GaussianMixture.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Cluster.GaussianMixture.fit/2","doc":"The function returns a struct with the following parameters:\n\n  * `:weights` - The fractions of data sampled from each Gaussian, respectively.\n\n  * `:means` - Means of the Gaussian components.\n\n  * `:covariances` - Covariance matrices of the Gaussian components.\n\n  * `:precisions_cholesky` - Cholesky decomposition of the precision matrices\n\t(inverses of covariances). This is useful for the model inference.","ref":"Scholar.Cluster.GaussianMixture.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.GaussianMixture.fit/2","doc":"iex> key = Nx.Random.key(12)\n    iex> x = Nx.tensor([[1, 2], [1, 4], [1, 0], [10, 2], [10, 4], [10, 0]])\n    iex> Scholar.Cluster.GaussianMixture.fit(x, num_gaussians: 2, key: key).means\n    Nx.tensor(\n      [\n        [1.0, 2.0],\n        [10.0, 2.0]\n      ]\n    )","ref":"Scholar.Cluster.GaussianMixture.html#fit/2-examples"},{"type":"function","title":"Scholar.Cluster.GaussianMixture.predict/2","doc":"Makes predictions with the given `model` on inputs `x`.","ref":"Scholar.Cluster.GaussianMixture.html#predict/2"},{"type":"function","title":"Return Values - Scholar.Cluster.GaussianMixture.predict/2","doc":"It returns a tensor with Gaussian assignments for every input point.","ref":"Scholar.Cluster.GaussianMixture.html#predict/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.GaussianMixture.predict/2","doc":"iex> key = Nx.Random.key(12)\n    iex> x = Nx.tensor([[1, 2], [1, 4], [1, 0], [10, 2], [10, 4], [10, 0]])\n    iex> model = Scholar.Cluster.GaussianMixture.fit(x, num_gaussians: 2, key: key)\n    iex> Scholar.Cluster.GaussianMixture.predict(model, Nx.tensor([[8, 1], [2, 3]]))\n    Nx.tensor(\n      [1, 0]\n    )","ref":"Scholar.Cluster.GaussianMixture.html#predict/2-examples"},{"type":"function","title":"Scholar.Cluster.GaussianMixture.predict_prob/2","doc":"Makes predictions with the given `model` on inputs `x`.","ref":"Scholar.Cluster.GaussianMixture.html#predict_prob/2"},{"type":"function","title":"Return Values - Scholar.Cluster.GaussianMixture.predict_prob/2","doc":"It returns a tensor probabilities of Gaussian assignments for every input point.","ref":"Scholar.Cluster.GaussianMixture.html#predict_prob/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.GaussianMixture.predict_prob/2","doc":"iex> key = Nx.Random.key(12)\n    iex> x = Nx.tensor([[1, 2], [1, 4], [1, 0], [10, 2], [10, 4], [10, 0]])\n    iex> model = Scholar.Cluster.GaussianMixture.fit(x, num_gaussians: 2, key: key)\n    iex> Scholar.Cluster.GaussianMixture.predict_prob(model, Nx.tensor([[8, 1], [2, 3]]))\n    Nx.tensor(\n      [\n        [0.0, 1.0],\n        [1.0, 0.0]\n      ]\n    )","ref":"Scholar.Cluster.GaussianMixture.html#predict_prob/2-examples"},{"type":"module","title":"Scholar.Cluster.Hierarchical","doc":"Performs [hierarchical, agglomerative clustering](https://en.wikipedia.org/wiki/Hierarchical_clustering#Agglomerative_clustering_example)\non a dataset.\n\nHierarchical clustering is good for when the number of clusters is not known ahead of time.\nIt also allows for the creation of a [dendrogram plot](https://en.wikipedia.org/wiki/Dendrogram)\n(regardless of the dimensionality of the dataset) which can be used to select the number of\nclusters in a post-processing step.","ref":"Scholar.Cluster.Hierarchical.html"},{"type":"module","title":"Limitations - Scholar.Cluster.Hierarchical","doc":"Due to the requirements of the current implementation, only these options are supported:\n\n  * `dissimilarity: :euclidean`\n  * `linkage: :average | :complete | :single | :weighted`\n\nOur current algorithm is $O(\\frac{n^2}{p} \\cdot \\log(n))$ where $n$ is the number of data points\nand $p$ is the number of processors.\nThis is better than the generic algorithm which is $O(n^3)$.\nIt is also parallel, which means that runtime decreases in direct proportion to the number of\nprocessors.\n\nHowever, the implementation requires certain theoretical properties of the dissimilarities and\nlinkages.\nAs such, we've restricted the options to only those combinations with the correct properties.\n\nIn the future, we plan to add additional algorithms which won't have the same restrictions.","ref":"Scholar.Cluster.Hierarchical.html#module-limitations"},{"type":"function","title":"Scholar.Cluster.Hierarchical.fit/2","doc":"Use hierarchical clustering to form the initial model to be clustered with `labels_list/2` or\n`labels_map/2`.","ref":"Scholar.Cluster.Hierarchical.html#fit/2"},{"type":"function","title":"Options - Scholar.Cluster.Hierarchical.fit/2","doc":"* `:dissimilarity` - Pairwise dissimilarity function: computes the 'dissimilarity' between each pair of data points.\n  Dissimilarity is analogous to distance, but without the expectation that the triangle\n  inequality holds.\n\n  Choices:\n\n    * `:euclidean` - L2 norm.\n\n  See \"Limitations\" in the moduledoc for an explanation of the lack of choices.\n\n  The default value is `:euclidean`.\n\n* `:linkage` - Linkage function: how to compute the intra-clade dissimilarity of two clades if they were\n  merged.\n\n  Choices:\n\n    * `:average` - The unweighted average dissimilarity across all pairs of points.\n\n    * `:complete` - (Historic name) The maximum dissimilarity across all pairs of points.\n\n    * `:single` - (Historic name) The minimum dissimilarity across all pairs of points.\n\n    * `:ward` - (Named for [Ward's method](https://en.wikipedia.org/wiki/Ward%27s_method))\n      The minimum increase in sum of squares (MISSQ) of dissimilarities.\n\n    * `:weighted` - The weighted average dissimilarity across all pairs of points.\n\n  The default value is `:single`.","ref":"Scholar.Cluster.Hierarchical.html#fit/2-options"},{"type":"function","title":"Return values - Scholar.Cluster.Hierarchical.fit/2","doc":"Returns a `Scholar.Cluster.Hierarchical` struct with the following fields:\n\n  * `clades` (`Nx.Tensor` with shape `{n - 1, 2}`) -\n    Contains the indices of the pair of clades merged at each step of the agglomerative\n    clustering process.\n\n    Agglomerative clustering starts by considering each datum in `data` its own singleton group\n    or [\"clade\"](https://en.wikipedia.org/wiki/Clade).\n    It then picks two clades to merge into a new clade containing the data from both.\n    It does this until there is a single clade remaining.\n\n    Since each datum starts as its own clade, e.g. `data[0]` is clade `0`, indexing of new clades\n    starts at `n` where `n` is the size of the original `data` tensor.\n    If `clades[k] == [i, j]`, then clades `i` and `j` were merged to form `k + n`.\n\n  * `dissimilarities` (`Nx.Tensor` with shape `{n - 1}`) -\n    Contains a metric that measures the intra-clade closeness of each newly formed clade.\n    Represented by the heights of each clade in a dendrogram plot.\n    Determined by both the `:dissimilarity` and `:linkage` options.\n\n  * `num_points` (`pos_integer/0`) -\n    Number of points in the dataset.\n    Must be $\\geq 3$.\n\n  * `sizes` (`Nx.Tensor` with shape `{n - 1}`) -\n    `sizes[i]` is the size of clade `i`.\n    If clade `k` was created by merging clades `i` and `j`, then\n    `sizes[k] == sizes[i] + sizes[j]`.","ref":"Scholar.Cluster.Hierarchical.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.Hierarchical.fit/2","doc":"iex> data = Nx.tensor([[2], [7], [9], [0], [3]])\n    iex> Hierarchical.fit(data)\n    %Scholar.Cluster.Hierarchical{\n      clades: Nx.tensor([[0, 4], [1, 2], [3, 5], [6, 7]]),\n      dissimilarities: Nx.tensor([1.0, 2.0, 2.0, 4.0]),\n      num_points: 5,\n      sizes: Nx.tensor([2, 2, 3, 5])\n    }","ref":"Scholar.Cluster.Hierarchical.html#fit/2-examples"},{"type":"function","title":"Scholar.Cluster.Hierarchical.labels_list/2","doc":"Cluster a `Scholar.Cluster.Hierarchical` struct into a list of cluster labels.","ref":"Scholar.Cluster.Hierarchical.html#labels_list/2"},{"type":"function","title":"Options - Scholar.Cluster.Hierarchical.labels_list/2","doc":"* `:cluster_by` (non-empty `t:keyword/0`) - Required. How to select which clades from the dendrogram should form the final clusters.\n  Must provide either a height or a number of clusters.\n\n  * `:height` (`t:float/0`) - Height of the dendrogram to use as the split point for clusters.\n\n  * `:num_clusters` (`t:pos_integer/0`) - Number of clusters to form.","ref":"Scholar.Cluster.Hierarchical.html#labels_list/2-options"},{"type":"function","title":"Return values - Scholar.Cluster.Hierarchical.labels_list/2","doc":"Returns a list of length `n` and values `0..(k - 1)` where `n` is the number of data points and\n`k` is the number of clusters formed.\nThe `i`th element of the result list is the label of the `i`th data point's cluster.\n\nCluster labels are arbitrary, but deterministic.","ref":"Scholar.Cluster.Hierarchical.html#labels_list/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.Hierarchical.labels_list/2","doc":"iex> data = Nx.tensor([[5], [5], [5], [10], [10]])\n    iex> model = Hierarchical.fit(data)\n    iex> Hierarchical.labels_list(model, cluster_by: [num_clusters: 2])\n    [0, 0, 0, 1, 1]","ref":"Scholar.Cluster.Hierarchical.html#labels_list/2-examples"},{"type":"function","title":"Scholar.Cluster.Hierarchical.labels_map/2","doc":"Cluster a `Scholar.Cluster.Hierarchical` struct into a map of cluster labels to member indices.","ref":"Scholar.Cluster.Hierarchical.html#labels_map/2"},{"type":"function","title":"Options - Scholar.Cluster.Hierarchical.labels_map/2","doc":"* `:cluster_by` (non-empty `t:keyword/0`) - Required. How to select which clades from the dendrogram should form the final clusters.\n  Must provide either a height or a number of clusters.\n\n  * `:height` (`t:float/0`) - Height of the dendrogram to use as the split point for clusters.\n\n  * `:num_clusters` (`t:pos_integer/0`) - Number of clusters to form.","ref":"Scholar.Cluster.Hierarchical.html#labels_map/2-options"},{"type":"function","title":"Return values - Scholar.Cluster.Hierarchical.labels_map/2","doc":"Returns a map where the keys are integers from `0..(k - 1)` where `k` is the number of clusters.\nEach value is a cluster represented by a list of member indices.\nE.g. if the result map was `%{0 => [0, 1], 1 => [2]}`, then elements `[0, 1]` of the data would\nbe in cluster `0` and the singleton element `[2]` would be in cluster `1`.\n\nCluster labels are arbitrary, but deterministic.","ref":"Scholar.Cluster.Hierarchical.html#labels_map/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.Hierarchical.labels_map/2","doc":"iex> data = Nx.tensor([[5], [5], [5], [10], [10]])\n    iex> model = Hierarchical.fit(data)\n    iex> Hierarchical.labels_map(model, cluster_by: [num_clusters: 2])\n    %{0 => [0, 1, 2], 1 => [3, 4]}","ref":"Scholar.Cluster.Hierarchical.html#labels_map/2-examples"},{"type":"module","title":"Scholar.Cluster.KMeans","doc":"K-Means Algorithm.\n\nK-Means is simple clustering method that works iteratively [1]. In the first iteration,\ncentroids are chosen randomly from input data. It turned out that some initializations\nare especially effective. In 2007 David Arthur and Sergei Vassilvitskii proposed initialization\ncalled k-means++ which speed up convergence of algorithm drastically [2]. After initialization, from each centroid\nfind points that are the closest to that centroid. Then, for each centroid replace it with the\ncenter of mass of associated points. These two steps mentioned above are repeated until the solution\nconverges. Since some initializations are unfortunate and converge to sub-optimal results\nwe need repeat the whole procedure a few times and take the best result.\n\nAverage time complexity is $O(CKNI)$, where $C$ is the number of clusters, $N$ is the number of samples,\n$I$ is the number of iterations until convergence, and $K$ is the number of features. Space\ncomplexity is $O(K*(N+C))$.\n\nReference:\n\n* [1] - [K-Means Algorithm](https://cs.nyu.edu/~roweis/csc2515-2006/readings/lloyd57.pdf)\n* [2] - [K-Means++ Initialization](http://ilpubs.stanford.edu:8090/778/1/2006-13.pdf)","ref":"Scholar.Cluster.KMeans.html"},{"type":"function","title":"Scholar.Cluster.KMeans.fit/2","doc":"Fits a K-Means model for sample inputs `x`.","ref":"Scholar.Cluster.KMeans.html#fit/2"},{"type":"function","title":"Options - Scholar.Cluster.KMeans.fit/2","doc":"* `:num_clusters` (`t:pos_integer/0`) - Required. The number of clusters to form as well as the number of centroids to generate.\n\n* `:max_iterations` (`t:pos_integer/0`) - Maximum number of iterations of the k-means algorithm for a single run. The default value is `300`.\n\n* `:num_runs` (`t:pos_integer/0`) - Number of time the k-means algorithm will be run with different centroid seeds.\n  The final results will be the best output of num_runs runs in terms of inertia. The default value is `10`.\n\n* `:tol` - Relative tolerance with regards to Frobenius norm of the difference in\n  the cluster centers of two consecutive iterations to declare convergence. The default value is `0.0001`.\n\n* `:weights` - The weights for each observation in `x`. If equals to `nil`,\n  all observations are assigned equal weight.\n\n* `:init` - Method for centroid initialization, either of:\n\n  * `:k_means_plus_plus` - selects initial cluster centroids using sampling based\n    on an empirical probability distribution of the points' contribution to\n    the overall inertia. This technique speeds up convergence, and is\n    theoretically proven to be O(log(k))-optimal.\n\n  * `:random` - choose `:num_clusters` observations (rows) at random from data for\n    the initial centroids.\n\n  The default value is `:k_means_plus_plus`.\n\n* `:key` - Determines random number generation for centroid initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.","ref":"Scholar.Cluster.KMeans.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Cluster.KMeans.fit/2","doc":"The function returns a struct with the following parameters:\n\n  * `:clusters` - Coordinates of cluster centers.\n\n  * `:num_iterations` - Number of iterations run.\n\n  * `:inertia` - Sum of squared distances of samples to their closest cluster center.\n\n  * `:labels` - Labels of each point.","ref":"Scholar.Cluster.KMeans.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.KMeans.fit/2","doc":"iex> key = Nx.Random.key(42)\n    iex> x = Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]])\n    iex> Scholar.Cluster.KMeans.fit(x, num_clusters: 2, key: key)\n    %Scholar.Cluster.KMeans{\n      num_iterations: Nx.tensor(\n        2\n      ),\n      clusters: Nx.tensor(\n        [\n          [1.0, 2.5],\n          [2.0, 4.5]\n        ]\n      ),\n      inertia: Nx.tensor(\n        1.0\n      ),\n      labels: Nx.tensor(\n        [0, 1, 0, 1]\n      )\n    }","ref":"Scholar.Cluster.KMeans.html#fit/2-examples"},{"type":"function","title":"Scholar.Cluster.KMeans.predict/2","doc":"Makes predictions with the given `model` on inputs `x`.","ref":"Scholar.Cluster.KMeans.html#predict/2"},{"type":"function","title":"Return Values - Scholar.Cluster.KMeans.predict/2","doc":"It returns a tensor with clusters corresponding to the input.","ref":"Scholar.Cluster.KMeans.html#predict/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.KMeans.predict/2","doc":"iex> key = Nx.Random.key(42)\n    iex> x = Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]])\n    iex> model = Scholar.Cluster.KMeans.fit(x, num_clusters: 2, key: key)\n    iex> Scholar.Cluster.KMeans.predict(model, Nx.tensor([[1.9, 4.3], [1.1, 2.0]]))\n    Nx.tensor(\n      [1, 0]\n    )","ref":"Scholar.Cluster.KMeans.html#predict/2-examples"},{"type":"function","title":"Scholar.Cluster.KMeans.transform/2","doc":"Calculates distances between each sample from `x` and the calculated centroids.","ref":"Scholar.Cluster.KMeans.html#transform/2"},{"type":"function","title":"Return Values - Scholar.Cluster.KMeans.transform/2","doc":"It returns a tensor with corresponding distances.","ref":"Scholar.Cluster.KMeans.html#transform/2-return-values"},{"type":"function","title":"Examples - Scholar.Cluster.KMeans.transform/2","doc":"iex> key = Nx.Random.key(40)\n    iex> model =\n    ...>  Scholar.Cluster.KMeans.fit(Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]]),\n    ...>    num_clusters: 2,\n    ...>    key: key\n    ...>  )\n    iex> Scholar.Cluster.KMeans.transform(model, Nx.tensor([[1.0, 2.5]]))\n    Nx.tensor(\n      [\n        [2.2360680103302, 0.0]\n      ]\n    )","ref":"Scholar.Cluster.KMeans.html#transform/2-examples"},{"type":"module","title":"Scholar.Decomposition.PCA","doc":"Principal Component Analysis (PCA).\n\nThe main concept of PCA is to find components (i.e. columns of a matrix) which explain the most variance\nof data set [1]. The sample data is decomposed using linear combination of\nvectors that lie on the directions of those components.\n\nThe time complexity is $O(NP^2 + P^3)$ where $N$ is the number of samples and $P$ is the number of features.\nSpace complexity is $O(P * (P+N))$.\nReference:\n\n* [1] - [Principal Component Analysis](https://en.wikipedia.org/wiki/Principal_component_analysis)","ref":"Scholar.Decomposition.PCA.html"},{"type":"function","title":"Scholar.Decomposition.PCA.fit/2","doc":"Fits a PCA for sample inputs `x`.","ref":"Scholar.Decomposition.PCA.html#fit/2"},{"type":"function","title":"Options - Scholar.Decomposition.PCA.fit/2","doc":"* `:num_components` - Number of components to keep. If `:num_components` is not set, all components are kept\n  which is the minimum value from number of features and number of samples. The default value is `nil`.","ref":"Scholar.Decomposition.PCA.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Decomposition.PCA.fit/2","doc":"The function returns a struct with the following parameters:\n\n  * `:components` - Principal axes in feature space, representing the directions of maximum variance in the data.\n    Equivalently, the right singular vectors of the centered input data, parallel to its eigenvectors.\n    The components are sorted by `:explained_variance`.\n\n  * `:explained_variance` - The amount of variance explained by each of the selected components.\n    The variance estimation uses `:num_samples - 1` degrees of freedom.\n    Equal to `:num_components` largest eigenvalues of the covariance matrix of `x`.\n\n  * `:explained_variance_ratio` - Percentage of variance explained by each of the selected components.\n    If `:num_components` is not set then all components are stored and the sum of the ratios is equal to 1.0.\n\n  * `:singular_values` - The singular values corresponding to each of the selected components.\n    The singular values are equal to the 2-norms of the `:num_components` variables in the lower-dimensional space.\n\n  * `:mean` - Per-feature empirical mean, estimated from the training set.\n\n  * `:num_components` - It equals the parameter `:num_components`, or the lesser\n    value of `:num_features` and `:num_samples` if the parameter `:num_components` is `nil`.\n\n  * `:num_features` - Number of features in the training data.\n\n  * `:num_samples` - Number of samples in the training data.","ref":"Scholar.Decomposition.PCA.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Decomposition.PCA.fit/2","doc":"iex> x = Nx.tensor([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])\n    iex> Scholar.Decomposition.PCA.fit(x)\n    %Scholar.Decomposition.PCA{\n      components: Nx.tensor(\n        [\n          [-0.838727593421936, -0.5445511937141418],\n          [0.5445511937141418, -0.838727593421936]\n        ]\n      ),\n      explained_variance: Nx.tensor(\n        [7.939542293548584, 0.06045711785554886]\n      ),\n      explained_variance_ratio: Nx.tensor(\n        [0.9924428462982178, 0.007557140197604895]\n      ),\n      singular_values: Nx.tensor(\n        [6.300611972808838, 0.5498050451278687]\n      ),\n      mean: Nx.tensor(\n        [0.0, 0.0]\n      ),\n      num_components: 2,\n      num_features: Nx.tensor(\n        2\n      ),\n      num_samples: Nx.tensor(\n        6\n      )\n    }","ref":"Scholar.Decomposition.PCA.html#fit/2-examples"},{"type":"function","title":"Scholar.Decomposition.PCA.fit_transform/2","doc":"Fit the model with `x` and apply the dimensionality reduction on `x`.\n\nThis function is analogous to calling `fit/2` and then\n`transform/3`, but it is calculated more efficiently.","ref":"Scholar.Decomposition.PCA.html#fit_transform/2"},{"type":"function","title":"Options - Scholar.Decomposition.PCA.fit_transform/2","doc":"* `:whiten` (`t:boolean/0`) - When true the result is multiplied by the square root of `:num_samples` and then\n  divided by the `:singular_values` to ensure uncorrelated outputs with unit component-wise variances.\n\n  Whitening will remove some information from the transformed signal (the relative variance scales of the components)\n  but can sometime improve the predictive accuracy of the downstream estimators by making their data respect some hard-wired assumptions.\n\n  The default value is `false`.","ref":"Scholar.Decomposition.PCA.html#fit_transform/2-options"},{"type":"function","title":"Return Values - Scholar.Decomposition.PCA.fit_transform/2","doc":"The function returns a tensor with decomposed data.","ref":"Scholar.Decomposition.PCA.html#fit_transform/2-return-values"},{"type":"function","title":"Examples - Scholar.Decomposition.PCA.fit_transform/2","doc":"iex> x = Nx.tensor([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])\n    iex> Scholar.Decomposition.PCA.fit_transform(x)\n    Nx.tensor(\n      [\n        [1.3819537162780762, 0.2936314642429352],\n        [2.2231407165527344, -0.25125157833099365],\n        [3.6050944328308105, 0.04237968474626541],\n        [-1.3819535970687866, -0.29363128542900085],\n        [-2.2231407165527344, 0.2512516379356384],\n        [-3.6050944328308105, -0.04237968474626541]\n      ]\n    )","ref":"Scholar.Decomposition.PCA.html#fit_transform/2-examples"},{"type":"function","title":"Scholar.Decomposition.PCA.transform/3","doc":"For a fitted `model` performs a decomposition.","ref":"Scholar.Decomposition.PCA.html#transform/3"},{"type":"function","title":"Options - Scholar.Decomposition.PCA.transform/3","doc":"* `:whiten` (`t:boolean/0`) - When true the result is multiplied by the square root of `:num_samples` and then\n  divided by the `:singular_values` to ensure uncorrelated outputs with unit component-wise variances.\n\n  Whitening will remove some information from the transformed signal (the relative variance scales of the components)\n  but can sometime improve the predictive accuracy of the downstream estimators by making their data respect some hard-wired assumptions.\n\n  The default value is `false`.","ref":"Scholar.Decomposition.PCA.html#transform/3-options"},{"type":"function","title":"Return Values - Scholar.Decomposition.PCA.transform/3","doc":"The function returns a tensor with decomposed data.","ref":"Scholar.Decomposition.PCA.html#transform/3-return-values"},{"type":"function","title":"Examples - Scholar.Decomposition.PCA.transform/3","doc":"iex> x = Nx.tensor([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])\n    iex> model = Scholar.Decomposition.PCA.fit(x)\n    iex> Scholar.Decomposition.PCA.transform(model, x)\n    Nx.tensor(\n      [\n        [1.3832788467407227, 0.2941763997077942],\n        [2.222006320953369, -0.25037479400634766],\n        [3.605285167694092, 0.04380160570144653],\n        [-1.3832788467407227, -0.2941763997077942],\n        [-2.222006320953369, 0.25037479400634766],\n        [-3.605285167694092, -0.04380160570144653]\n      ]\n    )","ref":"Scholar.Decomposition.PCA.html#transform/3-examples"},{"type":"module","title":"Scholar.Integrate","doc":"Module for numerical integration.","ref":"Scholar.Integrate.html"},{"type":"function","title":"Scholar.Integrate.simpson/3","doc":"Integrate `y` along the given axis using the simpson's rule.\nThe integration happens in sequence along elements of `x`.","ref":"Scholar.Integrate.html#simpson/3"},{"type":"function","title":"Options - Scholar.Integrate.simpson/3","doc":"* `:axis` - Axis along which to integrate. The default value is `-1`.\n\n* `:keep_axis` (`t:boolean/0`) - If set to true, the axis which is reduced is kept. The default value is `false`.\n\n* `:even` - If set to `:avg`, the average of the first and last interval is used in the integration.\n  If set to `:first`, the first interval is used in the integration.\n  If set to `:last`, the last interval is used in the integration. The default value is `:avg`.","ref":"Scholar.Integrate.html#simpson/3-options"},{"type":"function","title":"Examples - Scholar.Integrate.simpson/3","doc":"iex> y = Nx.tensor([1, 2, 3])\n    iex> x = Nx.tensor([4, 5, 6])\n    iex> Scholar.Integrate.simpson(y, x)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> x = Nx.tensor([[1, 2, 3], [1, 2, 3]])\n    iex> Scholar.Integrate.simpson(y, x)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> x = Nx.tensor([[1, 1, 1], [2, 2, 2]])\n    iex> Scholar.Integrate.simpson(y, x, axis: 0)\n    #Nx.Tensor","ref":"Scholar.Integrate.html#simpson/3-examples"},{"type":"function","title":"Scholar.Integrate.simpson_uniform/2","doc":"Integrate `y` along the given axis using the composite trapezoidal rule.\n\nThis is a simplified version of `trapezoidal/3` that assumes `x` is\na uniform tensor along `axis` with step size equal to `dx`.","ref":"Scholar.Integrate.html#simpson_uniform/2"},{"type":"function","title":"Options - Scholar.Integrate.simpson_uniform/2","doc":"* `:axis` - Axis along which to integrate. The default value is `-1`.\n\n* `:keep_axis` (`t:boolean/0`) - If set to true, the axis which is reduced is kept. The default value is `false`.\n\n* `:even` - If set to `:avg`, the average of the first and last interval is used in the integration.\n  If set to `:first`, the first interval is used in the integration.\n  If set to `:last`, the last interval is used in the integration. The default value is `:avg`.\n\n* `:dx` - The spacing between samples. The default value is `1.0`.","ref":"Scholar.Integrate.html#simpson_uniform/2-options"},{"type":"function","title":"Examples - Scholar.Integrate.simpson_uniform/2","doc":"iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Integrate.simpson_uniform(y)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Integrate.simpson_uniform(y, dx: 2)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> Scholar.Integrate.simpson_uniform(y, dx: 2, axis: 0)\n    #Nx.Tensor","ref":"Scholar.Integrate.html#simpson_uniform/2-examples"},{"type":"function","title":"Scholar.Integrate.trapezoidal/3","doc":"Integrate `y` along the given axis using the composite trapezoidal rule.\nThe integration happens in sequence along elements of `x`.","ref":"Scholar.Integrate.html#trapezoidal/3"},{"type":"function","title":"Options - Scholar.Integrate.trapezoidal/3","doc":"* `:axis` - Axis along which to integrate. The default value is `-1`.\n\n* `:keep_axis` (`t:boolean/0`) - If set to true, the axis which is reduced is kept. The default value is `false`.","ref":"Scholar.Integrate.html#trapezoidal/3-options"},{"type":"function","title":"Examples - Scholar.Integrate.trapezoidal/3","doc":"iex> y = Nx.tensor([1, 2, 3])\n    iex> x = Nx.tensor([4, 5, 6])\n    iex> Scholar.Integrate.trapezoidal(y, x)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> x = Nx.tensor([[1, 2, 3], [1, 2, 3]])\n    iex> Scholar.Integrate.trapezoidal(y, x)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> x = Nx.tensor([[1, 1, 1], [2, 2, 2]])\n    iex> Scholar.Integrate.trapezoidal(y, x, axis: 0)\n    #Nx.Tensor","ref":"Scholar.Integrate.html#trapezoidal/3-examples"},{"type":"function","title":"Scholar.Integrate.trapezoidal_uniform/2","doc":"Integrate `y` along the given axis using the composite trapezoidal rule.\n\nThis is a simplified version of `trapezoidal/3` that assumes `x` is\na uniform tensor along `axis` with step size equal to `dx`.","ref":"Scholar.Integrate.html#trapezoidal_uniform/2"},{"type":"function","title":"Options - Scholar.Integrate.trapezoidal_uniform/2","doc":"* `:axis` - Axis along which to integrate. The default value is `-1`.\n\n* `:keep_axis` (`t:boolean/0`) - If set to true, the axis which is reduced is kept. The default value is `false`.\n\n* `:dx` - The spacing between samples. The default value is `1.0`.","ref":"Scholar.Integrate.html#trapezoidal_uniform/2-options"},{"type":"function","title":"Examples - Scholar.Integrate.trapezoidal_uniform/2","doc":"iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Integrate.trapezoidal_uniform(y)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Integrate.trapezoidal_uniform(y, dx: 2)\n    #Nx.Tensor \n\n    iex> y = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> Scholar.Integrate.trapezoidal_uniform(y, dx: 2, axis: 0)\n    #Nx.Tensor","ref":"Scholar.Integrate.html#trapezoidal_uniform/2-examples"},{"type":"module","title":"Scholar.Interpolation.BezierSpline","doc":"Cubic Bezier Spline interpolation.\n\nThis kind of interpolation is calculated by fitting a set of\ncontinuous cubic polynomials of which the first and second\nderivatives are also continuous (which makes them curves of class $C^2$).\nIn contrast to the `Scholar.Interpolation.CubicSpline` algorithm,\nthe Bezier curves aren't necessarily of class $C^2$, but this interpolation\nforces this so it yields a smooth function as the result.\n\nThe interpolated curve also has local control,\nmeaning that a point that would be problematic for other\ninterpolations, such as the `Scholar.Interpolation.CubicSpline`\nor `Scholar.Interpolation.Linear` algorithms, will only affect\nthe segments right next to it, instead of affecting the curve as a whole.\n\nComputing Bezier curve is $O(N^2)$ where $N$ is the number of points.\n\nReference:\n\n  * [1] - [Bezier theory](https://en.wikipedia.org/wiki/B%C3%A9zier_curve)\n  * [2] - [Spline equation derivation](https://www.particleincell.com/2012/bezier-splines/)","ref":"Scholar.Interpolation.BezierSpline.html"},{"type":"function","title":"Scholar.Interpolation.BezierSpline.fit/2","doc":"Fits a cubic Bezier spline interpolation of the given `(x, y)` points.\n\nInputs are expected to be rank-1 tensors with the same shape\nand at least 4 entries.","ref":"Scholar.Interpolation.BezierSpline.html#fit/2"},{"type":"function","title":"Examples - Scholar.Interpolation.BezierSpline.fit/2","doc":"iex> x = Nx.iota({4})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0, 0.5])\n    iex> Scholar.Interpolation.BezierSpline.fit(x, y)\n    %Scholar.Interpolation.BezierSpline{\n      coefficients: Nx.tensor(\n        [\n          [\n            [0.0, 2.0],\n            [0.3333331048488617, 1.0333333015441895],\n            [0.6666665077209473, 0.06666667759418488],\n            [1.0, 0.0]\n          ],\n          [\n            [1.0, 0.0],\n            [1.3333334922790527, -0.06666667759418488],\n            [1.6666665077209473, 0.7666666507720947],\n            [2.0, 1.0]\n          ],\n          [\n            [2.0, 1.0],\n            [2.3333334922790527, 1.2333333492279053],\n            [2.6666667461395264, 0.8666666746139526],\n            [3.0, 0.5]\n          ]\n        ]\n      ),\n      k: Nx.tensor(\n        [\n          [0.0, 2.0],\n          [1.0, 0.0],\n          [2.0, 1.0],\n          [3.0, 0.5]\n        ]\n      )\n    }","ref":"Scholar.Interpolation.BezierSpline.html#fit/2-examples"},{"type":"function","title":"Scholar.Interpolation.BezierSpline.predict/3","doc":"Returns the value fit by `fit/2` corresponding to the `target_x` input.","ref":"Scholar.Interpolation.BezierSpline.html#predict/3"},{"type":"function","title":"Options - Scholar.Interpolation.BezierSpline.predict/3","doc":"* `:max_iter` (`t:pos_integer/0`) - determines the maximum iterations for converging the t parameter for each spline polynomial The default value is `15`.\n\n* `:eps` (`t:float/0`) - determines the tolerance to be used for converging the t parameter for each spline polynomial The default value is `1.0e-6`.","ref":"Scholar.Interpolation.BezierSpline.html#predict/3-options"},{"type":"function","title":"Examples - Scholar.Interpolation.BezierSpline.predict/3","doc":"iex> x = Nx.iota({4})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0, 0.5])\n    iex> model = Scholar.Interpolation.BezierSpline.fit(x, y)\n    iex> Scholar.Interpolation.BezierSpline.predict(model, Nx.tensor([3.0, 4.0, 2.0, 7.0]))\n    Nx.tensor(\n      [0.5000335574150085, -4.2724612285383046e-5, 0.9999786615371704, 34.5]\n    )","ref":"Scholar.Interpolation.BezierSpline.html#predict/3-examples"},{"type":"module","title":"Scholar.Interpolation.CubicSpline","doc":"Cubic Spline interpolation.\n\nThis kind of interpolation is calculated by fitting a set of\ncontinuous cubic polynomials of which the first and second\nderivatives are also continuous.\n\nThe interpolated curve is smooth and mitigates oscillations that\ncould appear if a single n-th degree polynomial were to be\nfitted over all of the points.\n\nCubic spline interpolation is $O(N)$ where $N$ is the number of points.\n\nReference:\n\n  * [1] - [Cubic Spline Interpolation theory](https://en.wikiversity.org/wiki/Cubic_Spline_Interpolation)\n  * [2] - [SciPy implementation](https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html)","ref":"Scholar.Interpolation.CubicSpline.html"},{"type":"function","title":"Scholar.Interpolation.CubicSpline.fit/3","doc":"Fits a cubic spline interpolation of the given `(x, y)` points\n\nInputs are expected to be rank-1 tensors with the same shape\nand at least 3 entries.","ref":"Scholar.Interpolation.CubicSpline.html#fit/3"},{"type":"function","title":"Options - Scholar.Interpolation.CubicSpline.fit/3","doc":"* `:boundary_condition` (`t:atom/0`) - one of :not_a_knot or :natural The default value is `:not_a_knot`.","ref":"Scholar.Interpolation.CubicSpline.html#fit/3-options"},{"type":"function","title":"Examples - Scholar.Interpolation.CubicSpline.fit/3","doc":"iex> x = Nx.iota({3})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0])\n    iex> Scholar.Interpolation.CubicSpline.fit(x, y)\n    %Scholar.Interpolation.CubicSpline{\n      coefficients: Nx.tensor(\n        [\n          [0.0, 1.5, -3.5, 2.0],\n          [0.0, 1.5, -0.5, 0.0]\n        ]\n      ),\n      x: Nx.tensor(\n        [0, 1, 2]\n      )\n    }","ref":"Scholar.Interpolation.CubicSpline.html#fit/3-examples"},{"type":"function","title":"Scholar.Interpolation.CubicSpline.predict/3","doc":"Returns the value fit by `fit/2` corresponding to the `target_x` input","ref":"Scholar.Interpolation.CubicSpline.html#predict/3"},{"type":"function","title":"Options - Scholar.Interpolation.CubicSpline.predict/3","doc":"* `:extrapolate` (`t:boolean/0`) - if false, out-of-bounds x return NaN. The default value is `true`.","ref":"Scholar.Interpolation.CubicSpline.html#predict/3-options"},{"type":"function","title":"Examples - Scholar.Interpolation.CubicSpline.predict/3","doc":"iex> x = Nx.iota({3})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0])\n    iex> model = Scholar.Interpolation.CubicSpline.fit(x, y)\n    iex> Scholar.Interpolation.CubicSpline.predict(model, Nx.tensor([[1.0, 4.0], [3.0, 7.0]]))\n    Nx.tensor(\n      [\n        [0.0, 12.0],\n        [5.0, 51.0]\n      ]\n    )","ref":"Scholar.Interpolation.CubicSpline.html#predict/3-examples"},{"type":"module","title":"Scholar.Interpolation.Linear","doc":"Linear interpolation.\n\nThis kind of interpolation is calculated by fitting polynomials\nof the first degree between each pair of given points.\n\nThis means that for points $(x_0, y_0), (x_1, y_1)$, the\npredictive polynomial will be given by:\n$$\n\\begin{cases}\ny = ax + b \\newline\na = \\dfrac{y_1 - y_0}{x_1 - x_0} \\newline\nb = y_1 - ax_1 = y_0 - ax_0\n\\end{cases}\n$$\n\nLinear interpolation has $O(N)$ time and space complexity where $N$ is the number of points.","ref":"Scholar.Interpolation.Linear.html"},{"type":"function","title":"Scholar.Interpolation.Linear.fit/2","doc":"Fits a linear interpolation of the given `(x, y)` points\n\nInputs are expected to be rank-1 tensors with the same shape\nand at least 2 entries.","ref":"Scholar.Interpolation.Linear.html#fit/2"},{"type":"function","title":"Examples - Scholar.Interpolation.Linear.fit/2","doc":"iex> x = Nx.iota({3})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0])\n    iex> Scholar.Interpolation.Linear.fit(x, y)\n    %Scholar.Interpolation.Linear{\n      coefficients: Nx.tensor(\n        [\n          [-2.0, 2.0],\n          [1.0, -1.0]\n        ]\n      ),\n      x: Nx.tensor(\n        [0, 1, 2]\n      )\n    }","ref":"Scholar.Interpolation.Linear.html#fit/2-examples"},{"type":"function","title":"Scholar.Interpolation.Linear.predict/3","doc":"Returns the value fit by `fit/2` corresponding to the `target_x` input.","ref":"Scholar.Interpolation.Linear.html#predict/3"},{"type":"function","title":"Examples - Scholar.Interpolation.Linear.predict/3","doc":"iex> x = Nx.iota({3})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0])\n    iex> model = Scholar.Interpolation.Linear.fit(x, y)\n    iex> Scholar.Interpolation.Linear.predict(model, Nx.tensor([[1.0, 4.0], [3.0, 7.0]]))\n    Nx.tensor(\n      [\n        [0.0, 3.0],\n        [2.0, 6.0]\n      ]\n    )\n\n    iex> x = Nx.iota({5})\n    iex> y = Nx.tensor([2.0, 0.0, 1.0, 3.0, 4.0])\n    iex> model = Scholar.Interpolation.Linear.fit(x, y)\n    iex> target_x = Nx.tensor([-2, -1, 1.25, 3, 3.25, 5.0])\n    iex> Scholar.Interpolation.Linear.predict(model, target_x, left: 0.0, right: 10.0)\n    #Nx.Tensor","ref":"Scholar.Interpolation.Linear.html#predict/3-examples"},{"type":"type","title":"Scholar.Interpolation.Linear.t/0","doc":"","ref":"Scholar.Interpolation.Linear.html#t:t/0"},{"type":"module","title":"Scholar.Linear.BayesianRidgeRegression","doc":"Bayesian ridge regression: A fully probabilistic linear model with parameter regularization.\n\nIn order to obtain a fully probabilistic linear model,\nwe declare the precision parameter in the model: $\\alpha$,\nThis parameter describes the dispersion of the data around the mean.\n\n$$\np(y | X, w, \\alpha) = \\mathcal{N}(y | Xw, \\alpha^{-1})\n$$\n\nWhere:\n* $X$ is an input data\n\n* $y$ is an input target\n\n* $w$ is the model weights matrix\n\n* $\\alpha$ is the precision parameter of the target and $\\alpha^{-1} = \\sigma^{2}$, the variance.\n\nIn order to obtain a fully probabilistic regularized linear model,\nwe declare the distribution of the model weights matrix\nwith it's corresponding precision parameter:\n\n$$\np(w | \\lambda) = \\mathcal{N}(w, \\lambda^{-1})\n$$\n\nWhere $\\lambda$ is the precision parameter of the weights matrix.\n\nBoth $\\alpha$ and $\\lambda$ are choosen to have prior gamma distributions,\ncontrolled through hyperparameters $\\alpha_1$, $\\alpha_2$, $\\lambda_1$, $\\lambda_2$.\nThese parameters are set by default to non-informative\n$\\alpha_1 = \\alpha_2 = \\lambda_1 = \\lambda_2 = 1^{-6}$.\n\nThis model is similar to the classical ridge regression.\nConfusingly the classical ridge regression's $\\alpha$ parameter is the Bayesian ridge's $\\lambda$ parameter.\n\nOther than that, the differences between alorithms are:\n* The matrix weight regularization parameter is estimated from data,\n* The precision of the target is estimated.\n\nAs such, Bayesian ridge is more flexible to the data at hand.\nThese features come at higher computational cost.\n\nThis implementation is ported from Python's scikit-learn.\nIt uses the algorithm described in (Tipping, 2001)\nand regularization parameters are updated as by (MacKay, 1992).\n\nReferences:\n\nD. J. C. MacKay, Bayesian Interpolation, Computation and Neural Systems,\nVol. 4, No. 3, 1992.\n\nM. E. Tipping, Sparse Bayesian Learning and the Relevance Vector Machine,\nJournal of Machine Learning Research, Vol. 1, 2001.\n\nPedregosa et al., Scikit-learn: Machine Learning in Python,\nJMLR 12, pp. 2825-2830, 2011.","ref":"Scholar.Linear.BayesianRidgeRegression.html"},{"type":"function","title":"Scholar.Linear.BayesianRidgeRegression.fit/3","doc":"Fits a Bayesian ridge model for sample inputs `x` and\nsample targets `y`.","ref":"Scholar.Linear.BayesianRidgeRegression.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.BayesianRidgeRegression.fit/3","doc":"* `:iterations` (`t:pos_integer/0`) - Maximum number of iterations before stopping the fitting algorithm.\n  The number of iterations may be lower is parameters converge. The default value is `300`.\n\n* `:sample_weights` - The weights for each observation. If not provided,\n  all observations are assigned equal weight.\n\n* `:fit_intercept?` (`t:boolean/0`) - If set to `true`, a model will fit the intercept. Otherwise,\n  the intercept is set to `0.0`. The intercept is an independent term\n  in a linear model. Specifically, it is the expected mean value\n  of targets for a zero-vector on input. The default value is `true`.\n\n* `:compute_scores?` (`t:boolean/0`) - If set to `true`, the log marginal likelihood will be computed\n  at each iteration of the algorithm. The default value is `false`.\n\n* `:alpha_init` - The initial value for alpha. This parameter influences the precision of the noise.\n  `:alpha` must be a non-negative float i.e. in [0, inf).\n  Defaults to 1/Var(y).\n\n* `:lambda_init` - The initial value for lambda. This parameter influences the precision of the weights.\n  `:lambda` must be a non-negative float i.e. in [0, inf).\n  Defaults to 1. The default value is `1.0`.\n\n* `:alpha_1` - Hyper-parameter : shape parameter for the Gamma distribution prior      \n  over the alpha parameter. The default value is `1.0e-6`.\n\n* `:alpha_2` - Hyper-parameter : inverse scale (rate) parameter for the Gamma distribution prior\n  over the alpha parameter. The default value is `1.0e-6`.\n\n* `:lambda_1` - Hyper-parameter : shape parameter for the Gamma distribution prior\n  over the lambda parameter. The default value is `1.0e-6`.\n\n* `:lambda_2` - Hyper-parameter : inverse scale (rate) parameter for the Gamma distribution prior      \n  over the lambda parameter. The default value is `1.0e-6`.\n\n* `:eps` (`t:float/0`) - The convergence tolerance. When `Nx.sum(Nx.abs(coef - coef_new)) < :eps`, the algorithm is considered to have converged. The default value is `1.0e-8`.","ref":"Scholar.Linear.BayesianRidgeRegression.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.BayesianRidgeRegression.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:coefficients` - Estimated coefficients for the linear regression problem.\n\n  * `:intercept` - Independent term in the linear model.\n\n  * `:alpha` - Estimated precision of the noise.\n\n  * `:lambda` - Estimated precision of the weights.\n\n  * `:sigma` - Estimated variance covariance matrix of weights with shape (n_features, n_features).\n\n  * `:iterations` - How many times the optimization algorithm was computed.\n\n  * `:has_converged` - Whether the coefficients converged during the optimization algorithm.\n\n  * `:scores` - Value of the log marginal likelihood at each iteration during the optimization.","ref":"Scholar.Linear.BayesianRidgeRegression.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.BayesianRidgeRegression.fit/3","doc":"iex> x = Nx.tensor([[1], [2], [6], [8], [10]])\n    iex> y = Nx.tensor([1, 2, 6, 8, 10])\n    iex> model = Scholar.Linear.BayesianRidgeRegression.fit(x, y)\n    iex> model.coefficients\n    #Nx.Tensor \n    iex> model.intercept\n    #Nx.Tensor","ref":"Scholar.Linear.BayesianRidgeRegression.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.BayesianRidgeRegression.predict/2","doc":"Makes predictions with the given `model` on input `x`.","ref":"Scholar.Linear.BayesianRidgeRegression.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.BayesianRidgeRegression.predict/2","doc":"iex> x = Nx.tensor([[1], [2], [6], [8], [10]])\n    iex> y = Nx.tensor([1, 2, 6, 8, 10])\n    iex> model = Scholar.Linear.BayesianRidgeRegression.fit(x, y)\n    iex> Scholar.Linear.BayesianRidgeRegression.predict(model, Nx.tensor([[1], [3], [4]]))\n    Nx.tensor(\n      [1.02969491481781, 3.0161972045898438, 4.009448528289795]  \n    )","ref":"Scholar.Linear.BayesianRidgeRegression.html#predict/2-examples"},{"type":"module","title":"Scholar.Linear.IsotonicRegression","doc":"Isotonic regression is a method of fitting a free-form line to a set of\nobservations by solving a convex optimization problem. It is a form of\nregression analysis that can be used as an alternative to polynomial\nregression to fit nonlinear data.\n\nTime complexity of isotonic regression is $O(N^2)$ where $N$ is the\nnumber of points.","ref":"Scholar.Linear.IsotonicRegression.html"},{"type":"function","title":"Scholar.Linear.IsotonicRegression.check_preprocess/1","doc":"","ref":"Scholar.Linear.IsotonicRegression.html#check_preprocess/1"},{"type":"function","title":"Scholar.Linear.IsotonicRegression.fit/3","doc":"Fits a isotonic regression model for sample inputs `x` and\nsample targets `y`.","ref":"Scholar.Linear.IsotonicRegression.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.IsotonicRegression.fit/3","doc":"* `:y_min` (`t:float/0`) - Lower bound on the lowest predicted value. If if not provided, the lower bound\n  is set to `Nx.Constant.neg_infinity()`.\n\n* `:y_max` (`t:float/0`) - Upper bound on the highest predicted value. If if not provided, the lower bound\n  is set to `Nx.Constant.infinity()`.\n\n* `:increasing` - Whether the isotonic regression should be fit with the constraint that the\n  function is monotonically increasing. If `false`, the constraint is that\n  the function is monotonically decreasing. If `:auto`, the constraint is\n  determined automatically based on the data. The default value is `:auto`.\n\n* `:out_of_bounds` - How to handle out-of-bounds points. If `:clip`, out-of-bounds points are\n  mapped to the nearest valid value. If `:nan`, out-of-bounds points are\n  replaced with `Nx.Constant.nan()`. The default value is `:nan`.\n\n* `:sample_weights` - The weights for each observation. If not provided,\n  all observations are assigned equal weight.","ref":"Scholar.Linear.IsotonicRegression.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.IsotonicRegression.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:x_min` - Minimum value of input tensor `x`.\n\n  * `:x_max` - Maximum value of input tensor `x`.\n\n  * `:x_thresholds` - Thresholds used for predictions.\n\n  * `:y_thresholds` - Predicted values associated with each threshold.\n\n  * `:increasing` - Whether the isotonic regression is increasing.\n\n  * `:cutoff_index` - The index of the last valid threshold. Rest elements are placeholders\n    for the sake of preserving shape of tensor.\n\n  * `:preprocess` - Interpolation function to be applied on input tensor `x`. Before `preprocess/1`\n    is applied it is set to {}","ref":"Scholar.Linear.IsotonicRegression.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.IsotonicRegression.fit/3","doc":"iex> x = Nx.tensor([1, 4, 7, 9, 10, 11])\n    iex> y = Nx.tensor([1, 3, 6, 8, 9, 10])\n    iex> Scholar.Linear.IsotonicRegression.fit(x, y)\n    %Scholar.Linear.IsotonicRegression{\n      x_min: Nx.tensor(\n        1.0\n      ),\n      x_max: Nx.tensor(\n        11.0\n      ),\n      x_thresholds: Nx.tensor(\n        [1.0, 4.0, 7.0, 9.0, 10.0, 11.0]\n      ),\n      y_thresholds: Nx.tensor(\n        [1.0, 3.0, 6.0, 8.0, 9.0, 10.0]\n      ),\n      increasing: Nx.u8(1),\n      cutoff_index: Nx.tensor(\n        5\n      ),\n      preprocess: {}\n    }","ref":"Scholar.Linear.IsotonicRegression.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.IsotonicRegression.predict/2","doc":"Makes predictions with the given `model` on input `x` and interpolating `function`.","ref":"Scholar.Linear.IsotonicRegression.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.IsotonicRegression.predict/2","doc":"iex> x = Nx.tensor([1, 4, 7, 9, 10, 11])\n    iex> y = Nx.tensor([1, 3, 6, 8, 9, 10])\n    iex> model = Scholar.Linear.IsotonicRegression.fit(x, y)\n    iex> model = Scholar.Linear.IsotonicRegression.preprocess(model)\n    iex> to_predict = Nx.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])\n    iex> Scholar.Linear.IsotonicRegression.predict(model, to_predict)\n    #Nx.Tensor","ref":"Scholar.Linear.IsotonicRegression.html#predict/2-examples"},{"type":"function","title":"Scholar.Linear.IsotonicRegression.preprocess/2","doc":"Preprocesses the `model` for prediction.\n\nReturns an updated `model`.","ref":"Scholar.Linear.IsotonicRegression.html#preprocess/2"},{"type":"function","title":"Examples - Scholar.Linear.IsotonicRegression.preprocess/2","doc":"iex> x = Nx.tensor([1, 4, 7, 9, 10, 11])\n    iex> y = Nx.tensor([1, 3, 6, 8, 9, 10])\n    iex> model = Scholar.Linear.IsotonicRegression.fit(x, y)\n    iex> Scholar.Linear.IsotonicRegression.preprocess(model)\n    %Scholar.Linear.IsotonicRegression{\n      x_min: Nx.tensor(\n        1.0\n      ),\n      x_max: Nx.tensor(\n        11.0\n      ),\n      x_thresholds: Nx.tensor(\n        [1.0, 4.0, 7.0, 9.0, 10.0, 11.0]\n      ),\n      y_thresholds: Nx.tensor(\n        [1.0, 3.0, 6.0, 8.0, 9.0, 10.0]\n      ),\n      increasing: Nx.u8(1),\n      cutoff_index: Nx.tensor(\n        5\n      ),\n      preprocess: %Scholar.Interpolation.Linear{\n        coefficients: Nx.tensor(\n          [\n            [0.6666666865348816, 0.3333333134651184],\n            [1.0, -1.0],\n            [1.0, -1.0],\n            [1.0, -1.0],\n            [1.0, -1.0]\n          ]\n        ),\n        x: Nx.tensor(\n          [1.0, 4.0, 7.0, 9.0, 10.0, 11.0]\n        )\n      }\n    }","ref":"Scholar.Linear.IsotonicRegression.html#preprocess/2-examples"},{"type":"type","title":"Scholar.Linear.IsotonicRegression.t/0","doc":"","ref":"Scholar.Linear.IsotonicRegression.html#t:t/0"},{"type":"module","title":"Scholar.Linear.LinearRegression","doc":"Ordinary least squares linear regression.\n\nTime complexity of linear regression is $O((K^2) * (K+N))$ where $N$ is the number of samples\nand $K$ is the number of features.","ref":"Scholar.Linear.LinearRegression.html"},{"type":"function","title":"Scholar.Linear.LinearRegression.fit/3","doc":"Fits a linear regression model for sample inputs `x` and\nsample targets `y`.","ref":"Scholar.Linear.LinearRegression.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.LinearRegression.fit/3","doc":"* `:sample_weights` - The weights for each observation. If not provided,\n  all observations are assigned equal weight.\n\n* `:fit_intercept?` (`t:boolean/0`) - If set to `true`, a model will fit the intercept. Otherwise,\n  the intercept is set to `0.0`. The intercept is an independent term\n  in a linear model. Specifically, it is the expected mean value\n  of targets for a zero-vector on input. The default value is `true`.","ref":"Scholar.Linear.LinearRegression.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.LinearRegression.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:coefficients` - Estimated coefficients for the linear regression problem.\n\n  * `:intercept` - Independent term in the linear model.","ref":"Scholar.Linear.LinearRegression.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.LinearRegression.fit/3","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> model = Scholar.Linear.LinearRegression.fit(x, y)\n    iex> model.coefficients\n    #Nx.Tensor \n    iex> model.intercept\n    #Nx.Tensor","ref":"Scholar.Linear.LinearRegression.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.LinearRegression.predict/2","doc":"Makes predictions with the given `model` on input `x`.","ref":"Scholar.Linear.LinearRegression.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.LinearRegression.predict/2","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> model = Scholar.Linear.LinearRegression.fit(x, y)\n    iex> Scholar.Linear.LinearRegression.predict(model, Nx.tensor([[2.0, 1.0]]))\n    Nx.tensor(\n      [4.200936794281006]\n    )","ref":"Scholar.Linear.LinearRegression.html#predict/2-examples"},{"type":"module","title":"Scholar.Linear.LogisticRegression","doc":"Logistic regression in both binary and multinomial variants.\n\nTime complexity is $O(N * K * I)$ where $N$ is the number of samples, $K$ is the number of features, and $I$ is the number of iterations.","ref":"Scholar.Linear.LogisticRegression.html"},{"type":"function","title":"Scholar.Linear.LogisticRegression.fit/3","doc":"Fits a logistic regression model for sample inputs `x` and sample\ntargets `y`.\n\nDepending on number of classes the function chooses either binary\nor multinomial logistic regression.","ref":"Scholar.Linear.LogisticRegression.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.LogisticRegression.fit/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. number of classes contained in the input tensors.\n\n* `:iterations` (`t:pos_integer/0`) - number of iterations of gradient descent performed inside logistic\n  regression. The default value is `1000`.\n\n* `:learning_loop_unroll` (`t:boolean/0`) - If `true`, the learning loop is unrolled. The default value is `false`.\n\n* `:optimizer` - The optimizer name or {init, update} pair of functions (see `Polaris.Optimizers` for more details). The default value is `:sgd`.\n\n* `:eps` (`t:float/0`) - The convergence tolerance. If the `abs(loss) < size(x) * :eps`, the algorithm is considered to have converged. The default value is `1.0e-8`.","ref":"Scholar.Linear.LogisticRegression.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.LogisticRegression.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:coefficients` - Coefficient of the features in the decision function.\n\n  * `:bias` - Bias added to the decision function.\n\n  * `:mode` - Indicates whether the problem is binary classification (`:num_classes` set to 2)\n    or multinomial (`:num_classes` is bigger than 2). For binary classification set to `:binary`, otherwise\n    set to `:multinomial`.","ref":"Scholar.Linear.LogisticRegression.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.LogisticRegression.fit/3","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([1, 0, 1])\n    iex> Scholar.Linear.LogisticRegression.fit(x, y, num_classes: 2)\n    %Scholar.Linear.LogisticRegression{\n      coefficients: Nx.tensor(\n        [\n          [2.5531527996063232, -0.5531544089317322],\n          [-0.35652396082878113, 2.3565237522125244]\n        ]\n      ),\n      bias: Nx.tensor(\n        [-0.28847914934158325, 0.28847917914390564]\n      )\n    }","ref":"Scholar.Linear.LogisticRegression.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.LogisticRegression.predict/2","doc":"Makes predictions with the given `model` on inputs `x`.","ref":"Scholar.Linear.LogisticRegression.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.LogisticRegression.predict/2","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([1, 0, 1])\n    iex> model = Scholar.Linear.LogisticRegression.fit(x, y, num_classes: 2)\n    iex> Scholar.Linear.LogisticRegression.predict(model, Nx.tensor([[-3.0, 5.0]]))\n    #Nx.Tensor","ref":"Scholar.Linear.LogisticRegression.html#predict/2-examples"},{"type":"function","title":"Scholar.Linear.LogisticRegression.predict_probability/2","doc":"Calculates probabilities of predictions with the given `model` on inputs `x`.","ref":"Scholar.Linear.LogisticRegression.html#predict_probability/2"},{"type":"function","title":"Examples - Scholar.Linear.LogisticRegression.predict_probability/2","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([1, 0, 1])\n    iex> model = Scholar.Linear.LogisticRegression.fit(x, y, num_classes: 2)\n    iex> Scholar.Linear.LogisticRegression.predict_probability(model, Nx.tensor([[-3.0, 5.0]]))\n    #Nx.Tensor","ref":"Scholar.Linear.LogisticRegression.html#predict_probability/2-examples"},{"type":"module","title":"Scholar.Linear.PolynomialRegression","doc":"Least squares polynomial regression.\n\nTime complexity of polynomial regression is $O((K^2) * (K+N))$ where $N$ is the number of samples and $K$ is the number of features.","ref":"Scholar.Linear.PolynomialRegression.html"},{"type":"function","title":"Scholar.Linear.PolynomialRegression.fit/3","doc":"Fits a polynomial regression model for sample inputs `x` and\nsample targets `y`.","ref":"Scholar.Linear.PolynomialRegression.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.PolynomialRegression.fit/3","doc":"* `:sample_weights` - The weights for each observation. If not provided,\n  all observations are assigned equal weight.\n\n* `:degree` (`t:pos_integer/0`) - The degree of the feature matrix to return. Must be an integer equal or greater than 1. 1\n  returns the input matrix. The default value is `2`.\n\n* `:fit_intercept?` (`t:boolean/0`) - If set to `true`, a model will fit the intercept. Otherwise,\n  the intercept is set to `0.0`. The intercept is an independent term\n  in a linear model. Specifically, it is the expected mean value\n  of targets for a zero-vector on input. The default value is `true`.","ref":"Scholar.Linear.PolynomialRegression.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.PolynomialRegression.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:coefficients` - Estimated coefficients for the polynomial regression problem.\n\n  * `:intercept` - Independent term in the polynomial model.","ref":"Scholar.Linear.PolynomialRegression.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.PolynomialRegression.fit/3","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> model = Scholar.Linear.PolynomialRegression.fit(x, y, degree: 1)\n    iex> model.coefficients\n    #Nx.Tensor \n    iex> model.intercept\n    #Nx.Tensor \n    iex> model.degree\n    1\n\n    iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> model = Scholar.Linear.PolynomialRegression.fit(x, y, degree: 2)\n    iex> model.coefficients\n    #Nx.Tensor \n    iex> model.intercept\n    #Nx.Tensor \n    iex> model.degree\n    2","ref":"Scholar.Linear.PolynomialRegression.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.PolynomialRegression.predict/2","doc":"Makes predictions with the given `model` on input `x`.","ref":"Scholar.Linear.PolynomialRegression.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.PolynomialRegression.predict/2","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> model = Scholar.Linear.PolynomialRegression.fit(x, y, degree: 2)\n    iex> Scholar.Linear.PolynomialRegression.predict(model, Nx.tensor([[2.0, 1.0]]))\n    #Nx.Tensor","ref":"Scholar.Linear.PolynomialRegression.html#predict/2-examples"},{"type":"function","title":"Scholar.Linear.PolynomialRegression.transform/2","doc":"Computes the feature matrix for polynomial regression.\n\n* `:degree` (`t:pos_integer/0`) - The degree of the feature matrix to return. Must be an integer equal or greater than 1. 1\n  returns the input matrix. The default value is `2`.\n\n* `:fit_intercept?` (`t:boolean/0`) - If set to `true`, a model will fit the intercept. Otherwise,\n  the intercept is set to `0.0`. The intercept is an independent term\n  in a linear model. Specifically, it is the expected mean value\n  of targets for a zero-vector on input. The default value is `true`.","ref":"Scholar.Linear.PolynomialRegression.html#transform/2"},{"type":"function","title":"Examples - Scholar.Linear.PolynomialRegression.transform/2","doc":"iex> x = Nx.tensor([[2]])\n    iex> Scholar.Linear.PolynomialRegression.transform(x, degree: 0)\n    ** (NimbleOptions.ValidationError) invalid value for :degree option: expected positive integer, got: 0\n\n    iex> x = Nx.tensor([[2]])\n    iex> Scholar.Linear.PolynomialRegression.transform(x, degree: 5, fit_intercept?: false)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[2, 3]])\n    iex> Scholar.Linear.PolynomialRegression.transform(x)\n    #Nx.Tensor \n\n    iex> x = Nx.iota({3, 2})\n    iex> Scholar.Linear.PolynomialRegression.transform(x, fit_intercept?: false)\n    #Nx.Tensor","ref":"Scholar.Linear.PolynomialRegression.html#transform/2-examples"},{"type":"module","title":"Scholar.Linear.RidgeRegression","doc":"Linear least squares with $L_2$ regularization.\n\nMinimizes the objective function:\n$$\n||y - Xw||^2\\_2 + \\alpha||w||^2\\_2\n$$\n\nWhere:\n* $X$ is an input data\n\n* $y$ is an input target\n\n* $w$ is the model weights matrix\n\n* $\\alpha$ is the parameter that controls level of regularization\n\nTime complexity is $O(N^2)$ for `:cholesky` solver and $O((N^2) * (K + N))$ for `:svd` solver,\nwhere $N$ is the number of observations and $K$ is the number of features.","ref":"Scholar.Linear.RidgeRegression.html"},{"type":"function","title":"Scholar.Linear.RidgeRegression.fit/3","doc":"Fits a Ridge regression model for sample inputs `x` and\nsample targets `y`.","ref":"Scholar.Linear.RidgeRegression.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.RidgeRegression.fit/3","doc":"* `:sample_weights` - The weights for each observation. If not provided,\n  all observations are assigned equal weight.\n\n* `:fit_intercept?` (`t:boolean/0`) - If set to `true`, a model will fit the intercept. Otherwise,\n  the intercept is set to `0.0`. The intercept is an independent term\n  in a linear model. Specifically, it is the expected mean value\n  of targets for a zero-vector on input. The default value is `true`.\n\n* `:solver` - Solver to use in the computational routines:\n\n  * `:svd` - Uses a Singular Value Decomposition of A to compute the Ridge coefficients.\n  In particular, it is more stable for singular matrices than `:cholesky` at the cost of being slower.\n\n  * `:cholesky` - Uses the standard `Nx.LinAlg.solve` function to obtain a closed-form solution.\n\n  The default value is `:svd`.\n\n* `:alpha` - Constant that multiplies the $L_2$ term, controlling regularization strength.\n  `:alpha` must be a non-negative float i.e. in [0, inf).\n\n  If `:alpha` is set to 0.0 the objective is the ordinary least squares regression.\n  In this case, for numerical reasons, you should use `Scholar.Linear.LinearRegression` instead.\n\n  The default value is `1.0`.","ref":"Scholar.Linear.RidgeRegression.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.RidgeRegression.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:coefficients` - Estimated coefficients for the linear regression problem.\n\n  * `:intercept` - Independent term in the linear model.","ref":"Scholar.Linear.RidgeRegression.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.RidgeRegression.fit/3","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> Scholar.Linear.RidgeRegression.fit(x, y)\n    %Scholar.Linear.RidgeRegression{\n      coefficients: Nx.tensor(\n        [-0.4237867593765259, -0.6891377568244934]\n      ),\n      intercept: Nx.tensor(\n        5.6569366455078125\n      )\n    }","ref":"Scholar.Linear.RidgeRegression.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.RidgeRegression.predict/2","doc":"Makes predictions with the given `model` on input `x`.","ref":"Scholar.Linear.RidgeRegression.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.RidgeRegression.predict/2","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([4.0, 3.0, -1.0])\n    iex> model = Scholar.Linear.RidgeRegression.fit(x, y)\n    iex> Scholar.Linear.RidgeRegression.predict(model, Nx.tensor([[2.0, 1.0]]))\n    Nx.tensor(\n      [4.120225429534912]\n    )","ref":"Scholar.Linear.RidgeRegression.html#predict/2-examples"},{"type":"module","title":"Scholar.Linear.SVM","doc":"Support Vector Machine linear classifier.\n\nIt uses the One-vs-Rest strategy to handle both binary and multinomial classification.\nThis implementation uses stochastic gradient descent from default or any other optimizer\navailable in `Polaris`. This makes it similar to a sklearn SGDClassifier [1].\n\nOn average it is slower than algorithms that use QP and kernel trick (LIBSVM [2]) or\nCoordinate Descent Algorithm (LIBLINEAR [3]). It also cannot use different kernels\nlike in LIBSVM, but you can use any type of optimizer available in `Polaris`.\n\nTime complexity is $O(N * K * I * C)$ where $N$ is the number of samples, $K$ is the\nnumber of features $I$ is the number of iterations and $C$ is the number of classes.\n\n[1] - https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDClassifier.html\n[2] - https://www.csie.ntu.edu.tw/~cjlin/libsvm/\n[3] - https://www.csie.ntu.edu.tw/~cjlin/liblinear/","ref":"Scholar.Linear.SVM.html"},{"type":"function","title":"Scholar.Linear.SVM.fit/3","doc":"Fits an SVM model for sample inputs `x` and sample\ntargets `y`.","ref":"Scholar.Linear.SVM.html#fit/3"},{"type":"function","title":"Options - Scholar.Linear.SVM.fit/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. number of classes contained in the input tensors.\n\n* `:iterations` (`t:pos_integer/0`) - number of iterations of gradient descent performed inside SVM. The default value is `1000`.\n\n* `:learning_loop_unroll` (`t:boolean/0`) - If `true`, the learning loop is unrolled. The default value is `false`.\n\n* `:optimizer` - The optimizer name or {init, update} pair of functions (see `Polaris.Optimizers` for more details). The default value is `:sgd`.\n\n* `:eps` (`t:float/0`) - The convergence tolerance. If the `abs(loss) < size(x) * :eps`, the algorithm is considered to have converged. The default value is `1.0e-8`.\n\n* `:loss_fn` - The loss function that is used in the algorithm. Functions should take two arguments: `y_predicted` and `y_true`.\n  If now provided it is set to highe loss without regularization. The default value is `nil`.","ref":"Scholar.Linear.SVM.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Linear.SVM.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:coefficients` - Coefficient of the features in the decision function.\n\n  * `:bias` - Bias added to the decision function.","ref":"Scholar.Linear.SVM.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Linear.SVM.fit/3","doc":"iex> x = Nx.tensor([[1.0, 2.0, 2.1], [3.0, 2.0, 1.4], [4.0, 7.0, 5.3], [3.0, 4.0, 6.3]])\n    iex> y = Nx.tensor([1, 0, 1, 1])\n    iex> Scholar.Linear.SVM.fit(x, y, num_classes: 2)\n    %Scholar.Linear.SVM{\n      coefficients: Nx.tensor(\n        [\n          [1.6899993419647217, 1.4599995613098145, 1.322001338005066],\n          [1.4799995422363281, 1.9599990844726562, 2.0080013275146484]\n        ]\n      ),\n      bias: Nx.tensor(\n        [0.23000003397464752, 0.4799998104572296]\n      )\n    }","ref":"Scholar.Linear.SVM.html#fit/3-examples"},{"type":"function","title":"Scholar.Linear.SVM.hinge_loss/3","doc":"","ref":"Scholar.Linear.SVM.html#hinge_loss/3"},{"type":"function","title":"Scholar.Linear.SVM.loss_function/1","doc":"","ref":"Scholar.Linear.SVM.html#loss_function/1"},{"type":"function","title":"Scholar.Linear.SVM.predict/2","doc":"Makes predictions with the given model on inputs `x`.","ref":"Scholar.Linear.SVM.html#predict/2"},{"type":"function","title":"Examples - Scholar.Linear.SVM.predict/2","doc":"iex> x = Nx.tensor([[1.0, 2.0], [3.0, 2.0], [4.0, 7.0]])\n    iex> y = Nx.tensor([1, 0, 1])\n    iex> model = Scholar.Linear.SVM.fit(x, y, num_classes: 2)\n    iex> Scholar.Linear.SVM.predict(model, Nx.tensor([[-3.0, 5.0]]))\n    #Nx.Tensor","ref":"Scholar.Linear.SVM.html#predict/2-examples"},{"type":"module","title":"Scholar.Manifold.MDS","doc":"Multidimensional scaling (MDS) seeks a low-dimensional representation of the data\nin which the distances respect well the distances in the original high-dimensional space.","ref":"Scholar.Manifold.MDS.html"},{"type":"function","title":"Scholar.Manifold.MDS.fit/1","doc":"Fits MDS for sample inputs `x`. It is simpyfied version of `fit/3` function.","ref":"Scholar.Manifold.MDS.html#fit/1"},{"type":"function","title":"Options - Scholar.Manifold.MDS.fit/1","doc":"* `:num_components` (`t:pos_integer/0`) - Dimension of the embedded space. The default value is `2`.\n\n* `:metric` (`t:boolean/0`) - If `true`, use dissimilarities as metric distances in the embedding space. The default value is `true`.\n\n* `:normalized_stress` (`t:boolean/0`) - If `true`, normalize the stress by the sum of squared dissimilarities.\n  Only valid if `metric` is `false`. The default value is `false`.\n\n* `:eps` (`t:float/0`) - Tolerance for stopping criterion. The default value is `0.001`.\n\n* `:max_iter` (`t:pos_integer/0`) - Maximum number of iterations for the optimization. The default value is `300`.\n\n* `:key` - Determines random number generation for centroid initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.\n\n* `:n_init` (`t:pos_integer/0`) - Number of times the embedding will be computed with different centroid seeds.\n  The final embedding is the embedding with the lowest stress. The default value is `8`.","ref":"Scholar.Manifold.MDS.html#fit/1-options"},{"type":"function","title":"Return Values - Scholar.Manifold.MDS.fit/1","doc":"Returns struct with embedded data, stress value, and number of iterations for best run.","ref":"Scholar.Manifold.MDS.html#fit/1-return-values"},{"type":"function","title":"Examples - Scholar.Manifold.MDS.fit/1","doc":"iex> x = Nx.iota({4,5})\n    iex> key = Nx.Random.key(42)\n    iex> Scholar.Manifold.MDS.fit(x, key: key)\n    %Scholar.Manifold.MDS{\n      embedding: Nx.tensor(\n        [\n          [13.072145462036133, -10.424199104309082],\n          [5.13038969039917, -2.341259479522705],\n          [-5.651908874511719, 1.7662434577941895],\n          [-12.550626754760742, 10.999215126037598]\n        ]\n      ),\n      stress: Nx.tensor(\n        0.36994707584381104\n      ),\n      n_iter: Nx.tensor(\n        20\n      )\n    }","ref":"Scholar.Manifold.MDS.html#fit/1-examples"},{"type":"function","title":"Scholar.Manifold.MDS.fit/2","doc":"Fits MDS for sample inputs `x`. It is simpyfied version of `fit/3` function.","ref":"Scholar.Manifold.MDS.html#fit/2"},{"type":"function","title":"Options - Scholar.Manifold.MDS.fit/2","doc":"* `:num_components` (`t:pos_integer/0`) - Dimension of the embedded space. The default value is `2`.\n\n* `:metric` (`t:boolean/0`) - If `true`, use dissimilarities as metric distances in the embedding space. The default value is `true`.\n\n* `:normalized_stress` (`t:boolean/0`) - If `true`, normalize the stress by the sum of squared dissimilarities.\n  Only valid if `metric` is `false`. The default value is `false`.\n\n* `:eps` (`t:float/0`) - Tolerance for stopping criterion. The default value is `0.001`.\n\n* `:max_iter` (`t:pos_integer/0`) - Maximum number of iterations for the optimization. The default value is `300`.\n\n* `:key` - Determines random number generation for centroid initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.\n\n* `:n_init` (`t:pos_integer/0`) - Number of times the embedding will be computed with different centroid seeds.\n  The final embedding is the embedding with the lowest stress. The default value is `8`.","ref":"Scholar.Manifold.MDS.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Manifold.MDS.fit/2","doc":"Returns struct with embedded data, stress value, and number of iterations for best run.","ref":"Scholar.Manifold.MDS.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Manifold.MDS.fit/2","doc":"iex> x = Nx.iota({4,5})\n    iex> init = Nx.reverse(Nx.iota({4,2}))\n    iex> Scholar.Manifold.MDS.fit(x, init)\n    %Scholar.Manifold.MDS{\n      embedding: Nx.tensor(\n        [\n          [11.858541488647461, 11.858541488647461],\n          [3.9528470039367676, 3.9528470039367676],\n          [-3.9528470039367676, -3.9528470039367676],\n          [-11.858541488647461, -11.858541488647461]\n        ]\n      ),\n      stress: Nx.tensor(\n        0.0\n      ),\n      n_iter: Nx.tensor(\n        3\n      )\n    }","ref":"Scholar.Manifold.MDS.html#fit/2-examples"},{"type":"function","title":"Scholar.Manifold.MDS.fit/3","doc":"Fits MDS for sample inputs `x`. It is simpyfied version of `fit/3` function.","ref":"Scholar.Manifold.MDS.html#fit/3"},{"type":"function","title":"Options - Scholar.Manifold.MDS.fit/3","doc":"* `:num_components` (`t:pos_integer/0`) - Dimension of the embedded space. The default value is `2`.\n\n* `:metric` (`t:boolean/0`) - If `true`, use dissimilarities as metric distances in the embedding space. The default value is `true`.\n\n* `:normalized_stress` (`t:boolean/0`) - If `true`, normalize the stress by the sum of squared dissimilarities.\n  Only valid if `metric` is `false`. The default value is `false`.\n\n* `:eps` (`t:float/0`) - Tolerance for stopping criterion. The default value is `0.001`.\n\n* `:max_iter` (`t:pos_integer/0`) - Maximum number of iterations for the optimization. The default value is `300`.\n\n* `:key` - Determines random number generation for centroid initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.\n\n* `:n_init` (`t:pos_integer/0`) - Number of times the embedding will be computed with different centroid seeds.\n  The final embedding is the embedding with the lowest stress. The default value is `8`.","ref":"Scholar.Manifold.MDS.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Manifold.MDS.fit/3","doc":"Returns struct with embedded data, stress value, and number of iterations for best run.","ref":"Scholar.Manifold.MDS.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Manifold.MDS.fit/3","doc":"iex> x = Nx.iota({4,5})\n    iex> key = Nx.Random.key(42)\n    iex> init = Nx.reverse(Nx.iota({4,3}))\n    iex> Scholar.Manifold.MDS.fit(x, init, num_components: 3, key: key)\n    %Scholar.Manifold.MDS{\n      embedding: Nx.tensor(\n        [\n          [9.682458877563477, 9.682458877563477, 9.682458877563477],\n          [3.2274858951568604, 3.2274858951568604, 3.2274858951568604],\n          [-3.2274863719940186, -3.2274863719940186, -3.2274863719940186],\n          [-9.682458877563477, -9.682458877563477, -9.682458877563477]\n        ]\n      ),\n      stress: Nx.tensor(\n        9.094947017729282e-12\n      ),\n      n_iter: Nx.tensor(\n        3\n      )\n    }","ref":"Scholar.Manifold.MDS.html#fit/3-examples"},{"type":"module","title":"Scholar.Manifold.TSNE","doc":"t-SNE (t-Distributed Stochastic Neighbor Embedding) is a nonlinear dimensionality reduction technique.\n\nThis is an exact implementation of t-SNE and therefore it has time complexity is $O(N^2)$ for $N$ samples.","ref":"Scholar.Manifold.TSNE.html"},{"type":"module","title":"Reference - Scholar.Manifold.TSNE","doc":"* [Van der Maaten, L., & Hinton, G. (2008). Visualizing data using t-SNE. Journal of machine learning research, 9(11).](http://www.jmlr.org/papers/volume9/vandermaaten08a/vandermaaten08a.pdf)","ref":"Scholar.Manifold.TSNE.html#module-reference"},{"type":"function","title":"Scholar.Manifold.TSNE.fit/2","doc":"Fits tSNE for sample inputs `x`.","ref":"Scholar.Manifold.TSNE.html#fit/2"},{"type":"function","title":"Options - Scholar.Manifold.TSNE.fit/2","doc":"* `:num_components` (`t:pos_integer/0`) - Dimension of the embedded space. The default value is `2`.\n\n* `:perplexity` (`t:pos_integer/0`) - The perplexity is related to the number of nearest neighbors that is used in other manifold learning algorithms.\n  Larger datasets usually require a larger perplexity.\n  Consider selecting a value between 5 and 50. The default value is `30`.\n\n* `:learning_rate` - The learning rate for t-SNE is usually in the range [10.0, 1000.0].\n  If the learning rate is too high, the data may look like a 'ball' with any point approximately equidistant from its nearest neighbors.\n  If the learning rate is too low, most points may look compressed in a dense cloud with few outliers.\n  If the cost function gets stuck in a bad local minimum increasing the learning rate may help. The default value is `500`.\n\n* `:num_iters` (`t:pos_integer/0`) - Maximum number of iterations for the optimization. The default value is `500`.\n\n* `:key` - Determines random number generation for centroid initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.\n\n* `:init` - Initialization of embedding.\n  Possible options are `:random` and `:pca`. The default value is `:pca`.\n\n* `:metric` - Metric used to compute the distances. The default value is `:squared_euclidean`.\n\n* `:exaggeration` - Controls how tight natural clusters in the original space are in the embedded space and\n  how much space will be between them. For larger values, the space between natural clusters will be larger in the embedded space. The default value is `10.0`.\n\n* `:learning_loop_unroll` (`t:boolean/0`) - If `true`, the learning loop is unrolled. The default value is `false`.","ref":"Scholar.Manifold.TSNE.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Manifold.TSNE.fit/2","doc":"Returns the embedded data `y`.","ref":"Scholar.Manifold.TSNE.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Manifold.TSNE.fit/2","doc":"iex> x = Nx.iota({4,5})\n    iex> key = Nx.Random.key(42)\n    iex> Scholar.Manifold.TSNE.fit(x, num_components: 2, key: key)\n    #Nx.Tensor","ref":"Scholar.Manifold.TSNE.html#fit/2-examples"},{"type":"function","title":"Scholar.Manifold.TSNE.pairwise_dist/2","doc":"","ref":"Scholar.Manifold.TSNE.html#pairwise_dist/2"},{"type":"module","title":"Scholar.Manifold.Trimap","doc":"TriMap: Large-scale Dimensionality Reduction Using Triplets.\n\nTriMap is a dimensionality reduction method that uses triplet constraints to form a low-dimensional embedding of a set of points.\nThe triplet constraints are of the form \"point i is closer to point j than point k\".\nThe triplets are sampled from the high-dimensional representation of the points and a weighting scheme is used to reflect the importance of each triplet.\n\nTriMap provides a significantly better global view of the data than the other dimensionality reduction methods such t-SNE, LargeVis.\n\nThe global structure includes relative distances of the clusters, multiple scales in the data, and the existence of possible outliers.\nWe define a global score to quantify the quality of an embedding in reflecting the global structure of the data.\n\nBased on: https://github.com/google-research/google-research/blob/master/trimap/trimap.py\nSource: https://arxiv.org/pdf/1910.00204.pdf","ref":"Scholar.Manifold.Trimap.html"},{"type":"function","title":"Scholar.Manifold.Trimap.transform/2","doc":"Embeds the given inputs using the TriMap algorithm.","ref":"Scholar.Manifold.Trimap.html#transform/2"},{"type":"function","title":"Options - Scholar.Manifold.Trimap.transform/2","doc":"* `:num_components` (`t:pos_integer/0`) - Dimension of the embedded space. The default value is `2`.\n\n* `:num_inliers` (`t:pos_integer/0`) - Number of inliners to sample. The default value is `10`.\n\n* `:num_outliers` (`t:pos_integer/0`) - Number of outliers to sample. The default value is `5`.\n\n* `:num_random` (`t:pos_integer/0`) - Number of random triplets to sample. The default value is `3`.\n\n* `:weight_temp` (`t:float/0`) - Temperature for the tempered log. The default value is `0.5`.\n\n* `:learning_rate` (`t:float/0`) - Learning rate. The default value is `0.1`.\n\n* `:num_iters` (`t:pos_integer/0`) - Number of iterations. The default value is `400`.\n\n* `:init_embedding_type` - Method to initialize the embedding. 0 for PCA, 1 for random. The default value is `0`.\n\n* `:key` - Random key used in Trimap.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.\n\n* `:triplets` (`t:term/0`) - Triplets to use in the Trimap algorithm.\n\n* `:weights` (`t:term/0`) - Weights to use in the Trimap algorithm.\n\n* `:init_embeddings` (`t:term/0`) - Initial embeddings to use in the Trimap algorithm.\n\n* `:metric` - Metric used to compute the distances. The default value is `:squared_euclidean`.\n\n* `:knn_algorithm` - Algorithm used to compute the nearest neighbors. Possible values:\n  * `:nndescent` - Nearest Neighbors Descent. See `Scholar.Neighbors.NNDescent` for more details.\n\n  * `:large_vis` - LargeVis algorithm. See `Scholar.Neighbors.LargeVis` for more details.\n\n  * `:brute` - Brute force algorithm. See `Scholar.Neighbors.BruteKNN` for more details.\n\n  * `:auto` - Automatically selects the algorithm based on the number of points.\n\n  The default value is `:auto`.","ref":"Scholar.Manifold.Trimap.html#transform/2-options"},{"type":"function","title":"Examples - Scholar.Manifold.Trimap.transform/2","doc":"iex> {inputs, key} = Nx.Random.uniform(Nx.Random.key(42), shape: {30, 5})\n    iex> Scholar.Manifold.Trimap.transform(inputs, num_components: 2, num_inliers: 3, num_outliers: 1, key: key, knn_algorithm: :nndescent)","ref":"Scholar.Manifold.Trimap.html#transform/2-examples"},{"type":"function","title":"Scholar.Manifold.Trimap.trimap_loss/1","doc":"","ref":"Scholar.Manifold.Trimap.html#trimap_loss/1"},{"type":"module","title":"Scholar.NaiveBayes.Complement","doc":"The Complement Naive Bayes classifier.\n\nIt was designed to correct the assumption of Multinomial Naive Bayes\nthat each class has roughly the same representation. It is particularly\nsuited for imbalanced data sets.\n\nTime complexity is $O(K * N * C)$ where $N$ is the number of samples and $K$ is the number of features,\nand $C$ is the number of classes.\n\nReference:\n\n* [1] - [Paper about Complement Naive Bayes Algorithm](https://cdn.aaai.org/ICML/2003/ICML03-081.pdf)","ref":"Scholar.NaiveBayes.Complement.html"},{"type":"function","title":"Scholar.NaiveBayes.Complement.fit/3","doc":"The multinomial Naive Bayes classifier is suitable for classification with\ndiscrete features (e.g., word counts for text classification)","ref":"Scholar.NaiveBayes.Complement.html#fit/3"},{"type":"function","title":"Options - Scholar.NaiveBayes.Complement.fit/3","doc":"* `:alpha` - Additive (Laplace/Lidstone) smoothing parameter\n  (set alpha to 0.0 and force_alpha to true, for no smoothing). The default value is `1.0`.\n\n* `:force_alpha` (`t:boolean/0`) - If False and alpha is less than 1e-10, it will set alpha to\n  1e-10. If True, alpha will remain unchanged. This may cause\n  numerical errors if alpha is too close to 0. The default value is `true`.\n\n* `:fit_priors` (`t:boolean/0`) - Whether to learn class prior probabilities or not.\n  If false, a uniform prior will be used. The default value is `true`.\n\n* `:priors` - Prior probabilities of the classes. If specified, the priors are not\n  adjusted according to the data.\n\n* `:num_classes` (`t:pos_integer/0`) - Required. Number of different classes used in training.\n\n* `:sample_weights` - List of `n_samples` elements.\n  A list of 1.0 values is used if none is given.\n\n* `:norm` (`t:boolean/0`) - Whether or not a second normalization of the weights is performed. The default value is `false`.","ref":"Scholar.NaiveBayes.Complement.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.NaiveBayes.Complement.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:feature_log_probability` - Empirical log probability of features\n      given a class, ``P(x_i|y)``.\n\n  * `:class_count` - Number of samples encountered for each class during fitting. This\n      value is weighted by the sample weight when provided.\n\n  * `:class_log_priors` - Smoothed empirical log probability for each class.\n\n  * `:classes` - class labels known to the classifier.\n\n  * `:feature_count` - Number of samples encountered for each (class, feature)\n      during fitting. This value is weighted by the sample weight when\n      provided.\n\n  * `:feature_all` - Number of samples encountered for each feature during fitting.\n      This value is weighted by the `sample_weights` when provided.","ref":"Scholar.NaiveBayes.Complement.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Complement.fit/3","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> Scholar.NaiveBayes.Complement.fit(x, y, num_classes: 3)\n    %Scholar.NaiveBayes.Complement{\n      feature_log_probability: Nx.tensor(\n        [\n          [1.3062516450881958, 1.0986123085021973, 0.9267619848251343],\n          [1.2452157735824585, 1.0986123085021973, 0.9707789421081543],\n          [1.3499267101287842, 1.0986123085021973, 0.8979415893554688]\n        ]\n      ),\n      class_count: Nx.tensor([1.0, 1.0, 2.0]),\n      class_log_priors: Nx.tensor([-1.3862943649291992, -1.3862943649291992, -0.6931471824645996]),\n      classes: Nx.tensor([0, 1, 2]),\n      feature_count: Nx.tensor(\n        [\n          [6.0, 7.0, 8.0],\n          [0.0, 1.0, 2.0],\n          [12.0, 14.0, 16.0]\n        ]\n      ),\n      feature_all: Nx.tensor([18.0, 22.0, 26.0])\n    }\n    iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> Scholar.NaiveBayes.Complement.fit(x, y, num_classes: 3, sample_weights: [1, 6, 2, 3])\n    %Scholar.NaiveBayes.Complement{\n      feature_log_probability: Nx.tensor(\n        [\n          [1.2953225374221802, 1.0986123085021973, 0.9343092441558838],\n          [1.2722758054733276, 1.0986123085021973, 0.9506921768188477],\n          [1.3062516450881958, 1.0986123085021973, 0.9267619848251343]\n        ]\n      ),\n      class_count: Nx.tensor([2.0, 1.0, 9.0]),\n      class_log_priors: Nx.tensor([-1.7917594909667969, -2.4849066734313965, -0.28768205642700195]),\n      classes: Nx.tensor([0, 1, 2]),\n      feature_count: Nx.tensor(\n        [\n          [12.0, 14.0, 16.0],\n          [0.0, 1.0, 2.0],\n          [45.0, 54.0, 63.0]\n        ]\n      ),\n      feature_all: Nx.tensor([57.0, 69.0, 81.0])\n    }","ref":"Scholar.NaiveBayes.Complement.html#fit/3-examples"},{"type":"function","title":"Scholar.NaiveBayes.Complement.predict/2","doc":"Perform classification on an array of test vectors `x` using `model`.","ref":"Scholar.NaiveBayes.Complement.html#predict/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Complement.predict/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Complement.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Complement.predict(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Complement.html#predict/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Complement.predict_joint_log_probability/2","doc":"Return joint log probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Complement.html#predict_joint_log_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Complement.predict_joint_log_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Complement.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Complement.predict_joint_log_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Complement.html#predict_joint_log_probability/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Complement.predict_log_probability/2","doc":"Return log-probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Complement.html#predict_log_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Complement.predict_log_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Complement.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Complement.predict_log_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Complement.html#predict_log_probability/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Complement.predict_probability/2","doc":"Return probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Complement.html#predict_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Complement.predict_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Complement.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Complement.predict_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Complement.html#predict_probability/2-examples"},{"type":"module","title":"Scholar.NaiveBayes.Gaussian","doc":"Gaussian Naive Bayes algorithm for classification.\n\nThe likelihood of the features is assumed to be Gaussian:\n$$ P(x\\_{i} | y) = \\frac{1}{\\sqrt{2\\pi\\sigma\\_{y}^{2}}} \\exp \\left(-\\frac{(x\\_{i} - \\mu\\_{y})^2}{2\\sigma\\_{y}^{2}}\\right) $$\n\nThe parameters $\\sigma\\_{y}$ and $\\mu\\_{y}$ are estimated using maximum likelihood.\n\nTime complexity is $O(K * N * C)$ where $N$ is the number of samples and $K$ is the number of features,\nand $C$ is the number of classes.\n\nReference:\n\n* [1] - [Detailed explanation of algorithm used to update feature means and variance online by Chan, Golub, and LaVeque](http://i.stanford.edu/pub/cstr/reports/cs/tr/79/773/CS-TR-79-773.pdf)","ref":"Scholar.NaiveBayes.Gaussian.html"},{"type":"function","title":"Scholar.NaiveBayes.Gaussian.fit/3","doc":"Gaussian Naive Bayes.","ref":"Scholar.NaiveBayes.Gaussian.html#fit/3"},{"type":"function","title":"Options - Scholar.NaiveBayes.Gaussian.fit/3","doc":"* `:var_smoothing` (`t:float/0`) - Portion of the largest variance of all features that is added to\n  variances for calculation stability. The default value is `1.0e-9`.\n\n* `:priors` - Prior probabilities of the classes. If specified, the priors are not\n  adjusted according to the data. We assume that priors are correct and\n  sum(priors) == 1.\n\n* `:sample_weights` - List of `n_samples` elements.\n\n  A list of 1.0 values is used if none is given.\n\n* `:num_classes` (`t:pos_integer/0`) - Required. Number of different classes used in training.","ref":"Scholar.NaiveBayes.Gaussian.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.NaiveBayes.Gaussian.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:theta` - mean of each feature per class.\n\n  * `:var` - Variance of each feature per class.\n\n  * `:class_count` - number of training samples observed in each class.\n\n  * `:class_priors` - probability of each class.\n\n  * `:classes` - class labels known to the classifier.\n\n  * `:epsilon` - absolute additive value to variances.","ref":"Scholar.NaiveBayes.Gaussian.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Gaussian.fit/3","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> Scholar.NaiveBayes.Gaussian.fit(x, y, num_classes: 3)\n    %Scholar.NaiveBayes.Gaussian{\n      theta: Nx.tensor(\n        [\n          [6.0, 7.0, 8.0],\n          [0.0, 1.0, 2.0],\n          [6.0, 7.0, 8.0]\n        ]\n      ),\n      var: Nx.tensor(\n        [\n          [1.1250000042650754e-8, 1.1250000042650754e-8, 1.1250000042650754e-8],\n          [1.1250000042650754e-8, 1.1250000042650754e-8, 1.1250000042650754e-8],\n          [9.0, 9.0, 9.0]\n        ]\n      ),\n      class_count: Nx.tensor([1.0, 1.0, 2.0]),\n      class_priors: Nx.tensor([0.25, 0.25, 0.5]),\n      classes: Nx.tensor([0, 1, 2]),\n      epsilon: Nx.tensor(1.1250000042650754e-8)\n    }\n\n    iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> Scholar.NaiveBayes.Gaussian.fit(x, y, num_classes: 3, sample_weights: [1, 6, 2, 3])\n    %Scholar.NaiveBayes.Gaussian{\n      theta: Nx.tensor(\n        [\n          [6.0, 7.0, 8.0],\n          [0.0, 1.0, 2.0],\n          [5.0, 6.0, 7.0]\n        ]\n      ),\n      var: Nx.tensor(\n        [\n          [1.1250000042650754e-8, 1.1250000042650754e-8, 1.1250000042650754e-8],\n          [1.1250000042650754e-8, 1.1250000042650754e-8, 1.1250000042650754e-8],\n          [8.0, 8.0, 8.0]\n        ]\n        ),\n      class_count: Nx.tensor([2.0, 1.0, 9.0]),\n      class_priors: Nx.tensor([0.1666666716337204, 0.0833333358168602, 0.75]),\n      classes: Nx.tensor([0, 1, 2]),\n      epsilon: Nx.tensor(1.1250000042650754e-8)\n    }","ref":"Scholar.NaiveBayes.Gaussian.html#fit/3-examples"},{"type":"function","title":"Scholar.NaiveBayes.Gaussian.predict/2","doc":"Perform classification on an array of test vectors `x` using `model`.","ref":"Scholar.NaiveBayes.Gaussian.html#predict/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Gaussian.predict/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Gaussian.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Gaussian.predict(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Gaussian.html#predict/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Gaussian.predict_joint_log_probability/2","doc":"Return joint log probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Gaussian.html#predict_joint_log_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Gaussian.predict_joint_log_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Gaussian.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Gaussian.predict_joint_log_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Gaussian.html#predict_joint_log_probability/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Gaussian.predict_log_probability/2","doc":"Return log-probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Gaussian.html#predict_log_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Gaussian.predict_log_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Gaussian.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Gaussian.predict_log_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Gaussian.html#predict_log_probability/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Gaussian.predict_probability/2","doc":"Return probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Gaussian.html#predict_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Gaussian.predict_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Gaussian.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Gaussian.predict_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Gaussian.html#predict_probability/2-examples"},{"type":"module","title":"Scholar.NaiveBayes.Multinomial","doc":"Naive Bayes classifier for multinomial models.\n\nThe multinomial Naive Bayes classifier is suitable for classification with\ndiscrete features (e.g., word counts for text classification)\n\nTime complexity is $O(K * N * C)$ where $N$ is the number of samples and $K$ is the number of features,\nand $C$ is the number of classes.","ref":"Scholar.NaiveBayes.Multinomial.html"},{"type":"function","title":"Scholar.NaiveBayes.Multinomial.fit/3","doc":"Fits a naive Bayes model. The function assumes that targets `y` are integers\nbetween 0 and `num_classes` - 1 (inclusive). Otherwise, those samples will not\ncontribute to `class_count`.","ref":"Scholar.NaiveBayes.Multinomial.html#fit/3"},{"type":"function","title":"Options - Scholar.NaiveBayes.Multinomial.fit/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of different classes used in training.\n\n* `:alpha` - Additive (Laplace/Lidstone) smoothing parameter\n  (set alpha to 0.0 and force_alpha to true, for no smoothing). The default value is `1.0`.\n\n* `:force_alpha` (`t:boolean/0`) - If `false` and alpha is less than 1e-10, it will set alpha to\n  1e-10. If `true`, alpha will remain unchanged. This may cause\n  numerical errors if alpha is too close to 0. The default value is `true`.\n\n* `:fit_priors` (`t:boolean/0`) - Whether to learn class prior probabilities or not.\n  If `false`, a uniform prior will be used. The default value is `true`.\n\n* `:class_priors` - Prior probabilities of the classes. If specified, the priors are not\n  adjusted according to the data.\n\n* `:sample_weights` - List of `num_samples` elements.\n  A list of 1.0 values is used if none is given.","ref":"Scholar.NaiveBayes.Multinomial.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.NaiveBayes.Multinomial.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:class_count` - Number of samples encountered for each class during fitting. This\n      value is weighted by the sample weight when provided.\n\n  * `:class_log_priors` - Smoothed empirical log probability for each class.\n\n  * `:feature_count` - Number of samples encountered for each (class, feature)\n      during fitting. This value is weighted by the sample weight when\n      provided.\n\n  * `:feature_log_probability` - Empirical log probability of features\n      given a class, ``P(x_i|y)``.","ref":"Scholar.NaiveBayes.Multinomial.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Multinomial.fit/3","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> Scholar.NaiveBayes.Multinomial.fit(x, y, num_classes: 3)\n    %Scholar.NaiveBayes.Multinomial{\n      feature_count: Nx.tensor(\n        [\n          [6.0, 7.0, 8.0],\n          [0.0, 1.0, 2.0],\n          [12.0, 14.0, 16.0]\n        ]\n      ),\n      class_count: Nx.tensor(\n        [1.0, 1.0, 2.0]\n      ),\n      class_log_priors: Nx.tensor(\n        [-1.3862943649291992, -1.3862943649291992, -0.6931471824645996]\n      ),\n      feature_log_probability: Nx.tensor(\n        [\n          [-1.232143759727478, -1.0986123085021973, -0.9808292388916016],\n          [-1.7917594909667969, -1.0986123085021973, -0.6931471824645996],\n          [-1.241713285446167, -1.0986123085021973, -0.9734492301940918]\n        ]\n      )\n    }\n\n    iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> Scholar.NaiveBayes.Multinomial.fit(x, y, num_classes: 3, sample_weights: [1, 6, 2, 3])\n    %Scholar.NaiveBayes.Multinomial{\n      feature_count: Nx.tensor(\n        [\n          [12.0, 14.0, 16.0],\n          [0.0, 1.0, 2.0],\n          [45.0, 54.0, 63.0]\n        ]\n      ),\n      class_count: Nx.tensor(\n        [2.0, 1.0, 9.0]\n      ),\n      class_log_priors: Nx.tensor(\n        [-1.7917594909667969, -2.4849066734313965, -0.28768205642700195]\n      ),\n      feature_log_probability: Nx.tensor(\n        [\n          [-1.241713285446167, -1.0986123085021973, -0.9734492301940918],\n          [-1.7917594909667969, -1.0986123085021973, -0.6931471824645996],\n          [-1.2773041725158691, -1.0986123085021973, -0.9470624923706055]\n        ]\n      )\n    }","ref":"Scholar.NaiveBayes.Multinomial.html#fit/3-examples"},{"type":"function","title":"Scholar.NaiveBayes.Multinomial.predict/2","doc":"Perform classification on an array of test vectors `x` using `model`.","ref":"Scholar.NaiveBayes.Multinomial.html#predict/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Multinomial.predict/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Multinomial.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Multinomial.predict(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Multinomial.html#predict/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Multinomial.predict_joint_log_probability/2","doc":"Return joint log probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Multinomial.html#predict_joint_log_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Multinomial.predict_joint_log_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Multinomial.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Multinomial.predict_joint_log_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Multinomial.html#predict_joint_log_probability/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Multinomial.predict_log_probability/2","doc":"Return log-probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Multinomial.html#predict_log_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Multinomial.predict_log_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Multinomial.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Multinomial.predict_log_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Multinomial.html#predict_log_probability/2-examples"},{"type":"function","title":"Scholar.NaiveBayes.Multinomial.predict_probability/2","doc":"Return probability estimates for the test vector `x` using `model`.","ref":"Scholar.NaiveBayes.Multinomial.html#predict_probability/2"},{"type":"function","title":"Examples - Scholar.NaiveBayes.Multinomial.predict_probability/2","doc":"iex> x = Nx.iota({4, 3})\n    iex> y = Nx.tensor([1, 2, 0, 2])\n    iex> model = Scholar.NaiveBayes.Multinomial.fit(x, y, num_classes: 3)\n    iex> Scholar.NaiveBayes.Multinomial.predict_probability(model, Nx.tensor([[6, 2, 4], [8, 5, 9]]))\n    #Nx.Tensor","ref":"Scholar.NaiveBayes.Multinomial.html#predict_probability/2-examples"},{"type":"module","title":"Scholar.Neighbors.BruteKNN","doc":"Brute-Force k-Nearest Neighbor Search Algorithm.\n\nIn order to find the k-nearest neighbors the algorithm calculates\nthe distance between the query point and each of the data samples.\nTherefore, its time complexity is $O(MN)$ for $N$ samples and $M$ query points.\nIt uses $O(BN)$ memory for batch size $B$.\nLarger batch sizes will lead to faster predictions, but will consume more memory.","ref":"Scholar.Neighbors.BruteKNN.html"},{"type":"function","title":"Scholar.Neighbors.BruteKNN.fit/2","doc":"Fits a brute-force k-NN model.","ref":"Scholar.Neighbors.BruteKNN.html#fit/2"},{"type":"function","title":"Options - Scholar.Neighbors.BruteKNN.fit/2","doc":"* `:num_neighbors` (`t:pos_integer/0`) - Required. The number of nearest neighbors.\n\n* `:metric` - The function that measures the pairwise distance between two points. Possible values:\n\n  * `{:minkowski, p}` - Minkowski metric. By changing value of `p` parameter (a positive number or `:infinity`)\n  we can set Manhattan (`1`), Euclidean (`2`), Chebyshev (`:infinity`), or any arbitrary $L_p$ metric.\n\n  * `:cosine` - Cosine metric.\n\n  * `:euclidean` - Euclidean metric.\n\n  * `:squared_euclidean` - Squared Euclidean metric.\n\n  * `:manhattan` - Manhattan metric.\n\n  * Anonymous function of arity 2 that takes two rank-2 tensors.\n\n  The default value is `&Scholar.Metrics.Distance.pairwise_minkowski/2`.\n\n* `:batch_size` (`t:pos_integer/0`) - The number of samples in a batch.","ref":"Scholar.Neighbors.BruteKNN.html#fit/2-options"},{"type":"function","title":"Examples - Scholar.Neighbors.BruteKNN.fit/2","doc":"iex> data = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> model = Scholar.Neighbors.BruteKNN.fit(data, num_neighbors: 2)\n    iex> model.num_neighbors\n    2\n    iex> model.data\n    #Nx.Tensor","ref":"Scholar.Neighbors.BruteKNN.html#fit/2-examples"},{"type":"function","title":"Scholar.Neighbors.BruteKNN.get_batches/2","doc":"","ref":"Scholar.Neighbors.BruteKNN.html#get_batches/2"},{"type":"function","title":"Scholar.Neighbors.BruteKNN.predict/2","doc":"Computes nearest neighbors of query tensor using brute-force search.\nReturns the neighbors indices and distances from query points.","ref":"Scholar.Neighbors.BruteKNN.html#predict/2"},{"type":"function","title":"Examples - Scholar.Neighbors.BruteKNN.predict/2","doc":"iex> data = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> model = Scholar.Neighbors.BruteKNN.fit(data, num_neighbors: 2)\n    iex> query = Nx.tensor([[1, 3], [4, 2], [3, 6]])\n    iex> {neighbors, distances} = Scholar.Neighbors.BruteKNN.predict(model, query)\n    iex> neighbors\n    #Nx.Tensor \n    iex> distances\n    #Nx.Tensor","ref":"Scholar.Neighbors.BruteKNN.html#predict/2-examples"},{"type":"function","title":"Scholar.Neighbors.BruteKNN.predict_n/2","doc":"","ref":"Scholar.Neighbors.BruteKNN.html#predict_n/2"},{"type":"module","title":"Scholar.Neighbors.KDTree","doc":"Implements a k-d tree, a space-partitioning data structure for organizing points\nin a k-dimensional space.\n\nIt can be used to predict the K-Nearest Neighbors of a given input.\n\nThis is implemented as one-dimensional tensor with indices pointed to highest\ndimension of the given tensor. Traversal starts by calling `root/0` and then\naccessing the `left_child/1` and `right_child/1`. The tree is left-balanced.\n\nEach level traverses over the last axis of tensor, the index for a level can be\ncomputed as: `rem(level, Nx.axis_size(tensor, -1))`.","ref":"Scholar.Neighbors.KDTree.html"},{"type":"module","title":"References - Scholar.Neighbors.KDTree","doc":"* [GPU-friendly, Parallel, and (Almost-)In-Place Construction of Left-Balanced k-d Trees](https://arxiv.org/pdf/2211.00120.pdf).","ref":"Scholar.Neighbors.KDTree.html#module-references"},{"type":"function","title":"Scholar.Neighbors.KDTree.fit/2","doc":"Builds a KDTree.","ref":"Scholar.Neighbors.KDTree.html#fit/2"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.fit/2","doc":"iex> tree = Scholar.Neighbors.KDTree.fit(Nx.iota({5, 2}))\n    iex> tree.data\n    Nx.tensor(\n      [\n        [0, 1],\n        [2, 3],\n        [4, 5],\n        [6, 7],\n        [8, 9]\n      ]\n    )\n    iex> tree.levels\n    3\n    iex> tree.indices\n    Nx.u32([3, 1, 4, 0, 2])","ref":"Scholar.Neighbors.KDTree.html#fit/2-examples"},{"type":"function","title":"Scholar.Neighbors.KDTree.left_child/1","doc":"Returns the index of the left child of i.\n\nIt is your responsibility to guarantee the result\nis not greater than the leading axis of the tensor.","ref":"Scholar.Neighbors.KDTree.html#left_child/1"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.left_child/1","doc":"iex> Scholar.Neighbors.KDTree.left_child(0)\n    1\n    iex> Scholar.Neighbors.KDTree.left_child(1)\n    3\n\n    iex> Scholar.Neighbors.KDTree.left_child(Nx.u32(3))\n    #Nx.Tensor","ref":"Scholar.Neighbors.KDTree.html#left_child/1-examples"},{"type":"function","title":"Scholar.Neighbors.KDTree.levels/1","doc":"Returns the number of resulting levels in a KDTree for `tensor`.","ref":"Scholar.Neighbors.KDTree.html#levels/1"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.levels/1","doc":"iex> Scholar.Neighbors.KDTree.levels(Nx.iota({10, 3}))\n    4","ref":"Scholar.Neighbors.KDTree.html#levels/1-examples"},{"type":"function","title":"Scholar.Neighbors.KDTree.parent/1","doc":"Returns the parent of child `i`.\n\nIt is your responsibility to guarantee the result is positive.","ref":"Scholar.Neighbors.KDTree.html#parent/1"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.parent/1","doc":"iex> Scholar.Neighbors.KDTree.parent(1)\n    0\n    iex> Scholar.Neighbors.KDTree.parent(2)\n    0\n\n    iex> Scholar.Neighbors.KDTree.parent(Nx.u32(3))\n    #Nx.Tensor","ref":"Scholar.Neighbors.KDTree.html#parent/1-examples"},{"type":"function","title":"Scholar.Neighbors.KDTree.predict/2","doc":"Predict the K nearest neighbors of `x_predict` in KDTree.","ref":"Scholar.Neighbors.KDTree.html#predict/2"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.predict/2","doc":"iex> x = Nx.iota({10, 2})\n    iex> x_predict = Nx.tensor([[2, 5], [1, 9], [6, 4]])\n    iex> kdtree = Scholar.Neighbors.KDTree.fit(x, num_neighbors: 3)\n    iex> {indices, distances} = Scholar.Neighbors.KDTree.predict(kdtree, x_predict)\n    iex> indices\n    #Nx.Tensor \n    iex> distances\n    #Nx.Tensor \n\n    iex> x = Nx.iota({10, 2})\n    iex> x_predict = Nx.tensor([[2, 5], [1, 9], [6, 4]])\n    iex> kdtree = Scholar.Neighbors.KDTree.fit(x, num_neighbors: 3, metric: {:minkowski, 1})\n    iex> {indices, distances} = Scholar.Neighbors.KDTree.predict(kdtree, x_predict)\n    iex> indices\n    #Nx.Tensor \n    iex> distances\n    #Nx.Tensor","ref":"Scholar.Neighbors.KDTree.html#predict/2-examples"},{"type":"function","title":"Scholar.Neighbors.KDTree.right_child/1","doc":"Returns the index of the right child of i.\n\nIt is your responsibility to guarantee the result\nis not greater than the leading axis of the tensor.","ref":"Scholar.Neighbors.KDTree.html#right_child/1"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.right_child/1","doc":"iex> Scholar.Neighbors.KDTree.right_child(0)\n    2\n    iex> Scholar.Neighbors.KDTree.right_child(1)\n    4\n\n    iex> Scholar.Neighbors.KDTree.right_child(Nx.u32(3))\n    #Nx.Tensor","ref":"Scholar.Neighbors.KDTree.html#right_child/1-examples"},{"type":"function","title":"Scholar.Neighbors.KDTree.root/0","doc":"Returns the root index.","ref":"Scholar.Neighbors.KDTree.html#root/0"},{"type":"function","title":"Examples - Scholar.Neighbors.KDTree.root/0","doc":"iex> Scholar.Neighbors.KDTree.root()\n    0","ref":"Scholar.Neighbors.KDTree.html#root/0-examples"},{"type":"module","title":"Scholar.Neighbors.KNNClassifier","doc":"K-Nearest Neighbors Classifier.\n\nPerforms classifiction by computing the (weighted) majority voting among k-nearest neighbors.","ref":"Scholar.Neighbors.KNNClassifier.html"},{"type":"function","title":"Scholar.Neighbors.KNNClassifier.fit/3","doc":"Fits a k-NN classifier model.","ref":"Scholar.Neighbors.KNNClassifier.html#fit/3"},{"type":"function","title":"Options - Scholar.Neighbors.KNNClassifier.fit/3","doc":"* `:algorithm` (`t:atom/0`) - Algorithm used to compute the k-nearest neighbors. Possible values:\n\n    * `:brute` - Brute-force search. See `Scholar.Neighbors.BruteKNN` for more details.\n\n    * `:kd_tree` - k-d tree. See `Scholar.Neighbors.KDTree` for more details.\n\n    * `:random_projection_forest` - Random projection forest. See `Scholar.Neighbors.RandomProjectionForest` for more details.\n\n    * Module implementing `fit(data, opts)` and `predict(model, query)`. predict/2 must return tuple containing indices\n    of k-nearest neighbors of query points as well as distances between query points and their k-nearest neighbors.\n\n  The default value is `:brute`.\n\n* `:num_classes` (`t:pos_integer/0`) - Required. The number of possible classes.\n\n* `:weights` - Weight function used in prediction. Possible values:\n\n    * `:uniform` - uniform weights. All points in each neighborhood are weighted equally.\n\n    * `:distance` - weight points by the inverse of their distance. in this case, closer neighbors of\n      a query point will have a greater influence than neighbors which are further away.\n\n  The default value is `:uniform`.\n\n\n\nAlgorithm-specific options (e.g. `:num_neighbors`, `:metric`) should be provided together with the classifier options.","ref":"Scholar.Neighbors.KNNClassifier.html#fit/3-options"},{"type":"function","title":"Examples - Scholar.Neighbors.KNNClassifier.fit/3","doc":"iex> x = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y = Nx.tensor([0, 0, 0, 1, 1])\n    iex> model = Scholar.Neighbors.KNNClassifier.fit(x, y, num_neighbors: 3, num_classes: 2)\n    iex> model.algorithm\n    Scholar.Neighbors.BruteKNN.fit(x, num_neighbors: 3)\n    iex> model.labels\n    Nx.tensor([0, 0, 0, 1, 1])\n\n    iex> x = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y = Nx.tensor([0, 0, 0, 1, 1])\n    iex> model = Scholar.Neighbors.KNNClassifier.fit(x, y, algorithm: :kd_tree, num_neighbors: 3, metric: {:minkowski, 1}, num_classes: 2)\n    iex> model.algorithm\n    Scholar.Neighbors.KDTree.fit(x, num_neighbors: 3, metric: {:minkowski, 1})\n    iex> model.labels\n    Nx.tensor([0, 0, 0, 1, 1])\n\n    iex> x = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y = Nx.tensor([0, 0, 0, 1, 1])\n    iex> key = Nx.Random.key(12)\n    iex> model = Scholar.Neighbors.KNNClassifier.fit(x, y, algorithm: :random_projection_forest, num_neighbors: 2, num_classes: 2, num_trees: 4, key: key)\n    iex> model.algorithm\n    Scholar.Neighbors.RandomProjectionForest.fit(x, num_neighbors: 2, num_trees: 4, key: key)\n    iex> model.labels\n    Nx.tensor([0, 0, 0, 1, 1])","ref":"Scholar.Neighbors.KNNClassifier.html#fit/3-examples"},{"type":"function","title":"Scholar.Neighbors.KNNClassifier.predict/2","doc":"Predicts classes using a k-NN classifier model.","ref":"Scholar.Neighbors.KNNClassifier.html#predict/2"},{"type":"function","title":"Examples - Scholar.Neighbors.KNNClassifier.predict/2","doc":"iex> x_train = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y_train = Nx.tensor([0, 0, 0, 1, 1])\n    iex> model = Scholar.Neighbors.KNNClassifier.fit(x_train, y_train, num_neighbors: 3, num_classes: 2)\n    iex> x = Nx.tensor([[1, 3], [4, 2], [3, 6]])\n    iex> Scholar.Neighbors.KNNClassifier.predict(model, x)\n    Nx.tensor([0, 0, 1])","ref":"Scholar.Neighbors.KNNClassifier.html#predict/2-examples"},{"type":"function","title":"Scholar.Neighbors.KNNClassifier.predict_probability/2","doc":"Predicts class probabilities using a k-NN classifier model.","ref":"Scholar.Neighbors.KNNClassifier.html#predict_probability/2"},{"type":"function","title":"Examples - Scholar.Neighbors.KNNClassifier.predict_probability/2","doc":"iex> x_train = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y_train = Nx.tensor([0, 0, 0, 1, 1])\n    iex> model = Scholar.Neighbors.KNNClassifier.fit(x_train, y_train, num_neighbors: 3, num_classes: 2)\n    iex> x = Nx.tensor([[1, 3], [4, 2], [3, 6]])\n    iex> Scholar.Neighbors.KNNClassifier.predict_probability(model, x)\n    Nx.tensor(\n      [\n        [1.0, 0.0],\n        [1.0, 0.0],\n        [0.3333333432674408, 0.6666666865348816]\n      ]\n    )","ref":"Scholar.Neighbors.KNNClassifier.html#predict_probability/2-examples"},{"type":"module","title":"Scholar.Neighbors.KNNRegressor","doc":"K-Nearest Neighbors Regressor.\n\nPerforms regression by computing the (weighted) mean of k-nearest neighbor labels.","ref":"Scholar.Neighbors.KNNRegressor.html"},{"type":"function","title":"Scholar.Neighbors.KNNRegressor.fit/3","doc":"Fits a k-NN regressor model.","ref":"Scholar.Neighbors.KNNRegressor.html#fit/3"},{"type":"function","title":"Options - Scholar.Neighbors.KNNRegressor.fit/3","doc":"* `:algorithm` (`t:atom/0`) - Algorithm used to compute the k-nearest neighbors. Possible values:\n\n    * `:brute` - Brute-force search. See `Scholar.Neighbors.BruteKNN` for more details.\n\n    * `:kd_tree` - k-d tree. See `Scholar.Neighbors.KDTree` for more details.\n\n    * `:random_projection_forest` - Random projection forest. See `Scholar.Neighbors.RandomProjectionForest` for more details.\n\n    * Module implementing `fit(data, opts)` and `predict(model, query)`. predict/2 must return tuple containing indices\n    of k-nearest neighbors of query points as well as distances between query points and their k-nearest neighbors.\n\n  The default value is `:brute`.\n\n* `:weights` - Weight function used in prediction. Possible values:\n\n    * `:uniform` - uniform weights. All points in each neighborhood are weighted equally.\n\n    * `:distance` - weight points by the inverse of their distance. in this case, closer neighbors of\n      a query point will have a greater influence than neighbors which are further away.\n\n  The default value is `:uniform`.\n\n\n\nAlgorithm-specific options (e.g. `:num_neighbors`, `:metric`) should be provided together with the regressor options.","ref":"Scholar.Neighbors.KNNRegressor.html#fit/3-options"},{"type":"function","title":"Examples - Scholar.Neighbors.KNNRegressor.fit/3","doc":"iex> x = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y = Nx.tensor([[1], [2], [3], [4], [5]])\n    iex> model = Scholar.Neighbors.KNNRegressor.fit(x, y, num_neighbors: 3)\n    iex> model.algorithm\n    Scholar.Neighbors.BruteKNN.fit(x, num_neighbors: 3)\n    iex> model.labels\n    Nx.tensor([[1], [2], [3], [4], [5]])\n\n    iex> x = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y = Nx.tensor([[1], [2], [3], [4], [5]])\n    iex> model = Scholar.Neighbors.KNNRegressor.fit(x, y, algorithm: :kd_tree, num_neighbors: 3, metric: {:minkowski, 1})\n    iex> model.algorithm\n    Scholar.Neighbors.KDTree.fit(x, num_neighbors: 3, metric: {:minkowski, 1})\n    iex> model.labels\n    Nx.tensor([[1], [2], [3], [4], [5]])\n\n    iex> x = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y = Nx.tensor([[1], [2], [3], [4], [5]])\n    iex> key = Nx.Random.key(12)\n    iex> model = Scholar.Neighbors.KNNRegressor.fit(x, y, algorithm: :random_projection_forest, num_neighbors: 2, num_trees: 4, key: key)\n    iex> model.algorithm\n    Scholar.Neighbors.RandomProjectionForest.fit(x, num_neighbors: 2, num_trees: 4, key: key)\n    iex> model.labels\n    Nx.tensor([[1], [2], [3], [4], [5]])","ref":"Scholar.Neighbors.KNNRegressor.html#fit/3-examples"},{"type":"function","title":"Scholar.Neighbors.KNNRegressor.predict/2","doc":"Predicts labels using a k-NN regressor model.","ref":"Scholar.Neighbors.KNNRegressor.html#predict/2"},{"type":"function","title":"Examples - Scholar.Neighbors.KNNRegressor.predict/2","doc":"iex> x_train = Nx.tensor([[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])\n    iex> y_train = Nx.tensor([[1], [2], [3], [4], [5]])\n    iex> model = Scholar.Neighbors.KNNRegressor.fit(x_train, y_train, num_neighbors: 3)\n    iex> x = Nx.tensor([[1, 3], [4, 2], [3, 6]])\n    iex> Scholar.Neighbors.KNNRegressor.predict(model, x)\n    Nx.tensor([[2.0], [2.0], [4.0]])","ref":"Scholar.Neighbors.KNNRegressor.html#predict/2-examples"},{"type":"module","title":"Scholar.Neighbors.LargeVis","doc":"LargeVis algorithm for approximate k-nearest neighbor (k-NN) graph construction.\n\nThe algorithms works in the following way. First, the approximate k-NN graph is constructed\nusing a random projection forest. Then, the graph is refined by looking at the neighbors of\nneighbors of every point for a fixed number of iterations. This step is called NN-expansion.","ref":"Scholar.Neighbors.LargeVis.html"},{"type":"module","title":"References - Scholar.Neighbors.LargeVis","doc":"* [Visualizing Large-scale and High-dimensional Data](https://arxiv.org/abs/1602.00370).","ref":"Scholar.Neighbors.LargeVis.html#module-references"},{"type":"function","title":"Scholar.Neighbors.LargeVis.expand/3","doc":"","ref":"Scholar.Neighbors.LargeVis.html#expand/3"},{"type":"function","title":"Scholar.Neighbors.LargeVis.fit/2","doc":"Constructs the approximate k-NN graph with LargeVis.\n\nReturns neighbor indices and distances.","ref":"Scholar.Neighbors.LargeVis.html#fit/2"},{"type":"function","title":"Examples - Scholar.Neighbors.LargeVis.fit/2","doc":"iex> key = Nx.Random.key(12)\n    iex> tensor = Nx.iota({5, 2})\n    iex> {graph, distances} = Scholar.Neighbors.LargeVis.fit(tensor, num_neighbors: 2, metric: :squared_euclidean, min_leaf_size: 2, num_trees: 3, key: key)\n    iex> graph\n    #Nx.Tensor \n    iex> distances\n    #Nx.Tensor","ref":"Scholar.Neighbors.LargeVis.html#fit/2-examples"},{"type":"module","title":"Scholar.Neighbors.NNDescent","doc":"Nearest Neighbors Descent (NND) is an algorithm that calculates Approximated Nearest Neighbors (ANN)\nfor a given set of points[1].\n\nIt is implemented using Random Projection Forest[2].","ref":"Scholar.Neighbors.NNDescent.html"},{"type":"module","title":"References - Scholar.Neighbors.NNDescent","doc":"[1] [Efficient K-Nearest Neighbor Graph Construction for Generic Similarity Measures](https://www.cs.princeton.edu/cass/papers/www11.pdf).\n  [2] [Randomized partition trees for nearest neighbor search](https://cseweb.ucsd.edu/~dasgupta/papers/exactnn-algo.pdf)","ref":"Scholar.Neighbors.NNDescent.html#module-references"},{"type":"function","title":"Scholar.Neighbors.NNDescent.fit/2","doc":"Calculates the approximate nearest neighbors for a given set of points. It\nreturns a struct containing two tensors:\n\n* `nearest_neighbors` - the indices of the nearest neighbors\n* `distances` - the distances to the nearest neighbors","ref":"Scholar.Neighbors.NNDescent.html#fit/2"},{"type":"function","title":"Examples - Scholar.Neighbors.NNDescent.fit/2","doc":"iex> data = Nx.iota({10, 5})\n    iex> key = Nx.Random.key(12)\n    iex> Scholar.Neighbors.NNDescent.fit(data, num_neighbors: 3, key: key)","ref":"Scholar.Neighbors.NNDescent.html#fit/2-examples"},{"type":"function","title":"Scholar.Neighbors.NNDescent.heap_size/1","doc":"","ref":"Scholar.Neighbors.NNDescent.html#heap_size/1"},{"type":"function","title":"Scholar.Neighbors.NNDescent.in1d/2","doc":"","ref":"Scholar.Neighbors.NNDescent.html#in1d/2"},{"type":"function","title":"Scholar.Neighbors.NNDescent.min_heap_insert/3","doc":"","ref":"Scholar.Neighbors.NNDescent.html#min_heap_insert/3"},{"type":"function","title":"Scholar.Neighbors.NNDescent.min_heap_pop/1","doc":"","ref":"Scholar.Neighbors.NNDescent.html#min_heap_pop/1"},{"type":"function","title":"Scholar.Neighbors.NNDescent.query/3","doc":"Query the training data for the nearest neighbors of the query data.\n\nUse this function if you query for the first time on a training data.\nIt will compute search graph that can be reused in the future queries.","ref":"Scholar.Neighbors.NNDescent.html#query/3"},{"type":"function","title":"Examples - Scholar.Neighbors.NNDescent.query/3","doc":"iex> data = Nx.iota({10, 5})\n    iex> query_data = Nx.tensor([[1,7,32,6,2], [1,4,5,2,5], [1,3,67,12,4]])\n    iex> key = Nx.Random.key(12)\n    iex> nn = Scholar.Neighbors.NNDescent.fit(data, num_neighbors: 3)\n    iex> Scholar.Neighbors.NNDescent.query(nn, query_data, num_neighbors: 5, key: key)","ref":"Scholar.Neighbors.NNDescent.html#query/3-examples"},{"type":"function","title":"Scholar.Neighbors.NNDescent.query/4","doc":"Query the training data for the nearest neighbors of the query data.\n\nUse this function if you have already queried training data.\nIt will compute search graph that can be reused in the future queries.","ref":"Scholar.Neighbors.NNDescent.html#query/4"},{"type":"function","title":"Examples - Scholar.Neighbors.NNDescent.query/4","doc":"iex> data = Nx.iota({10, 5})\n    iex> query_data = Nx.tensor([[1,7,32,6,2], [1,4,5,2,5], [1,3,67,12,4]])\n    iex> key = Nx.Random.key(12)\n    iex> nn = Scholar.Neighbors.NNDescent.fit(data, num_neighbors: 3)\n    iex> {_indices, _distances, search_graph} = Scholar.Neighbors.NNDescent.query(nn, query_data, num_neighbors: 5, key: key)\n    iex> query_data2 = Nx.tensor([[2,7,32,6,2], [1,4,6,2,5], [1,3,67,12,3]])\n    iex> Scholar.Neighbors.NNDescent.query(nn, query_data2, search_graph, num_neighbors: 5, key: key)","ref":"Scholar.Neighbors.NNDescent.html#query/4-examples"},{"type":"function","title":"Scholar.Neighbors.NNDescent.query_n/7","doc":"","ref":"Scholar.Neighbors.NNDescent.html#query_n/7"},{"type":"module","title":"Scholar.Neighbors.RadiusNearestNeighbors","doc":"The Radius Nearest Neighbors.\n\nIt implements both classification and regression.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html"},{"type":"function","title":"Scholar.Neighbors.RadiusNearestNeighbors.fit/3","doc":"Fit the Radius nearest neighbors classifier from the training data set.\n\nFor classification, provided labels need to be consecutive non-negative integers. If your labels does\nnot meet this condition please use `Scholar.Preprocessing.ordinal_encode`\n\nCurrently 2D labels are only supported for regression tasks.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#fit/3"},{"type":"function","title":"Options - Scholar.Neighbors.RadiusNearestNeighbors.fit/3","doc":"* `:radius` - Radius of neighborhood The default value is `1.0`.\n\n* `:num_classes` (`t:pos_integer/0`) - Number of classes in provided labels\n\n* `:weights` - Weight function used in prediction. Possible values:\n\n    * `:uniform` - uniform weights. All points in each neighborhood are weighted equally.\n\n    * `:distance` - weight points by the inverse of their distance. in this case, closer neighbors of\n      a query point will have a greater influence than neighbors which are further away.\n\n  The default value is `:uniform`.\n\n* `:metric` - The function that measures the pairwise distance between two points. Possible values:\n\n  * `{:minkowski, p}` - Minkowski metric. By changing value of `p` parameter (a positive number or `:infinity`)\n  we can set Manhattan (`1`), Euclidean (`2`), Chebyshev (`:infinity`), or any arbitrary $L_p$ metric.\n\n  * `:cosine` - Cosine metric.\n\n  * Anonymous function of arity 2 that takes two rank-2 tensors.\n\n  The default value is `&Scholar.Metrics.Distance.pairwise_minkowski/2`.\n\n* `:task` - Task that will be performed using Radius Nearest Neighbors. Possible values:\n\n  * `:classification` - Classifier implementing the Radius Nearest Neighbors vote.\n\n  * `:regression` - Regression based on Radius Nearest Neighbors.\n    The target is predicted by local interpolation of the targets associated of the nearest neighbors in the training set.\n\n  The default value is `:classification`.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#fit/3-options"},{"type":"function","title":"Return Values - Scholar.Neighbors.RadiusNearestNeighbors.fit/3","doc":"The function returns a struct with the following parameters:\n\n  * `:data` - Training data.\n\n  * `:labels` - Labels of each point.\n\n  * `:weights` - Weight function used in prediction.\n\n  * `:num_classes` - Number of classes in provided labels.\n\n  * `:task` - Task that will be performed using Radius Nearest Neighbors.\n  For `:classification` task, model will be a classifier implementing the Radius Nearest Neighbors vote.\n  For `:regression` task, model is a regressor based on Radius Nearest Neighbors.\n\n  * `:metric` - The metric function used.\n\n  * `:radius` - Radius of neighborhood.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#fit/3-return-values"},{"type":"function","title":"Examples - Scholar.Neighbors.RadiusNearestNeighbors.fit/3","doc":"iex> x = Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]])\n    iex> y = Nx.tensor([1, 0, 1, 1])\n    iex> Scholar.Neighbors.RadiusNearestNeighbors.fit(x, y, num_classes: 2)\n    %Scholar.Neighbors.RadiusNearestNeighbors{\n      data: Nx.tensor(\n        [\n          [1, 2],\n          [2, 4],\n          [1, 3],\n          [2, 5]\n        ]\n      ),\n      labels: Nx.tensor(\n        [1, 0, 1, 1]\n      ),\n      weights: :uniform,\n      num_classes: 2,\n      task: :classification,\n      metric: &Scholar.Metrics.Distance.pairwise_minkowski/2,\n      radius: 1.0\n    }","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#fit/3-examples"},{"type":"function","title":"Scholar.Neighbors.RadiusNearestNeighbors.predict/2","doc":"Makes predictions with the given `model` on inputs `x`.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#predict/2"},{"type":"function","title":"Return Values - Scholar.Neighbors.RadiusNearestNeighbors.predict/2","doc":"It returns a tensor with predicted class labels.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#predict/2-return-values"},{"type":"function","title":"Examples - Scholar.Neighbors.RadiusNearestNeighbors.predict/2","doc":"iex> x = Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]])\n    iex> y = Nx.tensor([1, 0, 1, 1])\n    iex> model = Scholar.Neighbors.RadiusNearestNeighbors.fit(x, y, num_classes: 2)\n    iex> Scholar.Neighbors.RadiusNearestNeighbors.predict(model, Nx.tensor([[1.9, 4.3], [1.1, 2.0]]))\n    Nx.tensor(\n      [0, 1]\n    )","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#predict/2-examples"},{"type":"function","title":"Scholar.Neighbors.RadiusNearestNeighbors.predict_probability/2","doc":"Return probability estimates for the test data `x`.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#predict_probability/2"},{"type":"function","title":"Return Values - Scholar.Neighbors.RadiusNearestNeighbors.predict_probability/2","doc":"It returns a typle with tensor with probabilities of classes and mask of outliers.\n  They are arranged in lexicographic order.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#predict_probability/2-return-values"},{"type":"function","title":"Examples - Scholar.Neighbors.RadiusNearestNeighbors.predict_probability/2","doc":"iex> x = Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]])\n    iex> y = Nx.tensor([1, 0, 1, 1])\n    iex> model = Scholar.Neighbors.RadiusNearestNeighbors.fit(x, y, num_classes: 2)\n    iex> Scholar.Neighbors.RadiusNearestNeighbors.predict_probability(model, Nx.tensor([[1.9, 4.3], [1.1, 2.0]]))\n    {Nx.tensor(\n      [\n        [0.5, 0.5],\n        [0.0, 1.0]\n      ]\n    ),\n    Nx.tensor(\n      [0, 0], type: :u8\n    )}","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#predict_probability/2-examples"},{"type":"function","title":"Scholar.Neighbors.RadiusNearestNeighbors.radius_neighbors/2","doc":"Find the Radius neighbors of a point.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#radius_neighbors/2"},{"type":"function","title":"Return Values - Scholar.Neighbors.RadiusNearestNeighbors.radius_neighbors/2","doc":"Returns indices of the selected neighbor points as a mask  (1 if a point is a neighbor, 0 otherwise) and their respective distances.","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#radius_neighbors/2-return-values"},{"type":"function","title":"Examples - Scholar.Neighbors.RadiusNearestNeighbors.radius_neighbors/2","doc":"iex> x = Nx.tensor([[1, 2], [2, 4], [1, 3], [2, 5]])\n    iex> y = Nx.tensor([1, 0, 1, 1])\n    iex> model = Scholar.Neighbors.RadiusNearestNeighbors.fit(x, y, num_classes: 2)\n    iex> Scholar.Neighbors.RadiusNearestNeighbors.radius_neighbors(model, Nx.tensor([[1.9, 4.3], [1.1, 2.0]]))\n    {Nx.tensor(\n      [\n        [2.469818353652954, 0.3162313997745514, 1.5811394453048706, 0.7071067690849304],\n        [0.10000114142894745, 2.1931710243225098, 1.0049877166748047, 3.132091760635376]\n      ]\n    ),\n    Nx.tensor(\n      [\n        [0, 1, 0, 1],\n        [1, 0, 0, 0]\n      ], type: :u8\n    )}","ref":"Scholar.Neighbors.RadiusNearestNeighbors.html#radius_neighbors/2-examples"},{"type":"module","title":"Scholar.Neighbors.RandomProjectionForest","doc":"Random Projection Forest for k-Nearest Neighbor Search.\n\nEach tree in a forest is constructed using a divide and conquer approach.\nWe start with the entire dataset and at every node we project the data onto a random\nhyperplane and split it in the following way: the points with the projection smaller\nthan or equal to the median are put into the left subtree and the points with projection\ngreater than the median are put into the right subtree. We then proceed\nrecursively with the left and right subtree.\n\nIn this implementation the trees are complete, i.e. there are 2^l nodes at level l.\nThe leaves of the trees are arranged as blocks in the field `indices`. We use the same\nhyperplane for all nodes on the same level as in [2].\n\n* [1] - Randomized partition trees for nearest neighbor search\n* [2] - Fast Nearest Neighbor Search through Sparse Random Projections and Voting","ref":"Scholar.Neighbors.RandomProjectionForest.html"},{"type":"function","title":"Scholar.Neighbors.RandomProjectionForest.fit/2","doc":"Grows a random projection forest.","ref":"Scholar.Neighbors.RandomProjectionForest.html#fit/2"},{"type":"function","title":"Options - Scholar.Neighbors.RandomProjectionForest.fit/2","doc":"* `:num_neighbors` (`t:pos_integer/0`) - Required. The number of nearest neighbors.\n\n* `:metric` - The function that measures the distance between two points. The default value is `:euclidean`.\n\n* `:min_leaf_size` (`t:pos_integer/0`) - The minumum number of points in the leaf.\n\n* `:num_trees` (`t:pos_integer/0`) - Required. The number of trees in the forest.\n\n* `:key` - Used for random number generation in hyperplane initialization.\n  If the key is not provided, it is set to `Nx.Random.key(System.system_time())`.","ref":"Scholar.Neighbors.RandomProjectionForest.html#fit/2-options"},{"type":"function","title":"Examples - Scholar.Neighbors.RandomProjectionForest.fit/2","doc":"iex> key = Nx.Random.key(12)\n    iex> tensor = Nx.iota({5, 2})\n    iex> forest = Scholar.Neighbors.RandomProjectionForest.fit(tensor, num_neighbors: 2, num_trees: 3, key: key)\n    iex> forest.indices\n    #Nx.Tensor","ref":"Scholar.Neighbors.RandomProjectionForest.html#fit/2-examples"},{"type":"function","title":"Scholar.Neighbors.RandomProjectionForest.fit_n/3","doc":"","ref":"Scholar.Neighbors.RandomProjectionForest.html#fit_n/3"},{"type":"function","title":"Scholar.Neighbors.RandomProjectionForest.predict/2","doc":"Computes approximate nearest neighbors of query tensor using random projection forest.\nReturns the neighbor indices and distances from query points.","ref":"Scholar.Neighbors.RandomProjectionForest.html#predict/2"},{"type":"function","title":"Examples - Scholar.Neighbors.RandomProjectionForest.predict/2","doc":"iex> key = Nx.Random.key(12)\n    iex> tensor = Nx.iota({5, 2})\n    iex> forest = Scholar.Neighbors.RandomProjectionForest.fit(tensor, num_neighbors: 2, metric: :squared_euclidean, num_trees: 3, key: key)\n    iex> query = Nx.tensor([[3, 4]])\n    iex> {neighbors, distances} = Scholar.Neighbors.RandomProjectionForest.predict(forest, query)\n    iex> neighbors\n    #Nx.Tensor \n    iex> distances\n    #Nx.Tensor","ref":"Scholar.Neighbors.RandomProjectionForest.html#predict/2-examples"},{"type":"module","title":"Scholar.Impute.SimpleImputer","doc":"Univariate imputer for completing missing values with simple strategies.","ref":"Scholar.Impute.SimpleImputer.html"},{"type":"function","title":"Scholar.Impute.SimpleImputer.fit/2","doc":"Univariate imputer for completing missing values with simple strategies.","ref":"Scholar.Impute.SimpleImputer.html#fit/2"},{"type":"function","title":"Options - Scholar.Impute.SimpleImputer.fit/2","doc":"* `:missing_values` - The placeholder for the missing values. All occurrences of `:missing_values` will be imputed. The default value is `:nan`.\n\n* `:strategy` - The imputation strategy.\n\n  * `:mean` - replace missing values using the mean along each column.\n\n  * `:median` - replace missing values using the median along each column.\n\n  * `:mode` - replace missing using the most frequent value along each column.\n    If there is more than one such value, only the smallest is returned.\n\n  * `:constant` - replace missing values with `:fill_value`.\n\n  The default value is `:mean`.\n\n* `:fill_value` - When strategy is set to `:constant`, `:fill_value` is used to replace all occurrences of `:missing_values`. The default value is `0.0`.","ref":"Scholar.Impute.SimpleImputer.html#fit/2-options"},{"type":"function","title":"Return Values - Scholar.Impute.SimpleImputer.fit/2","doc":"The function returns a struct with the following parameters:\n\n  * `:missing_values` - the same value as in `:missing_values`\n\n  * `:statistics` - The imputation fill value for each feature. Computing statistics can result in\n  [`Nx.Constant.nan/0`](https://hexdocs.pm/nx/Nx.Constants.html#nan/0) values.","ref":"Scholar.Impute.SimpleImputer.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Impute.SimpleImputer.fit/2","doc":"iex> x = Nx.tensor([[1, 2, :nan], [3, 7, :nan], [:nan, 4, 5]])\n    iex> Scholar.Impute.SimpleImputer.fit(x, strategy: :mean)\n    %Scholar.Impute.SimpleImputer{\n      statistics: Nx.tensor(\n        [2.0, 4.333333492279053, 5.0]\n      ),\n      missing_values: :nan\n    }","ref":"Scholar.Impute.SimpleImputer.html#fit/2-examples"},{"type":"function","title":"Scholar.Impute.SimpleImputer.transform/2","doc":"Impute all missing values in `x` using fitted imputer.","ref":"Scholar.Impute.SimpleImputer.html#transform/2"},{"type":"function","title":"Return Values - Scholar.Impute.SimpleImputer.transform/2","doc":"The function returns input tensor with NaN replaced with values saved in fitted imputer.","ref":"Scholar.Impute.SimpleImputer.html#transform/2-return-values"},{"type":"function","title":"Examples - Scholar.Impute.SimpleImputer.transform/2","doc":"iex> x = Nx.tensor([[1, 2, :nan], [3, 7, :nan], [:nan, 4, 5]])\n    iex> imputer = Scholar.Impute.SimpleImputer.fit(x, strategy: :mean)\n    iex> Scholar.Impute.SimpleImputer.transform(imputer, x)\n    Nx.tensor(\n      [\n        [1.0, 2.0, 5.0],\n        [3.0, 7.0, 5.0],\n        [2.0, 4.0, 5.0]\n      ]\n    )\n\n    iex> x = Nx.tensor([[1, 2, :nan], [3, 7, :nan], [:nan, 4, 5]])\n    iex> y = Nx.tensor([[7, :nan, 6], [6, 9, :nan], [8, :nan, 1]])\n    iex> imputer = Scholar.Impute.SimpleImputer.fit(x, strategy: :median)\n    iex> Scholar.Impute.SimpleImputer.transform(imputer, y)\n    Nx.tensor(\n      [\n        [7.0, 4.0, 6.0],\n        [6.0, 9.0, 5.0],\n        [8.0, 4.0, 1.0]\n      ]\n    )","ref":"Scholar.Impute.SimpleImputer.html#transform/2-examples"},{"type":"module","title":"Scholar.Metrics.Classification","doc":"Classification Metric functions.\n\nMetrics are used to measure the performance and compare\nthe performance of any kind of classifier in\neasy-to-understand terms.\n\nAll of the functions in this module are implemented as\nnumerical functions and can be JIT or AOT compiled with\nany supported `Nx` compiler.","ref":"Scholar.Metrics.Classification.html"},{"type":"function","title":"Scholar.Metrics.Classification.accuracy/3","doc":"Computes the accuracy of the given predictions\nfor binary and multi-class classification problems.","ref":"Scholar.Metrics.Classification.html#accuracy/3"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.accuracy/3","doc":"iex> Scholar.Metrics.Classification.accuracy(Nx.tensor([1, 0, 0]), Nx.tensor([1, 0, 1]))\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.accuracy(y_true, y_pred)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.accuracy(y_true, y_pred, normalize: false)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#accuracy/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.auc/2","doc":"Computes area under the curve (AUC) using the trapezoidal rule.\n\nThis is a general function, given points on a curve.","ref":"Scholar.Metrics.Classification.html#auc/2"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.auc/2","doc":"iex> y = Nx.tensor([0, 0, 1, 1])\n    iex> pred = Nx.tensor([0.1, 0.4, 0.35, 0.8])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(pred)\n    iex> {fpr, tpr, _thresholds} = Scholar.Metrics.Classification.roc_curve(y, pred, distinct_value_indices)\n    iex> Scholar.Metrics.Classification.auc(fpr, tpr)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#auc/2-examples"},{"type":"function","title":"Scholar.Metrics.Classification.average_precision_score/4","doc":"Compute average precision (AP) from prediction scores.\n\nAP summarizes a precision-recall curve as the weighted mean of precisions achieved at\neach threshold, with the increase in recall from the previous threshold used as the weight:\n$$ AP = sum_n (R_n - R_{n-1}) P_n $$\n\nwhere $ P_n $ and $ R_n $ are the precision and recall at the nth threshold.","ref":"Scholar.Metrics.Classification.html#average_precision_score/4"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.average_precision_score/4","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1])\n    iex> scores = Nx.tensor([0.1, 0.4, 0.35, 0.8])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(scores)\n    iex> weights = Nx.tensor([1, 1, 2, 2])\n    iex> ap = Scholar.Metrics.Classification.average_precision_score(y_true, scores, distinct_value_indices, weights)\n    iex> ap\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#average_precision_score/4-examples"},{"type":"function","title":"Scholar.Metrics.Classification.balanced_accuracy_score/3","doc":"Computes the balanced accuracy score for multi-class classification","ref":"Scholar.Metrics.Classification.html#balanced_accuracy_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.balanced_accuracy_score/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:sample_weights` - Sample weights of the observations.\n\n* `:adjusted` (`t:boolean/0`) - If `true`, the balanced accuracy is adjusted for chance\n  (depends on the number of classes). The default value is `false`.","ref":"Scholar.Metrics.Classification.html#balanced_accuracy_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.balanced_accuracy_score/3","doc":"iex> y_true = Nx.tensor([0, 1, 2, 0, 1, 2], type: {:u, 32})\n    iex> y_pred = Nx.tensor([0, 2, 1, 0, 0, 1], type: {:u, 32})\n    iex> Scholar.Metrics.Classification.balanced_accuracy_score(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([0, 1, 2, 0, 1, 2], type: {:u, 32})\n    iex> y_pred = Nx.tensor([0, 2, 1, 0, 0, 1], type: {:u, 32})\n    iex> sample_weights = [1, 1, 1, 2, 2, 2]\n    iex> Scholar.Metrics.Classification.balanced_accuracy_score(y_true, y_pred, num_classes: 3, sample_weights: sample_weights, adjusted: true)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#balanced_accuracy_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.binary_precision/2","doc":"Computes the precision of the given predictions with respect to\nthe given targets for binary classification problems.\n\nIf the sum of true positives and false positives is 0, then the\nresult is 0 to avoid zero division.","ref":"Scholar.Metrics.Classification.html#binary_precision/2"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.binary_precision/2","doc":"iex> Scholar.Metrics.Classification.binary_precision(Nx.tensor([0, 1, 1, 1]), Nx.tensor([1, 0, 1, 1]))\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#binary_precision/2-examples"},{"type":"function","title":"Scholar.Metrics.Classification.binary_recall/2","doc":"Computes the recall of the given predictions with respect to\nthe given targets for binary classification problems.\n\nIf the sum of true positives and false negatives is 0, then the\nresult is 0 to avoid zero division.","ref":"Scholar.Metrics.Classification.html#binary_recall/2"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.binary_recall/2","doc":"iex> Scholar.Metrics.Classification.binary_recall(Nx.tensor([0, 1, 1, 1]), Nx.tensor([1, 0, 1, 1]))\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#binary_recall/2-examples"},{"type":"function","title":"Scholar.Metrics.Classification.binary_sensitivity/2","doc":"Computes the sensitivity of the given predictions with respect\nto the given targets for binary classification problems.","ref":"Scholar.Metrics.Classification.html#binary_sensitivity/2"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.binary_sensitivity/2","doc":"iex> Scholar.Metrics.Classification.binary_sensitivity(Nx.tensor([0, 1, 1, 1]), Nx.tensor([1, 0, 1, 1]))\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#binary_sensitivity/2-examples"},{"type":"function","title":"Scholar.Metrics.Classification.binary_specificity/2","doc":"Computes the specificity of the given predictions with respect\nto the given targets for binary classification problems.\n\nIf the sum of true negatives and false positives is 0, then the\nresult is 0 to avoid zero division.","ref":"Scholar.Metrics.Classification.html#binary_specificity/2"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.binary_specificity/2","doc":"iex> Scholar.Metrics.Classification.binary_specificity(Nx.tensor([0, 1, 1, 1]), Nx.tensor([1, 0, 1, 1]))\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#binary_specificity/2-examples"},{"type":"function","title":"Scholar.Metrics.Classification.brier_score_loss/3","doc":"Compute the Brier score loss.\n\nThe smaller the Brier score loss, the better, hence the naming with \"loss\".\nThe Brier score measures the mean squared difference between the predicted\nprobability and the actual outcome. The Brier score always\ntakes on a value between zero and one, since this is the largest\npossible difference between a predicted probability (which must be\nbetween zero and one) and the actual outcome (which can take on values\nof only 0 and 1). It can be decomposed as the sum of refinement loss and\ncalibration loss. If predicted probabilities are not in the interval\n[0, 1], they will be clipped.\n\nThe Brier score is appropriate only for binary outcomes.","ref":"Scholar.Metrics.Classification.html#brier_score_loss/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.brier_score_loss/3","doc":"* `:sample_weights` - Sample weights of the observations. The default value is `1.0`.\n\n* `:pos_label` (`t:integer/0`) - Label of the positive class. The default value is `1`.","ref":"Scholar.Metrics.Classification.html#brier_score_loss/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.brier_score_loss/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 0])\n    iex> y_prob = Nx.tensor([0.1, 0.9, 0.8, 0.3])\n    iex> Scholar.Metrics.Classification.brier_score_loss(y_true, y_prob)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#brier_score_loss/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.cohen_kappa_score/3","doc":"Compute Cohen's kappa: a statistic that measures inter-annotator agreement.","ref":"Scholar.Metrics.Classification.html#cohen_kappa_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.cohen_kappa_score/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:weights` - Weighting to calculate the score.\n\n* `:weighting_type` - Weighting type to calculate the score.","ref":"Scholar.Metrics.Classification.html#cohen_kappa_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.cohen_kappa_score/3","doc":"iex> y1 = Nx.tensor([0, 1, 1, 0, 1, 2])\n    iex> y2 = Nx.tensor([0, 2, 1, 0, 0, 1])\n    iex> Scholar.Metrics.Classification.cohen_kappa_score(y1, y2, num_classes: 3)\n    #Nx.Tensor \n\n    iex> y1 = Nx.tensor([0, 1, 1, 0, 1, 2])\n    iex> y2 = Nx.tensor([0, 2, 1, 0, 0, 1])\n    iex> Scholar.Metrics.Classification.cohen_kappa_score(y1, y2, num_classes: 3, weighting_type: :linear)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#cohen_kappa_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.confusion_matrix/3","doc":"Calculates the confusion matrix given rank-1 tensors which represent\nthe expected (`y_true`) and predicted (`y_pred`) classes.","ref":"Scholar.Metrics.Classification.html#confusion_matrix/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.confusion_matrix/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors","ref":"Scholar.Metrics.Classification.html#confusion_matrix/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.confusion_matrix/3","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1, 2, 2], type: :u32)\n    iex> y_pred = Nx.tensor([0, 1, 0, 2, 2, 2], type: :u32)\n    iex> Scholar.Metrics.Classification.confusion_matrix(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([0, 0, 1, 1, 2, 2], type: {:u, 32})\n    iex> y_pred = Nx.tensor([0, 1, 0, 2, 2, 2], type: {:u, 32})\n    iex> sample_weights = [2, 5, 1, 1.5, 2, 8]\n    iex> Scholar.Metrics.Classification.confusion_matrix(y_true, y_pred, num_classes: 3, sample_weights: sample_weights, normalize: :predicted)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#confusion_matrix/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.det_curve/4","doc":"Compute error rates for different probability thresholds (DET).\n\nNote: This metric is used for evaluation of ranking and error tradeoffs of\na binary classification task.","ref":"Scholar.Metrics.Classification.html#det_curve/4"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.det_curve/4","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1])\n    iex> y_score = Nx.tensor([0.1, 0.4, 0.35, 0.8])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(y_score)\n    iex> {fpr, fnr, thresholds} = Scholar.Metrics.Classification.det_curve(y_true, y_score, distinct_value_indices)\n    iex> fpr\n    #Nx.Tensor \n    iex> fnr\n    #Nx.Tensor \n    iex> thresholds\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1])\n    iex> y_score = Nx.tensor([0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4, 0.4, 0.5, 0.5, 0.6, 0.7, 0.7, 0.8, 0.9])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(y_score)\n    iex> {fpr, fnr, thresholds} = Scholar.Metrics.Classification.det_curve(y_true, y_score, distinct_value_indices)\n    iex> fpr\n    #Nx.Tensor \n    iex> fnr\n    #Nx.Tensor \n    iex> thresholds\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#det_curve/4-examples"},{"type":"function","title":"Scholar.Metrics.Classification.distinct_value_indices/1","doc":"It's a helper function for `Scholar.Metrics.Classification.roc_curve` and `Scholar.Metrics.Classification.roc_auc_score` functions.\nYou should call it and use as follows:\n\n    distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(scores)\n    {fpr, tpr, thresholds} = Scholar.Metrics.Classification.roc_curve(y_true, scores, distinct_value_indices, weights)","ref":"Scholar.Metrics.Classification.html#distinct_value_indices/1"},{"type":"function","title":"Scholar.Metrics.Classification.f1_score/3","doc":"Calculates F1 score given rank-1 tensors which represent\nthe expected (`y_true`) and predicted (`y_pred`) classes.\n\nIf all examples are true negatives, then the result is 0 to\navoid zero division.","ref":"Scholar.Metrics.Classification.html#f1_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.f1_score/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:average` - This determines the type of averaging performed on the data.\n\n  * `:macro` - Calculate metrics for each label, and find their unweighted mean.\n  This does not take label imbalance into account.\n\n  * `:weighted` - Calculate metrics for each label, and find their average weighted by\n  support (the number of true instances for each label).\n\n  * `:micro` - Calculate metrics globally by counting the total true positives,\n  false negatives and false positives.\n\n  * `:none` - The F-score values for each class are returned.\n\n  The default value is `:none`.","ref":"Scholar.Metrics.Classification.html#f1_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.f1_score/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.f1_score(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.f1_score(y_true, y_pred, num_classes: 3, average: :macro)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.f1_score(y_true, y_pred, num_classes: 3, average: :weighted)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.f1_score(y_true, y_pred, num_classes: 3, average: :micro)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.f1_score(Nx.tensor([1, 0, 1, 0]), Nx.tensor([0, 1, 0, 1]), num_classes: 2, average: :none)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#f1_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.fbeta_score/4","doc":"Calculates F-beta score given rank-1 tensors which represent\nthe expected (`y_true`) and predicted (`y_pred`) classes.\n\nIf all examples are true negatives, then the result is 0 to\navoid zero division.\n\n$$F_\\beta = (1 + \\beta^2) \\cdot \\frac{\\mathrm{precision} \\cdot \\mathrm{recall}}{(\\beta^2 \\cdot \\mathrm{precision}) + \\mathrm{recall}}$$","ref":"Scholar.Metrics.Classification.html#fbeta_score/4"},{"type":"function","title":"Options - Scholar.Metrics.Classification.fbeta_score/4","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:average` - This determines the type of averaging performed on the data.\n\n  * `:macro` - Calculate metrics for each label, and find their unweighted mean.\n  This does not take label imbalance into account.\n\n  * `:weighted` - Calculate metrics for each label, and find their average weighted by\n  support (the number of true instances for each label).\n\n  * `:micro` - Calculate metrics globally by counting the total true positives,\n  false negatives and false positives.\n\n  * `:none` - The F-score values for each class are returned.\n\n  The default value is `:none`.","ref":"Scholar.Metrics.Classification.html#fbeta_score/4-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.fbeta_score/4","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.fbeta_score(y_true, y_pred, Nx.u32(1), num_classes: 3)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(y_true, y_pred, Nx.u32(2), num_classes: 3)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(y_true, y_pred, Nx.f32(0.5), num_classes: 3)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(y_true, y_pred, Nx.u32(2), num_classes: 3, average: :macro)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(y_true, y_pred, Nx.u32(2), num_classes: 3, average: :weighted)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(y_true, y_pred, Nx.f32(0.5), num_classes: 3, average: :micro)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(Nx.tensor([1, 0, 1, 0]), Nx.tensor([0, 1, 0, 1]), Nx.tensor(0.5), num_classes: 2, average: :none)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.fbeta_score(Nx.tensor([1, 0, 1, 0]), Nx.tensor([0, 1, 0, 1]), 0.5, num_classes: 2, average: :none)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#fbeta_score/4-examples"},{"type":"function","title":"Scholar.Metrics.Classification.log_loss/3","doc":"Computes the log loss, aka logistic loss or cross-entropy loss.\n\nThe log-loss is a measure of how well a forecaster performs, with smaller\nvalues being better. For each sample, a forecaster outputs a probability for\neach class, from which the log loss is computed by averaging the negative log\nof the probability forecasted for the true class over a number of samples.\n\n`y_true` should contain `num_classes` unique values, and the sum of `y_prob`\nalong axis 1 should be 1 to respect the law of total probability.","ref":"Scholar.Metrics.Classification.html#log_loss/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.log_loss/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:normalize` (`t:boolean/0`) - If `true`, return the mean loss over the samples.\n  Otherwise, return the sum of losses over the samples. The default value is `true`.\n\n* `:sample_weights` - Sample weights of the observations. The default value is `1.0`.","ref":"Scholar.Metrics.Classification.html#log_loss/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.log_loss/3","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1])\n    iex> y_prob = Nx.tensor([[0.9, 0.1], [0.8, 0.2], [0.3, 0.7], [0.01, 0.99]])\n    iex> Scholar.Metrics.Classification.log_loss(y_true, y_prob, num_classes: 2)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Classification.log_loss(y_true, y_prob, num_classes: 2, normalize: false)\n    #Nx.Tensor \n    iex> weights = Nx.tensor([0.7, 2.3, 1.3, 0.34])\n    iex(361)> Scholar.Metrics.Classification.log_loss(y_true, y_prob, num_classes: 2, sample_weights: weights)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#log_loss/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.mcc/2","doc":"Matthews Correlation Coefficient (MCC) provides a measure of the quality of binary classifications.\n\nIt returns a value between -1 and 1 where 1 represents a perfect prediction, 0 represents no better\nthan random prediction, and -1 indicates total disagreement between prediction and observation.","ref":"Scholar.Metrics.Classification.html#mcc/2"},{"type":"function","title":"Scholar.Metrics.Classification.precision/3","doc":"Computes the precision of the given predictions with respect to\nthe given targets for multi-class classification problems.\n\nIf the sum of true positives and false positives is 0, then the\nresult is 0 to avoid zero division.","ref":"Scholar.Metrics.Classification.html#precision/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.precision/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors","ref":"Scholar.Metrics.Classification.html#precision/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.precision/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.precision(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#precision/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.precision_recall_curve/4","doc":"Compute precision-recall pairs for different probability thresholds.\n\nNote: this implementation is restricted to the binary classification task.","ref":"Scholar.Metrics.Classification.html#precision_recall_curve/4"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.precision_recall_curve/4","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1])\n    iex> scores = Nx.tensor([0.1, 0.4, 0.35, 0.8])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(scores)\n    iex> weights = Nx.tensor([1, 1, 2, 2])\n    iex> {precision, recall, thresholds} = Scholar.Metrics.Classification.precision_recall_curve(y_true, scores, distinct_value_indices, weights)\n    iex> precision\n    #Nx.Tensor \n    iex> recall\n    #Nx.Tensor \n    iex> thresholds\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#precision_recall_curve/4-examples"},{"type":"function","title":"Scholar.Metrics.Classification.precision_recall_fscore_support/3","doc":"Calculates precision, recall, F-score and support for each\nclass. It also supports a `beta` argument which weights\nrecall more than precision by it's value.","ref":"Scholar.Metrics.Classification.html#precision_recall_fscore_support/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.precision_recall_fscore_support/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:average` - This determines the type of averaging performed on the data.\n\n  * `:macro` - Calculate metrics for each label, and find their unweighted mean.\n  This does not take label imbalance into account.\n\n  * `:weighted` - Calculate metrics for each label, and find their average weighted by\n  support (the number of true instances for each label).\n\n  * `:micro` - Calculate metrics globally by counting the total true positives,\n  false negatives and false positives.\n\n  * `:none` - The F-score values for each class are returned.\n\n  The default value is `:none`.\n\n* `:beta` - Determines the weight of recall in the combined score.\n  For values of `beta` > 1 it gives more weight to recall, while `beta` < 1 favors precision. The default value is `1`.","ref":"Scholar.Metrics.Classification.html#precision_recall_fscore_support/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.precision_recall_fscore_support/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.precision_recall_fscore_support(y_true, y_pred, num_classes: 3)\n    {Nx.f32([0.6666666865348816, 1.0, 0.25]),\n     Nx.f32([0.6666666865348816, 0.5, 1.0]),\n     Nx.f32([0.6666666865348816, 0.6666666865348816, 0.4000000059604645]),\n     Nx.u64([3, 6, 1])}\n    iex> Scholar.Metrics.Classification.precision_recall_fscore_support(y_true, y_pred, num_classes: 3, average: :macro)\n    {Nx.f32([0.6666666865348816, 1.0, 0.25]),\n     Nx.f32([0.6666666865348816, 0.5, 1.0]),\n     Nx.f32(0.5777778029441833),\n     Nx.Constants.nan()}\n    iex> Scholar.Metrics.Classification.precision_recall_fscore_support(y_true, y_pred, num_classes: 3, average: :weighted)\n    {Nx.f32([0.6666666865348816, 1.0, 0.25]),\n     Nx.f32([0.6666666865348816, 0.5, 1.0]),\n     Nx.f32(0.6399999856948853),\n     Nx.Constants.nan()}\n    iex> Scholar.Metrics.Classification.precision_recall_fscore_support(y_true, y_pred, num_classes: 3, average: :micro)\n    {Nx.f32(0.6000000238418579),\n     Nx.f32(0.6000000238418579),\n     Nx.f32(0.6000000238418579),\n     Nx.Constants.nan()}\n\n    iex> y_true = Nx.tensor([1, 0, 1, 0], type: :u32)\n    iex> y_pred = Nx.tensor([0, 1, 0, 1], type: :u32)\n    iex> opts = [beta: 2, num_classes: 2, average: :none]\n    iex> Scholar.Metrics.Classification.precision_recall_fscore_support(y_true, y_pred, opts)\n    {Nx.f32([0.0, 0.0]),\n     Nx.f32([0.0, 0.0]),\n     Nx.f32([0.0, 0.0]),\n     Nx.u64([2, 2])}","ref":"Scholar.Metrics.Classification.html#precision_recall_fscore_support/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.recall/3","doc":"Computes the recall of the given predictions with respect to\nthe given targets for multi-class classification problems.\n\nIf the sum of true positives and false negatives is 0, then the\nresult is 0 to avoid zero division.","ref":"Scholar.Metrics.Classification.html#recall/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.recall/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors","ref":"Scholar.Metrics.Classification.html#recall/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.recall/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.recall(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#recall/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.roc_auc_score/4","doc":"Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores.\n\nNote: this implementation is restricted to the binary classification task.","ref":"Scholar.Metrics.Classification.html#roc_auc_score/4"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.roc_auc_score/4","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1])\n    iex> scores = Nx.tensor([0.1, 0.4, 0.35, 0.8])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(scores)\n    iex> weights = Nx.tensor([1, 1, 2, 2])\n    iex> Scholar.Metrics.Classification.roc_auc_score(y_true, scores, distinct_value_indices, weights)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#roc_auc_score/4-examples"},{"type":"function","title":"Scholar.Metrics.Classification.roc_curve/4","doc":"Compute Receiver operating characteristic (ROC).\n\nNote: this implementation is restricted to the binary classification task.","ref":"Scholar.Metrics.Classification.html#roc_curve/4"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.roc_curve/4","doc":"iex> y_true = Nx.tensor([0, 0, 1, 1])\n    iex> scores = Nx.tensor([0.1, 0.4, 0.35, 0.8])\n    iex> distinct_value_indices = Scholar.Metrics.Classification.distinct_value_indices(scores)\n    iex> weights = Nx.tensor([1, 1, 2, 2])\n    iex> {fpr, tpr, thresholds} = Scholar.Metrics.Classification.roc_curve(y_true, scores, distinct_value_indices, weights)\n    iex> fpr\n    #Nx.Tensor \n    iex> tpr\n    #Nx.Tensor \n    iex> thresholds\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#roc_curve/4-examples"},{"type":"function","title":"Scholar.Metrics.Classification.sensitivity/3","doc":"Computes the sensitivity of the given predictions with respect\nto the given targets for multi-class classification problems.","ref":"Scholar.Metrics.Classification.html#sensitivity/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.sensitivity/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors","ref":"Scholar.Metrics.Classification.html#sensitivity/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.sensitivity/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.sensitivity(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#sensitivity/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.specificity/3","doc":"Computes the specificity of the given predictions with respect\nto the given targets for multi-class classification problems.\n\nIf the sum of true negatives and false positives is 0, then the\nresult is 0 to avoid zero division.","ref":"Scholar.Metrics.Classification.html#specificity/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.specificity/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors","ref":"Scholar.Metrics.Classification.html#specificity/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.specificity/3","doc":"iex> y_true = Nx.tensor([0, 1, 1, 1, 1, 0, 2, 1, 0, 1], type: :u32)\n    iex> y_pred = Nx.tensor([0, 2, 1, 1, 2, 2, 2, 0, 0, 1], type: :u32)\n    iex> Scholar.Metrics.Classification.specificity(y_true, y_pred, num_classes: 3)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#specificity/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.top_k_accuracy_score/3","doc":"Top-k Accuracy classification score.\n\nThis metric computes the number of times where the correct label is\namong the top k labels predicted (ranked by predicted scores).\n\nFor binary task assumed that y_score have values from 0 to 1.","ref":"Scholar.Metrics.Classification.html#top_k_accuracy_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.top_k_accuracy_score/3","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes contained in the input tensors\n\n* `:k` (`t:integer/0`) - Number of top elements to look at for computing accuracy. The default value is `5`.\n\n* `:normalize` (`t:boolean/0`) - If `true`, return the fraction of correctly classified samples.\n  Otherwise, return the number of correctly classified samples. The default value is `true`.","ref":"Scholar.Metrics.Classification.html#top_k_accuracy_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Classification.top_k_accuracy_score/3","doc":"iex> y_true = Nx.tensor([0, 1, 2, 2, 0])\n    iex> y_score = Nx.tensor([[0.5, 0.2, 0.1], [0.3, 0.4, 0.5], [0.4, 0.3, 0.2], [0.1, 0.3, 0.6], [0.9, 0.1, 0.0]])\n    iex> Scholar.Metrics.Classification.top_k_accuracy_score(y_true, y_score, k: 2, num_classes: 3)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([0, 1, 2, 2, 0])\n    iex> y_score = Nx.tensor([[0.5, 0.2, 0.1], [0.3, 0.4, 0.5], [0.4, 0.3, 0.2], [0.1, 0.3, 0.6], [0.9, 0.1, 0.0]])\n    iex> Scholar.Metrics.Classification.top_k_accuracy_score(y_true, y_score, k: 2, num_classes: 3, normalize: false)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([0, 1, 0, 1, 0])\n    iex> y_score = Nx.tensor([0.55, 0.3, 0.1, -0.2, 0.99])\n    iex> Scholar.Metrics.Classification.top_k_accuracy_score(y_true, y_score, k: 1, num_classes: 2)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#top_k_accuracy_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Classification.zero_one_loss/3","doc":"Zero-one classification loss.","ref":"Scholar.Metrics.Classification.html#zero_one_loss/3"},{"type":"function","title":"Options - Scholar.Metrics.Classification.zero_one_loss/3","doc":"* `:normalize` (`t:boolean/0`) - If `true`, return the fraction of incorrectly classified samples.\n  Otherwise, return the number of incorrectly classified samples. The default value is `true`.\n\n\n\n# Examples\n\n    iex> y_pred = Nx.tensor([1, 2, 3, 4])\n    iex> y_true = Nx.tensor([2, 2, 3, 4])\n    iex> Scholar.Metrics.Classification.zero_one_loss(y_true, y_pred)\n    #Nx.Tensor \n\n    iex> y_pred = Nx.tensor([1, 2, 3, 4])\n    iex> y_true = Nx.tensor([2, 2, 3, 4])\n    iex> Scholar.Metrics.Classification.zero_one_loss(y_true, y_pred, normalize: false)\n    #Nx.Tensor","ref":"Scholar.Metrics.Classification.html#zero_one_loss/3-options"},{"type":"module","title":"Scholar.Metrics.Clustering","doc":"Metrics related to clustering algorithms.","ref":"Scholar.Metrics.Clustering.html"},{"type":"function","title":"Scholar.Metrics.Clustering.silhouette_samples/3","doc":"Compute the Silhouette Coefficient for each sample.\n\nThe silhouette value is a measure of how similar an object is to its own cluster (cohesion)\ncompared to other clusters (separation). The silhouette ranges from −1 to +1, where a high\nvalue indicates that the object is well matched to its own cluster and poorly\nmatched to neighboring clusters. If most objects have a high value, then the\nclustering configuration is appropriate. If many points have a low or negative\nvalue, then the clustering configuration may have too many or too few clusters.\n\nTime complexity of silhouette score is $O(N^2)$ where $N$ is the number of samples.","ref":"Scholar.Metrics.Clustering.html#silhouette_samples/3"},{"type":"function","title":"Options - Scholar.Metrics.Clustering.silhouette_samples/3","doc":"* `:num_clusters` (`t:pos_integer/0`) - Required. Number of clusters in clustering.","ref":"Scholar.Metrics.Clustering.html#silhouette_samples/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Clustering.silhouette_samples/3","doc":"iex> x = Nx.tensor([[0, 0], [1, 0], [1, 1], [3, 3], [4, 4.5]])\n    iex> labels = Nx.tensor([0, 0, 0, 1, 1])\n    iex> Scholar.Metrics.Clustering.silhouette_samples(x, labels, num_clusters: 2)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[0.1, 0], [0, 1], [22, 65], [42, 3], [4.2, 51]])\n    iex> labels = Nx.tensor([0, 1, 2, 1, 1])\n    iex> Scholar.Metrics.Clustering.silhouette_samples(x, labels, num_clusters: 3)\n    #Nx.Tensor","ref":"Scholar.Metrics.Clustering.html#silhouette_samples/3-examples"},{"type":"function","title":"Scholar.Metrics.Clustering.silhouette_score/3","doc":"Compute the mean Silhouette Coefficient of all samples.","ref":"Scholar.Metrics.Clustering.html#silhouette_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Clustering.silhouette_score/3","doc":"* `:num_clusters` (`t:pos_integer/0`) - Required. Number of clusters in clustering.","ref":"Scholar.Metrics.Clustering.html#silhouette_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Clustering.silhouette_score/3","doc":"iex> x = Nx.tensor([[0, 0], [1, 0], [1, 1], [3, 3], [4, 4.5]])\n    iex> labels = Nx.tensor([0, 0, 0, 1, 1])\n    iex> Scholar.Metrics.Clustering.silhouette_score(x, labels, num_clusters: 2)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[0.1, 0], [0, 1], [22, 65], [42, 3], [4.2, 51]])\n    iex> labels = Nx.tensor([0, 1, 2, 1, 1])\n    iex> Scholar.Metrics.Clustering.silhouette_score(x, labels, num_clusters: 3)\n    #Nx.Tensor","ref":"Scholar.Metrics.Clustering.html#silhouette_score/3-examples"},{"type":"module","title":"Scholar.Metrics.Distance","doc":"Distance metrics between multi-dimensional tensors.\nThey all support distance calculations between any subset of axes.","ref":"Scholar.Metrics.Distance.html"},{"type":"function","title":"Scholar.Metrics.Distance.chebyshev/3","doc":"Chebyshev or $L_{\\infty}$ distance.\n\n$$\nD(x, y) = \\max_i |x_i - y_i|\n$$","ref":"Scholar.Metrics.Distance.html#chebyshev/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.chebyshev/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#chebyshev/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.chebyshev/3","doc":"iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([3, 2])\n    iex> Scholar.Metrics.Distance.chebyshev(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2])\n    iex> Scholar.Metrics.Distance.chebyshev(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.chebyshev(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 5], [3, 4, 3]])\n    iex> y = Nx.tensor([[8, 3, 1], [2, 5, 2]])\n    iex> Scholar.Metrics.Distance.chebyshev(x, y, axes: [1])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.chebyshev(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#chebyshev/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.cosine/3","doc":"Cosine distance.\n\n$$\nD(u, v) = 1 - \\frac{u \\cdot v}{\\|u\\|_2 \\|v\\|_2}\n$$","ref":"Scholar.Metrics.Distance.html#cosine/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.cosine/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#cosine/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.cosine/3","doc":"iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([5, 2])\n    iex> Scholar.Metrics.Distance.cosine(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.cosine(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 3], [0, 0, 0], [5, 2, 4]])\n    iex> y = Nx.tensor([[1, 5, 2], [2, 4, 1], [0, 0, 0]])\n    iex> Scholar.Metrics.Distance.cosine(x, y, axes: [1])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.cosine(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#cosine/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.euclidean/3","doc":"Standard euclidean distance ($L_{2}$ distance).\n\n$$\nD(x, y) = \\sqrt{\\sum_i (x_i - y_i)^2}\n$$","ref":"Scholar.Metrics.Distance.html#euclidean/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.euclidean/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#euclidean/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.euclidean/3","doc":"iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([3, 2])\n    iex> Scholar.Metrics.Distance.euclidean(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2])\n    iex> Scholar.Metrics.Distance.euclidean(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.euclidean(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 5], [3, 4, 3]])\n    iex> y = Nx.tensor([[8, 3, 1], [2, 5, 2]])\n    iex> Scholar.Metrics.Distance.euclidean(x, y, axes: [0])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.euclidean(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#euclidean/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.hamming/3","doc":"Hamming distance.\n\n$$\nhamming(x, y) = \\frac{\\left \\lvert x\\_{i, j, ...} \\neq y\\_{i, j, ...}\\right \\rvert}{\\left \\lvert x\\_{i, j, ...}\\right \\rvert}\n$$\nwhere $i, j, ...$ are the aggregation axes","ref":"Scholar.Metrics.Distance.html#hamming/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.hamming/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#hamming/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.hamming/3","doc":"iex> x = Nx.tensor([1, 0, 0])\n    iex> y = Nx.tensor([0, 1, 0])\n    iex> Scholar.Metrics.Distance.hamming(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.hamming(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 3], [0, 0, 0], [5, 2, 4]])\n    iex> y = Nx.tensor([[1, 5, 2], [2, 4, 1], [0, 0, 0]])\n    iex> Scholar.Metrics.Distance.hamming(x, y, axes: [1])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.hamming(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#hamming/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.hamming/4","doc":"Hamming distance in weighted version.","ref":"Scholar.Metrics.Distance.html#hamming/4"},{"type":"function","title":"Options - Scholar.Metrics.Distance.hamming/4","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#hamming/4-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.hamming/4","doc":"iex> x = Nx.tensor([1, 0, 0])\n    iex> y = Nx.tensor([0, 1, 0])\n    iex> weights = Nx.tensor([1, 0.5, 0.5])\n    iex> Scholar.Metrics.Distance.hamming(x, y, weights)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> weights = Nx.tensor([1, 0.5, 0.5])\n    iex> Scholar.Metrics.Distance.hamming(x, y, weights, axes: [1])\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#hamming/4-examples"},{"type":"function","title":"Scholar.Metrics.Distance.manhattan/3","doc":"Manhattan, Taxicab, or $L_{1}$ distance.\n\n$$\nD(x, y) = \\sum_i |x_i - y_i|\n$$","ref":"Scholar.Metrics.Distance.html#manhattan/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.manhattan/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#manhattan/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.manhattan/3","doc":"iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([3, 2])\n    iex> Scholar.Metrics.Distance.manhattan(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1.0, 2.0])\n    iex> y = Nx.tensor([1, 2])\n    iex> Scholar.Metrics.Distance.manhattan(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.manhattan(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 5], [3, 4, 3]])\n    iex> y = Nx.tensor([[8, 3, 1], [2, 5, 2]])\n    iex> Scholar.Metrics.Distance.manhattan(x, y, axes: [0])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.manhattan(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#manhattan/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.minkowski/3","doc":"Minkowski distance.\n\n$$\nD(x, y) = \\left(\\sum_i |x_i - y_i|^p\\right)^{\\frac{1}{p}}\n$$","ref":"Scholar.Metrics.Distance.html#minkowski/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.minkowski/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.\n\n* `:p` - A positive parameter of Minkowski distance or :infinity (then Chebyshev metric computed). The default value is `2.0`.","ref":"Scholar.Metrics.Distance.html#minkowski/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.minkowski/3","doc":"iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([5, 2])\n    iex> Scholar.Metrics.Distance.minkowski(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2])\n    iex> Scholar.Metrics.Distance.minkowski(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.minkowski(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 5], [3, 4, 3]])\n    iex> y = Nx.tensor([[8, 3, 1], [2, 5, 2]])\n    iex> Scholar.Metrics.Distance.minkowski(x, y, p: 2.5, axes: [0])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.minkowski(x, y, p: 2.5)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#minkowski/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_cosine/1","doc":"Pairwise cosine distance. It is equivalent to\nScholar.Metrics.Distance.pairwise_euclidean(x, x)","ref":"Scholar.Metrics.Distance.html#pairwise_cosine/1"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_cosine/1","doc":"iex> x = Nx.iota({6, 6})\n    iex> Scholar.Metrics.Distance.pairwise_cosine(x)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_cosine/1-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_cosine/2","doc":"Pairwise cosine distance.","ref":"Scholar.Metrics.Distance.html#pairwise_cosine/2"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_cosine/2","doc":"iex> x = Nx.iota({6, 6})\n    iex> y = Nx.reverse(x)\n    iex> Scholar.Metrics.Distance.pairwise_cosine(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_cosine/2-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_euclidean/1","doc":"Pairwise euclidean distance. It is equivalent to\nScholar.Metrics.Distance.pairwise_euclidean(x, x)","ref":"Scholar.Metrics.Distance.html#pairwise_euclidean/1"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_euclidean/1","doc":"iex> x = Nx.iota({6, 6})\n    iex> Scholar.Metrics.Distance.pairwise_euclidean(x)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_euclidean/1-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_euclidean/2","doc":"Pairwise euclidean distance.","ref":"Scholar.Metrics.Distance.html#pairwise_euclidean/2"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_euclidean/2","doc":"iex> x = Nx.iota({2, 3})\n    iex> y = Nx.reverse(x)\n    iex> Scholar.Metrics.Distance.pairwise_euclidean(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_euclidean/2-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_minkowski/3","doc":"Computes the pairwise Minkowski distance.","ref":"Scholar.Metrics.Distance.html#pairwise_minkowski/3"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_minkowski/3","doc":"iex> x = Nx.iota({2, 3})\n    iex> y = Nx.reverse(x)\n    iex> Scholar.Metrics.Distance.pairwise_minkowski(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_minkowski/3-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_squared_euclidean/1","doc":"Pairwise squared euclidean distance. It is equivalent to\nScholar.Metrics.Distance.pairwise_squared_euclidean(x, x)","ref":"Scholar.Metrics.Distance.html#pairwise_squared_euclidean/1"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_squared_euclidean/1","doc":"iex> x = Nx.iota({6, 6})\n    iex> Scholar.Metrics.Distance.pairwise_squared_euclidean(x)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_squared_euclidean/1-examples"},{"type":"function","title":"Scholar.Metrics.Distance.pairwise_squared_euclidean/2","doc":"Pairwise squared euclidean distance.","ref":"Scholar.Metrics.Distance.html#pairwise_squared_euclidean/2"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.pairwise_squared_euclidean/2","doc":"iex> x = Nx.iota({6, 6})\n    iex> y = Nx.reverse(x)\n    iex> Scholar.Metrics.Distance.pairwise_squared_euclidean(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#pairwise_squared_euclidean/2-examples"},{"type":"function","title":"Scholar.Metrics.Distance.squared_euclidean/3","doc":"Squared euclidean distance.\n\n$$\nD(x, y) = \\sum_i (x_i - y_i)^2\n$$","ref":"Scholar.Metrics.Distance.html#squared_euclidean/3"},{"type":"function","title":"Options - Scholar.Metrics.Distance.squared_euclidean/3","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Distance.html#squared_euclidean/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Distance.squared_euclidean/3","doc":"iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([3, 2])\n    iex> Scholar.Metrics.Distance.squared_euclidean(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1.0, 2.0])\n    iex> Scholar.Metrics.Distance.squared_euclidean(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Distance.squared_euclidean(x, y)\n    ** (ArgumentError) tensors must be broadcast compatible, got tensors with shapes {2} and {3}\n\n    iex> x = Nx.tensor([[1, 2, 5], [3, 4, 3]])\n    iex> y = Nx.tensor([[8, 3, 1], [2, 5, 2]])\n    iex> Scholar.Metrics.Distance.squared_euclidean(x, y, axes: [0])\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[6, 2, 9], [2, 5, 3]])\n    iex> y = Nx.tensor([[8, 3, 1]])\n    iex> Scholar.Metrics.Distance.squared_euclidean(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Distance.html#squared_euclidean/3-examples"},{"type":"module","title":"Scholar.Metrics.Neighbors","doc":"Metrics for evaluating the results of approximate k-nearest neighbor search algorithms.","ref":"Scholar.Metrics.Neighbors.html"},{"type":"function","title":"Scholar.Metrics.Neighbors.recall/2","doc":"Computes the recall of predicted k-nearest neighbors given the true k-nearest neighbors.\nRecall is defined as the average fraction of nearest neighbors the algorithm predicted correctly.","ref":"Scholar.Metrics.Neighbors.html#recall/2"},{"type":"function","title":"Examples - Scholar.Metrics.Neighbors.recall/2","doc":"iex> neighbors_true = Nx.tensor([[0, 1], [1, 2], [2, 1]])\n    iex> Scholar.Metrics.Neighbors.recall(neighbors_true, neighbors_true)\n    #Nx.Tensor \n\n    iex> neighbors_true = Nx.tensor([[0, 1], [1, 2], [2, 1]])\n    iex> neighbors_pred = Nx.tensor([[0, 1], [1, 0], [2, 0]])\n    iex> Scholar.Metrics.Neighbors.recall(neighbors_true, neighbors_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Neighbors.html#recall/2-examples"},{"type":"module","title":"Scholar.Metrics.Ranking","doc":"Provides metrics and calculations related to ranking quality.\n\nRanking metrics evaluate the quality of ordered lists of items,\noften used in information retrieval and recommendation systems.\n\nThis module currently supports the following ranking metrics:\n  * Discounted Cumulative Gain (DCG)","ref":"Scholar.Metrics.Ranking.html"},{"type":"function","title":"Scholar.Metrics.Ranking.dcg/3","doc":"","ref":"Scholar.Metrics.Ranking.html#dcg/3"},{"type":"function","title":"Scholar.Metrics.Ranking.dcg_n/3","doc":"","ref":"Scholar.Metrics.Ranking.html#dcg_n/3"},{"type":"function","title":"Options - Scholar.Metrics.Ranking.dcg_n/3","doc":"* `:k` - Truncation parameter to consider only the top-k elements.\n\n\n\nComputes the DCG based on true relevance scores (`y_true`) and their respective predicted scores (`y_score`).","ref":"Scholar.Metrics.Ranking.html#dcg_n/3-options"},{"type":"function","title":"Scholar.Metrics.Ranking.ndcg_n/3","doc":"Computes the normalized discounted cumulative gain (NDCG) based on true relevance scores `y_true` and their respective predicted scores `y_score`.","ref":"Scholar.Metrics.Ranking.html#ndcg_n/3"},{"type":"function","title":"Options - Scholar.Metrics.Ranking.ndcg_n/3","doc":"* `:k` - Truncation parameter to consider only the top-k elements.","ref":"Scholar.Metrics.Ranking.html#ndcg_n/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Ranking.ndcg_n/3","doc":"iex> true_relevance = Nx.tensor([10, 0, 0, 1, 5])\n    iex> scores = Nx.tensor([0.1, 0.2, 0.3, 4, 70])\n    iex> Scholar.Metrics.Ranking.ndcg_n(true_relevance, scores)\n    #Nx.Tensor \n    iex> scores = Nx.tensor([0.05, 1.1, 1.0, 0.5, 0.0])\n    iex> Scholar.Metrics.Ranking.ndcg_n(true_relevance, scores)\n    #Nx.Tensor \n    iex> scores = Nx.tensor([0.05, 1.1, 1.0, 0.5, 0.0])\n    iex> Scholar.Metrics.Ranking.ndcg_n(true_relevance, scores, k: 4)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Ranking.ndcg_n(true_relevance, true_relevance, k: 4)\n    #Nx.Tensor","ref":"Scholar.Metrics.Ranking.html#ndcg_n/3-examples"},{"type":"module","title":"Scholar.Metrics.Regression","doc":"Regression Metric functions.\n\nMetrics are used to measure the performance and compare\nthe performance of any kind of regressor in\neasy-to-understand terms.\n\nAll of the functions in this module are implemented as\nnumerical functions and can be JIT or AOT compiled with\nany supported `Nx` compiler.","ref":"Scholar.Metrics.Regression.html"},{"type":"function","title":"Scholar.Metrics.Regression.d2_absolute_error_score/3","doc":"`D^2` regression score function, fraction of absolute error explained.\n\nBest possible score is 1.0 and it can be negative (because the model can be\narbitrarily worse). A model that always uses the empirical median of `y_true`\nas constant prediction, disregarding the input features,\ngets a `D^2` score of 0.0.","ref":"Scholar.Metrics.Regression.html#d2_absolute_error_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Regression.d2_absolute_error_score/3","doc":"#{NimbleOptions.docs(@d2_absolute_error_score_opts)}","ref":"Scholar.Metrics.Regression.html#d2_absolute_error_score/3-options"},{"type":"function","title":"Return Values - Scholar.Metrics.Regression.d2_absolute_error_score/3","doc":"It returns float or tensor of floats.","ref":"Scholar.Metrics.Regression.html#d2_absolute_error_score/3-return-values"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.d2_absolute_error_score/3","doc":"iex> y_true = Nx.tensor([1, 2, 3])\n    iex> y_pred = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Regression.d2_absolute_error_score(y_true, y_pred)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([1, 2, 3])\n    iex> y_pred = Nx.tensor([2, 2, 2])\n    iex> Scholar.Metrics.Regression.d2_absolute_error_score(y_true, y_pred)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([1, 2, 3])\n    iex> y_pred = Nx.tensor([3, 2, 1])\n    iex> Scholar.Metrics.Regression.d2_absolute_error_score(y_true, y_pred)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([3, -0.5, 2, 7])\n    iex> y_pred = Nx.tensor([2.5, 0.0, 2, 8])\n    iex> Scholar.Metrics.Regression.d2_absolute_error_score(y_true, y_pred)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([[0.5, 1], [-1, 1], [7, -6]])\n    iex> y_pred = Nx.tensor([[0, 2], [-1, 2], [8, -5]])\n    iex> Scholar.Metrics.Regression.d2_absolute_error_score(y_true, y_pred)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([[0.5, 1], [-1, 1], [7, -6]])\n    iex> y_pred = Nx.tensor([[0, 2], [-1, 2], [8, -5]])\n    iex> Scholar.Metrics.Regression.d2_absolute_error_score(y_true, y_pred, multioutput: :raw_values)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#d2_absolute_error_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.d2_absolute_error_score_n/3","doc":"","ref":"Scholar.Metrics.Regression.html#d2_absolute_error_score_n/3"},{"type":"function","title":"Scholar.Metrics.Regression.d2_pinball_score/3","doc":"`D^2` regression score function, fraction of pinball loss explained.\n\nBest possible score is 1.0 and it can be negative (because the model can be\narbitrarily worse). A model that always uses the empirical alpha-quantile of\n`y_true` as constant prediction, disregarding the input features,\ngets a `D^2` score of 0.0.","ref":"Scholar.Metrics.Regression.html#d2_pinball_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Regression.d2_pinball_score/3","doc":"#{NimbleOptions.docs(@d2_pinball_score_opts)}","ref":"Scholar.Metrics.Regression.html#d2_pinball_score/3-options"},{"type":"function","title":"Return Values - Scholar.Metrics.Regression.d2_pinball_score/3","doc":"It returns float or tensor of floats.","ref":"Scholar.Metrics.Regression.html#d2_pinball_score/3-return-values"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.d2_pinball_score/3","doc":"iex> y_true = Nx.tensor([1, 2, 3])\n    iex> y_pred = Nx.tensor([1, 3, 3])\n    iex> Scholar.Metrics.Regression.d2_pinball_score(y_true, y_pred)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Regression.d2_pinball_score(y_true, y_pred, alpha: 0.9)\n    #Nx.Tensor \n    iex> Scholar.Metrics.Regression.d2_pinball_score(y_true, y_true, alpha: 0.1)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#d2_pinball_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.d2_tweedie_score/3","doc":"$D^2$ regression score function, fraction of Tweedie\ndeviance explained.\n\nBest possible score is 1.0, lower values are worse and it\ncan also be negative.\n\nSince it uses the mean Tweedie deviance, it also includes\nthe Gaussian, Poisson, Gamma and inverse-Gaussian\ndistribution families as special cases.","ref":"Scholar.Metrics.Regression.html#d2_tweedie_score/3"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.d2_tweedie_score/3","doc":"iex> y_true = Nx.tensor([1, 1, 1, 1, 1, 2, 2, 1, 3, 1], type: :u32)\n    iex> y_pred = Nx.tensor([2, 2, 1, 1, 2, 2, 2, 1, 3, 1], type: :u32)\n    iex> Scholar.Metrics.Regression.d2_tweedie_score(y_true, y_pred, 1)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#d2_tweedie_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.explained_variance_score/3","doc":"Explained variance regression score function.\n\nBest possible score is 1.0, lower values are worse.","ref":"Scholar.Metrics.Regression.html#explained_variance_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Regression.explained_variance_score/3","doc":"* `:force_finite` (`t:boolean/0`) - Flag indicating if NaN and -Inf scores resulting from constant data should be replaced with real numbers\n  (1.0 if prediction is perfect, 0.0 otherwise) The default value is `true`.","ref":"Scholar.Metrics.Regression.html#explained_variance_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.explained_variance_score/3","doc":"iex> y_true = Nx.tensor([3, -0.5, 2, 7], type: {:f, 32})\n    iex> y_pred = Nx.tensor([2.5, 0.0, 2, 8], type: {:f, 32})\n    iex> Scholar.Metrics.Regression.explained_variance_score(y_true, y_pred)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0], type: :f64)\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0 + 1.0e-8], type: :f64)\n    iex> Scholar.Metrics.Regression.explained_variance_score(y_true, y_pred, force_finite: true)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0], type: :f64)\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0 + 1.0e-8], type: :f64)\n    iex> Scholar.Metrics.Regression.explained_variance_score(y_true, y_pred, force_finite: false)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> Scholar.Metrics.Regression.explained_variance_score(y_true, y_pred, force_finite: false)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> Scholar.Metrics.Regression.explained_variance_score(y_true, y_pred, force_finite: true)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#explained_variance_score/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.max_residual_error/2","doc":"Calculates the maximum residual error.\n\nThe residual error is defined as $$|y - \\hat{y}|$$ where $y$ is a true value\nand $\\hat{y}$ is a predicted value.\nThis function returns the maximum residual error over all samples in the\ninput: $max(|y_i - \\hat{y_i}|)$. For perfect predictions, the maximum\nresidual error is `0.0`.","ref":"Scholar.Metrics.Regression.html#max_residual_error/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.max_residual_error/2","doc":"iex> y_true = Nx.tensor([3, -0.5, 2, 7])\n    iex> y_pred = Nx.tensor([2.5, 0.0, 2, 8.5])\n    iex> Scholar.Metrics.Regression.max_residual_error(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#max_residual_error/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_absolute_error/2","doc":"Calculates the mean absolute error of predictions\nwith respect to targets.\n\n$$MAE = \\frac{\\sum_{i=1}^{n} |\\hat{y_i} - y_i|}{n}$$","ref":"Scholar.Metrics.Regression.html#mean_absolute_error/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_absolute_error/2","doc":"iex> y_true = Nx.tensor([[0.0, 1.0], [0.0, 0.0]])\n    iex> y_pred = Nx.tensor([[1.0, 1.0], [1.0, 0.0]])\n    iex> Scholar.Metrics.Regression.mean_absolute_error(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_absolute_error/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_absolute_percentage_error/2","doc":"Calculates the mean absolute percentage error of predictions\nwith respect to targets. If `y_true` values are equal or close\nto zero, it returns an arbitrarily large value.\n\n$$MAPE = \\frac{\\sum_{i=1}^{n} \\frac{|\\hat{y_i} - y_i|}{max(\\epsilon, \\hat{y_i})}}{n}$$","ref":"Scholar.Metrics.Regression.html#mean_absolute_percentage_error/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_absolute_percentage_error/2","doc":"iex> y_true = Nx.tensor([3, -0.5, 2, 7])\n    iex> y_pred = Nx.tensor([2.5, 0.0, 2, 8])\n    iex> Scholar.Metrics.Regression.mean_absolute_percentage_error(y_true, y_pred)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([1.0, 0.0, 2.4, 7.0])\n    iex> y_pred = Nx.tensor([1.2, 0.1, 2.4, 8.0])\n    iex> Scholar.Metrics.Regression.mean_absolute_percentage_error(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_absolute_percentage_error/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_gamma_deviance/2","doc":"Calculates the mean Gamma deviance of predictions\nwith respect to targets.","ref":"Scholar.Metrics.Regression.html#mean_gamma_deviance/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_gamma_deviance/2","doc":"iex> y_true = Nx.tensor([1, 1, 1, 1, 1, 2, 2, 1, 3, 1], type: :u32)\n    iex> y_pred = Nx.tensor([2, 2, 1, 1, 2, 2, 2, 1, 3, 1], type: :u32)\n    iex> Scholar.Metrics.Regression.mean_gamma_deviance(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_gamma_deviance/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_pinball_loss/3","doc":"Calculates the mean pinball loss to evaluate predictive performance of quantile regression models.\n\n$$pinball(y, \\hat{y}) = \\frac{1}{n) \\sum_{i=1}^{n} \\alpha max(\\hat{y_i} - y_i, 0) +\n(1 - \\alpha) max(\\hat{y_i} - y_i, 0)$$\n\nThe residual error is defined as $$|y - \\hat{y}|$$ where $y$ is a true value\nand $\\hat{y}$ is a predicted value.\n\n#{NimbleOptions.docs(@mean_pinball_loss_schema)}","ref":"Scholar.Metrics.Regression.html#mean_pinball_loss/3"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_pinball_loss/3","doc":"iex> y_true = Nx.tensor([1, 2, 3])\n    iex> y_pred = Nx.tensor([2, 3, 4])\n    iex> Scholar.Metrics.Regression.mean_pinball_loss(y_true, y_pred)\n    #Nx.Tensor \n    iex> y_true = Nx.tensor([[1, 0, 0, 1], [0, 1, 1, 1], [1, 1, 0, 1]])\n    iex> y_pred = Nx.tensor([[0, 0, 0, 1], [1, 0, 1, 1], [0, 0, 0, 1]])\n    iex> Scholar.Metrics.Regression.mean_pinball_loss(y_true, y_pred, alpha: 0.5, multioutput: :raw_values)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_pinball_loss/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_poisson_deviance/2","doc":"Calculates the mean Poisson deviance of predictions\nwith respect to targets.","ref":"Scholar.Metrics.Regression.html#mean_poisson_deviance/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_poisson_deviance/2","doc":"iex> y_true = Nx.tensor([1, 1, 1, 1, 1, 2, 2, 1, 3, 1], type: :u32)\n    iex> y_pred = Nx.tensor([2, 2, 1, 1, 2, 2, 2, 1, 3, 1], type: :u32)\n    iex> Scholar.Metrics.Regression.mean_poisson_deviance(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_poisson_deviance/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_square_error/2","doc":"Calculates the mean square error of predictions\nwith respect to targets.\n\n$$MSE = \\frac{\\sum_{i=1}^{n} (\\hat{y_i} - y_i)^2}{n}$$","ref":"Scholar.Metrics.Regression.html#mean_square_error/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_square_error/2","doc":"iex> y_true = Nx.tensor([[0.0, 2.0], [0.5, 0.0]])\n    iex> y_pred = Nx.tensor([[1.0, 1.0], [1.0, 0.0]])\n    iex> Scholar.Metrics.Regression.mean_square_error(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_square_error/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_square_log_error/2","doc":"Calculates the mean square logarithmic error of predictions\nwith respect to targets.\n\n$$MSLE = \\frac{\\sum_{i=1}^{n} (\\log(\\hat{y_i} + 1) - \\log(y_i + 1))^2}{n}$$","ref":"Scholar.Metrics.Regression.html#mean_square_log_error/2"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_square_log_error/2","doc":"iex> y_true = Nx.tensor([[0.0, 1.0], [0.0, 0.0]])\n    iex> y_pred = Nx.tensor([[1.0, 1.0], [1.0, 0.0]])\n    iex> Scholar.Metrics.Regression.mean_square_log_error(y_true, y_pred)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_square_log_error/2-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_tweedie_deviance/3","doc":"Calculates the mean Tweedie deviance of predictions\nwith respect to targets. Includes the Gaussian, Poisson,\nGamma and inverse-Gaussian families as special cases.\n\n$$d(y,\\mu) =\n\\begin{cases}\n(y-\\mu)^2, & \\text{for }p=0\\\\\\\\\n2(y \\log(y/\\mu) + \\mu - y), & \\text{for }p=1\\\\\\\\\n2(\\log(\\mu/y) + y/\\mu - 1), & \\text{for }p=2\\\\\\\\\n2\\left(\\frac{\\max(y,0)^{2-p}}{(1-p)(2-p)}-\\frac{y\\mu^{1-p}}{1-p}+\\frac{\\mu^{2-p}}{2-p}\\right), & \\text{for }p<0 \\vee p>2\n\\end{cases}$$","ref":"Scholar.Metrics.Regression.html#mean_tweedie_deviance/3"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_tweedie_deviance/3","doc":"iex> y_true = Nx.tensor([1, 1, 1, 1, 1, 2, 2, 1, 3, 1], type: :u32)\n    iex> y_pred = Nx.tensor([2, 2, 1, 1, 2, 2, 2, 1, 3, 1], type: :u32)\n    iex> Scholar.Metrics.Regression.mean_tweedie_deviance(y_true, y_pred, 1)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_tweedie_deviance/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.mean_tweedie_deviance!/3","doc":"Similar to `mean_tweedie_deviance/3` but raises `RuntimeError` if the\ninputs cannot be used with the given power argument.\n\nNote: This function cannot be used in `defn`.","ref":"Scholar.Metrics.Regression.html#mean_tweedie_deviance!/3"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.mean_tweedie_deviance!/3","doc":"iex> y_true = Nx.tensor([1, 1, 1, 1, 1, 2, 2, 1, 3, 1], type: :u32)\n    iex> y_pred = Nx.tensor([2, 2, 1, 1, 2, 2, 2, 1, 3, 1], type: :u32)\n    iex> Scholar.Metrics.Regression.mean_tweedie_deviance!(y_true, y_pred, 1)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#mean_tweedie_deviance!/3-examples"},{"type":"function","title":"Scholar.Metrics.Regression.quantile/2","doc":"","ref":"Scholar.Metrics.Regression.html#quantile/2"},{"type":"function","title":"Scholar.Metrics.Regression.r2_score/3","doc":"Calculates the $R^2$ score of predictions with respect to targets.\n\n$$R^2 = 1 - \\frac{\\sum (y_i - \\hat{y}_i)^2}{\\sum (y_i - \\bar{y})^2}$$","ref":"Scholar.Metrics.Regression.html#r2_score/3"},{"type":"function","title":"Options - Scholar.Metrics.Regression.r2_score/3","doc":"* `:force_finite` (`t:boolean/0`) - Flag indicating if NaN and -Inf scores resulting from constant data should be replaced with real numbers\n  (1.0 if prediction is perfect, 0.0 otherwise) The default value is `true`.","ref":"Scholar.Metrics.Regression.html#r2_score/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Regression.r2_score/3","doc":"iex> y_true = Nx.tensor([3, -0.5, 2, 7], type: {:f, 32})\n    iex> y_pred = Nx.tensor([2.5, 0.0, 2, 8], type: {:f, 32})\n    iex> Scholar.Metrics.Regression.r2_score(y_true, y_pred)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0], type: :f64)\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0 + 1.0e-8], type: :f64)\n    iex> Scholar.Metrics.Regression.r2_score(y_true, y_pred, force_finite: true)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0], type: :f64)\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0 + 1.0e-8], type: :f64)\n    iex> Scholar.Metrics.Regression.r2_score(y_true, y_pred, force_finite: false)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> Scholar.Metrics.Regression.r2_score(y_true, y_pred, force_finite: false)\n    #Nx.Tensor \n\n    iex> y_true = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> y_pred = Nx.tensor([-2.0, -2.0, -2.0])\n    iex> Scholar.Metrics.Regression.r2_score(y_true, y_pred, force_finite: true)\n    #Nx.Tensor","ref":"Scholar.Metrics.Regression.html#r2_score/3-examples"},{"type":"module","title":"Scholar.Metrics.Similarity","doc":"Similarity metrics between multi-dimensional tensors.","ref":"Scholar.Metrics.Similarity.html"},{"type":"function","title":"Scholar.Metrics.Similarity.binary_jaccard/3","doc":"Calculates Jaccard similarity based on binary attributes.\nIt assumes that inputs have the same shape.\n\n$$\nJ(X, Y) = \\frac{M\\_{11}}{M\\_{01} + M\\_{10} + M\\_{11}}\n$$\n\n\nWhere:\n\n* $M_{11}$ is the total number of attributes, for which both $X$ and $Y$ have 1.\n* $M_{10}$ is the total number of attributes, for which $X$ has 1 and $Y$ has 0.\n* $M_{01}$ is the total number of attributes, for which $X$ has 0 and $Y$ has 1.","ref":"Scholar.Metrics.Similarity.html#binary_jaccard/3"},{"type":"function","title":"Options - Scholar.Metrics.Similarity.binary_jaccard/3","doc":"* `:axis` - Axis over which the distance will be calculated. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Similarity.html#binary_jaccard/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Similarity.binary_jaccard/3","doc":"iex> x = Nx.tensor([1,0,0,1,1,1])\n    iex> y = Nx.tensor([0,0,1,1,1,0])\n    iex> Scholar.Metrics.Similarity.binary_jaccard(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[1,1,0,1], [1,1,0,1]])\n    iex> y = Nx.tensor([[1,1,0,1], [1,0,0,1]])\n    iex> Scholar.Metrics.Similarity.binary_jaccard(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[1,1,0,1], [1,1,0,1]])\n    iex> y = Nx.tensor([[1,1,0,1], [1,0,0,1]])\n    iex> Scholar.Metrics.Similarity.binary_jaccard(x, y, axis: 1)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 1])\n    iex> y = Nx.tensor([1, 0, 0])\n    iex> Scholar.Metrics.Similarity.binary_jaccard(x, y)\n    ** (ArgumentError) expected tensor to have shape {2}, got tensor with shape {3}","ref":"Scholar.Metrics.Similarity.html#binary_jaccard/3-examples"},{"type":"function","title":"Scholar.Metrics.Similarity.dice_coefficient/3","doc":"Calculates Dice coefficient.\n\nIt is a statistic used to gauge the similarity of two samples.\nMathematically, it is defined as:\n\n$$\nDice(A, B) = \\frac{2 \\mid A \\cap B \\mid}{\\mid A \\mid + \\mid B \\mid}\n$$\n\nwhere A and B are sets.","ref":"Scholar.Metrics.Similarity.html#dice_coefficient/3"},{"type":"function","title":"Options - Scholar.Metrics.Similarity.dice_coefficient/3","doc":"* `:axis` - Axis over which the distance will be calculated. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Similarity.html#dice_coefficient/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Similarity.dice_coefficient/3","doc":"iex> x = Nx.tensor([1.0, 5.0, 3.0, 6.7])\n    iex> y = Nx.tensor([5.0, 2.5, 3.1, 9.0])\n    iex> Scholar.Metrics.Similarity.dice_coefficient(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2, 3, 5, 7])\n    iex> y = Nx.tensor([1, 2, 4, 8, 9])\n    iex> Scholar.Metrics.Similarity.dice_coefficient(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.iota({2,3})\n    iex> y = Nx.tensor([[0, 3, 4], [3, 4, 8]])\n    iex> Scholar.Metrics.Similarity.dice_coefficient(x, y, axis: 1)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Similarity.dice_coefficient(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Similarity.html#dice_coefficient/3-examples"},{"type":"function","title":"Scholar.Metrics.Similarity.dice_coefficient_binary/3","doc":"Calculates Dice coefficient based on binary attributes.\nIt assumes that inputs have the same shape.","ref":"Scholar.Metrics.Similarity.html#dice_coefficient_binary/3"},{"type":"function","title":"Options - Scholar.Metrics.Similarity.dice_coefficient_binary/3","doc":"* `:axis` - Axis over which the distance will be calculated. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Similarity.html#dice_coefficient_binary/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Similarity.dice_coefficient_binary/3","doc":"iex> x = Nx.tensor([1,0,0,1,1,1])\n    iex> y = Nx.tensor([0,0,1,1,1,0])\n    iex> Scholar.Metrics.Similarity.dice_coefficient_binary(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[1,1,0,1], [1,1,0,1]])\n    iex> y = Nx.tensor([[1,1,0,1], [1,0,0,1]])\n    iex> Scholar.Metrics.Similarity.dice_coefficient_binary(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[1,1,0,1], [1,1,0,1]])\n    iex> y = Nx.tensor([[1,1,0,1], [1,0,0,1]])\n    iex> Scholar.Metrics.Similarity.dice_coefficient_binary(x, y, axis: 1)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 1])\n    iex> y = Nx.tensor([1, 0, 0])\n    iex> Scholar.Metrics.Similarity.dice_coefficient_binary(x, y)\n    ** (ArgumentError) expected tensor to have shape {2}, got tensor with shape {3}","ref":"Scholar.Metrics.Similarity.html#dice_coefficient_binary/3-examples"},{"type":"function","title":"Scholar.Metrics.Similarity.jaccard/3","doc":"Calculates Jaccard similarity (also known as Jaccard similarity coefficient, or Jaccard index).\n\nJaccard similarity is a statistic used to measure similarities between two sets. Mathematically, the calculation\nof Jaccard similarity is the ratio of set intersection over set union.\n\n$$\nJ(A, B) = \\frac{\\mid A \\cap B \\mid}{\\mid A \\cup B \\mid}\n$$\n\nwhere A and B are sets.","ref":"Scholar.Metrics.Similarity.html#jaccard/3"},{"type":"function","title":"Options - Scholar.Metrics.Similarity.jaccard/3","doc":"* `:axis` - Axis over which the distance will be calculated. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Metrics.Similarity.html#jaccard/3-options"},{"type":"function","title":"Examples - Scholar.Metrics.Similarity.jaccard/3","doc":"iex> x = Nx.tensor([1.0, 5.0, 3.0, 6.7])\n    iex> y = Nx.tensor([5.0, 2.5, 3.1, 9.0])\n    iex> Scholar.Metrics.Similarity.jaccard(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2, 3, 5, 7])\n    iex> y = Nx.tensor([1, 2, 4, 8, 9])\n    iex> Scholar.Metrics.Similarity.jaccard(x, y)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[0, 1, 2], [3, 4, 5]])\n    iex> y = Nx.tensor([[0, 3, 4], [3, 4, 8]])\n    iex> Scholar.Metrics.Similarity.jaccard(x, y, axis: 1)\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([1, 2])\n    iex> y = Nx.tensor([1, 2, 3])\n    iex> Scholar.Metrics.Similarity.jaccard(x, y)\n    #Nx.Tensor","ref":"Scholar.Metrics.Similarity.html#jaccard/3-examples"},{"type":"module","title":"Scholar.ModelSelection","doc":"Module containing cross validation, splitting function, and other model selection methods.","ref":"Scholar.ModelSelection.html"},{"type":"function","title":"Scholar.ModelSelection.cross_validate/4","doc":"General interface of cross validation.","ref":"Scholar.ModelSelection.html#cross_validate/4"},{"type":"function","title":"Examples - Scholar.ModelSelection.cross_validate/4","doc":"iex> folding_fun = fn x -> Scholar.ModelSelection.k_fold_split(x, 3) end\n    iex> scoring_fun = fn x, y ->\n    ...>   {x_train, x_test} = x\n    ...>   {y_train, y_test} = y\n    ...>   model = Scholar.Linear.LinearRegression.fit(x_train, y_train, fit_intercept?: true)\n    ...>   y_pred = Scholar.Linear.LinearRegression.predict(model, x_test)\n    ...>   mse = Scholar.Metrics.Regression.mean_square_error(y_test, y_pred)\n    ...>   mae = Scholar.Metrics.Regression.mean_absolute_error(y_test, y_pred)\n    ...>   [mse, mae]\n    ...> end\n    iex> x = Nx.iota({7, 2})\n    iex> y = Nx.tensor([0, 1, 2, 0, 1, 1, 0])\n    iex> Scholar.ModelSelection.cross_validate(x, y, folding_fun, scoring_fun)\n    #Nx.Tensor","ref":"Scholar.ModelSelection.html#cross_validate/4-examples"},{"type":"function","title":"Scholar.ModelSelection.grid_search/5","doc":"General interface of grid search.\n\nThe `opts` must be a keyword list of list values, which will become different\ncombinations to perform the grid search on.","ref":"Scholar.ModelSelection.html#grid_search/5"},{"type":"function","title":"Examples - Scholar.ModelSelection.grid_search/5","doc":"iex> folding_fun = fn x -> Scholar.ModelSelection.k_fold_split(x, 3) end\n    iex> scoring_fun = fn x, y, opts ->\n    ...>   {x_train, x_test} = x\n    ...>   {y_train, y_test} = y\n    ...>   model = Scholar.Linear.LogisticRegression.fit(x_train, y_train, opts)\n    ...>   y_pred = Scholar.Linear.LogisticRegression.predict(model, x_test)\n    ...>   mse = Scholar.Metrics.Regression.mean_square_error(y_test, y_pred)\n    ...>   mae = Scholar.Metrics.Regression.mean_absolute_error(y_test, y_pred)\n    ...>   [mse, mae]\n    ...> end\n    iex> x = Nx.iota({7, 2})\n    iex> y = Nx.tensor([0, 1, 2, 0, 1, 1, 0])\n    iex> opts = [\n    ...>   num_classes: [3],\n    ...>   iterations: [10, 20, 50],\n    ...>   optimizer: [Polaris.Optimizers.adam(learning_rate: 0.005), Polaris.Optimizers.adam(learning_rate: 0.01)],\n    ...> ]\n    iex> Scholar.ModelSelection.grid_search(x, y, folding_fun, scoring_fun, opts)","ref":"Scholar.ModelSelection.html#grid_search/5-examples"},{"type":"function","title":"Scholar.ModelSelection.k_fold_split/2","doc":"Perform K-Fold split on the given data.","ref":"Scholar.ModelSelection.html#k_fold_split/2"},{"type":"function","title":"Examples - Scholar.ModelSelection.k_fold_split/2","doc":"iex> x = Nx.iota({7, 2})\n    iex> Scholar.ModelSelection.k_fold_split(x, 2) |> Enum.to_list()\n    [\n      {Nx.tensor(\n        [\n          [6, 7],\n          [8, 9],\n          [10, 11]\n        ]\n      ),\n      Nx.tensor(\n        [\n          [0, 1],\n          [2, 3],\n          [4, 5]\n        ]\n      )},\n      {Nx.tensor(\n        [\n          [0, 1],\n          [2, 3],\n          [4, 5]\n        ]\n      ),\n      Nx.tensor(\n        [\n          [6, 7],\n          [8, 9],\n          [10, 11]\n        ]\n      )}\n    ]","ref":"Scholar.ModelSelection.html#k_fold_split/2-examples"},{"type":"function","title":"Scholar.ModelSelection.weighted_cross_validate/5","doc":"General interface of weighted cross validation.","ref":"Scholar.ModelSelection.html#weighted_cross_validate/5"},{"type":"function","title":"Examples - Scholar.ModelSelection.weighted_cross_validate/5","doc":"iex> folding_fun = fn x -> Scholar.ModelSelection.k_fold_split(x, 3) end\n    iex> scoring_fun = fn x, y, weights ->\n    ...>   {x_train, x_test} = x\n    ...>   {y_train, y_test} = y\n    ...>   {weights_train, _weights_test} = weights\n    ...>   model = Scholar.Linear.LinearRegression.fit(x_train, y_train, fit_intercept?: true, sample_weights: weights_train)\n    ...>   y_pred = Scholar.Linear.LinearRegression.predict(model, x_test)\n    ...>   mse = Scholar.Metrics.Regression.mean_square_error(y_test, y_pred)\n    ...>   mae = Scholar.Metrics.Regression.mean_absolute_error(y_test, y_pred)\n    ...>   [mse, mae]\n    ...> end\n    iex> x = Nx.iota({7, 2})\n    iex> y = Nx.tensor([0, 1, 2, 0, 1, 1, 0])\n    iex> weights = Nx.tensor([1, 2, 1, 2, 1, 2, 1])\n    iex> Scholar.ModelSelection.weighted_cross_validate(x, y, weights, folding_fun, scoring_fun)\n    #Nx.Tensor","ref":"Scholar.ModelSelection.html#weighted_cross_validate/5-examples"},{"type":"function","title":"Scholar.ModelSelection.weighted_grid_search/6","doc":"General interface of weighted grid search.\n\nIf you want to use `opts` in some functions inside `scoring_fun`, you need to pass it as a parameter\nlike in the example below.","ref":"Scholar.ModelSelection.html#weighted_grid_search/6"},{"type":"function","title":"Examples - Scholar.ModelSelection.weighted_grid_search/6","doc":"iex> folding_fun = fn x -> Scholar.ModelSelection.k_fold_split(x, 3) end\n    iex> scoring_fun = fn x, y, weights, opts ->\n    ...>   {x_train, x_test} = x\n    ...>   {y_train, y_test} = y\n    ...>   {weights_train, _weights_test} = weights\n    ...>   opts = Keyword.put(opts, :sample_weights, weights_train)\n    ...>   model = Scholar.Linear.RidgeRegression.fit(x_train, y_train, opts)\n    ...>   y_pred = Scholar.Linear.RidgeRegression.predict(model, x_test)\n    ...>   mse = Scholar.Metrics.Regression.mean_square_error(y_test, y_pred)\n    ...>   mae = Scholar.Metrics.Regression.mean_absolute_error(y_test, y_pred)\n    ...>   [mse, mae]\n    ...> end\n    iex> x = Nx.iota({7, 2})\n    iex> y = Nx.tensor([0, 1, 2, 0, 1, 1, 0])\n    iex> weights = [Nx.tensor([1, 2, 1, 2, 1, 2, 1]), Nx.tensor([2, 1, 2, 1, 2, 1, 2])]\n    iex> opts = [\n    ...>   alpha: [0, 1, 5],\n    ...>   fit_intercept?: [true, false],\n    ...> ]\n    iex> Scholar.ModelSelection.weighted_grid_search(x, y, weights, folding_fun, scoring_fun, opts)","ref":"Scholar.ModelSelection.html#weighted_grid_search/6-examples"},{"type":"module","title":"Scholar.Preprocessing","doc":"Set of functions for preprocessing data.","ref":"Scholar.Preprocessing.html"},{"type":"function","title":"Scholar.Preprocessing.binarize/2","doc":"Converts a tensor into binary values based on the given threshold.","ref":"Scholar.Preprocessing.html#binarize/2"},{"type":"function","title":"Options - Scholar.Preprocessing.binarize/2","doc":"* `:type` - Type of the resultant tensor. The default value is `:f32`.\n\n* `:threshold` - Feature values below or equal to this are replaced by 0, above it by 1. The default value is `0`.","ref":"Scholar.Preprocessing.html#binarize/2-options"},{"type":"function","title":"Examples - Scholar.Preprocessing.binarize/2","doc":"iex> Scholar.Preprocessing.binarize(Nx.tensor([[1.0, -1.0, 2.0], [2.0, 0.0, 0.0], [0.0, 1.0, -1.0]]))\n    #Nx.Tensor \n\n    iex> Scholar.Preprocessing.binarize(Nx.tensor([[1.0, -1.0, 2.0], [2.0, 0.0, 0.0], [0.0, 1.0, -1.0]]), threshold: 1.3, type: {:u, 8})\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#binarize/2-examples"},{"type":"function","title":"Scholar.Preprocessing.max_abs_scale/2","doc":"Scales a tensor by dividing each sample in batch by maximum absolute value in the batch.\n\nIt is a shortcut for `Scholar.Preprocessing.MaxAbsScaler.fit_transform/2`.\nSee `Scholar.Preprocessing.MaxAbsScaler` for more information.","ref":"Scholar.Preprocessing.html#max_abs_scale/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.max_abs_scale/2","doc":"iex> Scholar.Preprocessing.max_abs_scale(Nx.tensor([1, 2, 3]))\n    #Nx.Tensor \n\n    iex> Scholar.Preprocessing.max_abs_scale(Nx.tensor([[1, -1, 2], [3, 0, 0], [0, 1, -1], [2, 3, 1]]), axes: [0])\n    #Nx.Tensor \n\n    iex> Scholar.Preprocessing.max_abs_scale(42)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#max_abs_scale/2-examples"},{"type":"function","title":"Scholar.Preprocessing.min_max_scale/2","doc":"Scales a tensor by a given range.\n\nIt is a shortcut for `Scholar.Preprocessing.MinMaxScaler.fit_transform/2`.\nSee `Scholar.Preprocessing.MinMaxScaler` for more information.","ref":"Scholar.Preprocessing.html#min_max_scale/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.min_max_scale/2","doc":"iex> Scholar.Preprocessing.min_max_scale(Nx.tensor([1, 2, 3]))\n    #Nx.Tensor \n\n    iex> Scholar.Preprocessing.min_max_scale(42)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#min_max_scale/2-examples"},{"type":"function","title":"Scholar.Preprocessing.normalize/2","doc":"Normalize samples individually to unit norm.\n\nThe zero-tensors cannot be normalized and they stay the same\nafter normalization.\n\nIt is a shortcut for `Scholar.Preprocessing.Normalizer.fit_transform/2`.\nSee `Scholar.Preprocessing.Normalizer` for more information.","ref":"Scholar.Preprocessing.html#normalize/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.normalize/2","doc":"iex> Scholar.Preprocessing.normalize(Nx.tensor([[0, 0, 0], [3, 4, 5], [-2, 4, 3]]), axes: [1])\n    #Nx.Tensor \n\n    iex> Scholar.Preprocessing.normalize(Nx.tensor([[0, 0, 0], [3, 4, 5], [-2, 4, 3]]))\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#normalize/2-examples"},{"type":"function","title":"Scholar.Preprocessing.one_hot_encode/2","doc":"It is a shortcut for `Scholar.Preprocessing.OneHotEncoder.fit_transform/2`.\nSee `Scholar.Preprocessing.OneHotEncoder` for more information.","ref":"Scholar.Preprocessing.html#one_hot_encode/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.one_hot_encode/2","doc":"iex> Scholar.Preprocessing.one_hot_encode(Nx.tensor([2, 0, 3, 2, 1, 1, 0]), num_classes: 4)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#one_hot_encode/2-examples"},{"type":"function","title":"Scholar.Preprocessing.ordinal_encode/2","doc":"It is a shortcut for `Scholar.Preprocessing.OrdinalEncoder.fit_transform/2`.\nSee `Scholar.Preprocessing.OrdinalEncoder` for more information.","ref":"Scholar.Preprocessing.html#ordinal_encode/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.ordinal_encode/2","doc":"iex> Scholar.Preprocessing.ordinal_encode(Nx.tensor([3, 2, 4, 56, 2, 4, 2]), num_classes: 4)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#ordinal_encode/2-examples"},{"type":"function","title":"Scholar.Preprocessing.standard_scale/2","doc":"Standardizes the tensor by removing the mean and scaling to unit variance.\n\nIt is a shortcut for `Scholar.Preprocessing.StandardScale.fit_transform/3`.\nSee `Scholar.Preprocessing.StandardScale` for more information.","ref":"Scholar.Preprocessing.html#standard_scale/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.standard_scale/2","doc":"iex> Scholar.Preprocessing.standard_scale(Nx.tensor([1,2,3]))\n    #Nx.Tensor","ref":"Scholar.Preprocessing.html#standard_scale/2-examples"},{"type":"module","title":"Scholar.Preprocessing.MaxAbsScaler","doc":"Scales a tensor by dividing each sample in batch by maximum absolute value in the batch\n\nCentering and scaling happen independently on each feature by computing the relevant\nstatistics on the samples in the training set. Maximum absolute value then is\nstored to be used on new samples.","ref":"Scholar.Preprocessing.MaxAbsScaler.html"},{"type":"function","title":"Scholar.Preprocessing.MaxAbsScaler.fit/2","doc":"Compute the maximum absolute value of samples to be used for later scaling.","ref":"Scholar.Preprocessing.MaxAbsScaler.html#fit/2"},{"type":"function","title":"Options - Scholar.Preprocessing.MaxAbsScaler.fit/2","doc":"* `:axes` - Axes to calculate the max absolute value over. By default the absolute values\n  are calculated between the whole tensors.","ref":"Scholar.Preprocessing.MaxAbsScaler.html#fit/2-options"},{"type":"function","title":"Return values - Scholar.Preprocessing.MaxAbsScaler.fit/2","doc":"Returns a struct with the following parameters:\n\n  * `max_abs`: the calculated maximum absolute value of samples.","ref":"Scholar.Preprocessing.MaxAbsScaler.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Preprocessing.MaxAbsScaler.fit/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> Scholar.Preprocessing.MaxAbsScaler.fit(t)\n    %Scholar.Preprocessing.MaxAbsScaler{\n      max_abs: Nx.tensor(\n        [\n          [2]\n        ]\n      )\n    }","ref":"Scholar.Preprocessing.MaxAbsScaler.html#fit/2-examples"},{"type":"function","title":"Scholar.Preprocessing.MaxAbsScaler.fit_transform/2","doc":"Standardizes the tensor by removing the mean and scaling to unit variance.","ref":"Scholar.Preprocessing.MaxAbsScaler.html#fit_transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.MaxAbsScaler.fit_transform/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> Scholar.Preprocessing.MaxAbsScaler.fit_transform(t)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.MaxAbsScaler.html#fit_transform/2-examples"},{"type":"function","title":"Scholar.Preprocessing.MaxAbsScaler.transform/2","doc":"Performs the standardization of the tensor using a fitted scaler.","ref":"Scholar.Preprocessing.MaxAbsScaler.html#transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.MaxAbsScaler.transform/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> scaler = Scholar.Preprocessing.MaxAbsScaler.fit(t)\n    iex> Scholar.Preprocessing.MaxAbsScaler.transform(scaler, t)\n    #Nx.Tensor \n    iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> scaler = Scholar.Preprocessing.MaxAbsScaler.fit(t)\n    iex> new_tensor = Nx.tensor([[0.5, 1, -1], [0.3, 0.8, -1.6]])\n    iex> Scholar.Preprocessing.MaxAbsScaler.transform(scaler, new_tensor)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.MaxAbsScaler.html#transform/2-examples"},{"type":"module","title":"Scholar.Preprocessing.MinMaxScaler","doc":"Scales a tensor by dividing each sample in batch by maximum absolute value in the batch\n\nCentering and scaling happen independently on each feature by computing the relevant\nstatistics on the samples in the training set. Maximum absolute value then is\nstored to be used on new samples.","ref":"Scholar.Preprocessing.MinMaxScaler.html"},{"type":"function","title":"Scholar.Preprocessing.MinMaxScaler.fit/2","doc":"Compute the maximum absolute value of samples to be used for later scaling.","ref":"Scholar.Preprocessing.MinMaxScaler.html#fit/2"},{"type":"function","title":"Options - Scholar.Preprocessing.MinMaxScaler.fit/2","doc":"* `:axes` - Axes to calculate the max absolute value over. By default the absolute values\n  are calculated between the whole tensors.\n\n* `:min_bound` - The lower boundary of the desired range of transformed data. The default value is `0`.\n\n* `:max_bound` - The upper boundary of the desired range of transformed data. The default value is `1`.","ref":"Scholar.Preprocessing.MinMaxScaler.html#fit/2-options"},{"type":"function","title":"Return values - Scholar.Preprocessing.MinMaxScaler.fit/2","doc":"Returns a struct with the following parameters:\n\n  * `min_data`: the calculated minimum value of samples.\n\n  * `max_data`: the calculated maximum value of samples.\n\n  * `min_bound`: The lower boundary of the desired range of transformed data.\n\n  * `max_bound`: The upper boundary of the desired range of transformed data.","ref":"Scholar.Preprocessing.MinMaxScaler.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Preprocessing.MinMaxScaler.fit/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> Scholar.Preprocessing.MinMaxScaler.fit(t)\n    %Scholar.Preprocessing.MinMaxScaler{\n      min_data: Nx.tensor(\n        [\n          [-1]\n        ]\n      ),\n      max_data: Nx.tensor(\n        [\n          [2]\n        ]\n      ),\n      min_bound: Nx.tensor(\n        0\n      ),\n      max_bound: Nx.tensor(\n        1\n      )\n    }","ref":"Scholar.Preprocessing.MinMaxScaler.html#fit/2-examples"},{"type":"function","title":"Scholar.Preprocessing.MinMaxScaler.fit_transform/2","doc":"Standardizes the tensor by removing the mean and scaling to unit variance.","ref":"Scholar.Preprocessing.MinMaxScaler.html#fit_transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.MinMaxScaler.fit_transform/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> Scholar.Preprocessing.MinMaxScaler.fit_transform(t)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.MinMaxScaler.html#fit_transform/2-examples"},{"type":"function","title":"Scholar.Preprocessing.MinMaxScaler.transform/2","doc":"Performs the standardization of the tensor using a fitted scaler.","ref":"Scholar.Preprocessing.MinMaxScaler.html#transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.MinMaxScaler.transform/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> scaler = Scholar.Preprocessing.MinMaxScaler.fit(t)\n    iex> Scholar.Preprocessing.MinMaxScaler.transform(scaler, t)\n    #Nx.Tensor \n\n    iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> scaler = Scholar.Preprocessing.MinMaxScaler.fit(t)\n    iex> new_tensor = Nx.tensor([[0.5, 1, -1], [0.3, 0.8, -1.6]])\n    iex> Scholar.Preprocessing.MinMaxScaler.transform(scaler, new_tensor)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.MinMaxScaler.html#transform/2-examples"},{"type":"module","title":"Scholar.Preprocessing.Normalizer","doc":"Implements functionality for rescaling tensor to unit norm. It enables to apply normalization along any combination of axes.","ref":"Scholar.Preprocessing.Normalizer.html"},{"type":"function","title":"Scholar.Preprocessing.Normalizer.fit_transform/2","doc":"Normalize samples individually to unit norm.\n\nThe zero-tensors cannot be normalized and they stay the same\nafter normalization.","ref":"Scholar.Preprocessing.Normalizer.html#fit_transform/2"},{"type":"function","title":"Options - Scholar.Preprocessing.Normalizer.fit_transform/2","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.\n\n* `:norm` - The norm to use to normalize each non zero sample.\n  Possible options are `:euclidean`, `:manhattan`, and `:chebyshev` The default value is `:euclidean`.","ref":"Scholar.Preprocessing.Normalizer.html#fit_transform/2-options"},{"type":"function","title":"Examples - Scholar.Preprocessing.Normalizer.fit_transform/2","doc":"iex> t = Nx.tensor([[0, 0, 0], [3, 4, 5], [-2, 4, 3]])\n    iex> Scholar.Preprocessing.Normalizer.fit_transform(t, axes: [1])\n    #Nx.Tensor \n\n    iex> t = Nx.tensor([[0, 0, 0], [3, 4, 5], [-2, 4, 3]])\n    iex> Scholar.Preprocessing.Normalizer.fit_transform(t)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.Normalizer.html#fit_transform/2-examples"},{"type":"module","title":"Scholar.Preprocessing.OneHotEncoder","doc":"Implements encoder that converts integer value (substitute of categorical data in tensors) into 0-1 vector.\nThe index of 1 in the vector is aranged in sorted manner. This means that for x < y => one_index(x) < one_index(y).\n\nCurrently the module supports only 1D tensors.","ref":"Scholar.Preprocessing.OneHotEncoder.html"},{"type":"function","title":"Scholar.Preprocessing.OneHotEncoder.fit/2","doc":"Creates mapping from values into one-hot vectors.","ref":"Scholar.Preprocessing.OneHotEncoder.html#fit/2"},{"type":"function","title":"Options - Scholar.Preprocessing.OneHotEncoder.fit/2","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes to be encoded.","ref":"Scholar.Preprocessing.OneHotEncoder.html#fit/2-options"},{"type":"function","title":"Examples - Scholar.Preprocessing.OneHotEncoder.fit/2","doc":"iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> Scholar.Preprocessing.OneHotEncoder.fit(t, num_classes: 4)\n    %Scholar.Preprocessing.OneHotEncoder{\n      encoder: %Scholar.Preprocessing.OrdinalEncoder{\n        encoding_tensor: Nx.tensor(\n          [\n            [0, 2],\n            [1, 3],\n            [2, 4],\n            [3, 56]\n          ]\n        )\n      },\n      one_hot: Nx.tensor(\n        [\n          [1, 0, 0, 0],\n          [0, 1, 0, 0],\n          [0, 0, 1, 0],\n          [0, 0, 0, 1]\n        ], type: :u8\n      )\n    }","ref":"Scholar.Preprocessing.OneHotEncoder.html#fit/2-examples"},{"type":"function","title":"Scholar.Preprocessing.OneHotEncoder.fit_transform/2","doc":"Apply encoding on the provided tensor directly. It's equivalent to `fit/2` and then `transform/2` on the same data.","ref":"Scholar.Preprocessing.OneHotEncoder.html#fit_transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.OneHotEncoder.fit_transform/2","doc":"iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> Scholar.Preprocessing.OneHotEncoder.fit_transform(t, num_classes: 4)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.OneHotEncoder.html#fit_transform/2-examples"},{"type":"function","title":"Scholar.Preprocessing.OneHotEncoder.transform/2","doc":"Encode labels as a one-hot numeric tensor. All values provided to `transform/2` must be seen\nin `fit/2` function, otherwise an error occurs.","ref":"Scholar.Preprocessing.OneHotEncoder.html#transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.OneHotEncoder.transform/2","doc":"iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> enoder = Scholar.Preprocessing.OneHotEncoder.fit(t, num_classes: 4)\n    iex> Scholar.Preprocessing.OneHotEncoder.transform(enoder, t)\n    #Nx.Tensor \n\n    iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> enoder = Scholar.Preprocessing.OneHotEncoder.fit(t, num_classes: 4)\n    iex> new_tensor = Nx.tensor([2, 3, 4, 3, 4, 56, 2])\n    iex> Scholar.Preprocessing.OneHotEncoder.transform(enoder, new_tensor)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.OneHotEncoder.html#transform/2-examples"},{"type":"module","title":"Scholar.Preprocessing.OrdinalEncoder","doc":"Implements encoder that converts integer value (substitute of categorical data in tensors) into other integer value.\nThe values assigned starts from `0` and go up to `num_classes - 1`.They are maintained in sorted manner.\nThis means that for x < y => encoded_value(x) < encoded_value(y).\n\nCurrently the module supports only 1D tensors.","ref":"Scholar.Preprocessing.OrdinalEncoder.html"},{"type":"function","title":"Scholar.Preprocessing.OrdinalEncoder.fit/2","doc":"Fit the ordinal encoder to provided data. The labels are assigned in a sorted manner.","ref":"Scholar.Preprocessing.OrdinalEncoder.html#fit/2"},{"type":"function","title":"Options - Scholar.Preprocessing.OrdinalEncoder.fit/2","doc":"* `:num_classes` (`t:pos_integer/0`) - Required. Number of classes to be encoded.","ref":"Scholar.Preprocessing.OrdinalEncoder.html#fit/2-options"},{"type":"function","title":"Examples - Scholar.Preprocessing.OrdinalEncoder.fit/2","doc":"iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> Scholar.Preprocessing.OrdinalEncoder.fit(t, num_classes: 4)\n    %Scholar.Preprocessing.OrdinalEncoder{\n      encoding_tensor: Nx.tensor(\n        [\n          [0, 2],\n          [1, 3],\n          [2, 4],\n          [3, 56]\n        ]\n      )\n    }","ref":"Scholar.Preprocessing.OrdinalEncoder.html#fit/2-examples"},{"type":"function","title":"Scholar.Preprocessing.OrdinalEncoder.fit_transform/2","doc":"Apply encoding on the provided tensor directly. It's equivalent to `fit/2` and then `transform/2` on the same data.","ref":"Scholar.Preprocessing.OrdinalEncoder.html#fit_transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.OrdinalEncoder.fit_transform/2","doc":"iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> Scholar.Preprocessing.OridinalEncoder.fit_transform(t, num_classes: 4)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.OrdinalEncoder.html#fit_transform/2-examples"},{"type":"function","title":"Scholar.Preprocessing.OrdinalEncoder.transform/2","doc":"Encodes a tensor's values into integers from range 0 to `:num_classes - 1` or -1 if the value did not occur in fitting process.","ref":"Scholar.Preprocessing.OrdinalEncoder.html#transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.OrdinalEncoder.transform/2","doc":"iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> enoder = Scholar.Preprocessing.OrdinalEncoder.fit(t, num_classes: 4)\n    iex> Scholar.Preprocessing.OrdinalEncoder.transform(enoder, t)\n    #Nx.Tensor \n\n    iex> t = Nx.tensor([3, 2, 4, 56, 2, 4, 2])\n    iex> enoder = Scholar.Preprocessing.OrdinalEncoder.fit(t, num_classes: 4)\n    iex> new_tensor = Nx.tensor([2, 3, 4, 5, 4, 56, 2])\n    iex> Scholar.Preprocessing.OrdinalEncoder.transform(enoder, new_tensor)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.OrdinalEncoder.html#transform/2-examples"},{"type":"module","title":"Scholar.Preprocessing.StandardScaler","doc":"Standardizes the tensor by removing the mean and scaling to unit variance.\n\nFormula for input tensor $x$:\n\n$$\nz = \\frac{x - \\mu}{\\sigma}\n$$\n\nWhere $\\mu$ is the mean of the samples, and $\\sigma$ is the standard deviation.\nStandardization can be helpful in cases where the data follows\na Gaussian distribution (or Normal distribution) without outliers.\n\nCentering and scaling happen independently on each feature by computing the relevant\nstatistics on the samples in the training set. Mean and standard deviation are then\nstored to be used on new samples.","ref":"Scholar.Preprocessing.StandardScaler.html"},{"type":"function","title":"Scholar.Preprocessing.StandardScaler.fit/2","doc":"Compute the standard deviation and mean of samples to be used for later scaling.","ref":"Scholar.Preprocessing.StandardScaler.html#fit/2"},{"type":"function","title":"Options - Scholar.Preprocessing.StandardScaler.fit/2","doc":"* `:axes` - Axes to calculate the distance over. By default the distance\n  is calculated between the whole tensors.","ref":"Scholar.Preprocessing.StandardScaler.html#fit/2-options"},{"type":"function","title":"Return values - Scholar.Preprocessing.StandardScaler.fit/2","doc":"Returns a struct with the following parameters:\n\n  * `standard_deviation`: the calculated standard deviation of samples.\n\n  * `mean`: the calculated mean of samples.","ref":"Scholar.Preprocessing.StandardScaler.html#fit/2-return-values"},{"type":"function","title":"Examples - Scholar.Preprocessing.StandardScaler.fit/2","doc":"iex> Scholar.Preprocessing.StandardScaler.fit(Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]]))\n    %Scholar.Preprocessing.StandardScaler{\n      standard_deviation: Nx.tensor(\n        [\n          [1.0657403469085693]\n        ]\n      ),\n      mean: Nx.tensor(\n        [\n          [0.4444444477558136]\n        ]\n      )\n    }","ref":"Scholar.Preprocessing.StandardScaler.html#fit/2-examples"},{"type":"function","title":"Scholar.Preprocessing.StandardScaler.fit_transform/2","doc":"Standardizes the tensor by removing the mean and scaling to unit variance.","ref":"Scholar.Preprocessing.StandardScaler.html#fit_transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.StandardScaler.fit_transform/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> Scholar.Preprocessing.StandardScaler.fit_transform(t)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.StandardScaler.html#fit_transform/2-examples"},{"type":"function","title":"Scholar.Preprocessing.StandardScaler.transform/2","doc":"Performs the standardization of the tensor using a fitted scaler.","ref":"Scholar.Preprocessing.StandardScaler.html#transform/2"},{"type":"function","title":"Examples - Scholar.Preprocessing.StandardScaler.transform/2","doc":"iex> t = Nx.tensor([[1, -1, 2], [2, 0, 0], [0, 1, -1]])\n    iex> scaler = Scholar.Preprocessing.StandardScaler.fit(t)\n    %Scholar.Preprocessing.StandardScaler{\n      standard_deviation: Nx.tensor(\n        [\n          [1.0657403469085693]\n        ]\n      ),\n      mean: Nx.tensor(\n        [\n          [0.4444444477558136]\n        ]\n      )\n    }\n    iex> Scholar.Preprocessing.StandardScaler.transform(scaler, t)\n    #Nx.Tensor","ref":"Scholar.Preprocessing.StandardScaler.html#transform/2-examples"},{"type":"module","title":"Scholar.Stats","doc":"Statistical functions\n\nAll of the functions in this module are implemented as\nnumerical functions and can be JIT or AOT compiled with\nany supported `Nx` compiler.","ref":"Scholar.Stats.html"},{"type":"function","title":"Scholar.Stats.correlation_matrix/1","doc":"Computes correlation matrix for sample inputs `x`.\n\nThe value on the position $Corr_{ij}$ in the $Corr$ matrix is calculated using the formula:\n$$ Corr(X\\_i, X\\_j) = \\frac{Cov(X\\_i, X\\_j)}{\\sqrt{Cov(X\\_i, X\\_i)Cov(X\\_j, X\\_j)}} $$\nWhere:\n  * $X_i$ is a $i$th row of input\n\n  * $Cov(X\\_i, X\\_j)$ is covariance between features $X_i$ and $X_j$\n\n\nTime complexity of correlation estimation is $O(N * K^2)$ where $N$ is the number of samples\nand $K$ is the number of features.","ref":"Scholar.Stats.html#correlation_matrix/1"},{"type":"function","title":"Example - Scholar.Stats.correlation_matrix/1","doc":"iex> Scholar.Stats.correlation_matrix(Nx.tensor([[3, 6, 5], [26, 75, 3], [23, 4, 1]]))\n    #Nx.Tensor \n\n    iex> Scholar.Stats.correlation_matrix(Nx.tensor([[3, 6], [2, 3], [7, 9], [5, 3]]))\n    #Nx.Tensor \n\n    iex> x = Nx.tensor([[3, 6, 5], [26, 75, 3], [23, 4, 1]])\n    iex> means = Nx.mean(x, axes: [-2])\n    iex> Scholar.Stats.correlation_matrix(x, means)\n    #Nx.Tensor","ref":"Scholar.Stats.html#correlation_matrix/1-example"},{"type":"function","title":"Scholar.Stats.correlation_matrix/2","doc":"","ref":"Scholar.Stats.html#correlation_matrix/2"},{"type":"function","title":"Scholar.Stats.kurtosis/2","doc":"Computes the kurtosis (Fisher or Pearson) of a dataset.","ref":"Scholar.Stats.html#kurtosis/2"},{"type":"function","title":"Options - Scholar.Stats.kurtosis/2","doc":"* `:axes` - Axes to calculate the operation. If set to `nil` then\n  the operation is performed on the whole tensor. The default value is `[0]`.\n\n* `:keep_axes` (`t:boolean/0`) - If set to true, the axes which are reduced are left. The default value is `false`.\n\n* `:bias` (`t:boolean/0`) - If false, then the calculations are corrected for statistical bias. The default value is `true`.\n\n* `:variant` - If :fisher then Fisher's definition is used, if :pearson then Pearson's definition is used. The default value is `:fisher`.","ref":"Scholar.Stats.html#kurtosis/2-options"},{"type":"function","title":"Examples - Scholar.Stats.kurtosis/2","doc":"iex> x = Nx.tensor([[3, 5, 3], [2, 6, 1], [9, 3, 2], [1, 6, 8]])\n    iex> Scholar.Stats.kurtosis(x)\n    #Nx.Tensor","ref":"Scholar.Stats.html#kurtosis/2-examples"},{"type":"function","title":"Scholar.Stats.moment/3","doc":"Calculates the nth moment about the mean for a sample.","ref":"Scholar.Stats.html#moment/3"},{"type":"function","title":"Options - Scholar.Stats.moment/3","doc":"* `:axes` - Axes to calculate the operation. If set to `nil` then\n  the operation is performed on the whole tensor. The default value is `[0]`.\n\n* `:keep_axes` (`t:boolean/0`) - If set to true, the axes which are reduced are left. The default value is `false`.","ref":"Scholar.Stats.html#moment/3-options"},{"type":"function","title":"Examples - Scholar.Stats.moment/3","doc":"iex> x = Nx.tensor([[3, 5, 3], [2, 6, 1], [9, 3, 2], [1, 6, 8]])\n    iex> Scholar.Stats.moment(x, 2)\n    #Nx.Tensor","ref":"Scholar.Stats.html#moment/3-examples"},{"type":"function","title":"Scholar.Stats.skew/2","doc":"Computes the sample skewness of a data set.","ref":"Scholar.Stats.html#skew/2"},{"type":"function","title":"Options - Scholar.Stats.skew/2","doc":"* `:axes` - Axes to calculate the operation. If set to `nil` then\n  the operation is performed on the whole tensor. The default value is `[0]`.\n\n* `:keep_axes` (`t:boolean/0`) - If set to true, the axes which are reduced are left. The default value is `false`.\n\n* `:bias` (`t:boolean/0`) - If false, then the calculations are corrected for statistical bias. The default value is `true`.","ref":"Scholar.Stats.html#skew/2-options"},{"type":"function","title":"Examples - Scholar.Stats.skew/2","doc":"iex> x = Nx.tensor([[3, 5, 3], [2, 6, 1], [9, 3, 2], [1, 6, 8]])\n    iex> Scholar.Stats.skew(x)\n    #Nx.Tensor","ref":"Scholar.Stats.html#skew/2-examples"},{"type":"extras","title":"README","doc":"Traditional machine learning tools built on top of Nx. Scholar implements\nseveral algorithms for classification, regression, clustering, dimensionality\nreduction, metrics, and preprocessing.\n\nFor deep learning, see [Axon](https://github.com/elixir-nx/axon).\nFor decision trees/forests, see [EXGBoost](https://github.com/acalejos/exgboost).","ref":"readme.html"},{"type":"extras","title":"Installation - README","doc":"","ref":"readme.html#installation"},{"type":"extras","title":"Mix projects - README","doc":"Add to your `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:scholar, \"~> 0.3.0\"}\n  ]\nend\n```\n\nBesides Scholar, you will most likely want to use an existing Nx compiler/backend,\nsuch as EXLA:\n\n```elixir\ndef deps do\n  [\n    {:scholar, \"~> 0.3.0\"},\n    {:exla, \">= 0.0.0\"}\n  ]\nend\n```\n\nAnd then in your `config/config.exs` file:\n\n```elixir\nimport Config\nconfig :nx, :default_backend, EXLA.Backend\n# Client can also be set to :cuda / :rocm\nconfig :nx, :default_defn_options, [compiler: EXLA, client: :host]\n```\n\n> #### JIT required! {: .warning}\n>\n> It is important you set the `default_defn_options` as shown in the snippet above,\n> as many algorithms in Scholar use loops which are much more memory efficient when\n> JIT compiled.\n>\n> If for some reason you cannot set a default `defn` compiler, you can explicitly\n> JIT any function, for example: `EXLA.jit(&Scholar.Cluster.AffinityPropagation.fit/1)`.","ref":"readme.html#mix-projects"},{"type":"extras","title":"Notebooks - README","doc":"To use Scholar inside code notebooks, run:\n\n```elixir\nMix.install([\n  {:scholar, \"~> 0.3.0\"},\n  {:exla, \">= 0.0.0\"}\n])\n\nNx.global_default_backend(EXLA.Backend)\n# Client can also be set to :cuda / :rocm\nNx.Defn.global_default_options(compiler: EXLA, client: :host)\n```\n\n> #### JIT required! {: .warning}\n>\n> It is important you set the `Nx.Defn.global_default_options/1` as shown in the snippet\n> above, as many algorithms in Scholar use loops which are much more memory efficient\n> when JIT compiled.\n>\n> If for some reason you cannot set a default `defn` compiler, you can explicitly\n> JIT any function, for example: `EXLA.jit(&Scholar.Cluster.AffinityPropagation.fit/1)`.","ref":"readme.html#notebooks"},{"type":"extras","title":"License - README","doc":"Copyright (c) 2022 The Machine Learning Working Group of the Erlang Ecosystem Foundation\n\nLicensed under the Apache License, Version 2.0 (the \"License\");\nyou may not use this file except in compliance with the License.\nYou may obtain a copy of the License at [http://www.apache.org/licenses/LICENSE-2.0](http://www.apache.org/licenses/LICENSE-2.0)\n\nUnless required by applicable law or agreed to in writing, software\ndistributed under the License is distributed on an \"AS IS\" BASIS,\nWITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\nSee the License for the specific language governing permissions and\nlimitations under the License.","ref":"readme.html#license"},{"type":"extras","title":"Cross-validation with gradient boosting trees","doc":"# Cross-validation with gradient boosting trees\n\n```elixir\nMix.install([\n  {:scholar, \"~> 0.2.0\"},\n  {:exgboost, \"~> 0.5\"},\n  {:req, \"~> 0.3.9\"},\n  {:kino_vega_lite, \"~> 0.1.9\"},\n  {:kino, \"~> 0.10.0\"},\n  {:kino_explorer, \"~> 0.1.7\"},\n  {:nx, \"~> 0.7.2\"},\n  {:explorer, \"~> 0.8.2\"}\n])\n```","ref":"cv_gradient_boosting_tree.html"},{"type":"extras","title":"Setup - Cross-validation with gradient boosting trees","doc":"We will use Explorer in this notebook, so let's define aliases for its main modules:\n\n```elixir\nrequire Explorer.DataFrame, as: DF\nrequire Explorer.Series, as: S\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\n```\n\nWe are going to work with [Medical Cost Personal Datasets](https://www.kaggle.com/datasets/mirichoi0218/insurance) to predict medical charges that were applied to each person from the dataset. Let's download it:\n\n```elixir\ndata =\n  Req.get!(\n    \"https://gist.githubusercontent.com/meperezcuello/82a9f1c1c473d6585e750ad2e3c05a41/raw/d42d226d0dd64e7f5395a0eec1b9190a10edbc03/Medical_Cost.csv\"\n  ).body\n\ndf = DF.load_csv!(data)\n```\n\nThe dataset consists of 7 columns: age, sex, BMI (body mass index), children (number), smoker (yes-no), region (NE, NW, SE, SW), and charges that we want to predict. Since gradient boosting trees that we are using in our analysis accept numerical data, we need to process further three columns: sex, smoker and region and encode them from categorical to numerical.\n\n```elixir\ny = DF.select(df, \"charges\") |> Nx.concatenate()\n\nx =\n  df\n  |> DF.discard([\"charges\"])\n  |> DF.mutate(\n    sex: cast(sex, :category),\n    smoker: cast(smoker, :category),\n    region: cast(region, :category)\n  )\n  |> Nx.stack(axis: 1)\n\n{x, y}\n```\n\nBefore training our model, we separate the data between train and test sets.\n\n```elixir\n{x_train, x_test} = Nx.split(x, 0.8)\n{y_train, y_test} = Nx.split(y, 0.8)\n```","ref":"cv_gradient_boosting_tree.html#setup"},{"type":"extras","title":"Training a gradient boosting tree - Cross-validation with gradient boosting trees","doc":"Gradient boosting works by sequentially adding predictors to an ensemble, each one correcting its predecessor. Let's go through a simple regression example, using decision trees as the base predictors; this is called _gradient tree boosting_, or _gradient boosted regression trees_ (GBRT).\n\nEXGBoost provides an implementation of gradient boosting trees that accepts a wide range of hyperparameter configurations. For the full list of hyperparameters refer to the [EXGBoost](https://hexdocs.pm/exgboost/EXGBoost.html) docs.\n\n```elixir\ny_pred =\n  EXGBoost.train(\n    x_train,\n    y_train,\n    booster: :gbtree,\n    tree_method: :auto,\n    objective: :reg_squarederror,\n    num_boost_rounds: 100,\n    evals: [{x_train, y_train, \"training\"}],\n    verbose_eval: true\n  )\n  |> EXGBoost.predict(x_test)\n```\n\nHaving our predictions, we can measure performance by calculating the root mean squared error of predictions with respect to target values.\n\n```elixir\nalias Scholar.Metrics.Regression, as: Metrics\n\nMetrics.mean_square_error(y_test, y_pred)\n|> Nx.sqrt()\n|> Nx.to_number()\n```\n\nWith very little preprocessing we get similar results to the linear regression model. However, we can improve our model evaluation process by using cross-validation.","ref":"cv_gradient_boosting_tree.html#training-a-gradient-boosting-tree"},{"type":"extras","title":"Evaluating with cross-validation - Cross-validation with gradient boosting trees","doc":"_k_-fold cross-validation works by creating splits on the training set into _k_ smaller sets, so that the model is trained using $k - 1$ splits (folds) as training data and is validated on the remaining part of the data. When using this technique, the performance measure is the average of the values computed in each iteration.\n\n \n\n<!-- livebook:{\"break_markdown\":true} -->\n\nScholar provides tools for performing _k_-fold CV.\n\n```elixir\nalias Scholar.ModelSelection\n```\n\nFirst, we need to define a folding function that will perform the _k_-folds, and also a scoring function that will train the model and evaluate performance with each split.\n\n```elixir\nfolding_fn = fn x -> ModelSelection.k_fold_split(x, 5) end\n\nscoring_fn = fn x, y ->\n  {x_train, x_test} = x\n  {y_train, y_test} = y\n\n  y_pred =\n    EXGBoost.train(\n      x_train,\n      y_train,\n      booster: :gbtree,\n      tree_method: :auto,\n      objective: :reg_squarederror,\n      num_boost_rounds: 100,\n      evals: [{x_train, y_train, \"training\"}],\n      verbose_eval: true\n    )\n    |> EXGBoost.predict(x_test)\n\n  Metrics.mean_square_error(y_test, y_pred)\n  |> Nx.sqrt()\nend\n```\n\nNow let's run the cross-validation function and put the scores tensor in a series.\n\n```elixir\ncv_score =\n  ModelSelection.cross_validate(\n    x_train,\n    y_train,\n    folding_fn,\n    scoring_fn\n  )\n  |> Nx.squeeze()\n  |> S.from_tensor()\n```\n\nTaking the average results in the performance reported by cross-validation.\n\n```elixir\nS.mean(cv_score)\n```","ref":"cv_gradient_boosting_tree.html#evaluating-with-cross-validation"},{"type":"extras","title":"Fine-tuning our model with grid search - Cross-validation with gradient boosting trees","doc":"Finding the right configuration of hyperparameters is an important part of the process when selecting a model. One could try different combinations of hyperparameter values manually, but this can get tedious and time consuming. Instead, we can use the _grid search_ method, an iterative process for finding an optimal configuration of hyperparameter values for a given model.\n\nFirst, we need to provide a \"grid\" of hyperparameter values, so that the algorithm can train and evaluate our model with all possible combinations.\n\n```elixir\ngrid = [\n  booster: [:gbtree],\n  objective: [:reg_squarederror],\n  evals: [[{x_train, y_train, \"training\"}]],\n  verbose_eval: [true],\n  tree_method: [:approx, :exact],\n  max_depth: [2, 3, 4, 5, 6],\n  num_boost_rounds: [20, 50, 90],\n  subsample: [0.25, 0.5, 0.75, 1.0]\n]\n```\n\nWe also need to adapt our scoring function in order to use the hyperparameter values for each grid search iteration.\n\n```elixir\ngs_scoring_fn = fn x, y, hyperparams ->\n  {x_train, x_test} = x\n  {y_train, y_test} = y\n\n  y_pred =\n    x_train\n    |> EXGBoost.train(y_train, hyperparams)\n    |> EXGBoost.predict(x_test)\n\n  Metrics.mean_square_error(y_test, y_pred)\n  |> Nx.sqrt()\nend\n```\n\nLet's run the grid search and see the results. Remember that the more hyperparameter values you add to the grid, the more it will take the algorithm to end.\n\n```elixir\ngs_scores =\n  ModelSelection.grid_search(\n    x_train,\n    y_train,\n    folding_fn,\n    gs_scoring_fn,\n    grid\n  )\n```\n\nThe output is a list of maps, each corresponding to an iteration of the grid search algorithm. Every iteration yields a `score` calculated by our scoring function. Let's find the set of hyperparameters that optimizes the score.\n\n```elixir\nbest_config =\n  Enum.min_by(gs_scores, fn %{score: score} ->\n    score\n    |> Nx.squeeze()\n    |> Nx.to_number()\n  end)\n```\n\nFinally we train and evaluate a model using the best hyperparameter configuration found by grid search.\n\n```elixir\n%{hyperparameters: opts} = best_config\n\nmodel = EXGBoost.train(x_train, y_train, opts)\ny_pred = EXGBoost.predict(model, x_test)\n\nrmse =\n  Metrics.mean_square_error(y_test, y_pred)\n  |> Nx.sqrt()\n\n\"RMSE: #{Nx.to_number(rmse)}\"\n```","ref":"cv_gradient_boosting_tree.html#fine-tuning-our-model-with-grid-search"},{"type":"extras","title":"k-means clustering","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# k-means clustering\n\n```elixir\nMix.install([\n  {:scholar, \"~> 0.2.0\"},\n  {:exla, \"~> 0.6.0\"},\n  {:nx, \"~> 0.6.0\", override: true},\n  {:explorer, \"~> 0.6.1\"},\n  {:stb_image, \"~> 0.6.1\"},\n  {:scidata, \"~> 0.1.10\"},\n  {:req, \"~> 0.3.9\"},\n  {:kino, \"~> 0.10.0\"},\n  {:kino_vega_lite, \"~> 0.1.9\"},\n  {:tucan, \"~> 0.3.0\"}\n])\n```","ref":"k_means.html"},{"type":"extras","title":"Setup - k-means clustering","doc":"This notebook introduces the **KMeans** clustering algorithm. We will explore KMeans in three different use cases. Let's setup some aliases:\n\n```elixir\nalias Scholar.Cluster.KMeans\nrequire Explorer.DataFrame, as: DF\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\nkey = Nx.Random.key(42)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [0, 42]\n>\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"k_means.html#setup"},{"type":"extras","title":"Iris Dataset - k-means clustering","doc":"In the first example, we will focus on is the **Iris Dataset**. It is one of the most renowned datasets. It consists of 150 records describing three iris species: *Iris Setosa*, *Iris Virginica*, and *Iris Versicolor*. Our task will be to predict the species of given flowers.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nFirstly, we load the data, then we split it into Training Data (x) and Target (y) and cast those into Nx tensors.\n\n```elixir\ndf = Explorer.Datasets.iris()\nx = df |> DF.discard([\"species\"]) |> Nx.stack(axis: 1)\n\ny =\n  df[[\"species\"]]\n  |> DF.dummies([\"species\"])\n  |> Nx.stack(axis: 1)\n  |> Nx.argmax(axis: 1)\n\n{x, y}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [5.1, 3.5, 1.4, 0.2],\n     [4.9, 3.0, 1.4, 0.2],\n     [4.7, 3.2, 1.3, 0.2],\n     [4.6, 3.1, 1.5, 0.2],\n     [5.0, 3.6, 1.4, 0.2],\n     [5.4, 3.9, 1.7, 0.4],\n     [4.6, 3.4, 1.4, 0.3],\n     [5.0, 3.4, 1.5, 0.2],\n     [4.4, 2.9, 1.4, 0.2],\n     [4.9, 3.1, 1.5, 0.1],\n     [5.4, 3.7, 1.5, 0.2],\n     [4.8, 3.4, 1.6, 0.2],\n     [4.8, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ...]\n >}\n```","ref":"k_means.html#iris-dataset"},{"type":"extras","title":"Exploratory Data Analysis - k-means clustering","doc":"<!-- livebook:{\"break_markdown\":true} -->\n\nAn important part of Data Science workflow is so-called **Exploratory Data Analysis**. EDA helps us understand the data in a better way and suggests some efficient strategies to solve problems. There is no one specific course of action which defines good EDA. It should contain tabular summaries and plots showing relations between features.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nWe start our EDA by finding the mean values of each feature by species.\n\n```elixir\ngrouped_data = DF.group_by(df, \"species\")\n\nDF.summarise(\n  grouped_data,\n  petal_length: mean(petal_length),\n  petal_width: mean(petal_width),\n  sepal_width: mean(sepal_width),\n  sepal_length: mean(sepal_length)\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Explorer.DataFrame \n```\n\nWe see that `petal_length` and `petal_width` are the most distinguishing features. Let's explore them a little bit more.\n\n```elixir\nTucan.histogram(df, \"petal_length\", color_by: \"species\")\n|> Tucan.facet_by(:column, \"species\")\n|> Tucan.Scale.set_y_domain(0, 55)\n|> Tucan.set_size(200, 200)\n|> Tucan.set_title(\"Histograms of petal_length column by species\", offset: 25, anchor: :middle)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\n```elixir\nTucan.scatter(df, \"petal_length\", \"petal_width\", filled: true, color_by: \"species\")\n|> Tucan.set_size(300, 300)\n|> Tucan.set_title(\"Scatterplot of data samples projected on plane petal_width x petal_length\",\n  offset: 25\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\n```elixir\nTucan.scatter(df, \"petal_length\", \"petal_width\")\n|> Tucan.facet_by(:column, \"species\")\n|> Tucan.set_title(\n  \"Scatterplot of data samples projected on plane petal_width x petal_length by species\",\n  offset: 25\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nNow we have a better understanding of the data. Iris species have different petal widths and petal lengths. Iris *Setosa* has the smallest petal, *Versicolor* is medium size, and *Virginica* has the largest petal. We can ascertain that our analysis is correct and plot the so-called **Elbow plot**. The Elbow plot is a plot which presents Inertia vs the number of clusters. If there is a characteristic elbow, then we have a strong suggestion that the number of clusters is correct. Let's train KMeans models for a different number of clusters from range 1 to 11.\n\n```elixir\nclusterings = 1..11\n\nmodels =\n  for num_clusters <- clusterings do\n    KMeans.fit(x, num_clusters: num_clusters, key: key)\n  end\n\ninertias = for model <- models, do: Nx.to_number(model.inertia)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[680.8244, 152.36870647733906, 78.94084142614602, 57.44028021295475, 46.56163015873016,\n 38.95701115711985, 35.15943976939724, 30.324232174688056, 27.927083333333336, 26.371291306519566,\n 24.004956137000256]\n```\n\n```elixir\nTucan.lineplot([num_clusters: clusterings, inertia: inertias], \"num_clusters\", \"inertia\",\n  x: [type: :nominal, axis: [label_angle: 0]],\n  title: \"Elbow Plot\"\n)\n|> Tucan.Axes.set_xy_titles(\"Number of Clusters\", \"Inertia\")\n|> Tucan.set_size(600, 300)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nAs you can see, we have the elbow when the number of clusters equals three. So this value of the parameter seems to be the best.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nIn order to compare our clustering with the target labels, we need to ensure our clusters are in a matching order.\n\n```elixir\ndefmodule Iris.Clusters do\n  import Nx.Defn\n\n  defn sort_clusters(model) do\n    # We sort clusters by the first coordinate\n    order = Nx.argsort(model.clusters[[.., 0]])\n    labels_maping = Nx.argsort(order)\n\n    %{\n      model\n      | labels: Nx.take(labels_maping, model.labels),\n        clusters: Nx.take(model.clusters, order)\n    }\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Iris.Clusters, <<70, 79, 82, 49, 0, 0, 10, ...>>, true}\n```\n\n```elixir\nbest_model = Enum.at(models, 2)\nbest_model = Iris.Clusters.sort_clusters(best_model)\naccuracy = Scholar.Metrics.Classification.accuracy(best_model.labels, y)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  0.8933333158493042\n>\n```\n\nAccuracy is nearly 90% - that's pretty decent! Let's look at our results plotted on one of the previous plots.\n\n```elixir\ncoords = [\n  cluster_petal_length: best_model.clusters[[.., 2]] |> Nx.to_flat_list(),\n  cluster_petal_width: best_model.clusters[[.., 3]] |> Nx.to_flat_list()\n]\n\nTucan.layers([\n  Tucan.scatter(df, \"petal_length\", \"petal_width\", color_by: \"species\", filled: true),\n  Tucan.scatter(coords, \"cluster_petal_length\", \"cluster_petal_width\",\n    filled: true,\n    point_size: 100,\n    point_color: \"green\"\n  )\n])\n|> Tucan.set_size(300, 300)\n|> Tucan.set_title(\n  \"Scatterplot of data samples projected on plane petal_width x petal_length with calculated centroids\",\n  offset: 25\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nAs we expect 😎\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"k_means.html#exploratory-data-analysis"},{"type":"extras","title":"Clustering of pixel colors - k-means clustering","doc":"The other interesting use case of KMeans clustering is pixel clustering. This technique replaces all pixels with similar colors (similar in terms of euclidean distance between RGB) with a centroid related to them.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nLet us start with loading the referral image.\n\n```elixir\nurl =\n  \"https://pix4free.org/assets/library/2021-01-12/originals/san_francisco_california_golden_gate_bridge_water.jpg\"\n\n%{body: raw_image} = Req.get!(url)\nimage = StbImage.read_binary!(raw_image)\n\n{height, width, _num_channels} = image.shape\nimage = StbImage.resize(image, div(height, 3), div(width, 3))\nshape = image.shape\n\nimage_kino = image |> StbImage.to_binary(:jpg) |> Kino.Image.new(:jpeg)\n```\n\nNow we will try to use only ten colors to represent the same picture.\n\n```elixir\nx = image |> StbImage.to_nx() |> Nx.reshape({:auto, 3})\n\nmodel =\n  KMeans.fit(x,\n    num_clusters: 10,\n    num_runs: 10,\n    max_iterations: 200,\n    key: key\n  )\n\nrepainted_x = Nx.take(model.clusters, model.labels)\n\ntensor_to_image = fn x ->\n  x\n  |> Nx.reshape(shape)\n  |> Nx.round()\n  |> Nx.as_type({:u, 8})\n  |> StbImage.from_nx()\n  |> StbImage.to_binary(:jpg)\n  |> Kino.Image.new(:jpeg)\nend\n\nrepainted_x = tensor_to_image.(repainted_x)\n```\n\nLook that even though we use only ten colors, we can say without any doubt that this is the same image. Let's experiment more deeply. Now we will try 5, 10, 15, 20 and 40 colors and then compare the processed images with the original one.\n\n```elixir\nclusterings = [5, 10, 15, 20, 40]\n\nmodels =\n  for num_clusters <- clusterings do\n    KMeans.fit(x, num_clusters: num_clusters, key: key)\n  end\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      9\n    >,\n    clusters: #Nx.Tensor \n      [\n        [61.85893249511719, 60.239295959472656, 58.86780548095703],\n        [4.204394340515137, 87.57450866699219, 99.723876953125],\n        [136.80120849609375, 136.31080627441406, 133.2758026123047],\n        [8.59874153137207, 138.21804809570312, 150.6464385986328],\n        [212.19105529785156, 194.9630126953125, 186.4161376953125]\n      ]\n    >,\n    inertia: #Nx.Tensor \n      609477056.0\n    >,\n    labels: #Nx.Tensor \n      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      17\n    >,\n    clusters: #Nx.Tensor \n      [\n        [43.8863410949707, 51.81157302856445, 52.23817825317383],\n        [217.59837341308594, 208.80609130859375, 205.50840759277344],\n        [151.6479949951172, 163.68496704101562, 166.06430053710938],\n        [6.749063968658447, 148.70291137695312, 160.81790161132812],\n        [87.08784484863281, 72.5816879272461, 68.8161849975586],\n        [111.68388366699219, 126.14994049072266, 125.45569610595703],\n        [1.9535311460494995, 76.27875518798828, 88.1100845336914],\n        [193.17453002929688, 100.84982299804688, 74.338134765625],\n        [5.854827404022217, 107.62191009521484, 120.50939178466797],\n        [225.98037719726562, 179.45010375976562, 155.50418090820312]\n      ]\n    >,\n    inertia: #Nx.Tensor \n      271123392.0\n    >,\n    labels: #Nx.Tensor \n      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      25\n    >,\n    clusters: #Nx.Tensor \n      [\n        [37.49226379394531, 49.20851516723633, 49.95775604248047],\n        [68.16342163085938, 63.4951057434082, 62.18951416015625],\n        [222.685546875, 213.00254821777344, 209.28672790527344],\n        [1.7072257995605469, 74.72100830078125, 86.64070892333984],\n        [211.2851104736328, 104.6119613647461, 69.65288543701172],\n        [8.460535049438477, 161.30209350585938, 172.6630859375],\n        [168.0978240966797, 155.88272094726562, 149.5994873046875],\n        [80.14752960205078, 115.78617095947266, 120.41868591308594],\n        [107.40535736083984, 78.1260757446289, 71.10165405273438],\n        [233.2224578857422, 185.8236541748047, 160.95843505859375],\n        [1.716829776763916, 128.2496795654297, 141.5491180419922],\n        [118.86845397949219, 156.38539123535156, 165.07196044921875],\n        [175.82142639160156, 184.47825622558594, 189.05616760253906],\n        [132.0281982421875, 125.65474700927734, 118.93122863769531],\n        [4.0000224113464355, 101.1719741821289, 113.9442367553711]\n      ]\n    >,\n    inertia: #Nx.Tensor \n      177419888.0\n    >,\n    labels: #Nx.Tensor \n      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      22\n    >,\n    clusters: #Nx.Tensor \n      [\n        [1.4399189949035645, 68.7367935180664, 80.45700073242188],\n        [212.9929962158203, 166.43873596191406, 144.96112060546875],\n        [7.347901821136475, 162.13064575195312, 173.42539978027344],\n        [2.1622705459594727, 107.44270324707031, 120.4988784790039],\n        [134.57351684570312, 77.13945770263672, 64.61695098876953],\n        [220.81675720214844, 107.48382568359375, 68.0926742553711],\n        [221.37522888183594, 214.39996337890625, 212.31607055664062],\n        [180.4493408203125, 186.41665649414062, 190.1817169189453],\n        [61.370914459228516, 112.97090148925781, 120.70513916015625],\n        [2.7440736293792725, 87.0137710571289, 99.2638931274414],\n        [108.83963012695312, 138.17112731933594, 144.1614227294922],\n        [147.15538024902344, 131.68154907226562, 123.84530639648438],\n        [108.10203552246094, 109.16707611083984, 103.55908203125],\n        [155.11891174316406, 159.10829162597656, 159.69961547851562],\n        [239.6630096435547, 196.75611877441406, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      131876120.0\n    >,\n    labels: #Nx.Tensor \n      [15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      26\n    >,\n    clusters: #Nx.Tensor \n      [\n        [2.26080584526062, 85.23534393310547, 97.10236358642578],\n        [217.07537841796875, 176.8404083251953, 158.10992431640625],\n        [72.31938934326172, 168.48968505859375, 179.66793823242188],\n        [126.88947296142578, 86.30879211425781, 77.26168823242188],\n        [60.02167892456055, 54.381187438964844, 52.95188903808594],\n        [181.08859252929688, 173.72731018066406, 169.16676330566406],\n        [246.27549743652344, 195.74496459960938, 167.2862548828125],\n        [197.31752014160156, 200.42007446289062, 203.8845672607422],\n        [167.11375427246094, 153.47061157226562, 146.3658905029297],\n        [1.486596703529358, 126.65245056152344, 140.52223205566406],\n        [97.79022979736328, 103.88321685791016, 99.8821029663086],\n        [226.1820831298828, 220.26646423339844, 219.80264282226562],\n        [168.97100830078125, 110.16779327392578, 101.84797668457031],\n        [230.60598754882812, 206.01318359375, 192.64845275878906],\n        [5.403233528137207, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      64511020.0\n    >,\n    labels: #Nx.Tensor \n      [23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, ...]\n    >\n  }\n]\n```\n\n```elixir\nimage_boxes =\n  for {model, num_clusters} <- Enum.zip(models, clusterings) do\n    repainted_x = Nx.take(model.clusters, model.labels)\n\n    image_kino = tensor_to_image.(repainted_x)\n\n    Kino.Layout.grid(\n      [Kino.Markdown.new(\"","ref":"k_means.html#clustering-of-pixel-colors"},{"type":"extras","title":"Number of colors: #{num_clusters}\"), image_kino], - k-means clustering","doc":"boxed: true\n    )\n  end\n\nimage_box =\n  Kino.Layout.grid(\n    [Kino.Markdown.new(\"","ref":"k_means.html#number-of-colors-num_clusters-image_kino"},{"type":"extras","title":"Original image\"), image_kino], - k-means clustering","doc":"boxed: true\n  )\n\nKino.Layout.grid(image_boxes ++ [image_box], columns: 2)\n```\n\nLook that even with only five colors can recognize the Golden Gate Bridge in the image. On the other hand, with only 40 colors we keep almost all details except the sky and water surface. Sky and water do not map well because there is a small gradient in changing colors. Pixel clustering is a great way to compress images drastically with small integration in their appearance.\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"k_means.html#original-image-image_kino"},{"type":"extras","title":"Clustering images from Fashion-MNIST - k-means clustering","doc":"The last example is the clustering problem on the Fashion-MNIST Dataset. The dataset consists of 60000 images 28 by 28 pixels of ten different parts of clothing. Let's dive into this clustering problem.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nBefore we start, we define the StratifiedSplit module. The module trims input data and splits it, so the number of samples per class is the same for each.\n\n```elixir\ndefmodule StratifiedSplit do\n  import Nx.Defn\n\n  defn trim_samples(x, labels, opts \\\\ []) do\n    opts = keyword!(opts, [:num_classes, :samples_per_class])\n\n    num_classes = opts[:num_classes]\n    samples_per_class = opts[:samples_per_class]\n\n    membership_mask = Nx.iota({1, num_classes}) == Nx.reshape(labels, {:auto, 1})\n\n    indices =\n      membership_mask\n      |> Nx.argsort(axis: 0, direction: :desc)\n      |> Nx.slice_along_axis(0, samples_per_class, axis: 0)\n      |> Nx.flatten()\n\n    {Nx.take(x, indices), Nx.take(labels, indices)}\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, StratifiedSplit, <<70, 79, 82, 49, 0, 0, 13, ...>>, true}\n```\n\nFirstly, load the data and cast it into Nx tensors.\n\n```elixir\n{image_data, labels_data} = Scidata.FashionMNIST.download()\n\n{images_binary, images_type, images_shape} = image_data\n{num_samples, _num_channels = 1, image_height, image_width} = images_shape\n\nimages =\n  images_binary\n  |> Nx.from_binary(images_type)\n  |> Nx.reshape({num_samples, :auto})\n  |> Nx.divide(255)\n\n{labels_binary, labels_type, _shape} = labels_data\ntarget = Nx.from_binary(labels_binary, labels_type)\n\nnum_classes = 10\nsamples_per_class = 20\n\n{images, target} =\n  StratifiedSplit.trim_samples(images, target,\n    num_classes: num_classes,\n    samples_per_class: samples_per_class\n  )\n\nnum_images = num_classes * samples_per_class\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n200\n```\n\nLet's also define a function that will visualize an image in the tensor format for us.\n\n```elixir\ntensor_to_kino = fn x ->\n  x\n  |> Nx.reshape({image_height, image_width, 1})\n  # Replicate the value into 3 channels for PNG\n  |> Nx.broadcast({image_height, image_width, 3})\n  |> Nx.multiply(255)\n  |> Nx.as_type({:u, 8})\n  |> StbImage.from_nx()\n  |> StbImage.resize(112, 112)\n  |> StbImage.to_binary(:png)\n  |> Kino.Image.new(:png)\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Function<42.3316493/1 in :erl_eval.expr/6>\n```\n\nHere is one of the images.\n\n```elixir\ntensor_to_kino.(images[0])\n```\n\nWe will try some different numbers of clusters and then measure the quality of clustering.\n\n```elixir\nnums_clusters = 2..20\n\nmodels =\n  for num_clusters <- 2..20 do\n    KMeans.fit(images, num_clusters: num_clusters, key: key)\n  end\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      4\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 3.501400715322234e-5, 1.4005602861288935e-4, 3.851540677715093e-4, 3.501400933600962e-4, 7.002801285125315e-4, 4.901961074210703e-4, 0.009278712794184685, 0.04635854437947273, 0.09737396240234375, 0.22310924530029297, 0.29975488781929016, 0.32121849060058594, 0.3055672347545624, 0.3146008551120758, 0.36162465810775757, 0.32324934005737305, 0.3010154068470001, 0.20304621756076813, 0.06995797902345657, 0.019502801820635796, 0.003641456598415971, 0.003641456598415971, 0.0034313725773245096, 0.0025560224894434214, 0.0010504202218726277, 3.501400715322234e-5, 0.0, 0.0, 0.0, 3.501400715322234e-5, 4.901961074210703e-4, 5.252101109363139e-4, 0.006232493091374636, 0.05105042830109596, 0.13872550427913666, 0.2501050531864166, 0.37622547149658203, 0.533753514289856, 0.6370097994804382, 0.7304272055625916, 0.7347339391708374, 0.7232843637466431, 0.7482843399047852, 0.7121148109436035, 0.6053571105003357, 0.5261555314064026, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      10950.6201171875\n    >,\n    labels: #Nx.Tensor \n      [0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      3\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 5.9417710872367024e-5, 1.1883542174473405e-4, 2.376708434894681e-4, 2.376708434894681e-4, 4.753416869789362e-4, 2.376708434894681e-4, 0.013071895577013493, 0.05971479415893555, 0.12269756942987442, 0.27730244398117065, 0.3171122968196869, 0.2941770851612091, 0.2795603275299072, 0.28009507060050964, 0.31200236082077026, 0.2995246648788452, 0.3170528709888458, 0.25864526629447937, 0.07664884626865387, 0.02192513458430767, 1.7825313261710107e-4, 4.1592397610656917e-4, 3.565062361303717e-4, 1.1883542174473405e-4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.376708434894681e-4, 5.9417710872367024e-5, 0.009566251188516617, 0.05246583744883537, 0.11509210616350174, 0.23559121787548065, 0.3770647943019867, 0.5828877091407776, 0.6433154940605164, 0.7103387117385864, 0.7102198004722595, 0.6955437064170837, 0.7320261001586914, 0.675638735294342, 0.6002377271652222, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      9246.3125\n    >,\n    labels: #Nx.Tensor \n      [0, 0, 2, 0, 1, 1, 2, 1, 1, 1, 0, 0, 2, 0, 0, 1, 2, 1, 1, 2, 0, 0, 2, 0, 2, 1, 1, 1, 0, 1, 0, 0, 1, 0, 2, 1, 2, 1, 1, 1, 0, 0, 2, 0, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 1.8239855126012117e-4, 6.383949075825512e-4, 5.471956683322787e-4, 0.0010943913366645575, 7.295942050404847e-4, 0.004012768156826496, 0.02891017124056816, 0.06675787270069122, 0.19589604437351227, 0.31500229239463806, 0.3976288139820099, 0.38148659467697144, 0.3805745542049408, 0.4316461682319641, 0.3498404026031494, 0.2810761630535126, 0.13807569444179535, 0.05654355138540268, 0.008755129761993885, 5.471956101246178e-4, 8.20793560706079e-4, 0.0014591884100809693, 6.383949657902122e-4, 0.0, 9.119927563006058e-5, 0.0, 0.0, 0.0, 9.119927563006058e-5, 9.119927417486906e-4, 0.0012767899315804243, 0.0015503877075389028, 0.05243958532810211, 0.18467853963375092, 0.29092568159103394, 0.4202462434768677, 0.5493844747543335, 0.6998631954193115, 0.8229821920394897, 0.810123085975647, 0.7939808964729309, 0.8111263513565063, 0.7670770883560181, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      8447.5419921875\n    >,\n    labels: #Nx.Tensor \n      [3, 3, 0, 3, 1, 1, 0, 1, 2, 2, 3, 3, 0, 3, 3, 1, 0, 1, 1, 2, 3, 3, 0, 3, 0, 1, 1, 1, 3, 2, 3, 3, 1, 3, 0, 1, 0, 1, 2, 2, 3, 3, 0, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      6\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 1.9607844296842813e-4, 3.9215688593685627e-4, 7.843137718737125e-4, 7.843137718737125e-4, 3.9215688593685627e-4, 3.9215688593685627e-4, 0.036274511367082596, 0.14215686917304993, 0.21843135356903076, 0.31882351636886597, 0.3727450966835022, 0.29549020528793335, 0.2123529464006424, 0.22921571135520935, 0.29098039865493774, 0.3682352900505066, 0.40980392694473267, 0.2998039126396179, 0.1456862837076187, 0.05725490301847458, 1.9607844296842813e-4, 0.001372549100778997, 5.882353289052844e-4, 5.882353289052844e-4, 0.0, 1.9607844296842813e-4, 0.0, 0.0, 0.0, 0.0, 7.843137718737125e-4, 1.9607844296842813e-4, 0.0313725508749485, 0.16862747073173523, 0.32960787415504456, 0.522549033164978, 0.6184313893318176, 0.6737255454063416, 0.7147058844566345, 0.7619606852531433, 0.7511764764785767, 0.6862744688987732, 0.7696077823638916, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      7935.5498046875\n    >,\n    labels: #Nx.Tensor \n      [0, 3, 2, 3, 4, 4, 2, 4, 1, 1, 3, 3, 2, 3, 3, 4, 2, 4, 4, 2, 3, 3, 2, 3, 2, 4, 4, 4, 3, 1, 0, 3, 4, 3, 2, 4, 2, 4, 1, 1, 0, 3, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 1.9607844296842813e-4, 3.9215688593685627e-4, 7.843137718737125e-4, 7.843137718737125e-4, 3.9215688593685627e-4, 3.9215688593685627e-4, 0.036274511367082596, 0.14215686917304993, 0.21843135356903076, 0.31882351636886597, 0.3727450966835022, 0.29549020528793335, 0.2123529464006424, 0.22921571135520935, 0.29098039865493774, 0.3682352900505066, 0.40980392694473267, 0.2998039126396179, 0.1456862837076187, 0.05725490301847458, 1.9607844296842813e-4, 0.001372549100778997, 5.882353289052844e-4, 5.882353289052844e-4, 0.0, 1.9607844296842813e-4, 0.0, 0.0, 0.0, 0.0, 7.843137718737125e-4, 1.9607844296842813e-4, 0.0313725508749485, 0.16862747073173523, 0.32960787415504456, 0.522549033164978, 0.6184313893318176, 0.6737255454063416, 0.7147058844566345, 0.7619606852531433, 0.7511764764785767, 0.6862744688987732, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      7484.12109375\n    >,\n    labels: #Nx.Tensor \n      [0, 3, 2, 3, 4, 4, 2, 5, 1, 1, 3, 3, 2, 3, 3, 4, 2, 4, 4, 1, 3, 3, 2, 3, 2, 5, 4, 5, 3, 1, 0, 3, 4, 3, 2, 4, 2, 5, 1, 1, 0, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 1.9129604334011674e-4, 6.695361225865781e-4, 5.738881300203502e-4, 0.0011477762600407004, 5.738881300203502e-4, 0.004208512604236603, 0.03032042272388935, 0.0700143575668335, 0.20545193552970886, 0.3070301413536072, 0.3960784375667572, 0.3846963346004486, 0.38326162099838257, 0.43538981676101685, 0.3438546359539032, 0.2696317434310913, 0.12692491710186005, 0.043424203991889954, 0.009182210080325603, 5.738881300203502e-4, 7.65184173360467e-4, 0.001434720354154706, 6.69536180794239e-4, 0.0, 9.564802167005837e-5, 0.0, 0.0, 0.0, 9.564802167005837e-5, 9.564801584929228e-4, 0.001339072361588478, 0.0016260163392871618, 0.05499761179089546, 0.19368726015090942, 0.298995703458786, 0.41300809383392334, 0.5291248559951782, 0.6889526844024658, 0.8161645531654358, 0.8013391494750977, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      7091.55810546875\n    >,\n    labels: #Nx.Tensor \n      [2, 1, 0, 1, 6, 6, 0, 3, 4, 5, 1, 1, 0, 2, 1, 6, 0, 3, 6, 5, 2, 1, 0, 2, 0, 3, 6, 3, 1, 4, 2, 1, 6, 1, 0, 3, 0, 3, 4, 4, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      6868.54296875\n    >,\n    labels: #Nx.Tensor \n      [3, 4, 1, 6, 6, 6, 1, 7, 0, 5, 6, 4, 1, 4, 2, 6, 2, 7, 6, 5, 4, 4, 1, 4, 2, 7, 6, 7, 6, 0, 3, 4, 6, 6, 2, 7, 1, 7, 0, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      6\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      6582.2734375\n    >,\n    labels: #Nx.Tensor \n      [3, 4, 1, 6, 6, 8, 1, 7, 0, 5, 6, 4, 1, 4, 2, 8, 2, 7, 6, 5, 4, 4, 1, 4, 2, 8, 8, 7, 6, 0, 3, 4, 6, 6, 2, 8, 1, 7, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      6\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      6426.0517578125\n    >,\n    labels: #Nx.Tensor \n      [3, 4, 9, 6, 6, 8, 9, 7, 0, 5, 6, 4, 1, 4, 2, 8, 9, 7, 2, 5, 4, 4, 9, 4, 9, 8, 6, 7, 2, 0, 3, 4, 6, 6, 2, 8, 9, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      7\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      6238.46923828125\n    >,\n    labels: #Nx.Tensor \n      [3, 4, 9, 6, 6, 8, 9, 7, 0, 5, 6, 4, 1, 4, 2, 8, 9, 7, 2, 5, 4, 10, 9, 4, 9, 8, 6, 7, 2, 0, 3, 10, 6, 6, 2, 8, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 2.3068052541930228e-4, 0.0, 2.3068052541930228e-4, 0.06805074959993362, 0.1568627506494522, 0.17716263234615326, 0.17923875153064728, 0.18846596777439117, 0.18685120344161987, 0.18362168967723846, 0.22260668873786926, 0.21245676279067993, 0.20484431087970734, 0.21453288197517395, 0.07381777465343475, 0.017070358619093895, 2.3068052541930228e-4, 6.920415908098221e-4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.3068052541930228e-4, 6.920415326021612e-4, 0.04013840854167938, 0.30542102456092834, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      6081.9736328125\n    >,\n    labels: #Nx.Tensor \n      [7, 3, 2, 3, 5, 5, 6, 1, 8, 4, 5, 0, 2, 0, 3, 10, 6, 1, 5, 4, 0, 11, 2, 0, 2, 1, 5, 1, 9, 10, 7, 11, 3, 9, 3, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 4.6136105083860457e-4, 0.011995386332273483, 0.012687427923083305, 0.012456747703254223, 0.012456747703254223, 0.012456747703254223, 0.011072664521634579, 0.013148789294064045, 0.01361014973372221, 0.012687427923083305, 0.014994233846664429, 0.06528258323669434, 0.058362167328596115, 0.059284891933202744, 0.06920415163040161, 0.044290658086538315, 0.013840830884873867, 0.01407151110470295, 0.015224914066493511, 0.0117647061124444, 0.013840830884873867, 0.01568627543747425, 0.014532871544361115, 0.014763553626835346, 0.003921568859368563, 0.0, 0.0, 0.0, 0.013148789294064045, 0.024221453815698624, 0.02283737063407898, 0.025836216285824776, 0.022606689482927322, 0.021453287452459335, 0.018223760649561882, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5876.912109375\n    >,\n    labels: #Nx.Tensor \n      [8, 10, 2, 1, 1, 5, 2, 3, 4, 12, 1, 10, 2, 10, 11, 5, 11, 0, 6, 7, 10, 10, 2, 10, 2, 0, 0, 3, 1, 12, 8, 10, 1, 1, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 4.6136105083860457e-4, 0.011995386332273483, 0.012687427923083305, 0.012456747703254223, 0.012456747703254223, 0.012456747703254223, 0.011072664521634579, 0.013148789294064045, 0.01361014973372221, 0.012687427923083305, 0.014994233846664429, 0.06528258323669434, 0.058362167328596115, 0.059284891933202744, 0.06920415163040161, 0.044290658086538315, 0.013840830884873867, 0.01407151110470295, 0.015224914066493511, 0.0117647061124444, 0.013840830884873867, 0.01568627543747425, 0.014532871544361115, 0.014763553626835346, 0.003921568859368563, 0.0, 0.0, 0.0, 0.013148789294064045, 0.024221453815698624, 0.02283737063407898, 0.025836216285824776, 0.022606689482927322, 0.021453287452459335, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5797.251953125\n    >,\n    labels: #Nx.Tensor \n      [8, 10, 13, 1, 1, 5, 2, 3, 4, 12, 1, 10, 13, 10, 11, 5, 11, 0, 6, 7, 10, 10, 13, 10, 13, 0, 0, 3, 1, 12, 8, 10, 1, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      4\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 2.4509805371053517e-4, 2.4509805371053517e-4, 0.0, 2.4509805371053517e-4, 4.901961074210703e-4, 0.001470588380470872, 4.901961074210703e-4, 0.02401961013674736, 0.05441176891326904, 0.1200980469584465, 0.23112745583057404, 0.2237745076417923, 0.26225489377975464, 0.2404411882162094, 0.20392157137393951, 0.15514707565307617, 0.037254903465509415, 0.030637256801128387, 0.02549019642174244, 0.035784315317869186, 0.0416666679084301, 0.0365196093916893, 0.008333333767950535, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.4509805371053517e-4, 2.4509805371053517e-4, 2.4509805371053517e-4, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5662.34521484375\n    >,\n    labels: #Nx.Tensor \n      [8, 11, 2, 5, 0, 0, 2, 10, 13, 6, 5, 11, 14, 12, 5, 9, 2, 10, 0, 4, 12, 11, 14, 12, 1, 7, 0, 10, 5, 6, 8, 11, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 2.4509805371053517e-4, 2.4509805371053517e-4, 0.0, 2.4509805371053517e-4, 4.901961074210703e-4, 0.001470588380470872, 4.901961074210703e-4, 0.02401961013674736, 0.05441176891326904, 0.1200980469584465, 0.23112745583057404, 0.2237745076417923, 0.26225489377975464, 0.2404411882162094, 0.20392157137393951, 0.15514707565307617, 0.037254903465509415, 0.030637256801128387, 0.02549019642174244, 0.035784315317869186, 0.0416666679084301, 0.0365196093916893, 0.008333333767950535, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.4509805371053517e-4, 2.4509805371053517e-4, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5568.4345703125\n    >,\n    labels: #Nx.Tensor \n      [8, 11, 2, 5, 0, 0, 2, 10, 13, 6, 5, 11, 14, 12, 5, 9, 2, 10, 0, 4, 12, 11, 14, 15, 1, 7, 0, 10, 5, 6, 8, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 2.801120572257787e-4, 2.801120572257787e-4, 0.0, 2.801120572257787e-4, 5.602241144515574e-4, 0.001680672401562333, 5.602241144515574e-4, 0.027450982481241226, 0.06218487769365311, 0.13613446056842804, 0.2641456723213196, 0.24537815153598785, 0.24425771832466125, 0.2731092572212219, 0.23305322229862213, 0.17731094360351562, 0.04257703199982643, 0.020168067887425423, 5.602241144515574e-4, 5.602241144515574e-4, 2.801120572257787e-4, 0.0, 2.801120572257787e-4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.801120572257787e-4, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5398.72412109375\n    >,\n    labels: #Nx.Tensor \n      [8, 11, 2, 5, 0, 0, 2, 10, 13, 6, 5, 11, 14, 12, 5, 9, 2, 10, 16, 4, 12, 11, 14, 15, 1, 7, 0, 10, 5, 6, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 2.801120572257787e-4, 2.801120572257787e-4, 0.0, 2.801120572257787e-4, 5.602241144515574e-4, 0.001680672401562333, 5.602241144515574e-4, 0.027450982481241226, 0.06218487769365311, 0.13613446056842804, 0.2641456723213196, 0.24537815153598785, 0.24425771832466125, 0.2731092572212219, 0.23305322229862213, 0.17731094360351562, 0.04257703199982643, 0.020168067887425423, 5.602241144515574e-4, 5.602241144515574e-4, 2.801120572257787e-4, 0.0, 2.801120572257787e-4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5336.95263671875\n    >,\n    labels: #Nx.Tensor \n      [8, 11, 2, 5, 0, 0, 2, 10, 17, 6, 5, 11, 14, 12, 5, 9, 2, 10, 16, 4, 12, 11, 14, 15, 1, 7, 0, 10, 5, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 5.22875867318362e-4, 0.001568627660162747, 5.22875867318362e-4, 0.0313725508749485, 0.08000000566244125, 0.10849674046039581, 0.24653595685958862, 0.22901961207389832, 0.22797387838363647, 0.2549019753932953, 0.20392157137393951, 0.15973857045173645, 0.062745101749897, 0.033986929804086685, 0.02718954347074032, 0.0376470610499382, 0.04418300837278366, 0.038954250514507294, 0.00862745102494955, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5233.1396484375\n    >,\n    labels: #Nx.Tensor \n      [8, 11, 2, 5, 0, 0, 2, 10, 17, 6, 0, 11, 14, 12, 18, 9, 2, 10, 18, 4, 12, 11, 14, 15, 1, 7, 0, 10, ...]\n    >\n  },\n  %Scholar.Cluster.KMeans{\n    num_iterations: #Nx.Tensor \n      5\n    >,\n    clusters: #Nx.Tensor \n      [\n        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 5.22875867318362e-4, 0.001568627660162747, 5.22875867318362e-4, 0.0313725508749485, 0.08000000566244125, 0.10849674046039581, 0.24653595685958862, 0.22901961207389832, 0.22797387838363647, 0.2549019753932953, 0.20392157137393951, 0.15973857045173645, 0.062745101749897, 0.020130719989538193, 5.22875867318362e-4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ...],\n        ...\n      ]\n    >,\n    inertia: #Nx.Tensor \n      5173.603515625\n    >,\n    labels: #Nx.Tensor \n      [8, 19, 2, 5, 0, 0, 2, 10, 17, 6, 0, 11, 14, 12, 18, 9, 2, 10, 18, 4, 12, 11, 14, 15, 1, 7, 0, ...]\n    >\n  }\n]\n```\n\n```elixir\ndata = [\n  num_clusters: nums_clusters,\n  inertia: for(model <- models, do: Nx.to_number(model.inertia))\n]\n\nTucan.lineplot(data, \"num_clusters\", \"inertia\",\n  x: [type: :ordinal, axis: [label_angle: 0]],\n  width: 600,\n  height: 300\n)\n|> Tucan.Axes.set_xy_titles(\"Number of Clusters\", \"Inertia\")\n|> Tucan.Scale.set_y_domain(4800, 11500)\n|> Tucan.set_title(\"Elbow Plot\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nLook that this time there is no elbow on a plot. We need to use a different method to predict the number of classes. Now we will use *Silhouette Score*. It is a metric that indicates the quality of clustering. The higher score we achieve, the better clustering we get. However, we should be aware that Silhouette Score is just a heuristic and not always works.\n\n```elixir\nsilhouette_scores =\n  for {model, num_clusters} <- Enum.zip(models, nums_clusters) do\n    Scholar.Metrics.Clustering.silhouette_score(images, model.labels, num_clusters: num_clusters)\n    |> Nx.to_number()\n  end\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[0.1867797076702118, 0.19426067173480988, 0.18798942863941193, 0.16196762025356293,\n 0.14662104845046997, 0.15014168620109558, 0.1334874927997589, 0.12096332758665085,\n 0.12907366454601288, 0.12029680609703064, 0.12559780478477478, 0.12784500420093536,\n 0.12478214502334595, 0.1155780702829361, 0.11121585220098495, 0.11069852113723755,\n 0.10738851875066757, 0.10977344214916229, 0.10331624001264572]\n```\n\n```elixir\ndata = [num_clusters: nums_clusters, silhouette_scores: silhouette_scores]\n\nTucan.lineplot(data, \"num_clusters\", \"silhouette_scores\",\n  points: true,\n  point_color: \"darkBlue\",\n  x: [type: :ordinal, axis: [label_angle: 0]]\n)\n|> Tucan.Axes.set_xy_titles(\"Number of Clusters\", \"Silhouette score\")\n|> Tucan.Scale.set_y_domain(0.088, 0.205)\n|> Tucan.set_size(600, 300)\n|> Tucan.set_title(\"Silhouette score vs Number of Clusters\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nAs we can see, the model with num_clusters equal to 3 has the highest Silhouette Score. Now we will visualize this clusterization.\n\n```elixir\nbest_num_clusters = 3\nbest_model = Enum.at(models, 1)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Scholar.Cluster.KMeans{\n  num_iterations: #Nx.Tensor \n    3\n  >,\n  clusters: #Nx.Tensor \n    [\n      [0.0, 5.9417710872367024e-5, 1.1883542174473405e-4, 2.376708434894681e-4, 2.376708434894681e-4, 4.753416869789362e-4, 2.376708434894681e-4, 0.013071895577013493, 0.05971479415893555, 0.12269756942987442, 0.27730244398117065, 0.3171122968196869, 0.2941770851612091, 0.2795603275299072, 0.28009507060050964, 0.31200236082077026, 0.2995246648788452, 0.3170528709888458, 0.25864526629447937, 0.07664884626865387, 0.02192513458430767, 1.7825313261710107e-4, 4.1592397610656917e-4, 3.565062361303717e-4, 1.1883542174473405e-4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.376708434894681e-4, 5.9417710872367024e-5, 0.009566251188516617, 0.05246583744883537, 0.11509210616350174, 0.23559121787548065, 0.3770647943019867, 0.5828877091407776, 0.6433154940605164, 0.7103387117385864, 0.7102198004722595, 0.6955437064170837, 0.7320261001586914, 0.675638735294342, 0.6002377271652222, 0.5515151619911194, 0.35151517391204834, ...],\n      ...\n    ]\n  >,\n  inertia: #Nx.Tensor \n    9246.3125\n  >,\n  labels: #Nx.Tensor \n    [0, 0, 2, 0, 1, 1, 2, 1, 1, 1, 0, 0, 2, 0, 0, 1, 2, 1, 1, 2, 0, 0, 2, 0, 2, 1, 1, 1, 0, 1, 0, 0, 1, 0, 2, 1, 2, 1, 1, 1, 0, 0, 2, 0, 2, 1, ...]\n  >\n}\n```\n\n```elixir\npredicted_cluster_with_indices =\n  best_model.labels\n  |> Nx.to_flat_list()\n  |> Enum.with_index()\n  |> Enum.group_by(&elem(&1, 0), &elem(&1, 1))\n\nfor cluster <- 0..(best_num_clusters - 1) do\n  indices = predicted_cluster_with_indices[cluster]\n\n  boxes =\n    for index <- indices do\n      original_cluster = Nx.to_number(target[index])\n\n      Kino.Layout.grid([\n        Kino.Markdown.new(\"Original cluster: #{original_cluster}\"),\n        tensor_to_kino.(images[index])\n      ])\n    end\n\n  Kino.Layout.grid(\n    [\n      Kino.Markdown.new(\"","ref":"k_means.html#clustering-images-from-fashion-mnist"},{"type":"extras","title":"Cluster #{cluster}\"), - k-means clustering","doc":"Kino.Layout.grid(boxes, columns: 5)\n    ],\n    boxed: true\n  )\nend\n|> Kino.Layout.grid()\n```\n\nOops, it doesn't look right! That's because our algorithm for three clusters gathers images by colors rather than shapes. To spot this, let's plot the average image of each cluster.\n\n```elixir\nfor cluster <- 0..(best_num_clusters - 1) do\n  indices = predicted_cluster_with_indices[cluster]\n\n  mean_image =\n    indices\n    |> Enum.map(&images[&1])\n    |> Nx.stack()\n    |> Nx.mean(axes: [0])\n\n  tensor_to_kino.(mean_image)\nend\n|> Kino.Layout.grid(columns: 3)\n```\n\nOne of the images has a vertical line (something like trousers), the next image is almost all white (similar to a jumper), and the last one is mostly black. This time Silhouette Score turns out to be not the best indicator. To get better clustering, try to rerun the code with a higher number of clusters.","ref":"k_means.html#cluster-cluster"},{"type":"extras","title":"k-nearest neighbors (KNN)","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# k-nearest neighbors (KNN)\n\n```elixir\nMix.install([\n  {:scholar, \"~> 0.3.0\"},\n  {:explorer, \"~> 0.8.2\", override: true},\n  {:exla, \"~> 0.7.2\"},\n  {:nx, \"~> 0.7.2\"},\n  {:req, \"~> 0.4.14\"},\n  {:kino_vega_lite, \"~> 0.1.11\"},\n  {:kino, \"~> 0.12.3\"},\n  {:kino_explorer, \"~> 0.1.18\"},\n  {:tucan, \"~> 0.3.1\"}\n])\n```","ref":"k_nearest_neighbors.html"},{"type":"extras","title":"Setup - k-nearest neighbors (KNN)","doc":"We will use VegaLite, Explorer, and Scholar throughout this guide, so let's define some aliases:\n\n```elixir\nrequire Explorer.DataFrame, as: DF\nrequire Explorer.Series, as: S\nrequire Explorer.Query, as: Q\nalias Scholar.Neighbors.{KNNClassifier, KNNRegressor, BruteKNN}\nalias Scholar.Metrics.{Classification, Regression}\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\nseed = 42\nkey = Nx.Random.key(42)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [0, 42]\n>\n```","ref":"k_nearest_neighbors.html#setup"},{"type":"extras","title":"Use cases - k-nearest neighbors (KNN)","doc":"This notebook will cover the three primary applications of k-nearest Neighbors: classification, regression, and anomaly detection. Let's get started with a practical example. Imagine just moved to a new city, and you're here for the first time. Since you're an active person, you'd like to find a nearby gym with good facilities. What would you do? You'd probably start by searching for gyms on online maps. The search results might look something like this:\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n![](files/knn_gyms.png)\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nNow you can check out the gyms and eventually decide which one will be your regular spot. What did the search engine do? It calculated the nearest gyms (nearest neighbors) from your current location, which is the essence of finding nearest neighbors.\n\nNow let's move to a more abstract example. You're listening to your favorite rock playlist and you think, \"Yeah, these tracks are cool, but I've been listening to them on repeat. Maybe I should explore some new music.\" Searching for random songs might not be the most effective approach; you could end up with hip-hop or pop tracks, which you may not enjoy as much. However, it might also lead you to discover entirely new genres. A better approach could be to explore other rock playlists available online. While these playlists align with your preferred genre, they may not consider your unique tastes within rock music. Wouldn't it be great if there were a tool that could recommend new songs based on your previous playlists? Fortunately, such tools exist!\n\nOne type of [recommendation system](https://en.wikipedia.org/wiki/Recommender_system) relies on collaborative filtering: it recommends songs based on what other users with similar musical tastes (aka its neighbours) listen to. Another approach is to treat songs as points and then compute the closest songs to your favorites. Part of the challenge in solving these problems is how to model users and songs as points in space however, once that is done, KNN algorithms play an essential role in understanding the relationships between them. So let's take a look at some concrete examples.\n\n<!-- livebook:{\"branch_parent_index\":1} -->","ref":"k_nearest_neighbors.html#use-cases"},{"type":"extras","title":"Classification - k-nearest neighbors (KNN)","doc":"The process of classification using KNN is relatively straightforward. We start by working with a dataset where each sample is assigned a label. To predict the label of a new sample, we compute the distance between it and all the other samples in the dataset. Next, we select only the $k$ closest samples based on the distance metric, where $k$ is a user-defined parameter. We then examine the labels of these $k$ samples and choose the label that appears most frequently. This label is assigned as the predicted label for the new sample.\n\nLet's first grasp how kNN works visually with a small example.\n\n```elixir\ndata =\n  DF.new(\n    x: [-1, 0.2, -0.5, -2.1, -2.3, -2.2, 0.1, 0.3, 0.4, 0.7, 1.3],\n    y: [-0.1, 0.4, 0.5, 0.4, 1.1, -1.0, -0.1, 0.2, 1.5, 1.6, 0.9],\n    label: [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1]\n  )\n\npoint_to_predict = DF.new(x: [0.0], y: [0.0])\n\nk_eq_3 = DF.new(x: [0.45], y: [0.45], name: [\"k = 3\"])\nk_eq_5 = DF.new(x: [0.72], y: [0.72], name: [\"k = 5\"])\n\nradius = fn x, y -> :math.sqrt(x ** 2 + y ** 2) end\n\nTucan.layers([\n  Tucan.scatter(data, \"x\", \"y\",\n    point_size: 200,\n    filled: true,\n    shape_by: \"label\",\n    color_by: \"label\"\n  )\n  |> Tucan.Scale.set_color_scheme(:plasma),\n  Tucan.scatter(point_to_predict, \"x\", \"y\",\n    point_size: 400,\n    point_color: \"green\",\n    point_shape: \"triangle-up\",\n    filled: true\n  ),\n  Tucan.Geometry.circle({0, 0}, radius.(0.2, 0.4), line_color: \"brown\", stroke_width: 2),\n  Tucan.Geometry.circle({0, 0}, radius.(-1, 0.1), line_color: \"blue\", stroke_width: 2),\n  Tucan.annotate(Tucan.new(), 0.42, 0.42, \"k = 3\", color: \"brown\", size: 20),\n  Tucan.annotate(Tucan.new(), 0.81, 0.81, \"k = 5\", color: \"blue\", size: 20)\n])\n|> Tucan.Scale.set_xy_domain(-2.4, 1.7)\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Scatterplot showing KNN prediction process\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nNote that the final prediction may vary depending on the value of $k$. For example, with $k = 3$, we would predict the green triangle as an orange square (1), whereas with $k = 5$, the purple circle (0) would be the final prediction.\n\nLet's now test this using Scholar code. First, we will define our data.\n\n```elixir\nx = Nx.stack(DF.discard(data, \"label\"), axis: 1)\nlabels = Nx.stack(DF.select(data, \"label\"), axis: 1) |> Nx.squeeze(axes: [1])\nx_pred = Nx.stack(point_to_predict, axis: 1)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [0.0, 0.0]\n  ]\n>\n```\n\nLet's now try with $k = 3$.\n\n```elixir\nmodel = KNNClassifier.fit(x, labels, num_classes: 2, num_neighbors: 3, algorithm: :brute)\nKNNClassifier.predict(model, x_pred)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [1]\n>\n```\n\nAnd $k = 5$.\n\n```elixir\nmodel = KNNClassifier.fit(x, labels, num_classes: 2, num_neighbors: 5, algorithm: :brute)\nKNNClassifier.predict(model, x_pred)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [0]\n>\n```\n\nAs we can see, the predictions match our intuition from analyzing the plot.\n\nNow, let's try KNN on a more complicated dataset: the Wine dataset. Our task will be to classify the quality of the wine. Before we load the dataset into `Explorer.DataFrame` for more efficient exploration, let's check some more detailed information about the dataset.\n\n```elixir\ninfo =\n  Req.get!(\n    \"https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality.names\"\n  ).body\n\nKino.Markdown.new(info)\n```\n\n```elixir\ndata =\n  Req.get!(\n    \"https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-white.csv\"\n  ).body\n\ndf_data = DF.load_csv!(data, delimiter: \";\", dtypes: %{\"total sulfur dioxide\": :float})\ntensor_data = Nx.stack(df_data, axis: 1)\ndf_data\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nAs we can see, there are no null values in the dataset. Now, let's check the size of the dataset.\n\n```elixir\nDF.shape(df_data)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{4898, 12}\n```\n\nLet's check some statistical properties of the dataset. We will start with *skewness*, which measures the asymmetry of the probability distribution of a random variable about its mean. To better understand this concept, please take a look at the picture below.\n\n| ![Skewness](https://upload.wikimedia.org/wikipedia/commons/c/cc/Relationship_between_mean_and_median_under_different_skewness.png) |\n| :--------------------------------------------------------------------------------------------------------------------------------: |\n| Figure 1: A general relationship of mean and median under differently skewed unimodal distribution                                 |\n\nNow, let's check the skewness of our dataset using `Scholar.Stats.skew/1` function.\n\n```elixir\nScholar.Stats.skew(tensor_data)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [0.6475530855160632, 1.5764965159574844, 1.2815277799152376, 1.0767638711454446, 5.0217921696710315, 1.4063140718346212, 0.3905901775815236, 0.9774735389046988, 0.45764233925379794, 0.9768943947733456, 0.4871927332763434, 0.15574868141362447]\n>\n```\n\nAs we can see, all features have positive skewness, which means that their distributions are more similar to the left plot in the picture above.\n\nMoving on to another statistical function, let's discuss *kurtosis*. Kurtosis measures how much data is located in the tails of distributions. If the kurtosis is greater than 0, the distribution is said to be \"platykurtic\", indicating that it has more extreme values than a univariate normal distribution. Similarly, a \"leptokurtic\" distribution has positive kurtosis and less extreme values, while a \"mesokurtic\" distribution has the same kurtosis as a normal distribution. Let's check the kurtosis of our dataset.\n\n| ![Skewness](https://vitalflux.com/wp-content/uploads/2023/01/fourth-moment-kurtosis-640x334.png) |\n| :----------------------------------------------------------------------------------------------: |\n| Figure 2: Plot showing Platykurtic, Mesokurtic and Leptokurtic distributions                     |\n\n```elixir\nScholar.Stats.kurtosis(tensor_data)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [2.168736944824719, 5.08520490451785, 6.167374226819426, 3.4650542966046363, 37.52503905008619, 11.453415905047144, 0.5700448984658735, 9.78258726703508, 0.5290085383907339, 1.5880812942840778, -0.6989373013774784, 0.21508011570192975]\n>\n```\n\nAlmost all features have positive kurtosis (they are leptokurtic). `alcohol` has negative excess kurtosis, which means it is platykurtic. Below there is Kernel Density Estimate (KDE) to check how exactly this tail look like.\n\n```elixir\n# Increase the sample size (or use 1.0 to plot all data)\nsample = DF.sample(df_data, 0.5, seed: seed)\n\nTucan.hconcat([\n  Tucan.density(sample, \"alcohol\", only: [\"alcohol\"], fill_color: \"lightblue\")\n  |> Tucan.Axes.set_x_title(\"% of alcohol\")\n  |> Tucan.set_size(350, 300)\n  |> Tucan.set_title(\"KDE plot of alcohol feature\", offset: 20),\n  Tucan.density(sample, \"pH\", only: [\"pH\"], fill_color: \"lightgreen\")\n  |> Tucan.set_size(350, 300)\n  |> Tucan.set_title(\"KDE plot of pH feature\", offset: 20)\n])\n|> Tucan.set_title(\"KDE plots of Features\", anchor: :middle, offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nThe previous analysis indicates that the alcohol feature has more extreme values than pH. Now, we will create a correlation heatmap to investigate the relationships between the features in the dataset.\n\n```elixir\ncorrelation = Scholar.Stats.correlation_matrix(tensor_data, ddof: 1)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [1.000204206657137, -0.022701925084394028, 0.28923975031725996, 0.08903887998201279, 0.02309035789846668, -0.04940594604443256, 0.09108835320911231, 0.26538519619855744, -0.42594525408939937, -0.017146485732801857, -0.12090580792461156, -0.1136860414197186],\n    [-0.022701925084394028, 1.0002042066571377, -0.14950233378736394, 0.06429918773242695, 0.070525970411686, -0.0970317497617824, 0.08927873114082308, 0.027119382290178354, -0.03192188560415534, -0.035735442831260074, 0.06773177122976645, -0.19476273264768545],\n    [0.28923975031725996, -0.14950233378736394, 1.000204206657138, 0.09423086293135367, 0.11438780236851878, 0.09409643220951228, 0.12115553337968561, 0.14953309999019224, -0.16378164987548596, 0.062343668678184735, -0.07574419436799362, -0.009210971441640103],\n    [0.08903887998201279, 0.06429918773242695, 0.09423086293135367, 1.0002042066571373, 0.0887026459138499, 0.29915943156848096, 0.40152128773970025, 0.8391377774397895, -0.1941730973339338, -0.026669810894874028, -0.4507232439271815, -0.09759675473273582],\n    [0.02309035789846668, 0.070525970411686, ...],\n    ...\n  ]\n>\n```\n\n```elixir\n{corr_size, _} = Nx.shape(correlation)\ncorrelation_list = Nx.to_flat_list(correlation)\ncolumn_names = DF.names(df_data)\n\ncorr_to_plot =\n  DF.new(\n    x: List.flatten(List.duplicate(column_names, corr_size)),\n    y: List.flatten(for name <- column_names, do: List.duplicate(name, corr_size)),\n    corr_val: Enum.map(correlation_list, fn x -> Float.round(x, 2) end)\n  )\n\nTucan.heatmap(corr_to_plot, \"x\", \"y\", \"corr_val\", annotate: true, text_color: [{nil, 0, \"white\"}])\n|> Tucan.Scale.set_color_scheme(:inferno)\n|> Tucan.Axes.set_title(:color, \"Correlation\")\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Correlation Matrix for California Housing\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nWe can observe from the correlation heatmap that there is no strong correlation between the quality of wine and any single feature. However, we can see that there are two strong correlations: density is proportional to the amount of residual sugar and inversely proportional to the amount of alcohol.\n\nGiven that citric acid, free sulfur dioxide, and sulphates are not strongly correlated with the quality of wine or any other features, we can drop them from the dataset to simplify our analysis.\n\nBefore dropping the features, it is important to consider their potential impact on the final prediction. Even if they are not strongly correlated with the target variable (quality), they may still provide useful information for the KNN algorithm. It is recommended to experiment with different feature subsets to find the optimal combination for the given task. You can try running the notebook again without dropping these features.\n\n```elixir\ntensor_data =\n  DF.discard(df_data, [\"citric acid\", \"free sulfur dioxide\", \"sulphates\"]) |> Nx.stack(axis: 1)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [7.0, 0.27, 20.7, 0.045, 170.0, 1.001, 3.0, 8.8, 6.0],\n    [6.3, 0.3, 1.6, 0.049, 132.0, 0.994, 3.3, 9.5, 6.0],\n    [8.1, 0.28, 6.9, 0.05, 97.0, 0.9951, 3.26, 10.1, 6.0],\n    [7.2, 0.23, 8.5, 0.058, 186.0, 0.9956, 3.19, 9.9, 6.0],\n    [7.2, 0.23, 8.5, 0.058, 186.0, 0.9956, 3.19, 9.9, 6.0],\n    [8.1, 0.28, 6.9, 0.05, 97.0, ...],\n    ...\n  ]\n>\n```\n\nThe next step will be splitting the dataset into features and labels. To make the classification task easier, we will convert the labels into three categories: \"poor,\" \"medium,\" and \"excellent\". Additionally, we will normalize the features using the standard scaler.\n\n```elixir\nx = tensor_data[[.., 0..-2//1]]\ny = tensor_data[[.., -1]]\n\n# convert quality into labels quality: (0, 4) poor => 0, (5, 6) medium => 1, (7, 10) excelent => 2\n\npoor_and_better_mask = Nx.greater(y, 4)\nmedium_and_better_mask = Nx.greater(y, 6)\n\ny = Nx.select(poor_and_better_mask, Nx.select(medium_and_better_mask, 2, 1), 0)\nx = Scholar.Preprocessing.standard_scale(x)\n\n{x, y}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [-0.29385297541380284, -0.4368649169854585, -0.0027291121847267076, -0.44164614831586296, 3.1698834995014824, -0.4213312276408997, -0.37885264350988346, -0.2556031247705665],\n     [-0.30872791733061694, -0.43622741947473787, -0.40860252734351166, -0.4415611486477669, 2.3623866525887167, -0.4214799770600679, -0.3724776684026774, -0.24072818285375244],\n     [-0.27047806668738067, -0.43665241781521824, -0.29597796711620483, -0.4415398987307429, 1.6186395567480112, -0.42145660215134145, -0.37332766508363824, -0.22797823263934036],\n     [-0.2896029920089988, -0.4377149136664193, -0.2619780998777726, -0.44136989939455074, 3.509882171885805, -0.42144597719282945, -0.3748151592753196, -0.23222821604414437],\n     [-0.2896029920089988, -0.4377149136664193, -0.2619780998777726, -0.44136989939455074, 3.509882171885805, -0.42144597719282945, -0.3748151592753196, -0.23222821604414437],\n     [-0.27047806668738067, -0.43665241781521824, -0.29597796711620483, -0.4415398987307429, 1.6186395567480112, -0.42145660215134145, -0.37332766508363824, -0.22797823263934036],\n     [-0.310852909033019, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2, 1, 2, 1, 1, 2, 2, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 0, 1, ...]\n >}\n```\n\nAfter performing EDA on our wine dataset, we can now train our KNN model and make predictions. However, the question arises as to what value of neighbors should we choose for the KNN model. To answer this question, we will use *K-Fold Cross-Validation* (specifically, Nested Cross-Validation). K-Fold is a procedure performed on the training dataset where we split the data into k smaller subsets (folds). For each tested parameter value, we run k separate trainings and testings, where one fold is the validation (test) dataset, and the other k-1 folds are the training dataset. To make this process clear, there is a GIF below that explains it visually.\n\n| ![KFold](https://upload.wikimedia.org/wikipedia/commons/4/4b/KfoldCV.gif)                                                                                    |\n| :----------------------------------------------------------------------------------------------------------------------------------------------------------: |\n| Figure 3: Illustration of KFold Cross-Validation when n = 12 observations and k = 3. After data is shuffled, a total of 3 models will be trained and tested. |\n\n```elixir\ndefmodule Comb do\n  def combinations([]), do: [[]]\n\n  def combinations([{name, values} | opts]) do\n    for subcombination <- combinations(opts), value <- values do\n      [{name, value} | subcombination]\n    end\n  end\nend\n\nComb.combinations(task: [:classification], num_classes: [3], num_neighbors: Enum.to_list(1..30))\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[\n  [task: :classification, num_classes: 3, num_neighbors: 1],\n  [task: :classification, num_classes: 3, num_neighbors: 2],\n  [task: :classification, num_classes: 3, num_neighbors: 3],\n  [task: :classification, num_classes: 3, num_neighbors: 4],\n  [task: :classification, num_classes: 3, num_neighbors: 5],\n  [task: :classification, num_classes: 3, num_neighbors: 6],\n  [task: :classification, num_classes: 3, num_neighbors: 7],\n  [task: :classification, num_classes: 3, num_neighbors: 8],\n  [task: :classification, num_classes: 3, num_neighbors: 9],\n  [task: :classification, num_classes: 3, num_neighbors: 10],\n  [task: :classification, num_classes: 3, num_neighbors: 11],\n  [task: :classification, num_classes: 3, num_neighbors: 12],\n  [task: :classification, num_classes: 3, num_neighbors: 13],\n  [task: :classification, num_classes: 3, num_neighbors: 14],\n  [task: :classification, num_classes: 3, num_neighbors: 15],\n  [task: :classification, num_classes: 3, num_neighbors: 16],\n  [task: :classification, num_classes: 3, num_neighbors: 17],\n  [task: :classification, num_classes: 3, num_neighbors: 18],\n  [task: :classification, num_classes: 3, num_neighbors: 19],\n  [task: :classification, num_classes: 3, num_neighbors: 20],\n  [task: :classification, num_classes: 3, num_neighbors: 21],\n  [task: :classification, num_classes: 3, num_neighbors: 22],\n  [task: :classification, num_classes: 3, num_neighbors: 23],\n  [task: :classification, num_classes: 3, num_neighbors: 24],\n  [task: :classification, num_classes: 3, num_neighbors: 25],\n  [task: :classification, num_classes: 3, num_neighbors: 26],\n  [task: :classification, num_classes: 3, num_neighbors: 27],\n  [task: :classification, num_classes: 3, num_neighbors: 28],\n  [task: :classification, num_classes: 3, num_neighbors: 29],\n  [task: :classification, num_classes: 3, num_neighbors: 30]\n]\n```\n\nThe code implementing KFold cross-validation is located at the beginning of the notebook.\n\n```elixir\nmin_num_neighbors = 1\nparam_space_size = 30\n\n# K in KFold\nk = 10\ntrain_size = 4000\n\n{indices, _new_key} = Nx.Random.shuffle(key, Nx.iota(Nx.shape(y)))\nx = Nx.take(x, indices)\ny = Nx.take(y, indices)\n{train_data, test_data} = Nx.split(x, train_size)\n{train_labels, test_labels} = Nx.split(y, train_size)\n\nfolding_fun = fn x -> Scholar.ModelSelection.k_fold_split(x, k) end\n\nscoring_fun = fn x, y, opts ->\n  {x_train, x_test} = x\n  {y_train, y_test} = y\n\n  model =\n    KNNClassifier.fit(x_train, y_train,\n      num_classes: opts[:num_classes],\n      num_neighbors: opts[:num_neighbors],\n      algorithm: :brute\n    )\n\n  prediction = KNNClassifier.predict(model, x_test)\n\n  accuracy = Classification.accuracy(y_test, prediction)\n  [accuracy]\nend\n\nres =\n  Scholar.ModelSelection.grid_search(train_data, train_labels, folding_fun, scoring_fun,\n    num_classes: [3],\n    num_neighbors: Enum.to_list(min_num_neighbors..param_space_size)\n  )\n\nres = for(sub_res <- res, do: sub_res[:score]) |> Nx.concatenate() |> Nx.as_type(:f64)\n\nresults =\n  DF.new(accuracy: res, num_neighbors: Nx.add(Nx.iota({param_space_size}), min_num_neighbors))\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nTo potentially improve our KNN model's performance, we can try setting the weights parameter to `:distance` and use weighted version of grid search. Next, we can visualize the results of KFold using a plot.\n\n```elixir\nTucan.lineplot(results, \"num_neighbors\", \"accuracy\",\n  line_color: \"gray\",\n  points: true,\n  stroke_dash: [5, 5]\n)\n|> Tucan.Scale.set_x_domain(1, 31)\n|> Tucan.Scale.set_y_domain(0.7, 0.85)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Accuracy vs Number of Neighbors\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\n```elixir\nbest_neighbors = Nx.add(Nx.argmax(res), min_num_neighbors) |> Nx.to_number()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n1\n```\n\nOk, let's test our best model on test data.\n\n```elixir\nbest_model =\n  KNNClassifier.fit(train_data, train_labels,\n    num_neighbors: best_neighbors,\n    num_classes: 3,\n    algorithm: :brute\n  )\n\nfinal_prediction = KNNClassifier.predict(best_model, test_data)\nClassification.accuracy(test_labels, final_prediction)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  0.7672605514526367\n>\n```\n\nPretty well, now let's move on to a regression task using KNN.\n\n<!-- livebook:{\"branch_parent_index\":1} -->","ref":"k_nearest_neighbors.html#classification"},{"type":"extras","title":"Regression - k-nearest neighbors (KNN)","doc":"KNNs can also be used in regression tasks. We will explore this on a dataset about Airfoil Self-Noise. Here's a brief description of the dataset:\n\n**Data Set Information**:\n\nThe NASA data set comprises different size NACA 0012 airfoils at various wind tunnel speeds and angles of attack. The span of the airfoil and the observer position were the same in all of the experiments.\n\n**Attribute Information**:\n\nThis problem has the following inputs:\n\n1. Frequency, in Hertzs.\n2. Angle of attack, in degrees.\n3. Chord length, in meters.\n4. Free-stream velocity, in meters per second.\n5. Suction side displacement thickness, in meters.\n\nThe only output is:\n\n1. Scaled sound pressure level, in decibels.\n\nDua, D. and Graff, C. (2019). UCI Machine Learning Repository [http://archive.ics.uci.edu/ml]. Irvine, CA: University of California, School of Information and Computer Science.\n\n```elixir\ndataset =\n  Req.get!(\n    \"https://archive.ics.uci.edu/ml/machine-learning-databases/00291/airfoil_self_noise.dat\"\n  ).body\n\ncolumn_names = [\n  \"frequency\",\n  \"angle_of_attack\",\n  \"chord_length\",\n  \"free_stream_velocity\",\n  \"suction_side_displacement_thickness\",\n  \"scaled_sound_pressure_level\"\n]\n\ndataset_df = DF.load_csv!(dataset, header: false, delimiter: \"\\t\") |> DF.rename(column_names)\nlabels = DF.select(dataset_df, \"scaled_sound_pressure_level\")\ndata = DF.discard(dataset_df, \"scaled_sound_pressure_level\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nBelow are the KDE plots of the dataset's features. We will analyze them in the next step.\n\n```elixir\nTucan.concat(\n  [\n    Tucan.density(data, \"frequency\", fill_color: \"lightblue\", only: [\"frequency\"])\n    |> Tucan.Axes.set_x_title(\"Frequency (Hz)\")\n    |> Tucan.set_size(350, 300)\n    |> Tucan.set_title(\"KDE plot of Frequency feature\", offset: 20),\n    Tucan.density(data, \"free_stream_velocity\",\n      fill_color: \"purple\",\n      only: [\"free_stream_velocity\"]\n    )\n    |> Tucan.Axes.set_x_title(\"Angle of attack\")\n    |> Tucan.set_size(350, 300)\n    |> Tucan.set_title(\"KDE plot of Angle of Attack feature\", offset: 20),\n    Tucan.density(data, \"angle_of_attack\", fill_color: \"lightgreen\", only: [\"angle_of_attack\"])\n    |> Tucan.Axes.set_x_title(\"Angle of attack\")\n    |> Tucan.set_size(350, 300)\n    |> Tucan.set_title(\"KDE plot of Angle of Attack feature\", offset: 20),\n    Tucan.density(data, \"suction_side_displacement_thickness\",\n      fill_color: \"red\",\n      only: [\"suction_side_displacement_thickness\"]\n    )\n    |> Tucan.Axes.set_x_title(\"Suction Side Displacement Thickness\")\n    |> Tucan.set_size(350, 300)\n    |> Tucan.set_title(\"KDE plot of Suction Side Displacement\\nThickness feature\", offset: 20)\n  ],\n  columns: 2\n)\n|> Tucan.set_title(\"KDE plots of Features\", anchor: :middle, offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nThe Suction Side Displacement Thickness and Frequency features have asymmetric distributions with many observations having small values. To make these distributions less skewed, we will use the logarithms of these two features.\n\n```elixir\ndata =\n  DF.mutate(data,\n    frequency_log: log(frequency),\n    suction_side_displacement_thickness_log: log(suction_side_displacement_thickness)\n  )\n\ndata = DF.discard(data, [\"frequency\", \"suction_side_displacement_thickness\"])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nScale data and convert them to tensors.\n\n```elixir\nx = Nx.stack(data, axis: 1) |> Scholar.Preprocessing.standard_scale()\ny = Nx.stack(labels, axis: 1)\n{x, y}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.24959453243691035, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.2391306895323175, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.22866684662772468, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.21709086757891163, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.2066270246743188, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.19616318176972597, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.18532568847301975, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.17412335981632007, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.1636595169117272, -0.8410438169469526],\n     [-0.5630552019680845, -0.5487622533323925, 2.780406864058286, -0.15282202361502104, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [\n     [126.201],\n     [125.201],\n     [125.951],\n     [127.591],\n     [127.461],\n     [125.571],\n     [125.201],\n     [123.061],\n     [121.301],\n     [119.541],\n     [117.151],\n     [115.391],\n     [112.241],\n     [108.721],\n     [126.416],\n     [127.696],\n     [128.086],\n     [126.966],\n     [126.086],\n     [126.986],\n     [126.616],\n     [124.106],\n     [123.236],\n     [121.106],\n     [119.606],\n     [117.976],\n     [116.476],\n     [113.076],\n     [111.076],\n     [118.129],\n     [119.319],\n     [122.779],\n     [124.809],\n     [126.959],\n     [128.629],\n     [129.099],\n     [127.899],\n     [125.499],\n     [124.049],\n     [123.689],\n     [121.399],\n     [120.319],\n     [119.229],\n     [117.789],\n     [116.229],\n     [114.779],\n     [112.139],\n     [109.619],\n     ...\n   ]\n >}\n```\n\nIn this section, we will apply a similar KFold procedure as in the classification section to find the best value for the `:num_neighbors` parameter in the KNN regression model.\n\n```elixir\nmin_num_neighbors = 1\nparam_space_size = 30\n\n# K in KFold\nk = 10\ntrain_size = 1100\n\n{indices, _new_key} = Nx.Random.shuffle(key, Nx.iota({Nx.axis_size(y, 0)}))\nx_shuffled = Nx.take(x, indices)\ny_shuffled = Nx.take(y, indices)\n{train_data, test_data} = Nx.split(x_shuffled, train_size)\n{train_labels, test_labels} = Nx.split(y_shuffled, train_size)\n\nfolding_fun = fn x -> Scholar.ModelSelection.k_fold_split(x, k) end\n\nscoring_fun = fn x, y, opts ->\n  {x_train, x_test} = x\n  {y_train, y_test} = y\n\n  model =\n    KNNRegressor.fit(x_train, y_train,\n      num_neighbors: opts[:num_neighbors],\n      algorithm: :brute\n    )\n\n  prediction = KNNRegressor.predict(model, x_test)\n\n  rmse = Regression.mean_square_error(y_test, prediction) |> Nx.sqrt()\n  [rmse]\nend\n\nres =\n  Scholar.ModelSelection.grid_search(train_data, train_labels, folding_fun, scoring_fun,\n    num_neighbors: Enum.to_list(min_num_neighbors..param_space_size)\n  )\n\nres = for(sub_res <- res, do: sub_res[:score]) |> Nx.concatenate() |> Nx.as_type(:f64)\n\nresults = DF.new(RMSE: res, num_neighbors: Nx.add(Nx.iota({param_space_size}), min_num_neighbors))\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\n```elixir\nTucan.lineplot(results, \"num_neighbors\", \"RMSE\",\n  points: true,\n  stroke_dash: [5, 5],\n  line_color: \"gray\"\n)\n|> Tucan.Scale.set_x_domain(1, 31)\n|> Tucan.Scale.set_y_domain(2.6, 6.0)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"RMSE vs. Number of Neighbors\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\n```elixir\nbest_neighbors = Nx.add(Nx.argmin(res), min_num_neighbors) |> Nx.to_number()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n2\n```\n\nThe best value is 2. Now try this value to a final prediction.\n\n```elixir\nbest_model =\n  KNNRegressor.fit(train_data, train_labels,\n    num_neighbors: best_neighbors,\n    algorithm: :brute\n  )\n\nfinal_prediction = KNNRegressor.predict(best_model, test_data)\nRegression.mean_square_error(test_labels, final_prediction) |> Nx.sqrt()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  2.243907122203796\n>\n```","ref":"k_nearest_neighbors.html#regression"},{"type":"extras","title":"Outliers prediction - k-nearest neighbors (KNN)","doc":"<!-- livebook:{\"break_markdown\":true} -->\n\nKNNs can be used for outlier detection. In this task, we will also use the Airfoil Self-Noise dataset. We will calculate for each sample a mean value of distances to the three closest points from dataset.\n\n```elixir\nnum_samples = Nx.axis_size(x, 0)\n\n{indices, distances} =\n  BruteKNN.fit(x, num_neighbors: 4, metric: {:minkowski, 2}) |> BruteKNN.predict(x)\n\nmean_distances = Nx.mean(distances[[.., 1..-1//1]], axes: [1], keep_axes: true)\n\nsamples_mean_distances =\n  Nx.concatenate([mean_distances, Nx.iota({num_samples, 1})], axis: 1)\n  |> DF.new()\n  |> DF.rename([\"mean_distances\", \"index\"])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nNow, plot the mean values of distances.\n\n```elixir\nTucan.lineplot(samples_mean_distances, \"index\", \"mean_distances\", line_color: \"jet\")\n|> Tucan.hruler(0.018, line_color: \"red\", stroke_dash: [5, 5])\n|> Tucan.Scale.set_x_domain(0, 1510)\n|> Tucan.Scale.set_y_domain(0.0085, 0.0225)\n|> Tucan.set_size(800, 630)\n|> Tucan.set_title(\"Plot to check anomalies\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nWe can assume, that samples with average distance bigger than 0.018 can be treated as outliers.\n\n```elixir\ndata = DF.put(data, :is_outlier, Nx.greater(mean_distances, 0.018))\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\n```elixir\nalias VegaLite, as: Vl\n\nVl.new(\n  title: [\n    text: \"Logarithm of Frequency vs Logarithm of Suction Side Displacement Thickness\",\n    offset: 20\n  ],\n  width: 630,\n  height: 630\n)\n|> Vl.data_from_values(data)\n|> Vl.layers([\n  Vl.new()\n  |> Vl.param(\"brush\", select: :interval)\n  |> Vl.mark(:circle)\n  |> Vl.encode_field(:x, \"frequency_log\", type: :quantitative, scale: [domain: [4.5, 10.5]])\n  |> Vl.encode_field(:y, \"suction_side_displacement_thickness_log\",\n    type: :quantitative,\n    scale: [domain: [-8.0, -2.5]]\n  )\n  |> Vl.encode(:color,\n    condition: [param: \"brush\", field: \"is_outlier\", type: :nominal],\n    value: :gray\n  ),\n  Vl.new()\n  |> Vl.transform(filter: [param: \"brush\", field: \"is_outlier\", equal: true])\n  |> Vl.mark(:text)\n  |> Vl.encode(:x, value: 11)\n  |> Vl.encode(:y, value: -3.5)\n  |> Vl.encode_field(:text, \"filter\",\n    type: :quantitative,\n    aggregate: :count\n  )\n])\n```\n\n<!-- livebook:{\"output\":true} -->","ref":"k_nearest_neighbors.html#outliers-prediction"},{"type":"extras","title":"Linear regression in practice","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Linear regression in practice\n\n```elixir\nMix.install([\n  {:scholar, \"~> 0.2.0\"},\n  {:explorer, \"~> 0.6.1\"},\n  {:exla, \"~> 0.6.0\"},\n  {:nx, \"~> 0.6.0\", override: true},\n  {:req, \"~> 0.3.9\"},\n  {:kino_vega_lite, \"~> 0.1.9\"},\n  {:kino, \"~> 0.10.0\"},\n  {:kino_explorer, \"~> 0.1.7\"},\n  {:tucan, \"~> 0.3.0\"}\n])\n```","ref":"linear_regression.html"},{"type":"extras","title":"Setup - Linear regression in practice","doc":"In the livebook, we will cover the typical use cases of linear regression on practical examples.\n\n```elixir\nrequire Explorer.DataFrame, as: DF\nrequire Explorer.Series, as: S\nalias Scholar.Linear.LinearRegression, as: LR\nalias Scholar.Linear.PolynomialRegression, as: PR\nalias Scholar.Impute.SimpleImputer\nalias Scholar.Metrics.Regression\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\nseed = 42\nkey = Nx.Random.key(42)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [0, 42]\n>\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"linear_regression.html#setup"},{"type":"extras","title":"Linear regression on synthetic data - Linear regression in practice","doc":"Before we dive into real-life use cases of linear regression, we start with a simpler one. We will generate data with a linear pattern and then use `Scholar.Linear.LinearRegression` to compute regression.\n\nFirstly, we generate the data which simulates the function $f(x) = 3x + 4$ with added uniform, zero-mean noise. `Nx.Random.uniform` creates a tensor with a given shape and type.\n\n```elixir\ndefmodule LinearData do\n  import Nx.Defn\n\n  defn data do\n    key = Nx.Random.key(42)\n    size = 100\n    {x, new_key} = Nx.Random.uniform(key, 0, 2, shape: {size, 1}, type: :f64)\n    {noise, _} = Nx.Random.uniform(new_key, -0.5, 0.5, shape: {size, 1}, type: :f64)\n    y = 3 * x + 4 + noise\n    {x, y}\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, LinearData, <<70, 79, 82, 49, 0, 0, 11, ...>>, true}\n```\n\nNow let's plot the generated points.\n\n```elixir\nsize = 100\n{x, y} = LinearData.data()\n\ndf = DF.new(x: x, y: y)\n\nTucan.scatter(df, \"x\", \"y\", width: 630, height: 630, filled: true)\n|> Tucan.Scale.set_x_domain(-0.05, 2.05)\n|> Tucan.Scale.set_y_domain(2.5, 12)\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.set_title(\"Scatterplot of Generated Data\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nFor a regression task, we will use the `Scholar.Linear.LinearRegression` module.\n\n```elixir\nmodel = LR.fit(x, y)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Scholar.Linear.LinearRegression{\n  coefficients: #Nx.Tensor \n    [\n      [2.9578183089017]\n    ]\n  >,\n  intercept: #Nx.Tensor \n    [4.023067929858136]\n  >\n}\n```\n\nAs we can see, the coefficient is almost 3.0, and the intercept is nearly 4.0. Those are decent estimations. They are not exactly equal to 3.0 and 4.0 because we introduce noise to our samples.\n\nNow, let's plot the result of linear regression.\n\n```elixir\n[intercept] = Nx.to_flat_list(model.intercept)\n[coefficients] = Nx.to_flat_list(model.coefficients)\nx_1 = 0\nx_2 = 2\n\nline = %{\n  x: [x_1, x_2],\n  y: [x_1 * coefficients + intercept, x_2 * coefficients + intercept]\n}\n\nTucan.layers([\n  Tucan.scatter(df, \"x\", \"y\", filled: true),\n  Tucan.lineplot(line, \"x\", \"y\", line_color: \"green\")\n])\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.Scale.set_x_domain(-0.05, 2.05)\n|> Tucan.Scale.set_y_domain(2.5, 12)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Scatterplot of Generated Data\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nUsing `Scholar.Linear.LinearRegression.predict`, we can predict an expected value for a given input. However, we must remember that our prediction will be valid only if we consider linearly dependent data. Fortunately, our data set is perfect for this kind of prediction.\n\nNow we will predict one value and draw it on the previous graph in a different color.\n\n```elixir\nx_prediction = Nx.tensor([[0.83]])\n\n[y_prediction] =\n  model\n  |> LR.predict(x_prediction)\n  |> Nx.to_flat_list()\n\n[x_prediction] = Nx.to_flat_list(x_prediction)\n{x_prediction, y_prediction}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{0.8299999833106995, 6.478057076882628}\n```\n\n```elixir\nprediction = %{\n  x: [x_prediction],\n  y: [y_prediction]\n}\n\nTucan.layers([\n  Tucan.scatter(df, \"x\", \"y\", filled: true),\n  Tucan.lineplot(line, \"x\", \"y\", line_color: \"green\"),\n  Tucan.scatter(prediction, \"x\", \"y\", point_color: \"red\", point_size: 80, filled: true)\n])\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.Scale.set_x_domain(-0.05, 2.05)\n|> Tucan.Scale.set_y_domain(2.5, 12)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Scatterplot of Generated Data\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nAs we expected, the red dot lies on the regression line.\n\nThis implementation of linear regression is based on the so-called *Least Squares* method. In practice, the function computes $X^+y$ where $X^+$ is a pseudo-inverse matrix (more precisely, Moore-Penrose matrix). You can calculate the results using `Nx.LinAlg.pinv/2`.\n\n```elixir\nx_b = Nx.concatenate([Nx.broadcast(1.0, {size, 1}), x], axis: 1)\nx_b |> Nx.LinAlg.pinv() |> Nx.dot(y)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [4.023067929858136],\n    [2.957818308901698]\n  ]\n>\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"linear_regression.html#linear-regression-on-synthetic-data"},{"type":"extras","title":"Polynomial regression on synthetic data - Linear regression in practice","doc":"Before moving on to a more complex example, this section will briefly show how to use another regression method. While not strictly linear, the approach and calculations that go into it is similar enough that it makes sense for it to be explained alongside linear regression.\n\nInstead of the `Scholar.Linear.LinearRegression` module, the following example uses `Scholar.Linear.PolynomialRegression`. Polynomial and linear regression differ in one key way. Linear regression optimizes a function in the form of $f(x) = ax + b$, whereas polynomial regression\nuses the function $f(x) = b + a_1x^1 + a_2x^2 + … + a_nx^n$, where $n$ represents the degree. Notice how if the degree is $1$, the function represents linear regression.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nFor this example we will start by generating data simulating the function $f(x) = 2x^2 + 3x + 5$ of degree 2 with some added noise.\n\n```elixir\ndefmodule PolynomialData do\n  import Nx.Defn\n\n  defn data do\n    key = Nx.Random.key(42)\n    size = 100\n    {x, new_key} = Nx.Random.uniform(key, -2, 2, shape: {size, 1}, type: :f32)\n    {noise, _} = Nx.Random.uniform(new_key, -0.5, 0.5, shape: {size, 1}, type: :f32)\n    y = 2 * x ** 2 + 3 * x + 5 + noise\n    {x, y}\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, PolynomialData, <<70, 79, 82, 49, 0, 0, 11, ...>>, true}\n```\n\nNow let's plot the generated data:\n\n```elixir\n{x, y} = PolynomialData.data()\ndf = DF.new(x: Nx.to_flat_list(x), y: Nx.to_flat_list(y))\n\nTucan.scatter(df, \"x\", \"y\", filled: true)\n|> Tucan.Scale.set_x_domain(-2.05, 2.05)\n|> Tucan.Scale.set_y_domain(2.5, 20)\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Scatterplot of Generated Data\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nAs is clear in the picture, the plotted data would most likely not be accurately estimated by a straight line like the ones linear regression estimates, so polynomial regression will better fit the data in this case. We do this with the `Scholar.Linear.PolynomialRegression` module. If you're familiar with the `Scholar.Linear.LinearRegression` module, the next steps will feel familiar.\n\nTo more clearly show the results, we will plot both methods.\n\n```elixir\nx_start = -2\nx_end = 2\nprecision = 1000\nx_values = Enum.map((x_start * precision)..(x_end * precision), fn r -> r / precision end)\n\n# Linear model\nlinear_model = LR.fit(x, y)\n\ny_linear_values =\n  LR.predict(\n    linear_model,\n    Nx.tensor(x_values) |> Nx.reshape({:auto, 1})\n  )\n\ndf_linear_results =\n  DF.new(\n    x: x_values,\n    y: y_linear_values |> Nx.to_flat_list()\n  )\n\n# Polynomial model\nmodel = PR.fit(x, y, degree: 2)\n\ny_values =\n  PR.predict(\n    model,\n    Nx.tensor(x_values) |> Nx.reshape({:auto, 1})\n  )\n\ndf_results =\n  DF.new(\n    x: x_values,\n    y: y_values |> Nx.to_flat_list()\n  )\n\nTucan.layers([\n  Tucan.scatter(df, \"x\", \"y\", filled: true),\n  Tucan.lineplot(df_results, \"x\", \"y\", line_color: \"green\"),\n  Tucan.lineplot(df_linear_results, \"x\", \"y\", line_color: \"red\", clip: true)\n])\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Scatterplot of Generated Data\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nNotice how the process is mostly the same for fitting a model and making a prediction. The one key difference comes in the coefficients, and how the input data is handled. The polynomial model returns a number of coefficients depending on the number of variables and the degree. This is why there are two coefficients but only one variable. In this example, with one variable and degree 2, we get two coefficients:\n\n```elixir\nmodel.coefficients\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [3.023973226547241, 2.013108730316162]\n  ]\n>\n```\n\nThese coefficients will be used to optimize the function and correspond to the size of the transformed input data.\n\nFeel free, in the cell bellow, to play around with the number of variables and the degree of the transformation.\n\n```elixir\nn_variables = 1\nn_samples = 5\n\nNx.iota({n_samples, n_variables})\n|> PR.transform(degree: 3, fit_intercept?: false)\n|> dbg()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [0, 0, 0],\n    [1, 1, 1],\n    [2, 4, 8],\n    [3, 9, 27],\n    [4, 16, 64]\n  ]\n>\n```\n\nWe can make simple predictions, just like in `Scholar.Linear.LinearRegression`.\n\n```elixir\nx_prediction = Nx.tensor([[-0.83], [0.83]])\n\ny_predictions =\n  PR.predict(model, x_prediction)\n  |> Nx.to_flat_list()\n\nx_predictions = x_prediction |> Nx.to_flat_list()\n\n{x_predictions, y_predictions}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{[-0.8299999833106995, 0.8299999833106995], [3.8831818103790283, 8.902976989746094]}\n```\n\nAnd plot these predictions with the training data.\n\n```elixir\ndf_prediction =\n  DF.new(\n    x: x_predictions,\n    y: y_predictions\n  )\n\nx_scale = [domain: [-2.05, 2.05]]\ny_scale = [domain: [2.5, 20]]\n\nTucan.layers([\n  Tucan.scatter(df, \"x\", \"y\", filled: true),\n  Tucan.lineplot(df_results, \"x\", \"y\", line_color: \"green\"),\n  Tucan.scatter(df_prediction, \"x\", \"y\", point_color: \"red\", point_size: 80, filled: true)\n])\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Plot of Generated Data and Predictions\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nNow we are ready to go into a more complex example!\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"linear_regression.html#polynomial-regression-on-synthetic-data"},{"type":"extras","title":"California Housing - Linear regression in practice","doc":"In this section we will play with [California Housing Data Set](https://www.kaggle.com/datasets/camnugent/california-housing-prices). The data pertains to the houses found in a given California district and some summary stats about them based on the 1990 census data. Be warned the data aren't cleaned, so there are some preprocessing steps required! The columns are as follows (their names are pretty self explanatory):\n\n* `longitude`\n* `latitude`\n* `housing_median_age`\n* `total_rooms`\n* `total_bedrooms`\n* `population`\n* `households`\n* `median_income`\n* `median_house_value`\n* `ocean_proximity`\n\nThe main task of this section is to predict the median_house_income. However, before we use our linear regression for prediction, we need to learn more about the data.\n\n```elixir\ndata =\n  Req.get!(\n    \"https://raw.githubusercontent.com/sonarsushant/California-House-Price-Prediction/master/housing.csv\"\n  ).body\n\ndf = DF.load_csv!(data)\n```\n\nFirstly, let's look at the distribution of houses based on the distance to the ocean.\n\n```elixir\nS.frequencies(df[\"ocean_proximity\"])\n```\n\nNow, we will plot univariate histograms for each feature of the data set.\n\n```elixir\n# Increase the sample size (or use 1.0 to plot all data)\nsample = DF.sample(df, 0.2, seed: seed)\n\nhistograms =\n  for name <- List.delete(df.names, \"ocean_proximity\") do\n    Tucan.histogram(sample, name, maxbins: 50, fill_opacity: 1.0, only: [name])\n    |> Tucan.Axes.put_options(:x, ticks: false)\n  end\n\nTucan.concat(histograms, columns: 3)\n|> Tucan.set_title(\"Univariate Histograms of all features\", anchor: :middle)\n|> Tucan.set_size(500, 500)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nFrom histograms, we can spot that *median_income* and *median_house_values* have a similar distribution. Both of them are heavy-tailed with high skewness. We might speculate that those two features are strictly correlated. We will check that later on.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nNow let's render the houses as a scatter plot using the latitude and longitude, overlayed on top of California's map. Let's also use color to encode the house prices and use the circle size to indicate the population of districts.\n\n```elixir\nTucan.scatter(df, \"longitude\", \"latitude\",\n  filled: true,\n  tooltip: true,\n  only: ~w(latitude longitude median_house_value population)\n)\n|> Tucan.color_by(\"median_house_value\", type: :quantitative)\n|> Tucan.size_by(\"population\")\n|> Tucan.Scale.set_x_domain(-124.55, -113.80)\n|> Tucan.Scale.set_y_domain(32.45, 42.05)\n|> Tucan.Grid.set_enabled(false)\n|> Tucan.Scale.set_color_scheme(:viridis)\n|> Tucan.background_image(\n  \"https://raw.githubusercontent.com/ageron/handson-ml2/master/images/end_to_end_project/california.png\"\n)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Scatterplot of Generated Data\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nFrom This plot, we can read that prices are substantially dependent on **geolocalization** and **population**. For geolocalization, we see, those areas closer to the ocean are more expensive. But it's not a strict rule since houses on the northern bay of California are much more affordable than in in-land Mid California. For the population, there are two dense areas with expensive housing: Los Angeles Bay (In South California) and San Francisco Bay (in Mid Califonia). They are metropolises with a lot of different tech companies, and business and cultural institutions, so, logically, housing in those places will be expensive.\n\n \n \nHint:\n \n \n\n \n\nYou can try to add another feature by computing clustering on this data set. It might be a sum or power mean of distances to the clusters. We may predict that centroids will be located in San Francisco Bay and Los Angeles Bay. You can also pass population as weights to k-means.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nBefore we convert our data to tensor, we will add three more columns which might be informative:\n\n* `rooms_per_family`\n* `bedrooms_per_rooms`\n* `population_per_family`\n\nThe names of columns are self-describing. Now, add them to our data frame.\n\n```elixir\ndf =\n  DF.mutate(df,\n    rooms_per_family: total_rooms / households,\n    bedrooms_per_rooms: total_bedrooms / total_rooms,\n    population_per_family: population / households\n  )\n```\n\nIn the next step, we will find the correlation matrix. But to do this, we need to cast our data frame to Nx tensor and split data into train and test sets.\n\nFirst let's remove all `:nan`s from our data and then convert the \"ocean_proximity\" string column to a numerical one. Explorer supports a `:category` type but, in this case, we will do a custom conversion as we can consider the ocean proximity as ordinal data since we can order them by the distance to the ocean. The bigger value the further from the ocean.\n\n```elixir\n# Replace all nils with :nan so we are able to convert to tensor.\nnames =\n  df\n  |> DF.names()\n  |> List.delete(\"ocean_proximity\")\n\nafter_preprocessing = for name <- names, into: %{}, do: {name, S.fill_missing(df[name], :nan)}\n\npreprocessed_data = DF.new(after_preprocessing)\n\nmapping = %{\n  \"ISLAND\" => 0.0,\n  \"<1H OCEAN\" => 1.0,\n  \"NEAR OCEAN\" => 2.0,\n  \"NEAR BAY\" => 3.0,\n  \"INLAND\" => 4.0\n}\n\nmapped_location = S.transform(df[\"ocean_proximity\"], fn x -> Map.fetch!(mapping, x) end)\n\ndf = DF.put(preprocessed_data, :ocean_proximity, mapped_location)\n```\n\nNow we convert dataframes into tensors. We can do so by concatenating and stacking the columns accordingly:\n\n```elixir\n# Shuffle data to make splitting more resonable\n{num_rows, _num_cols} = DF.shape(df)\n\nindices = Nx.iota({num_rows})\n{permutation_indices, _} = Nx.Random.shuffle(key, Nx.iota({num_rows}), axis: 0)\n\ny =\n  df[[\"median_house_value\"]]\n  |> Nx.concatenate()\n  |> Nx.take(permutation_indices)\n\nx =\n  df\n  |> DF.discard(\"median_house_value\")\n  |> Nx.stack(axis: 1)\n  |> Nx.take(permutation_indices)\n\n{x, y}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [0.2707641196013289, 326.0, 15.0, 32.76, -117.02, 1.0278, 543.0, 1.665644171779141, 3.6932515337423313, 326.0, 1204.0, 1.0],\n     [0.2598343685300207, 230.0, 27.0, 38.44, -122.71, 1.7, 462.0, 2.008695652173913, 4.2, 251.0, 966.0, 1.0],\n     [0.2902033271719039, 793.0, 35.0, 34.09, -118.35, 3.0349, 1526.0, 1.9243379571248425, 3.41109709962169, 785.0, 2705.0, 1.0],\n     [0.2940552016985138, 549.0, 25.0, 33.91, -118.35, 2.8512, 1337.0, 2.435336976320583, 3.431693989071038, 554.0, 1884.0, 1.0],\n     [0.2067861321509945, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [1.542e5, 3.5e5, 2.667e5, 2.728e5, 1.163e5, 2.941e5, 4.178e5, 1.851e5, 5.68e4, 500001.0, 500001.0, 1.152e5, 2.343e5, 8.75e4, 8.37e4, 500001.0, 1.669e5, 3.003e5, 1.546e5, 6.69e4, 2.325e5, 1.367e5, 1.375e5, 1.27e5, 1.683e5, 1.441e5, 9.53e4, 1.834e5, 1.435e5, 6.85e4, 1.625e5, 1.661e5, 1.908e5, 2.431e5, 1.488e5, 3.036e5, 1.479e5, 1.5e5, 3.288e5, 7.08e4, 2.25e5, 1.375e5, 3.5e5, 3.742e5, 1.549e5, 4.5e5, 3.063e5, 2.051e5, ...]\n >}\n```\n\nSince we don't have a stratified split of data implemented (to learn more see [Stratified Sampling](https://en.wikipedia.org/wiki/Stratified_sampling)), we shuffle data set and take advantage of [Law of large numbers](https://en.wikipedia.org/wiki/Law_of_large_numbers). It says that the average of the results obtained from a large number of trials should be close to the expected value and tends to become closer to the expected value as more trials are performed. As we take a lot of samples from shuffled data it implies that the sampled data sets will be stratified. Now, we will split the data into training and test sets.\n\n```elixir\ntrain_ratio = 0.8\n\n{x_train, x_test} = Nx.split(x, train_ratio)\n{y_train, y_test} = Nx.split(y, train_ratio)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [1.542e5, 3.5e5, 2.667e5, 2.728e5, 1.163e5, 2.941e5, 4.178e5, 1.851e5, 5.68e4, 500001.0, 500001.0, 1.152e5, 2.343e5, 8.75e4, 8.37e4, 500001.0, 1.669e5, 3.003e5, 1.546e5, 6.69e4, 2.325e5, 1.367e5, 1.375e5, 1.27e5, 1.683e5, 1.441e5, 9.53e4, 1.834e5, 1.435e5, 6.85e4, 1.625e5, 1.661e5, 1.908e5, 2.431e5, 1.488e5, 3.036e5, 1.479e5, 1.5e5, 3.288e5, 7.08e4, 2.25e5, 1.375e5, 3.5e5, 3.742e5, 1.549e5, 4.5e5, 3.063e5, 2.051e5, 2.737e5, ...]\n >,\n #Nx.Tensor \n   [8.55e4, 1.743e5, 5.65e4, 1.308e5, 1.375e5, 2.472e5, 2.25e5, 1.164e5, 500001.0, 1.154e5, 1.269e5, 1.229e5, 7.04e4, 3.534e5, 7.14e4, 4.083e5, 3.0e5, 1.648e5, 1.125e5, 2.028e5, 1.139e5, 2.15e5, 500001.0, 1.48e5, 1.683e5, 2.819e5, 3.338e5, 1.078e5, 3.277e5, 3.412e5, 3.289e5, 2.879e5, 2.545e5, 2.667e5, 6.82e4, 1.406e5, 2.795e5, 2.25e5, 1.17e5, 1.375e5, 3.115e5, 2.423e5, 4.93e5, 3.458e5, 1.542e5, 3.728e5, 6.6e4, 1.281e5, ...]\n >}\n```\n\nBefore we compute the correlation matrix, we will check if we have NaNs (Not a Number) in the data set.\n\n```elixir\ny_nan_count = Nx.sum(Nx.is_nan(y))\nx_nan_count = Nx.sum(Nx.is_nan(x))\n{x_nan_count, y_nan_count}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   414\n >,\n #Nx.Tensor \n   0\n >}\n```\n\nUps, we have some. Fortunately, for y, we don't have any NaNs. If we dig a little bit more, it turns out that NaNs are in   bedrooms_per_rooms (1st row)      total_bedrooms (10th row)  \n\n```elixir\n{bedrooms_per_rooms_idx, total_bedrooms_idx} = {0, 9}\nbedrooms_per_rooms_nan_count = Nx.sum(Nx.is_nan(x[[.., bedrooms_per_rooms_idx]]))\ntotal_bedrooms_nan_count = Nx.sum(Nx.is_nan(x[[.., total_bedrooms_idx]]))\nNx.equal(x_nan_count, Nx.add(bedrooms_per_rooms_nan_count, total_bedrooms_nan_count))\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  1\n>\n```\n\nFor these two, we use `Scholar.Impute.SimpleImputer` with startegy set to median of values. Function `fit` learn the median of features and `transform` for trained model replace all NaNs with a given startegy. It is important that we perform imputation after splitting data because otherwise we will have a leakage of information from test data.\n\n```elixir\nx_train =\n  x_train\n  |> SimpleImputer.fit(strategy: :median)\n  |> SimpleImputer.transform(x_train)\n\nx_test =\n  x_test\n  |> SimpleImputer.fit(strategy: :median)\n  |> SimpleImputer.transform(x_test)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [0.20019126554032515, 548.0, 23.0, 36.32, -119.33, 2.5, 1446.0, 2.6386861313868613, 5.724452554744525, 628.0, 3137.0, 4.0],\n    [0.18461538461538463, 83.0, 34.0, 34.26, -118.44, 5.5124, 433.0, 5.216867469879518, 3.9156626506024095, 60.0, 325.0, 1.0],\n    [0.23362175525339926, 398.0, 35.0, 35.77, -119.25, 1.6786, 1449.0, 3.64070351758794, 4.065326633165829, 378.0, 1618.0, 4.0],\n    [0.20259128386336867, 313.0, 24.0, 37.8, -121.2, 3.5625, 927.0, 2.961661341853035, 5.424920127795527, 344.0, 1698.0, 4.0],\n    [0.42487046632124353, 155.0, ...],\n    ...\n  ]\n>\n```\n\nEventually, we can compute the correlation matrix. We will use `Scholar.Covariance` to calculate the correlation matrix.\n\n```elixir\ncorrelation =\n  Nx.concatenate([x_train, Nx.new_axis(y_train, 1)], axis: 1)\n  |> Scholar.Covariance.correlation_matrix(biased: true)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [1.0000000000000002, 0.06144282875654085, 0.13311283521287637, -0.12216156759996018, 0.09984860365269956, -0.612630933680883, 0.03517997111673592, 0.004697459810143742, -0.4078383723411937, 0.08095666109618091, -0.18612145325855575, -0.11835385450623682, -0.25385512001635363],\n    [0.06144282875654085, 1.0, -0.305555948690877, -0.07566125267999342, 0.06170866872453172, 0.01398821983308101, 0.907460734474932, -0.02634672373043427, -0.07388109216765114, 0.9735036322842058, 0.9198029415934222, -0.04626792658479704, 0.0639675034251392],\n    [0.13311283521287637, -0.305555948690877, 0.9999999999999998, 0.017212690983879612, -0.11382253728911094, -0.12557649464959478, -0.2961765721596695, 0.012442541870155687, -0.14834326174239934, -0.32094539339262423, -0.3602627671081426, -0.11940626101050672, 0.10251505008793772],\n    [-0.12216156759996018, -0.07566125267999342, 0.017212690983879612, 0.9999999999999992, -0.92526150146814, -0.0709535112221358, -0.11317680203176562, 0.0032929374670018535, 0.10562463526360089, -0.07174931649052743, -0.038747214812891013, ...],\n    ...\n  ]\n>\n```\n\nMaybe visual representation would be nicer. 😅\n\n```elixir\n{corr_size, _} = Nx.shape(correlation)\ncorrelation_list = Nx.to_flat_list(correlation)\n\nnames = [\n  \"Bedrooms per rooms\",\n  \"Households\",\n  \"Housing median age\",\n  \"Latitude\",\n  \"Longitude\",\n  \"Median income\",\n  \"Population\",\n  \"Population per family\",\n  \"Rooms per family\",\n  \"Total bedrooms\",\n  \"Total rooms\",\n  \"Ocean proximity\",\n  \"Median house value\"\n]\n\ncorr_to_plot =\n  DF.new(\n    x: List.flatten(List.duplicate(names, corr_size)),\n    y: List.flatten(for name <- names, do: List.duplicate(name, corr_size)),\n    corr_val: Enum.map(correlation_list, fn x -> Float.round(x, 2) end)\n  )\n\nTucan.heatmap(corr_to_plot, \"x\", \"y\", \"corr_val\",\n  annotate: true,\n  text_color: [{nil, 0, \"white\"}, {0, nil, \"black\"}]\n)\n|> Tucan.Scale.set_color_scheme(:viridis)\n|> Tucan.set_size(630, 630)\n|> Tucan.set_title(\"Correlation Matrix for California Housing\", offset: 20)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nWe can spot that _median_house_value_ is strongly correlated with _median_income_. It's pretty straightforward, the more money you have, the more expensive house you can buy. Non-obvious is a negative correlation with _bedrooms_per_rooms_. But it also can be explained. Bedrooms are the most crucial rooms in the house. Firstly, you need to guarantee that you have a house with enough bedrooms. If this condition is satisfied, then you can focus on \"additional rooms\" like a chill room, cabinets and so on. So if you buy a house with more additional rooms, then you decrease the ratio.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nNow we are ready to train a model for the _median_house_value_ prediction. We will use linear regression. In the first step, we create the model by calling the `fit` function.\n\n```elixir\nmodel = LR.fit(x_train, y_train)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Scholar.Linear.LinearRegression{\n  coefficients: #Nx.Tensor \n    [292661.1879978386, 111.46237796967083, 1170.1666131366273, -33322.29721781791, -35394.90464853701, 41384.64324873476, -41.31598131072877, 50.507900596456054, 3040.877192253343, 7.253624678472778, 3.272614960520949, -8059.519389730254]\n  >,\n  intercept: #Nx.Tensor \n    -3101555.563125743\n  >\n}\n```\n\nNow we can predict the values for the test set and measure the error of our prediction. We will calculate root mean square error (RMSE) and mean absolute error (MAE).\n\n```elixir\npredictions = LR.predict(model, x_test)\nrmse = Regression.mean_square_error(y_test, predictions) |> Nx.sqrt()\nmae = Regression.mean_absolute_error(y_test, predictions)\n{rmse, mae}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   67648.9435367406\n >,\n #Nx.Tensor \n   48942.11908533544\n >}\n```\n\nOk, but is it a good or poor estimation? Huh, check the mean value of the target and then compare it to the value of errors.\n\n```elixir\nNx.mean(y)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  206855.81690891474\n>\n```\n\nFor such a simple model as linear regression, it seems to be a pretty good result. But there is space to improve this result. You can, for example, add some additional features to the data set. In the future, you will be able to try more complicated models, such as random forests.","ref":"linear_regression.html#california-housing"},{"type":"extras","title":"Manifold learning","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Manifold learning\n\n```elixir\nMix.install([\n  {:scholar, github: \"elixir-nx/scholar\"},\n  {:explorer, \"~> 0.8.2\", override: true},\n  {:exla, \"~> 0.7.2\"},\n  {:nx, \"~> 0.7.2\"},\n  {:req, \"~> 0.4.14\"},\n  {:kino_vega_lite, \"~> 0.1.11\"},\n  {:kino, \"~> 0.12.3\"},\n  {:kino_explorer, \"~> 0.1.18\"},\n  {:tucan, \"~> 0.3.1\"}\n])\n```","ref":"manifold_learning.html"},{"type":"extras","title":"Setup - Manifold learning","doc":"We will use `Explorer` in this notebook, so let's define an alias for its main module DataFrame:\n\n```elixir\nrequire Explorer.DataFrame, as: DF\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\nExplorer.DataFrame\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[]\n```","ref":"manifold_learning.html#setup"},{"type":"extras","title":"Testing Manifold Learning Functionalities - Manifold learning","doc":"In this notebook, we test how manifold learning algorithms work and how to use them for dimensionality reduction.\n\nFirst, let's fetch the dataset that we experiment on. The data represents 3D coordinates of a mammoth. Below we include a figure of original dataset.\n\n![](files/mammoth.png)\n\n```elixir\nsource = \"https://raw.githubusercontent.com/MNoichl/UMAP-examples-mammoth-/master/mammoth_a.csv\"\n\ndata = Req.get!(source).body\n\ndf = DF.load_csv!(data)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nNow, convert the dataframe into tensor, so we can manipulate the data using `Scholar`.\n\n```elixir\ntensor_data = Nx.stack(df, axis: 1)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [58.823, 228.407, 79.843],\n    [59.197, 228.642, 77.478],\n    [58.734, 228.931, 78.515],\n    [59.043, 228.693, 78.571],\n    [59.223, 228.667, 78.611],\n    [59.103, 228.711, 78.305],\n    [58.854, 228.786, 78.597],\n    [59.123, 228.695, 77.371],\n    [59.002, 228.592, 78.925],\n    [58.368, 227.879, 81.155],\n    [59.168, 229.8, 74.95],\n    [58.798, 229.431, 76.296],\n    [59.257, 229.227, 76.144],\n    [58.408, 250.928, 93.15],\n    [58.575, 250.743, 93.323],\n    [71.011, 217.179, 62.859],\n    [70.259, 217.511, ...],\n    ...\n  ]\n>\n```\n\nSince there is almost 1 million data points and they are sorted, we shuffle dataset and then use only the part of the dataset.\n\n<!-- livebook:{\"branch_parent_index\":1} -->","ref":"manifold_learning.html#testing-manifold-learning-functionalities"},{"type":"extras","title":"Trimap - Manifold learning","doc":"We start with Trimap. It's a manifold learning algorithm that is based of nearest neighbors. It preserves the global structure of dataset, which can be used for understanding the overall data distribution. Let's look what will be the result of the Trimap on mammoth dataset.\n\n```elixir\n{tensor_data, key} = Nx.Random.shuffle(Nx.Random.key(42), tensor_data)\n\ntrimap_res =\n  Scholar.Manifold.Trimap.transform(tensor_data[[0..10000, ..]],\n    key: Nx.Random.key(55),\n    num_components: 2,\n    num_inliers: 12,\n    num_outliers: 4,\n    weight_temp: 0.5,\n    learning_rate: 0.1,\n    metric: :squared_euclidean\n  )\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [87.5960481212865, 107.83563946728007],\n    [95.93224095586187, 91.03811962459187],\n    [76.23538360750037, 101.22665484812988],\n    [91.28047447374527, 83.89099279844432],\n    [84.84272736715022, 63.05776975461835],\n    [77.47594824886791, 70.58007253201106],\n    [48.1726647167702, 83.59889134210331],\n    [87.09967163881679, 81.16403134057438],\n    [78.77111626020424, 89.5381523774902],\n    [97.90247495971414, 90.30331715233045],\n    [74.82692937810215, 88.72983366943957],\n    [84.41794589221618, 100.28068173065604],\n    [48.9525994243841, 65.31540933930391],\n    [97.74716685388529, 87.01213386339163],\n    [55.25583739452793, 78.09500152893669],\n    [78.8556707294727, 96.20470118706773],\n    [87.06284966061098, 92.95614239803128],\n    [98.14345384260274, 77.08288277313781],\n    [45.61527701121027, 86.90240376974904],\n    [109.77362315414614, 105.15212699721849],\n    [77.18580406991187, 90.15798204995377],\n    [72.62493132684811, 61.503358925355755],\n    [64.86122257994592, 56.26636625077798],\n    [107.15238372453034, 86.8502564790939],\n    [76.74146917610297, 92.78343841441362],\n    ...\n  ]\n>\n```\n\nNow, lets plot the results of Trimap algorithm\n\n```elixir\ncoords = [\n  x: trimap_res[[.., 0]] |> Nx.to_flat_list(),\n  y: trimap_res[[.., 1]] |> Nx.to_flat_list()\n]\n\nTucan.layers([\n  Tucan.scatter(coords, \"x\", \"y\", point_size: 1)\n])\n|> Tucan.set_size(300, 300)\n|> Tucan.set_title(\n  \"Mammoth dataset with reduced dimensionality using Trimap\",\n  offset: 25\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nFor sure, we can recognize mammoth on this picture. Trimap indeed preserved the global structure of data. Result is similar to the projection of 3D mammoth to the YZ plane. Now, plot this projection and compare these two plots.\n\n```elixir\ncoords = [\n  x: tensor_data[[0..10000, 1]] |> Nx.to_flat_list(),\n  y: tensor_data[[0..10000, 2]] |> Nx.to_flat_list()\n]\n\nTucan.layers([\n  Tucan.scatter(coords, \"x\", \"y\", point_size: 1)\n])\n|> Tucan.set_size(300, 300)\n|> Tucan.set_title(\n  \"Mammoth data set with reduced dimensionality using trimap\",\n  offset: 25\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nThese two plots are similiar but there are some important differences. Even if the second figure seems \"prettier\" it is less informative than the result of trimap. On the first figure, we can spot two tusks while one the second one they overlap and we see only one. Similarly, legs overlay on the first plot and one the second one they are spread and don't intersect with each other.","ref":"manifold_learning.html#trimap"},{"type":"extras","title":"t-SNE - Manifold learning","doc":"Now, lets try different algorithm: t-Distributed Stochastic Neighbor Embedding, mostly known as t-SNE. t-SNE is a nonlinear dimensionality reduction technique that emphasizes local relationships, making it useful for identifying patterns and subgroups within the data.\n\n```elixir\ntsne_res =\n  Scholar.Manifold.TSNE.fit(tensor_data[[0..2000, ..]],\n    key: Nx.Random.key(55),\n    num_components: 2,\n    perplexity: 125,\n    exaggeration: 10.0,\n    learning_rate: 500,\n    metric: :squared_euclidean\n  )\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\nwarning: Range.new/2 has a default step of -1, please call Range.new/3 explicitly passing the step of -1 instead\n  (nx 0.7.2) lib/nx/shape.ex:1811: Nx.Shape.validate_dot_axes!/6\n  (nx 0.7.2) lib/nx/shape.ex:1744: Nx.Shape.dot/8\n  (nx 0.7.2) lib/nx.ex:12525: Nx.dot/6\n  (nx 0.7.2) lib/nx/lin_alg/qr.ex:102: Nx.LinAlg.QR.\"__defn:norm__\"/1\n  (nx 0.7.2) lib/nx/lin_alg/qr.ex:109: Nx.LinAlg.QR.\"__defn:householder_reflector__\"/3\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [-7.720863452772271, -8.924593239946077],\n    [-7.627408661169662, -8.493499839300748],\n    [-7.655294023089491, -8.67975773066019],\n    [-7.660605743680346, -8.669241639954539],\n    [-7.6598753013087535, -8.658590299593522],\n    [-7.650612040457881, -8.62046834395759],\n    [-7.66115746267903, -8.687692549707776],\n    [-7.624018688200683, -8.485297780382929],\n    [-7.6767334531761025, -8.7359601794454],\n    [-7.810238386734004, -9.244003168397134],\n    [-7.5452610843798915, -8.215109191115873],\n    [-7.582641346103063, -8.371180999947315],\n    [-7.5790580771288925, -8.318735667656417],\n    [-6.330265165889219, -17.67182425161148],\n    [-6.350620100516294, -17.664527925102853],\n    [-8.511254173153981, 3.2484235457551365],\n    [-8.667916640544206, 3.2325600834700907],\n    [-8.652664730644178, 3.1774228289244664],\n    [-8.722131914178286, 3.1262080759539486],\n    [-8.772817730928294, 3.098786298010697],\n    [-2.6314317865997636, 3.17224901794698],\n    [-2.8022476551090123, 3.2279404380435994],\n    [-2.8261919490936998, 3.2470904773946208],\n    [-4.9968543194910575, 4.953954052553824],\n    [-6.961367201446683, -17.431908124735948],\n    ...\n  ]\n>\n```\n\n```elixir\ncoords = [\n  x: tsne_res[[.., 0]] |> Nx.to_flat_list(),\n  y: tsne_res[[.., 1]] |> Nx.to_flat_list()\n]\n\nTucan.layers([\n  Tucan.scatter(coords, \"x\", \"y\", point_size: 1)\n])\n|> Tucan.set_size(300, 300)\n|> Tucan.set_title(\n  \"Mammoth dataset with reduced dimensionality using Trimap\",\n  offset: 25\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n\n\nAs we see, t-SNE gives completely different results than Trimap. This is because t-SNE has a completely different mathematical background of computation. Also t-SNE is slower algorithm, so it can't be used on such big datasets as Trimap. However, t-SNE preserves some features of mammoth like small tusks, feets, and corp. You can experiment with parameter *perplexity* which can substantially change the output of the algorithm.","ref":"manifold_learning.html#t-sne"},{"type":"extras","title":"Multidimensional scaling (MDS)","doc":"# Multidimensional scaling (MDS)\n\n```elixir\nMix.install([\n  {:scholar, github: \"elixir-nx/scholar\", override: true},\n  {:kino, \"~> 0.11.3\", override: true},\n  {:kino_vega_lite, \"~> 0.1.9\"},\n  {:exla, \"~> 0.6.4\"},\n  {:tucan, \"~> 0.3.0\"},\n  {:explorer, \"~> 0.7.1\"},\n  {:req, \"~> 0.4.5\"},\n  {:nx, git: \"https://github.com/elixir-nx/nx.git\", sparse: \"nx\", override: true}\n])\n```","ref":"mds.html"},{"type":"extras","title":"Setup - Multidimensional scaling (MDS)","doc":"```elixir\nalias Scholar.Manifold.MDS\nalias VegaLite, as: Vl\nalias Explorer.DataFrame, as: DF\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"mds.html#setup"},{"type":"extras","title":"Swiss roll - Multidimensional scaling (MDS)","doc":"Multidimensional scaling (MDS) seeks a low-dimensional representation of the data in which the distances respect well the distances in the original high-dimensional space.\n\nThe first example demonstrating MDS will use so-called Swiss Roll dataset. It's call as such because it resembles a popular sponge cake. Let's have a look on this dataset.\n\n| ![Skewness](https://scikit-learn.org/stable/_images/sphx_glr_plot_swissroll_001.png)                        |\n| :---------------------------------------------------------------------------------------------------------: |\n| Figure 1: Swiss Roll (Source: https://scikit-learn.org/stable/auto_examples/manifold/plot_swissroll.html)   |\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nBecause there are no good swiss roll datasets available on the Internet directly, we'll load to tensor created swiss roll dataset with 1000 samples.\n\n```elixir\nswiss_roll =\n  Nx.tensor([\n    [-8.8313152, 2.10151711, -4.45670876],\n    [6.15941887, 5.17977657, 2.58770188],\n    [10.23410953, 19.38484837, -6.29460341],\n    [11.60978359, 8.25325228, 5.93527204],\n    [-9.44716177, 4.1405192, -1.75077878],\n    [-9.14066212, 10.03281252, 1.49703653],\n    [-8.30319363, 8.17661164, 3.53076713],\n    [-3.50163931, 0.06305818, -10.06943069],\n    [-5.36417404, 9.28502669, -8.97624018],\n    [-5.74223798, 4.48630607, -8.68466132],\n    [11.51680254, 1.85733993, -4.09883143],\n    [-3.67908164, 6.4411102, -9.98645122],\n    [-7.38759241, 6.97403281, -7.00970874],\n    [-5.01240594, 5.82777801, 6.84732906],\n    [-9.34088572, 10.66496615, 0.6029048],\n    [-5.31263585, 13.34134448, 6.66928822],\n    [10.78417439, 20.0123519, -5.47919839],\n    [7.43646909, 7.77331775, -9.01371917],\n    [5.35145344, 18.18259548, 4.48193741],\n    [-9.47674306, 1.42425568, -1.09677707],\n    [5.57085789, 18.54771524, 4.10142655],\n    [-4.80862233, 5.3346213, 6.95886807],\n    [11.91401456, 12.42519895, -3.10333003],\n    [3.52144312, 4.79076306, -4.1197149],\n    [-3.88239859, 1.62842458, 7.37931694],\n    [4.50836922, 16.70636732, 5.58104807],\n    [12.50699262, 15.08508412, 2.57579946],\n    [1.33187579, 11.11716535, 7.56329701],\n    [11.94849042, 12.24028964, 5.03235914],\n    [-7.62193067, 13.40510227, -6.69624307],\n    [4.31595182, 6.04745455, -3.55837632],\n    [-0.47401598, 8.12957525, -10.94201811],\n    [-7.54759353, 13.36704012, -6.79844625],\n    [-2.83238685, 20.62432552, 7.70208861],\n    [11.02660149, 9.67867052, -5.06808249],\n    [4.97693483, 6.31073777, 12.83620986],\n    [12.54070575, 16.70115382, -0.28211401],\n    [-3.12322929, 14.36409155, -10.23335682],\n    [6.15504322, 6.17994614, 12.20604373],\n    [-9.46928984, 7.89309513, -1.38427241],\n    [-3.57353186, 20.46941371, 7.49035496],\n    [0.74117884, 10.16360345, -11.03776613],\n    [5.75830231, 20.12651737, -1.68213898],\n    [0.33151929, 19.54850012, -4.77026848],\n    [4.06279356, 9.47926784, -3.76096312],\n    [4.78869214, 10.34172688, -3.10600293],\n    [7.85635468, 6.93968359, 10.99932961],\n    [-3.86985637, 2.22281317, -9.89271464],\n    [8.19692404, 7.06896747, -8.44132047],\n    [0.44250408, 16.90923521, -4.78419875],\n    [6.3446155, 16.42643105, 0.53255482],\n    [6.07222961, 0.55829229, 2.89848962],\n    [4.19249291, 1.48283922, 5.89768184],\n    [-5.49908409, 14.64743159, 6.55005453],\n    [9.53695925, 13.44975138, -7.15522323],\n    [-8.93428834, 14.84191116, 2.14788096],\n    [11.02594346, 13.52161869, 7.1507573],\n    [1.63692776, 8.98139788, 7.46052236],\n    [4.32153517, 12.7298557, 5.77309443],\n    [-8.70674945, 15.37617867, 2.72063783],\n    [-4.81699699, 10.70378252, 6.95442742],\n    [5.86026776, 15.15945121, -1.45905989],\n    [6.44342732, 14.61781518, -9.62821188],\n    [12.58899959, 5.56269559, 0.34418107],\n    [11.89886403, 5.75441231, -3.14647192],\n    [10.84038438, 12.56144376, 7.47955699],\n    [8.22545033, 12.69229784, 10.68067914],\n    [-8.29320994, 7.8364129, 3.5484666],\n    [-6.08674638, 4.12945347, 6.12651234],\n    [3.74341434, 6.38150565, 13.34880751],\n    [4.24853793, 16.95338914, 5.84433787],\n    [-9.30843557, 7.95630492, -2.78160162],\n    [6.34476788, 13.44014924, 1.44330463],\n    [-8.76058105, 17.72028715, 2.59489988],\n    [5.63857516, 12.40441207, -10.03157055],\n    [-5.14799209, 12.69503064, -9.130977],\n    [-4.31509327, 17.38568396, 7.1998396],\n    [5.76534121, 20.31141789, -9.97320641],\n    [8.51320446, 10.27123949, 10.41535388],\n    [4.95945381, 14.00276241, -2.9136638],\n    [2.7874212, 8.18363232, 6.93225434],\n    [6.82948167, 13.41956201, 11.77423634],\n    [1.27886402, 17.65631858, 7.57970442],\n    [-8.07758036, 14.30406837, -6.0028668],\n    [4.13266509, 16.6570385, -3.70747957],\n    [12.5825374, 2.50703498, 0.22986054],\n    [-2.2605078, 11.50682434, 7.81518798],\n    [3.37348866, 12.04572521, 6.56291498],\n    [-6.49904453, 20.73956309, -8.00862441],\n    [4.00047908, 18.37510258, -3.80717916],\n    [5.98031223, 2.33735624, 3.17611432],\n    [6.34227329, 14.14384321, 0.50119354],\n    [2.73379039, 3.88664814, 6.96233845],\n    [11.79080211, 18.99432667, -3.44035457],\n    [12.08528456, 9.28169687, -2.57456403],\n    [2.74475645, 4.98017065, -4.48851969],\n    [10.80000471, 20.49884326, -5.45345469],\n    [-9.28620542, 11.34853999, 0.89233151],\n    [-6.77355445, 11.66460162, 5.52425488],\n    [2.76730238, 3.66446413, -10.89839044],\n    [-5.9174973, 9.29464305, -8.539737],\n    [-2.66403786, 10.18609958, 7.73982977],\n    [1.8803158, 9.67455061, 13.87565254],\n    [4.4742536, 13.94489946, -10.48412406],\n    [0.83245284, 15.00524786, -4.81221014],\n    [-1.45908722, 18.99011584, 7.90298737],\n    [6.34678493, 15.23879447, 0.56368875],\n    [6.3059199, 2.99744911, 0.15547907],\n    [-9.07517772, 1.61968996, 1.72252085],\n    [11.54373672, 11.70252553, -4.03865376],\n    [-1.79745436, 17.91955908, -10.67859439],\n    [1.98102707, 10.23384359, -4.71087471],\n    [-7.38878555, 13.5510619, 4.85862952],\n    [1.99277958, 20.74908238, -10.9957415],\n    [-9.4754214, 11.41981006, -1.18301902],\n    [4.79069647, 11.90037425, 5.26090616],\n    [-0.4470774, 4.45873243, 7.89788473],\n    [12.60587502, 10.22424953, 1.03455834],\n    [-3.02284231, 13.71775046, -10.27395798],\n    [1.89692351, 6.52916473, -4.72759871],\n    [-8.98162829, 2.45226941, 2.01254378],\n    [-8.85074596, 2.16862452, -4.40568126],\n    [-9.09477444, 9.10920615, 1.65725701],\n    [-1.59420735, 15.92413901, -10.73030283],\n    [3.84253976, 19.747862, 6.20639199],\n    [2.99588934, 3.7427977, -4.38613248],\n    [-7.63287559, 6.76846841, 4.55180338],\n    [-9.44495828, 11.30192091, -1.77793528],\n    [-9.3714683, 13.50778735, -2.41065506],\n    [-5.66214464, 7.21900098, -8.74877148],\n    [4.89889258, 1.06334016, 12.87287917],\n    [2.48047748, 9.56255577, 7.09672506],\n    [2.28405917, 3.0025995, -10.96557981],\n    [-4.91976519, 8.21809381, 6.89894202],\n    [-6.77264339, 6.50043151, -7.72833737],\n    [5.68637128, 17.834621, 3.87542805],\n    [-7.36919831, 2.76915445, 4.88207307],\n    [-4.67551326, 7.5889942, -9.44150225],\n    [4.61082309, 9.74047523, 5.46932658],\n    [2.55781647, 17.46513455, -4.55502376],\n    [2.54333691, 7.75362726, 7.06453858],\n    [-7.93896942, 9.00625301, 4.12306274],\n    [4.0461184, 16.86758555, -3.77346553],\n    [-6.11865229, 0.83166964, 6.10127456],\n    [-1.2958277, 1.05183285, -10.79867562],\n    [8.13882818, 10.71772927, -8.48856341],\n    [-8.6676732, 18.34670324, 2.80876248],\n    [-9.19093659, 10.89141337, 1.30747332],\n    [2.83893255, 16.62682712, 13.63964549],\n    [8.22365353, 7.59402792, -8.41937623],\n    [-9.44834484, 10.16153756, -0.25531781],\n    [12.46867557, 5.30573332, 2.85570706],\n    [-0.00738733, 18.32033424, -10.99489992],\n    [3.77209917, 2.0015116, -10.68876556],\n    [-5.77501984, 4.93348532, 6.36060111],\n    [4.97557715, 16.12420438, 5.02821176],\n    [8.3576441, 12.9549197, -8.30735835],\n    [-5.50944353, 20.22917632, 6.54322691],\n    [6.2902454, 18.84446588, 12.1239157],\n    [-8.82155888, 14.86681432, 2.44575791],\n    [3.22446319, 16.79289452, 6.66425592],\n    [4.75460865, 0.00877746, -3.14235018],\n    [0.91481478, 16.05261378, -4.81403011],\n    [11.19712672, 9.66134372, 6.82622795],\n    [-9.46768405, 0.83689931, -1.42159092],\n    [-5.53375621, 18.3998009, 6.52711765],\n    [10.96044847, 6.71630255, -5.18400692],\n    [10.4351764, 15.26003256, -6.01297327],\n    [4.18087526, 18.69528983, -3.66952153],\n    [6.22937585, 19.98937415, -0.2926236],\n    [-9.46938658, 2.11172417, -1.38190959],\n    [6.35393757, 4.27216985, 0.68868349],\n    [6.20282647, 0.38603574, 2.40642951],\n    [12.51159145, 10.60661689, 2.53874859],\n    [2.2754204, 7.7811052, -4.64043337],\n    [10.91956822, 4.13148501, -5.25418863],\n    [5.01569359, 15.80616581, -2.84629111],\n    [3.78456699, 17.88768735, -3.95705292],\n    [5.30830065, 19.72249123, -2.45758728],\n    [9.17003027, 16.58250053, -7.54907087],\n    [3.67149946, 10.54910499, 6.34340481],\n    [-5.68172263, 6.40012611, -8.73322022],\n    [2.61491996, 18.1532473, -10.92190052],\n    [12.29800693, 7.2092361, -1.76413294],\n    [2.50773863, 18.38404008, 13.72935656],\n    [2.25900517, 7.69168124, -4.64485463],\n    [-8.04833423, 10.64461678, -6.05133859],\n    [6.16508487, 0.57625979, -0.56846685],\n    [12.29762626, 14.78252085, -1.76581264],\n    [-8.79717701, 13.65648542, 2.50630109],\n    [7.23127936, 16.81039757, 11.48913947],\n    [6.16294153, 2.8500571, 2.57379403],\n    [-8.51342001, 16.81909389, -5.19360231],\n    [4.64255568, 7.24713667, -3.25744236],\n    [12.57202663, 14.81449512, 0.07383412],\n    [4.52682027, 3.2278555, 5.56127892],\n    [11.85103674, 12.62481635, 5.31396507],\n    [5.7433291, 10.23432855, -1.7129558],\n    [5.42176832, 2.52761295, -2.28632458],\n    [-9.45673506, 0.84492645, -0.37130913],\n    [11.28131752, 15.43200474, 6.65806894],\n    [11.25562486, 14.54476269, -4.64185679],\n    [5.53665318, 20.67696068, -2.09829731],\n    [6.24979379, 3.73527961, -0.19059094],\n    [6.05762713, 5.84889942, -9.83140315],\n    [11.34836526, 7.80851894, 6.51965563],\n    [4.62922328, 12.54565264, -3.27070438],\n    [3.88597741, 17.65496259, 6.17021748],\n    [6.09040706, 12.80852517, -0.83349675],\n    [4.98923228, 2.97183159, -2.8782543],\n    [4.31831015, 6.77654364, -10.53384521],\n    [4.30648895, 20.11727773, 5.78794618],\n    [2.59423262, 12.53698395, -4.54270217],\n    [3.95395557, 8.62692975, 6.11243484],\n    [-0.02429998, 7.4352724, 7.85703681],\n    [6.3398709, 1.58844699, 1.50746241],\n    [6.01869697, 16.47012319, 3.06507259],\n    [0.70415449, 7.7466722, -4.80654807],\n    [5.19192276, 4.55229095, 12.73206923],\n    [9.18178458, 13.32628919, 9.73366853],\n    [6.5502299, 8.93943139, 11.9598403],\n    [4.72367491, 9.6064527, -3.17478658],\n    [6.34581918, 14.64753812, 0.54955323],\n    [9.24500012, 11.79369975, 9.66393968],\n    [9.54218705, 18.37170203, 9.32230048],\n    [1.75081984, 15.02651711, 7.41840641],\n    [12.601052, 3.72181327, 1.3477256],\n    [12.51779379, 6.4347312, -0.48887104],\n    [4.42082809, 3.39523891, 13.08470694],\n    [12.43037807, 13.24812235, -1.09495397],\n    [11.09369122, 15.64626259, 7.02494162],\n    [-9.41659692, 6.77682741, -2.0676137],\n    [12.60580513, 10.18949219, 0.94048968],\n    [4.79385999, 1.62751429, 5.25709001],\n    [1.42409054, 19.99621531, -4.7940951],\n    [-5.9432868, 11.49566492, 6.23700443],\n    [9.84608487, 7.61662726, -6.7936431],\n    [-2.97012464, 20.87154095, 7.66840673],\n    [-3.78808993, 6.20248289, -9.93347304],\n    [-8.09109113, 13.5439202, 3.88812014],\n    [-1.94526534, 16.40543157, 7.8594572],\n    [3.37725903, 6.73427783, -4.20023195],\n    [2.18260996, 10.9808, -4.66465822],\n    [12.2538423, 14.17331298, -1.95224796],\n    [-3.58745667, 14.74504765, 7.48564561],\n    [6.2817215, 16.46669624, -0.00915335],\n    [4.60124278, 1.52827391, 5.47997853],\n    [-4.35618994, 14.37024175, -9.6315042],\n    [2.13533057, 15.40890678, 7.26044861],\n    [11.66390434, 5.01064442, 5.80382567],\n    [6.25953795, 15.03770352, 2.12531083],\n    [-6.97076522, 20.55338737, -7.5112793],\n    [5.48480022, 3.58901443, 12.5826365],\n    [-6.92549248, 12.20969922, -7.56199456],\n    [5.44653741, 16.31855163, 4.32387036],\n    [12.29441665, 17.75830125, 3.79029361],\n    [2.84603717, 2.94229036, 6.89869935],\n    [-1.24800989, 3.28858041, 7.9126011],\n    [2.56655831, 17.51901644, -4.55209358],\n    [-4.23407672, 8.86042395, 7.23563509],\n    [-9.42834435, 14.94327951, -0.03421047],\n    [-7.24138736, 2.7101173, -7.19317848],\n    [-9.19273969, 12.69473007, 1.30035888],\n    [-9.25063343, 20.26818152, 1.05783357],\n    [3.1527838, 20.68514372, -4.31424557],\n    [6.36074527, 17.5196472, 1.04554292],\n    [4.3772878, 10.34191952, -3.50546943],\n    [12.52915803, 7.23070999, -0.39024239],\n    [-3.77133484, 0.87217498, -9.94171654],\n    [-6.30584026, 17.99294701, 5.94813829],\n    [12.2693674, 8.84908718, 3.89931927],\n    [3.86110068, 19.64905301, 13.30585603],\n    [12.48585667, 5.97180791, 2.73585372],\n    [1.52311473, 15.68329522, 7.50055389],\n    [-6.61668276, 6.95179101, -7.89072957],\n    [5.45751823, 13.78570209, -2.22950852],\n    [5.57564272, 19.69873111, -2.03065558],\n    [-9.46772669, 10.53690656, -0.56910505],\n    [-9.47114535, 1.9365499, -0.65339504],\n    [7.21026985, 0.10135008, 11.50459741],\n    [6.36096027, 6.47941012, 1.01172813],\n    [-5.25528799, 9.10220351, -9.05521994],\n    [1.00409689, 11.59121384, -4.81440925],\n    [0.34952479, 12.79532865, -11.02173525],\n    [7.07135617, 5.90826433, 11.60525068],\n    [9.10414674, 2.12121685, -7.61610888],\n    [-9.13399816, 0.25393018, 1.52100015],\n    [4.15161677, 12.78203788, -3.69266238],\n    [5.10011276, 13.18989975, -2.74104088],\n    [-3.61799236, 1.88819289, 7.47522145],\n    [-6.47177577, 5.53708564, 5.80480478],\n    [-7.96813607, 6.14447986, 4.07925246],\n    [-9.46311926, 15.77750804, -0.47685411],\n    [10.00346835, 9.53470295, 8.74124915],\n    [11.72898907, 6.91332187, -3.5987422],\n    [5.9470408, 12.7901538, 12.32825978],\n    [12.26320873, 9.46255155, 3.92547593],\n    [-9.38822975, 4.48344336, 0.29864154],\n    [9.78991479, 3.22969719, -6.86155209],\n    [-6.30827905, 1.29334933, 5.94608448],\n    [12.361408, 11.64379715, -1.46748043],\n    [11.30324182, 7.50511035, 6.61326375],\n    [5.99099333, 20.22954069, 3.14584202],\n    [2.55340503, 6.36581665, -10.93078332],\n    [-8.28459554, 5.99454125, -5.64063984],\n    [-0.37540677, 3.32674247, 7.89258732],\n    [2.44500437, 16.3876331, 7.11455853],\n    [-3.21696928, 16.14180269, -10.19436565],\n    [3.10091446, 10.59185756, -4.3387045],\n    [-7.81614116, 5.61665964, 4.30163126],\n    [0.87011369, 5.12098907, -4.81321782],\n    [-1.37994813, 18.6569221, -10.78029786],\n    [5.97314654, 1.72113245, 3.19616827],\n    [-9.02535826, 5.11330826, -3.90236601],\n    [7.95426716, 7.35276076, 10.91705601],\n    [-7.92300608, 15.40103177, -6.25234414],\n    [-7.81568354, 1.55017873, 4.30227939],\n    [-6.52807472, 0.98103109, 5.75447049],\n    [5.74224614, 0.10326456, -1.71516719],\n    [-6.75182519, 1.59190599, 5.54536644],\n    [0.5991863, 4.09046489, 14.08190054],\n    [7.52957906, 1.40338925, 11.26278159],\n    [-7.9410127, 4.76624703, -6.22410593],\n    [0.54610649, 18.93896027, -4.79479657],\n    [5.71716963, 6.33811849, 3.81145832],\n    [3.47614624, 14.01604482, 6.48991841],\n    [7.09570829, 2.58574642, 11.58779908],\n    [11.77508873, 20.12065245, 5.52000733],\n    [-7.13864965, 0.63987368, 5.14613712],\n    [0.89565756, 4.17831196, -11.04025234],\n    [4.15011698, 2.12071053, -3.69383987],\n    [4.86894048, 17.37555409, 12.88679422],\n    [11.6913201, 15.43163044, -3.69219589],\n    [3.08433424, 1.35716429, 6.75477706],\n    [6.05238036, 7.36953597, -0.9531857],\n    [11.79590859, 19.14943813, 5.46458363],\n    [-4.20142934, 15.86864982, -9.71820285],\n    [3.73555399, 17.99281478, -3.98895873],\n    [9.17794322, 13.13426478, 9.73787396],\n    [3.9634721, 5.38358719, 6.10422933],\n    [7.59716226, 11.3187506, 11.20966007],\n    [-0.01103023, 10.57013661, 7.85537805],\n    [-4.51770169, 13.92496597, -9.53731381],\n    [-0.90475834, 19.06530699, 7.91624394],\n    [-6.1946827, 1.78627047, 6.04013266],\n    [-9.37443576, 8.35945887, 0.39441501],\n    [5.84155867, 3.26076758, 3.53371073],\n    [2.50670401, 20.35062954, -10.93729414],\n    [2.39628476, 15.93318147, 13.75758567],\n    [-3.02053692, 7.88756689, -10.27487648],\n    [-4.82170487, 20.88248788, 6.95192576],\n    [5.85884411, 5.92115656, -1.46234971],\n    [12.60399379, 3.93198339, 1.21795084],\n    [12.26568109, 7.87485496, -1.90309687],\n    [-8.82042808, 20.52858812, 2.44859376],\n    [2.3772035, 2.11304822, -4.61169746],\n    [0.06347708, 0.05088757, -4.72539502],\n    [-5.90551322, 5.82067884, -8.54985496],\n    [4.34659564, 12.02800538, -10.52501035],\n    [7.89846129, 17.33800614, -8.67762127],\n    [7.12876977, 14.69956143, 11.56397524],\n    [8.28299708, 16.80131632, -8.37018192],\n    [-2.19313184, 12.03371615, 7.82571806],\n    [-9.47534143, 9.25007197, -0.80221252],\n    [-5.79121753, 20.25213489, -8.64480938],\n    [1.64658927, 11.86884501, 13.92241276],\n    [12.05525654, 16.85064776, 4.69676261],\n    [5.80764757, 2.73570108, -9.95331037],\n    [6.35900294, 8.69661188, 0.82866805],\n    [-6.72997637, 16.21024445, 5.56644629],\n    [-9.46929759, 17.90890879, -1.3840838],\n    [-5.6947518, 16.2829144, -8.72282809],\n    [12.14933621, 17.26681116, 4.37109197],\n    [12.23583401, 7.1123822, 4.03889424],\n    [2.03068953, 3.49669124, -4.70029534],\n    [4.53077988, 5.85285972, -3.36593169],\n    [4.38959297, 2.78110543, 5.70480142],\n    [-9.07486681, 3.12246385, 1.72354229],\n    [5.32909709, 14.41859745, 4.51775806],\n    [6.02755813, 14.62919674, 3.0385062],\n    [-4.90210598, 15.46834082, 6.90860782],\n    [-0.99808115, 19.21451886, -10.85813638],\n    [9.20902847, 16.72393706, 9.70373876],\n    [11.46872605, 15.43412385, -4.2041334],\n    [12.34279577, 17.12224186, -1.55833575],\n    [-9.43790935, 15.33828436, -1.85908374],\n    [1.99016484, 8.11488289, -4.70896748],\n    [-3.83241639, 7.27475864, -9.9114869],\n    [2.99627921, 0.02675155, -10.85900779],\n    [12.33643285, 6.58530293, -1.58862717],\n    [6.05436989, 16.65693804, 2.9557556],\n    [1.05202781, 16.82225469, -11.04056566],\n    [-1.66199583, 19.32661758, 7.88845224],\n    [5.80712869, 2.13211514, -9.95355567],\n    [-6.2964307, 20.19366583, 5.95604802],\n    [-5.14501005, 6.30899751, 6.7707572],\n    [-5.9666599, 20.09866973, 6.21933018],\n    [7.50833768, 11.59772557, 11.27933456],\n    [2.36291408, 2.84039175, -4.61586971],\n    [6.1310463, 18.24281247, 2.69540916],\n    [-9.39227363, 7.97328725, 0.26914986],\n    [-9.40321575, 13.49050263, 0.18556303],\n    [-9.46509247, 9.16614047, -0.51417707],\n    [5.76390055, 18.5098781, 3.71098125],\n    [1.59744539, 10.63620367, -11.02433937],\n    [1.79854898, 6.52625605, -11.01156319],\n    [5.85603983, 10.37783546, 3.49905603],\n    [-8.50271151, 9.86210371, -5.21582292],\n    [6.25985428, 6.17816142, 2.12354282],\n    [6.15791552, 9.28390112, 2.59359896],\n    [11.2397899, 1.99368337, -4.67268335],\n    [9.38162489, 10.78751905, 9.50978695],\n    [5.13648602, 19.69917224, -10.24477884],\n    [0.28665043, 17.79842244, -11.01785707],\n    [1.86140801, 17.70020377, 7.37550435],\n    [3.83204229, 13.94018893, 6.21504835],\n    [11.46178071, 8.76352008, -4.21912963],\n    [2.43912605, 0.67611828, -4.59309238],\n    [11.34364384, 1.05018713, 6.52953999],\n    [-1.32375274, 10.33976747, 7.90979692],\n    [2.89460039, 11.95668048, -4.42926939],\n    [7.8589346, 6.8281969, -8.70774987],\n    [9.99859098, 6.14712165, -6.6039573],\n    [3.46544093, 2.75516009, 6.49765572],\n    [-9.47161899, 8.81006602, -0.66678511],\n    [10.72045337, 10.04258051, -5.58138651],\n    [12.03931017, 1.89335016, 4.74896161],\n    [10.01173798, 9.7489887, 8.73020013],\n    [3.99796396, 7.57771853, 6.07424541],\n    [-9.32683759, 9.35436205, -2.68192582],\n    [5.51926424, 7.23713298, -2.12780584],\n    [0.59878822, 16.24718722, -11.03355475],\n    [7.81590619, 3.76358506, 11.03285825],\n    [12.60505652, 8.71746935, 1.14541992],\n    [-9.00606252, 20.06165813, -3.96260683],\n    [0.35454903, 4.64864947, -4.77337996],\n    [-8.14689924, 12.92547141, -5.88543586],\n    [1.37525108, 7.4975037, 7.54955642],\n    [-5.47016524, 6.7296145, 6.56899967],\n    [5.38624612, 15.05021673, -2.34137344],\n    [-7.25126369, 8.45229537, -7.18105549],\n    [1.98976683, 0.73151342, -4.70905093],\n    [-9.47714535, 11.06303073, -0.94141357],\n    [7.27530961, 10.55097285, -9.12304797],\n    [-3.53097316, 0.01383815, 7.50457704],\n    [3.04359826, 3.98678095, 6.78025648],\n    [11.63679475, 18.62236351, 5.87017886],\n    [2.91401955, 11.868894, -10.87371853],\n    [5.82280496, 16.1509102, 3.57780421],\n    [12.60238486, 14.5760398, 1.29598465],\n    [0.1176621, 11.53465036, -11.00563601],\n    [4.83888272, 5.03822054, 5.20217929],\n    [-2.93194237, 15.45594742, -10.30970463],\n    [0.48833024, 0.46655802, -4.78917125],\n    [-9.24269689, 17.21720597, 1.09286587],\n    [9.20316205, 15.70190032, -7.51495278],\n    [-8.11047755, 12.99992421, -5.94759208],\n    [-5.56341682, 11.64913545, -8.8260334],\n    [0.89170045, 9.11693995, -11.04021566],\n    [0.8646575, 9.06777345, -4.81309017],\n    [-4.82983027, 4.81220433, -9.34407639],\n    [0.56019485, 8.48648438, 7.76174327],\n    [4.37026041, 7.31233976, -3.51161109],\n    [10.8547143, 15.24431357, 7.45493612],\n    [-9.47718815, 19.83729967, -1.03939175],\n    [-9.29031076, 13.00373561, -2.87449023],\n    [-6.36275192, 15.36969199, 5.89980572],\n    [12.60012589, 14.14331356, 0.61450063],\n    [5.81088842, 6.13381954, 3.60537995],\n    [3.87696013, 5.59306171, -3.89481542],\n    [11.10792913, 19.44556782, -4.92135198],\n    [2.86219443, 19.08417168, -4.4425404],\n    [-9.04903241, 4.30529595, -3.8265574],\n    [3.74239517, 7.85611232, -3.98455075],\n    [12.60357326, 12.34155331, 0.75313191],\n    [-9.32339508, 10.10850062, -2.70102594],\n    [10.59578128, 18.21811348, -5.77498426],\n    [-5.82897872, 0.51797505, 6.321654],\n    [3.52145295, 12.22948375, 6.45684404],\n    [6.73503903, 18.8845359, 11.83812236],\n    [-4.26024317, 4.42108256, 7.22418645],\n    [1.92345172, 7.06781192, -4.72248495],\n    [4.83891432, 13.95433521, -3.05124739],\n    [12.36657513, 11.77361769, 3.4481163],\n    [3.29979683, 0.09416538, 13.5000047],\n    [6.06021219, 19.02784327, 2.93722125],\n    [12.31246447, 6.38994103, -1.69951716],\n    [4.20787683, 3.15661216, -10.56756465],\n    [-3.50992086, 15.26826299, 7.51151708],\n    [3.67866439, 5.69096418, -4.02505182],\n    [4.23700065, 1.61247185, -3.62421353],\n    [1.27488189, 7.32462386, -11.03718705],\n    [10.29556247, 4.34428818, -6.21033762],\n    [-0.16432605, 17.50756184, -10.97937897],\n    [4.75209949, 7.1181983, -10.38933435],\n    [0.44346509, 6.10329572, -4.78430769],\n    [-7.63301286, 2.55616106, -6.68077377],\n    [-9.47498197, 8.15792291, -1.20397191],\n    [8.7979396, 1.59395401, 10.13686886],\n    [6.69524513, 2.18849218, 11.86469681],\n    [-6.74889247, 20.60677591, 5.54820453],\n    [12.16243832, 12.84567217, 4.32302805],\n    [6.18323611, 3.06971496, 2.49105418],\n    [6.01887607, 17.0489327, -1.05208758],\n    [3.54992646, 17.63831891, -4.10310671],\n    [6.08936601, 0.10844492, 2.84176066],\n    [-5.04920645, 0.56150948, 6.8263997],\n    [-9.46578573, 13.42557794, -0.52800229],\n    [6.12076913, 17.49023208, 2.73270283],\n    [-6.17954918, 3.87426946, 6.05241633],\n    [-8.60224837, 17.20952077, -5.00356863],\n    [-3.79156681, 20.79433091, 7.41341343],\n    [-7.51321729, 8.75562251, -6.84481511],\n    [1.10145608, 11.05790765, -11.04020413],\n    [4.56746174, 5.4577204, -3.33099388],\n    [6.17453501, 3.72451825, 12.19433571],\n    [-9.43859471, 16.81870388, -0.14025644],\n    [6.89462801, 6.6286802, 11.72948947],\n    [3.69406382, 1.40513988, -4.01538034],\n    [-7.34432123, 13.74035769, 4.9116108],\n    [-6.85639863, 5.18748609, -7.63810342],\n    [-7.34860447, 16.44101179, 4.90654387],\n    [8.29074022, 13.9049743, 10.62182902],\n    [5.44662705, 6.94592814, -2.24697224],\n    [-1.46734963, 20.11162257, -10.76045855],\n    [9.02392798, 19.55197246, -7.69631424],\n    [5.97771219, 6.5877462, -1.16650159],\n    [7.47988343, 12.30306392, -8.98358819],\n    [1.82503545, 17.38688446, 13.88708684],\n    [0.17767577, 2.41797825, -4.74648035],\n    [-1.63810091, 14.12366422, -10.71949123],\n    [5.88638256, 17.4340991, -9.91571993],\n    [12.49876229, 0.02031415, 2.6400017],\n    [8.51989295, 18.56741116, 10.40899947],\n    [0.8673597, 2.69609331, 7.69296694],\n    [0.34873548, 0.6232146, -4.77260547],\n    [4.06956032, 18.4227781, -3.75586119],\n    [4.29158857, 14.28829805, 13.13828865],\n    [-9.25933893, 17.77338568, -3.0231234],\n    [-3.2456347, 7.52485346, 7.59334776],\n    [5.21265378, 17.57705164, 4.69687797],\n    [3.54084664, 10.21214184, 13.41992889],\n    [-2.95299672, 4.10610572, 7.67273347],\n    [12.11884289, 1.30642208, -2.46036719],\n    [-9.47554847, 16.40000633, -0.81268906],\n    [0.92782626, 8.68732058, -11.04049778],\n    [3.70172911, 10.57273094, -4.01053907],\n    [12.3872413, 12.94397538, 3.34060489],\n    [6.3596338, 16.41864444, 0.85616289],\n    [-7.9111896, 8.23933225, -6.27076269],\n    [-6.26566138, 16.58858831, 5.98175358],\n    [3.37812941, 3.83734792, 13.47452059],\n    [-7.94594789, 0.70253853, 4.11263163],\n    [-2.7944145, 20.06634664, 7.71092918],\n    [-1.5540607, 20.12695528, 7.89683053],\n    [2.49865902, 3.01916596, -10.93839546],\n    [-8.36800472, 12.55981801, 3.41352341],\n    [5.22721792, 9.83689994, -2.57234913],\n    [10.47084609, 4.01465493, 8.07577536],\n    [3.07560641, 20.69897246, -10.84422151],\n    [-9.43967166, 11.18912572, -0.1521841],\n    [4.79579307, 14.74894143, 5.25475545],\n    [0.79062295, 1.97199731, -4.81074344],\n    [12.16699566, 11.33081205, 4.30613589],\n    [-3.75412428, 10.73107315, 7.42711622],\n    [4.32886347, 1.63803329, 5.76582894],\n    [-3.3013088, 16.669283, -10.15838327],\n    [5.45608624, 15.80270468, -2.23181256],\n    [5.00111016, 15.36512194, -2.86396549],\n    [4.93866614, 5.74507531, 12.85426572],\n    [-9.4234495, 17.49104976, 0.01244129],\n    [-6.87873729, 17.52868308, -7.61366481],\n    [-9.46866135, 17.5374019, -1.39928829],\n    [-0.90392149, 12.87036609, 7.91623464],\n    [6.18951066, 16.26833148, -0.4702797],\n    [-0.85552941, 19.37277951, 7.91554637],\n    [12.33187964, 2.79376583, 3.61842486],\n    [5.25588882, 11.04386537, -2.53244505],\n    [11.16860606, 16.41953337, 6.8818557],\n    [-6.54016179, 12.211683, 5.74354796],\n    [1.73961061, 12.17746339, -4.75488612],\n    [-0.28488445, 8.73968421, -10.96590111],\n    [11.7142917, 18.70734494, 5.67762732],\n    [3.55024469, 16.03396452, -10.74332526],\n    [12.60436753, 11.12709945, 0.79837264],\n    [6.32414558, 12.04173024, -9.69303326],\n    [9.37394658, 19.63008045, 9.51857862],\n    [-5.14641819, 18.16308386, 6.76992677],\n    [0.98496101, 11.49504399, -4.81446724],\n    [6.05594593, 18.31928124, -0.94231887],\n    [11.11159075, 14.042427, -4.91463133],\n    [6.35869106, 20.40877215, 1.15983541],\n    [-9.37410419, 16.86405121, -2.39295929],\n    [-9.47680298, 12.4590394, -1.09105948],\n    [-9.43862046, 6.03791649, -1.85124712],\n    [6.17519366, 20.20104478, 2.52439778],\n    [2.20009617, 18.29589992, -4.66023726],\n    [5.99578554, 1.65576537, -1.11715678],\n    [11.18317097, 3.84372312, -4.78117205],\n    [5.53466904, 1.61257183, 4.16820673],\n    [-5.2487189, 12.5521913, -9.05991653],\n    [4.93303926, 9.71290717, 5.08356735],\n    [3.74382626, 14.7536552, -3.9836268],\n    [3.13111586, 7.65415415, 13.55314343],\n    [4.66159147, 3.84270901, 5.4121466],\n    [2.81572649, 14.84206724, 6.91613947],\n    [7.10412348, 2.99496085, 11.58174949],\n    [-1.88592718, 3.06986314, -10.65476532],\n    [9.63660653, 2.61086759, -7.04183732],\n    [6.16582042, 6.15217363, 12.19957572],\n    [9.92021403, 3.94659523, 8.85116779],\n    [-7.5188064, 19.52724364, -6.83731373],\n    [-9.47728176, 13.10625841, -1.00993349],\n    [-3.74244014, 14.93903458, 7.43135038],\n    [11.39015546, 12.65765968, 6.43122135],\n    [4.31572789, 17.13462906, -10.53464773],\n    [5.24643535, 19.85979079, 12.70492389],\n    [5.87582192, 6.92838142, 3.4508175],\n    [5.95335505, 2.08095667, -1.23097851],\n    [-9.28522659, 1.70164748, 0.89709069],\n    [10.49909571, 6.27361871, -5.91971986],\n    [-6.78846021, 2.78309506, -7.71146331],\n    [9.90086175, 5.07066187, 8.8763812],\n    [-8.67129047, 17.78578717, 2.80071005],\n    [2.19086918, 5.19600509, -4.66257833],\n    [12.43040252, 16.94196127, 3.09788978],\n    [5.10029457, 19.62225802, -10.25906684],\n    [-8.41650557, 1.21800784, -5.38977798],\n    [12.57832486, 1.57968741, 1.83144584],\n    [3.81913271, 0.93674051, -3.93409267],\n    [2.91528393, 7.11641766, 13.61771334],\n    [1.66609636, 7.24979965, 7.44993373],\n    [-6.75627239, 1.48906974, 5.54105764],\n    [5.62326217, 20.90913604, -1.9450963],\n    [-5.05757277, 13.22078267, -9.19326522],\n    [4.61694024, 11.25573834, 5.46250247],\n    [-6.34412953, 16.08907921, -8.15820808],\n    [10.63920934, 20.68478738, 7.81290083],\n    [-8.16111094, 10.63996389, 3.77427726],\n    [-5.42201444, 17.98986971, -8.93340439],\n    [-9.12578158, 6.83357564, 1.55020747],\n    [6.88306428, 13.08051611, 11.73747302],\n    [6.05684674, 18.19987502, -0.93956248],\n    [6.28141425, 20.08377426, -0.01106354],\n    [-1.06500681, 9.70301364, 7.91639313],\n    [9.79709465, 5.34802496, -6.85292853],\n    [-9.22591665, 0.7652322, 1.16494521],\n    [5.93862884, 8.20388954, 3.2900805],\n    [-5.25605952, 14.1519058, 6.70413007],\n    [6.3837549, 5.31305765, -9.66086697],\n    [-9.34318851, 10.64116378, -2.58801175],\n    [-0.70795571, 10.45229065, -10.90781136],\n    [6.05630832, 1.45919688, 2.94962688],\n    [-1.85212118, 13.67564763, -10.6639658],\n    [6.22064833, 13.66297803, 2.3246352],\n    [-3.199839, 5.94272537, 7.60654191],\n    [-9.46267974, 9.94678496, -0.46890345],\n    [-2.00044279, 2.89967645, 7.85262463],\n    [0.45599332, 3.63610625, 14.09745963],\n    [2.41802752, 17.02242174, 13.75215548],\n    [5.9278506, 15.56156951, 12.33928764],\n    [4.93413221, 13.423613, -10.32284735],\n    [0.99915092, 16.35520192, 14.03054334],\n    [-9.32311682, 5.68672998, 0.70262972],\n    [-2.46486278, 20.20524868, -10.47854364],\n    [-3.17244161, 20.0109507, -10.21301767],\n    [10.29306801, 11.93640366, 8.33927129],\n    [-5.37675505, 6.99307973, 6.6290553],\n    [1.35589495, 17.78749875, 13.97482237],\n    [4.57628448, 1.58475759, 5.50752676],\n    [-3.17525198, 14.5783939, -10.21184745],\n    [7.09731258, 8.16235268, 11.58664653],\n    [12.50114725, 4.38415572, 2.62166143],\n    [5.29660167, 6.43599442, -2.47451684],\n    [-3.72491348, 14.04319513, -9.96436467],\n    [9.38792604, 16.87281961, -7.31953179],\n    [4.16077212, 12.93563346, 5.92736846],\n    [-5.3976888, 15.81872297, 6.61574638],\n    [6.18277881, 13.6931034, -0.49804191],\n    [-7.96057023, 8.17733191, -6.19319822],\n    [6.15580942, 11.286278, 2.60182253],\n    [5.56776118, 13.18907695, -2.04449842],\n    [5.95559246, 10.63665137, -1.22514796],\n    [-9.03053575, 15.70837466, -3.88597032],\n    [3.06473602, 18.33178946, -4.35535679],\n    [12.58280032, 14.09650402, 0.23417547],\n    [0.90799653, 14.40025496, -4.8139331],\n    [-8.06500986, 10.13458813, 3.92956857],\n    [6.22136812, 7.84474729, 2.32122306],\n    [11.04498948, 11.28130852, -5.03532394],\n    [6.36051762, 7.20954733, 0.90949498],\n    [2.6357712, 14.66611826, 7.01582793],\n    [11.09544351, 12.31031817, -4.94419279],\n    [-4.93935905, 5.77351781, -9.27261397],\n    [5.00971792, 4.74601751, -2.85355087],\n    [11.43831001, 6.63807929, 6.32713239],\n    [6.07283571, 15.97730192, -0.88987287],\n    [6.0335698, 4.67858974, 3.02026837],\n    [1.28431482, 4.2938609, 13.98675865],\n    [-2.27992619, 11.96177583, 7.81204479],\n    [1.18143677, 20.73034349, -4.81026869],\n    [6.43961185, 7.68935974, 12.03065569],\n    [0.23839815, 19.10432274, 7.81987179],\n    [-6.51512802, 16.30336814, -7.99273038],\n    [-8.31391581, 16.98680447, -5.58635483],\n    [9.06033909, 8.62488821, 9.86490498],\n    [-9.47625216, 0.4635807, -0.85402702],\n    [4.02145845, 17.62702022, 6.05360022],\n    [8.16548244, 0.88843843, -8.46696497],\n    [-8.9838694, 10.13967428, 2.00596379],\n    [-0.83623931, 2.35037183, -10.88684357],\n    [2.78976271, 0.4273365, 6.93092751],\n    [-7.40749902, 8.41097489, 4.83607613],\n    [-4.07236922, 10.39055184, -9.78791913],\n    [-6.20607875, 3.56672804, -8.2863697],\n    [1.35033842, 5.84902076, -4.80029411],\n    [5.83151255, 3.82552406, -1.52448351],\n    [-8.28483628, 11.70322644, -5.64019737],\n    [2.99787415, 6.23336687, -10.85871631],\n    [2.17921721, 19.50976176, 7.24081747],\n    [-9.38784164, 2.97615214, -2.29679281],\n    [5.96390174, 5.99667829, 3.22174855],\n    [-8.04949257, 13.81871622, 3.9539914],\n    [11.28454038, 17.42551197, -4.58499959],\n    [0.5921885, 16.43592456, -11.03331483],\n    [-2.09286511, 9.76988448, 7.84031106],\n    [-7.56980225, 10.91098349, 4.63380879],\n    [6.70242802, 2.11246651, -9.48109822],\n    [11.97525385, 7.1122447, 4.95114575],\n    [-1.853731, 2.44455543, 7.86993848],\n    [12.60589483, 14.64181732, 0.96656689],\n    [5.09490443, 14.59220534, -2.74768404],\n    [-0.22053996, 0.21078255, -10.97326308],\n    [-7.56082812, 19.15814931, -6.78044607],\n    [2.00301756, 13.8367927, 13.84943815],\n    [-5.51732243, 19.33516012, 6.5380196],\n    [0.77454046, 17.09259726, -4.81008202],\n    [11.88357976, 7.05959255, 5.22223515],\n    [12.10322728, 15.80150215, 4.53472121],\n    [9.26921032, 16.18634071, -7.44611168],\n    [-9.4661205, 16.94277537, -0.53482799],\n    [-6.46176534, 17.08161524, 5.81366242],\n    [5.82307852, 15.52884317, -1.54327674],\n    [-3.48058838, 10.45205717, 7.52108215],\n    [3.6106127, 6.64516035, 6.39017315],\n    [1.59031817, 13.02416839, -4.77600803],\n    [4.20132619, 5.81720564, -10.56952623],\n    [5.57424669, 6.58946441, -2.03311391],\n    [-7.78475074, 4.11542664, 4.34580707],\n    [-2.91347949, 7.7876348, -10.31684812],\n    [0.54947172, 11.25606715, 7.76390739],\n    [11.18817476, 15.84956135, 6.84375884],\n    [5.77607797, 4.77319603, -1.64495333],\n    [12.54244574, 5.85757716, 2.26213447],\n    [1.3151522, 11.24810046, -11.03609724],\n    [6.35285163, 10.44839605, 1.31060098],\n    [6.1728297, 19.3850004, 2.53405652],\n    [-4.03095697, 7.1572094, 7.32090959],\n    [5.65572309, 4.4004285, 3.93744533],\n    [12.15472852, 11.65721866, 4.35139945],\n    [12.2649096, 19.40163358, -1.90632692],\n    [1.61127087, 20.69977191, 7.46972463],\n    [-4.30383035, 12.89401082, 7.2048776],\n    [5.89220996, 9.32936465, -1.38377349],\n    [-8.66797021, 8.55688838, 2.80810209],\n    [-6.06824121, 12.63313092, -8.40976077],\n    [6.32716094, 8.21372886, 1.64536444],\n    [4.6702708, 20.51725842, 12.97689368],\n    [3.61089891, 1.28044345, 6.38995569],\n    [3.7491159, 17.69158421, 6.28228048],\n    [11.51609826, 8.88434517, 6.15369484],\n    [1.61567566, 17.08598593, -11.02333218],\n    [5.46987073, 3.68405937, -10.10637165],\n    [2.80336154, 16.46977888, 6.92319941],\n    [-1.84757928, 3.74639429, -10.66519292],\n    [-3.67775637, 12.69670072, 7.45443219],\n    [12.57236122, 4.48004318, 0.07838943],\n    [4.52766414, 14.64131342, -10.46652292],\n    [0.45728431, 4.12921238, 14.09732598],\n    [-3.48669783, 0.51691372, 7.51909994],\n    [-6.72289977, 3.39064286, -7.78091197],\n    [2.2685462, 2.74231692, -10.96738138],\n    [-8.48742605, 1.81028308, 3.18593572],\n    [-1.49512168, 2.46051283, 7.90078523],\n    [6.34239074, 1.45029468, 1.47550467],\n    [3.6809035, 14.13394809, 13.3711321],\n    [3.34173768, 4.28502893, -10.79028548],\n    [7.6600934, 12.73571569, 11.15956488],\n    [6.23874165, 10.48190969, -0.24692683],\n    [7.82081145, 13.93686441, -8.73655807],\n    [9.4114013, 18.36473565, -7.29405697],\n    [-7.20874646, 10.44706769, -7.2329754],\n    [-8.27209696, 16.90406282, 3.58559094],\n    [-9.04368029, 5.59278093, 1.82391426],\n    [2.655589, 12.13740042, -10.9158361],\n    [-9.39760807, 3.92064544, -2.22379997],\n    [0.76452381, 5.45531427, 7.71746368],\n    [12.60143647, 6.62935714, 0.66039499],\n    [12.10614689, 5.79345452, -2.50405231],\n    [11.96759072, 1.05526872, -2.94645934],\n    [4.85962149, 19.69529851, 5.17650123],\n    [2.665549, 8.41914546, -4.51768818],\n    [-9.37679259, 4.7405001, 0.37853799],\n    [2.85086189, 13.69305981, -4.44712134],\n    [-9.31459748, 0.57699202, 0.74833698],\n    [0.28797263, 13.84688223, 14.11382866],\n    [5.25792698, 19.80607923, 4.62866321],\n    [10.36814562, 6.70638885, 8.22965894],\n    [5.06194113, 13.29591228, -2.78926485],\n    [5.74519213, 8.77304827, -1.70914605],\n    [0.90307573, 14.12080951, 7.68410811],\n    [11.68536201, 14.70021788, 5.75055147],\n    [3.53317689, 16.24064654, 6.44819806],\n    [6.35614972, 19.12805125, 1.23692659],\n    [4.19013107, 20.24238803, -3.66213329],\n    [4.02378074, 2.6718563, 13.24447984],\n    [1.06633343, 18.82718163, -11.04048377],\n    [-8.73986013, 7.71880889, 2.64392147],\n    [-9.38860265, 20.82373441, -2.29125443],\n    [3.66363184, 16.69917135, 6.3495058],\n    [-9.47012158, 10.6023231, -0.62611705],\n    [7.50322226, 13.78849333, -8.96726925],\n    [2.90117132, 5.17725021, -4.42654732],\n    [9.95020824, 18.00163836, 8.81184046],\n    [-4.14922269, 7.61963527, -9.74668352],\n    [11.13528086, 15.31303164, 6.94603751],\n    [-9.34975224, 3.23567856, 0.55063758],\n    [5.57921356, 10.4399444, 4.08575269],\n    [6.15843267, 18.32898384, 2.59157293],\n    [2.4120058, 11.75523405, 13.75366311],\n    [-9.04635914, 4.22084065, 1.81544943],\n    [10.45356386, 13.80161056, -5.98633972],\n    [4.40796752, 11.08258169, 5.68604361],\n    [3.72575869, 11.18310085, 6.30085517],\n    [-3.34255814, 16.86798991, 7.56446859],\n    [1.67174315, 15.52183638, -4.76506206],\n    [3.98750217, 11.0080951, 13.25837003],\n    [5.0044622, 16.20853306, -2.85991582],\n    [-8.89589192, 13.81296773, 2.2529198],\n    [1.76551138, 7.73381367, -4.75074957],\n    [2.31853658, 12.33138393, 7.17623474],\n    [-7.47900512, 15.01466393, -6.89041981],\n    [5.82461393, 8.60363142, -1.53986838],\n    [-3.74746931, 1.80940914, 7.42953032],\n    [1.48102878, 8.062214, 7.51484426],\n    [12.39198912, 0.13903103, -1.31005014],\n    [6.22501762, 13.88620367, 2.30378269],\n    [1.87624155, 15.2125167, 7.3695971],\n    [-5.84847976, 13.44961807, -8.59758378],\n    [3.57680664, 9.07684902, 13.40756063],\n    [6.23403815, 15.34535655, -0.27009539],\n    [-0.0283997, 16.30207145, 7.85754462],\n    [-9.45532029, 11.53357657, -1.64034427],\n    [3.59681448, 13.41922739, 6.40063061],\n    [4.45391987, 20.55396369, -3.43711525],\n    [0.78075959, 3.2573418, -4.81034422],\n    [0.47311198, 2.77387042, 7.77886216],\n    [6.03680126, 9.54348831, -0.99987462],\n    [-0.18672304, 0.78150116, 7.87547339],\n    [2.97616419, 10.38512203, -4.39473153],\n    [-0.6569106, 3.51600339, 7.9096116],\n    [-1.02037396, 5.19592954, 7.91667739],\n    [2.8006065, 3.57206066, 6.92476815],\n    [0.58296908, 11.27885667, -4.79801226],\n    [-9.41908556, 10.02109474, 0.05225261],\n    [-7.06045336, 12.14730312, 5.23103075],\n    [-7.23818091, 20.75138294, 5.03474367],\n    [9.54433054, 1.78242614, 9.31974941],\n    [-2.96544787, 5.28489854, 7.66959206],\n    [-9.47678109, 17.4535852, -1.09318655],\n    [1.92079014, 3.17064829, -4.72300471],\n    [2.29680571, 9.30989509, -10.96408304],\n    [-0.89562274, 11.5732489, 7.91613766],\n    [6.16056055, 19.71582817, -9.77900671],\n    [-2.90544476, 9.89884562, -10.31994456],\n    [-6.103068, 3.87513355, 6.11363255],\n    [5.33925295, 0.44224098, -2.41216364],\n    [-6.46543583, 17.45931332, 5.81041782],\n    [10.62181766, 0.5204134, 7.84070021],\n    [-6.14388872, 8.77855441, 6.08113743],\n    [11.18969377, 3.05731189, -4.76880862],\n    [5.61589166, 13.40316314, -1.95856126],\n    [11.85155916, 19.00919151, -3.27798546],\n    [3.12835461, 7.7792221, 6.72682477],\n    [-8.18752325, 9.20184198, -5.81487252],\n    [5.02021859, 16.89350691, 12.81561363],\n    [-1.4085818, 5.7834015, -10.77388211],\n    [12.57724837, 15.85852768, 1.84755793],\n    [11.02018078, 12.48736056, 7.16130935],\n    [12.47065557, 13.32936423, -0.84136053],\n    [5.79926868, 2.36176081, 3.63195007],\n    [3.52338421, 13.23258192, -4.11859056],\n    [1.50273643, 5.9384242, 13.94913998],\n    [-5.17919659, 12.68524327, 6.75049211],\n    [6.24019288, 5.78462313, -0.23968494],\n    [-3.29673848, 4.53561858, -10.1603552],\n    [5.94696706, 8.42335455, -1.24752631],\n    [5.43682574, 7.07254093, 12.60772343],\n    [5.83603715, 20.05570828, -1.51432918],\n    [-9.21165711, 12.77590254, -3.23210049],\n    [-7.10545377, 14.45875101, 5.18245129],\n    [-8.30589034, 12.91679944, -5.60129365],\n    [8.91671755, 3.78304789, 10.01566435],\n    [-9.25397353, 4.60346437, 1.0428986],\n    [5.75202171, 13.30038507, -1.69512059],\n    [2.37407747, 16.45451768, 13.76309353],\n    [4.78227107, 16.15478589, 12.9265676],\n    [-4.5506191, 8.81520271, 7.08984212],\n    [0.95983301, 3.65733836, -4.81442809],\n    [4.20856222, 10.26099018, 5.88250466],\n    [-4.48273029, 10.51206298, 7.12246679],\n    [-4.38623592, 4.30524189, 7.16755271],\n    [-4.94497545, 10.84140858, -9.26889668],\n    [-3.15337977, 2.00129042, 7.6196332],\n    [2.32708204, 16.81850836, 7.17215977],\n    [6.11406143, 0.58132816, -0.75436705],\n    [-1.95156118, 16.97813048, -10.63656481],\n    [-4.63522155, 13.30046047, -9.46632649],\n    [6.89559337, 16.08113297, 11.72882219],\n    [2.60547148, 20.95452597, 7.03197819],\n    [11.96224339, 19.10529238, 4.99085171],\n    [-9.4772763, 10.93974307, -0.97608829],\n    [1.83017627, 16.44522632, -11.00922031],\n    [2.94455372, 11.63343377, 6.84068353],\n    [3.63314369, 18.50448852, -4.05321835],\n    [-5.73007393, 16.44690031, -8.69448192],\n    [0.28104078, 19.42323864, 14.11446032],\n    [10.03609414, 6.73687564, 8.69751784],\n    [5.24059175, 2.50383401, -2.55382496],\n    [-9.47426022, 12.70793063, -1.23445033],\n    [10.84103024, 0.48800361, -5.38605633],\n    [6.31362165, 7.66723108, 1.76575821],\n    [11.98365616, 12.69803211, 4.9252699],\n    [3.69710942, 3.97615552, 6.32342537],\n    [-9.43658231, 13.44263299, -1.87352076],\n    [1.3893681, 16.41538043, -4.79715323],\n    [12.45403973, 3.60403513, 2.95193129],\n    [6.34587712, 17.09102444, 0.55038829],\n    [-8.43812877, 12.96946242, -5.34693923],\n    [-8.32793578, 8.68041905, 3.48649454],\n    [-2.56860996, 2.96179631, 7.75956508],\n    [6.35161696, 13.24838702, 0.6430925],\n    [3.35921256, 14.922693, 6.57285694],\n    [2.91332736, 12.70367062, -4.42148382],\n    [2.12477287, 19.31235662, -4.67881007],\n    [4.98259723, 11.42409861, 5.01896571],\n    [6.28307295, 10.33002964, -0.00070569],\n    [4.13042193, 10.05683455, 5.9554379],\n    [4.57384431, 7.04109692, -3.32484941],\n    [0.00707521, 2.68084339, 7.8530775],\n    [-9.17843827, 7.48937662, 1.35614938],\n    [-3.87616772, 19.79081681, 7.38169477],\n    [2.56923907, 7.48151936, 7.05105578],\n    [-6.08311141, 10.1165004, -8.39666201],\n    [-9.09121543, 12.09778042, -3.68576978],\n    [-0.61306988, 8.92723948, 7.90762607],\n    [-2.4380857, 18.25407482, 7.78463196],\n    [5.43508502, 2.54775681, -10.12139265],\n    [11.68723835, 8.6986158, -3.7021906],\n    [9.85092052, 4.11975947, 8.94087536],\n    [5.05006632, 8.46914232, -2.80405123],\n    [6.04706331, 17.03691486, 2.9786743],\n    [0.44733292, 19.90373177, -4.78474409],\n    [2.72799771, 14.76401771, -4.4948139],\n    [1.27950318, 17.33199782, -4.80519287],\n    [11.85132612, 14.78501498, 5.31315893],\n    [9.97911072, 15.31811261, 8.77365454],\n    [4.55891885, 12.06777708, -10.45608692],\n    [10.9408185, 4.79416889, 7.30434388],\n    [5.38308599, 9.6637121, -2.34620525],\n    [8.02985923, 20.72195231, -8.57553242],\n    [7.4150793, 7.39107389, 11.35121],\n    [-3.52196407, 10.45529604, 7.50755464],\n    [-9.44620147, 7.04982356, -1.76273218],\n    [9.38340826, 8.4251251, -7.3244173],\n    [2.55860699, 16.24099748, -10.93004566],\n    [0.2326738, 18.42587479, 14.11877161],\n    [-9.401443, 13.48065195, -2.19391115],\n    [-7.62198324, 3.08572018, 4.56610795],\n    [-9.36039132, 11.82283746, -2.48268759],\n    [6.3609914, 0.20873186, 1.0007836],\n    [8.31257343, 13.25950889, -8.3454169],\n    [5.74292616, 13.38404266, -1.71377887],\n    [1.85450172, 19.1407703, -11.0073566],\n    [3.43440484, 15.38210781, -4.16902115],\n    [4.68187389, 17.80879432, -3.21780624],\n    [4.37413777, 12.38342191, 5.72047255],\n    [8.37687345, 5.71583271, -8.2910022],\n    [-8.92912508, 3.51695361, -4.1905712],\n    [5.45210322, 2.34019079, -2.23820856],\n    [5.14312669, 0.37008062, 12.7561125],\n    [-9.42951441, 20.93101766, -0.04570735],\n    [-5.33917271, 6.75316158, 6.65273389],\n    [5.75121537, 19.01304607, -1.69678142],\n    [10.28897652, 12.06908724, -6.2194425],\n    [-2.05223603, 15.17399051, 7.84585835],\n    [-1.42016368, 2.90483486, -10.77126387],\n    [-6.73512703, 9.31084134, -7.76805775],\n    [12.43958974, 9.51776908, 3.04242137],\n    [11.19924714, 17.51695551, 6.82206591]\n  ])\n```\n\nNow we will call `MDS.fit/1` on our dataset:\n\n```elixir\nkey = Nx.Random.key(42)\nembedding = MDS.fit(swiss_roll, key: key)\n```\n\nExtract only embedded data from struct\n\n```elixir\nembedded_data = embedding.embedding\n```\n\nNow we will plot the resulting embedded swiss roll\n\n```elixir\nembedded_data =\n  embedded_data\n  |> Nx.as_type(:f64)\n  |> DF.new()\n  |> DF.rename([\"x coordinate\", \"y coordinate\"])\n\nTucan.scatter(embedded_data, \"x coordinate\", \"y coordinate\", filled: true)\n|> Tucan.set_size(500, 500)\n|> Tucan.set_title(\"Scatterplot showing embedded Swiss roll data set\", offset: 20)\n```\n\nAs we see, MDS collapsed one dimension and what we see is similar to a cross section of the original dataset.\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"mds.html#swiss-roll"},{"type":"extras","title":"Digits dataset - Multidimensional scaling (MDS)","doc":"In the next section we change dataset to Digits dataset. It consists of almost 1800 8x8 images with digits.\n\n```elixir\ndigits_url =\n  \"https://github.com/scikit-learn/scikit-learn/raw/main/sklearn/datasets/data/digits.csv.gz\"\n\nimage_width = 8\nimage_height = 8\n\ndata =\n  Req.get!(digits_url).body\n\ndf_data = DF.load_csv!(data, delimiter: \",\", header: false)\nlabels = DF.select(df_data, \"column_65\") |> DF.rename([\"labels\"])\ndigits_data = Nx.stack(df_data, axis: 1)\n\n# remove labels from data\ndigits_data = digits_data[[.., 0..(image_width * image_height - 1)]]\nlabels\n```\n\nLet's look how one of the images look like\n\n```elixir\nimage =\n  Nx.reshape(digits_data[[1234, ..]], {image_height, image_width})\n  |> Nx.as_type({:u, 8})\n\nTucan.imshow(image, width: 200, height: 200, color_scheme: :greys, reverse: true)\n```\n\n```elixir\nembedding = MDS.fit(digits_data, key: key)\nembedded_data = embedding.embedding\n```\n\nAs we can observe, digits with the same label are embedded close to each other. The most problematic label is 5 which is similar to 9 and some samples lie close to each other.\n\n```elixir\nembedded_data =\n  embedded_data\n  |> Nx.as_type(:f64)\n  |> DF.new()\n  |> DF.rename([\"x coordinate\", \"y coordinate\"])\n\nembedded_data = DF.concat_columns([embedded_data, labels])\n\nTucan.scatter(embedded_data, \"x coordinate\", \"y coordinate\", filled: true, color_by: \"labels\")\n|> Tucan.set_size(500, 500)\n|> Tucan.Scale.set_color_scheme(:paired)\n|> Tucan.set_title(\"Scatterplot showing embedded Digits data set\", offset: 20)\n```","ref":"mds.html#digits-dataset"},{"type":"extras","title":"Nearest neighbors","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Nearest neighbors\n\n```elixir\nMix.install([\n  {:scholar, \"~> 0.3.0\"},\n  {:explorer, \"~> 0.8.1\"},\n  {:exla, \"~> 0.7.2\"},\n  {:nx, \"~> 0.7.2\"},\n  {:req, \"~> 0.4.14\"},\n  {:kino_vega_lite, \"~> 0.1.11\"},\n  {:kino, \"~> 0.12.3\"},\n  {:kino_explorer, \"~> 0.1.18\"}\n])\n```","ref":"nearest_neighbors.html"},{"type":"extras","title":"Setup - Nearest neighbors","doc":"We will use Explorer in this notebook, so let's define aliases for its main modules:\n\n```elixir\nrequire Explorer.DataFrame, as: DF\nrequire Explorer.Series, as: S\n```\n\nAnd let's configure `EXLA` as our default backend (where our tensors are stored) and compiler (which compiles Scholar code) across the notebook and all branched sections:\n\n```elixir\nNx.global_default_backend(EXLA.Backend)\nNx.Defn.global_default_options(compiler: EXLA)\n```","ref":"nearest_neighbors.html#setup"},{"type":"extras","title":"Definition of nearest neighbors - Nearest neighbors","doc":"We introduced [k-nearest neighbors (KNN) in a previous notebook](k_nearest_neighbors.livemd) alongside some example problems. In that guide, we used the default algorithm for KNN, which uses a brute-force approach. In this document, we will introduce alternative algorithms and their trade-offs, but to do so, we need to define exactly what \"nearest neighbors\" are.\n\nConsider a set of points ${p_1, p_2, \\ldots, p_N}$, where each point $p_i$ is represented as a vector of size $M$, with real-number coordinates: $p_i = (p_{i1}, p_{i2}, \\ldots, p_{iM})$. We say that the dimension of the space is $M$. For convenience, instead of listing the set of points as ${p_1, p_2, \\ldots, p_N}$, we can treat it as a matrix $P$ of size $N \\times M$, where $P_{ij}$ represents the $j$th coordinate of the $i$th point.\n\nA metric between two points $a$ and $b$ is defined as a function `d` that takes two points and returns a non-negative real number, subject to the following conditions for arbitrary points $a$, $b$, and $c$:\n\n* $d(a,b) = 0 \\iff a = b$ (identity)\n* $d(a,b) = d(b, a)$ (symmetry)\n* $d(a,b) \\leq d(a, c) + d(b, c)$ (triangle inequality)\n\nExamples of such metrics include the Euclidean metric and the Manhattan metric.\n\nGiven a set of points $P$, a point $q$, and a metric $d$, the $k$ nearest neighbors from $P$ to point $q$ are defined as a set $S \\subseteq P$, with $\\lvert S \\rvert = k$, where the distance $d$ between the points in $S$ and $q$ is minimized. In simpler terms, $S$ is a set of $k$ points from $P$ with the smallest distance to point $q$.\n\nIt's especially interesting to consider the case when point $q$ is part of the set $P$. In the remainder of the notebook, we will assume that $q$ is indeed in $P$, and if this is not the case, we will note it explicitly.","ref":"nearest_neighbors.html#definition-of-nearest-neighbors"},{"type":"extras","title":"Brute force approach - Nearest neighbors","doc":"Now let's consider how we can compute the $k$-nearest neighbors. The simplest method is to compute the distance using a metric $d$ for each pair of points in the dataset. This can be represented as a matrix $D$ of size $N \\times N$, where $D_{ij}$ indicates the distance between the $i$th and $j$th points. Given the symmetry of metric $d$, the matrix $D$ is also symmetric: $D_{ij} = D_{ji}$. To find the nearest neighbors, you would sort the distances for each point (sorting the rows of matrix $D$), and then return only the first $k$ columns from each row of the sorted matrix.\n\nWhile this method accurately computes the $k$-nearest neighbors, it has a major drawback: time complexity. The time complexity of this approach is $O(MN^2 + N\\log(N)) = O(MN^2)$, where $N$ is the number of points, and $M$ represents the number of dimensions (features) of each point. Even if we set $M = 1$, when $N$ is very large (like the number of songs on Spotify, which is on the order of $10^8$), the number of operations required becomes unfeasible, even if $k$ is relatively small. Given this, for large values of $N$, we need to consider other approaches. However, the brute-force method can still be useful in certain situations. If $N$ is small, the distance matrix $D$ can be computed in a reasonable amount of time. Additionally, if $k \\approx N$, this approach is optimal since you always need to compute $O(Nk)$, which leads to $O(N^2)$ operations. Computing the distance matrix is optimized in many frameworks, giving it an advantage over other methods when $k$ is on the order of magnitude of $N$.\n\nBrute force is the default algorithm used by `Scholar.Neighbors.KNNClassifier` and `Scholar.Neighbors.KNNRegressor`. However, different algorithms can be given via the `:algorithm` option. Let's take a look at one of them.\n\n<!-- livebook:{\"branch_parent_index\":1} -->","ref":"nearest_neighbors.html#brute-force-approach"},{"type":"extras","title":"KDtrees - Nearest neighbors","doc":"For large datasets, we need a much faster approach. It turns out that there is a data structure that allows querying for a single neighbor in $O(\\log(N))$ average time, resulting in a total computation time of $O(kN\\log(N))$, which is efficient for small values of $k$. But what is this data structure that allows querying faster than the brute-force approach? It's called a **KDTree**. KDTree, short for \"K-Dimensional Tree\", is a tree structure that organizes points in a K-dimensional space. More precisely, it's a binary tree where each non-leaf node has an associated splitting hyperplane. All points on the left side of this splitting hyperplane belong to the left subtree, while those on the right side belong to the right subtree. The dimension along which the split occurs is consistent across each level of the tree, and it changes as you move to different levels.\n\nTo better understand this, let's look at two graphical examples of KDTrees.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n| ![KDTree](https://upload.wikimedia.org/wikipedia/commons/b/b6/3dtree.png) |\n| :-----------------------------------------------------------------------: |\n| Figure 1: A 3-dimensional KDtree.                                         |\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nHere, the first split (a red vertical plane) divides the root cell (white) into two subcells. Then, each of these subcells is further split (by the green horizontal planes) into two additional subcells. Finally, the four resulting subcells are further split (by four blue vertical planes), creating a total of eight subcells. Since there are no more sectors to split, the algorithm terminates with these eight subcells.\n\nHere's another, more abstract example.\n\n| ![KDtree](https://upload.wikimedia.org/wikipedia/commons/4/4f/Kd-tree-example-with_records_as_leaf_nodes.png) |\n| :-----------------------------------------------------------------------------------------------------------: |\n| Figure 2: A different example of 3-dimensional KDtree.                                                        |\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nIn this example, the dimensions are abstract values like name, salary, and age. However, because we can define a linear order for each feature (such as a natural order for salary and age, and a lexicographic order for names), we can treat them like any other metric space.\n\nNow, let's move on to the code and implementation of KDTrees in Scholar!\n\n<!-- livebook:{\"branch_parent_index\":1} -->","ref":"nearest_neighbors.html#kdtrees"},{"type":"extras","title":"KDtrees in action: a practical example - Nearest neighbors","doc":"In this example, we'll use data from the [National Health and Nutrition Examination Survey (NHANES) 2013-2014](https://archive.ics.uci.edu/dataset/887/national+health+and+nutrition+health+survey+2013-2014+(nhanes)+age+prediction+subset). The dataset consists of 7 features that are biomedical indicators. For more information, click the link above.\n\nNow, let's read the data into an Explorer dataframe.\n\n```elixir\ndigits_url =\n  \"https://archive.ics.uci.edu/static/public/887/national+health+and+nutrition+health+survey+2013-2014+(nhanes)+age+prediction+subset.zip\"\n\ndata =\n  Req.get!(digits_url).body\n\n{:ok, [{_, data}]} = :zip.extract(data, [:memory])\ndf_data = DF.load_csv!(data)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nIn this analysis, we want to determine if the most similar people, in terms of medical parameters, belong to the same age group. In the dataset, we categorize people into two groups: adults (18-64 years) and seniors (65+ years). Intuitively, we might expect that people with similar medical results are often in the same age group, but we want to quantify the extent to which this is true. To do this, we compute the 5 closest results for each patient and then calculate the average for these two groups.\n\nFirst, we need to separate the medical results from the data that indicates the age group.\n\n```elixir\nx = df_data |> DF.discard([\"age_group\", \"SEQN\", \"RIDAGEYR\"]) |> Nx.stack(axis: 1)\ny = Nx.stack(S.cast(df_data[\"age_group\"], :category), axis: 1)\n\nx = Scholar.Preprocessing.StandardScaler.fit_transform(x)\n\n{x, y}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [-0.7098659661342087, -0.7098659661342087, -0.028236969791631317, 1.4745830429161875, -0.7098659661342087, 2.2836382314533714, -0.4487434040338326],\n     [-0.7098659661342087, -0.7098659661342087, -0.33972321737844713, 1.049829068934166, -0.7098659661342087, 0.8677916515132996, -0.672447163664364],\n     [-0.7300923458476383, -0.7098659661342087, -0.2810667162095013, 1.049829068934166, -0.7098659661342087, 0.6250750949521445, -0.6261287541206102],\n     [-0.7300923458476383, -0.7098659661342087, -0.16577635184295264, 1.35322476463561, -0.7098659661342087, 0.9486971703670181, -0.42366269318918],\n     [-0.7098659661342087, -0.7300923458476383, -0.024191693848945484, 1.3329983849221803, -0.7098659661342087, 0.8880180312267293, -0.5294466590904167],\n     [-0.7300923458476383, -0.7098659661342087, -0.2729761643241294, 1.4745830429161875, -0.7098659661342087, 1.2723192457818915, -0.627342336903416],\n     [-0.7300923458476383, -0.7098659661342087, 0.032442169348657464, 1.150960967501314, -0.7098659661342087, 3.3354099765517105, -0.3233398498105692],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [1],\n     [0],\n     [0],\n     [0],\n     [1],\n     [0],\n     [1],\n     [1],\n     [0],\n     [0],\n     [1],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [1],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [1],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     [0],\n     ...\n   ]\n >}\n```\n\nOnce the data is preprocessed, we can move on to the KDTree model. First, we need to set up the model using `Scholar.Neighbors.KDTree.fit/2`. This method initializes the tree structure for further processing.\n\n```elixir\nnum_neighbors = 6\n\nknn_model = Scholar.Neighbors.KDTree.fit(x, num_neighbors: 6, metric: {:minkowski, 2})\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Scholar.Neighbors.KDTree{\n  levels: 12,\n  indices: #Nx.Tensor \n    [257, 914, 1129, 225, 978, 301, 1131, 688, 481, 1786, 2135, 472, 307, 1606, 1498, 476, 716, 361, 367, 1627, 1615, 1562, 1660, 747, 776, 748, 811, 1690, 1803, 1678, 1687, 244, 925, 491, 1816, 206, 695, 229, 977, 1198, 1865, 1291, 1800, 1549, 2255, 1109, 2032, 597, ...]\n  >,\n  data: #Nx.Tensor \n    [\n      [-0.7098659661342087, -0.7098659661342087, -0.028236969791631317, 1.4745830429161875, -0.7098659661342087, 2.2836382314533714, -0.4487434040338326],\n      [-0.7098659661342087, -0.7098659661342087, -0.33972321737844713, 1.049829068934166, -0.7098659661342087, 0.8677916515132996, -0.672447163664364],\n      [-0.7300923458476383, -0.7098659661342087, -0.2810667162095013, 1.049829068934166, -0.7098659661342087, 0.6250750949521445, -0.6261287541206102],\n      [-0.7300923458476383, -0.7098659661342087, -0.16577635184295264, 1.35322476463561, -0.7098659661342087, 0.9486971703670181, -0.42366269318918],\n      [-0.7098659661342087, -0.7300923458476383, -0.024191693848945484, 1.3329983849221803, -0.7098659661342087, 0.8880180312267293, -0.5294466590904167],\n      [-0.7300923458476383, -0.7098659661342087, -0.2729761643241294, 1.4745830429161875, -0.7098659661342087, 1.2723192457818915, -0.627342336903416],\n      [-0.7300923458476383, -0.7098659661342087, 0.032442169348657464, 1.150960967501314, -0.7098659661342087, ...],\n      ...\n    ]\n  >,\n  num_neighbors: 6,\n  metric: #Function<5.83620768/2 in Scholar.Neighbors.Utils.metric/1>\n}\n```\n\nNow, we need to run a query to calculate the nearest neighbors for a given set of data. To compute the nearest neighbors, we call `Scholar.Neighbors.KDTree.predict`.\n\n```elixir\n{indices, distances} = Scholar.Neighbors.KDTree.predict(knn_model, x)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [0, 1367, 370, 119, 1158, 1423],\n     [1, 432, 1078, 1370, 1252, 358],\n     [2, 1932, 2179, 1516, 1089, 1821],\n     [3, 823, 510, 770, 552, 1066],\n     [4, 1029, 2103, 333, 823, 1063],\n     [5, 99, 1830, 1459, 701, 1767],\n     [6, 1060, 402, 557, 2274, 1733],\n     [7, 1631, 1227, 1493, 2209, 1991],\n     [8, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [\n     [0.0, 0.10564987631681173, 0.13647696316587582, 0.16467068264346493, 0.17265036785358587, 0.17355553880586283],\n     [0.0, 0.05084833013473313, 0.05862161772775837, 0.059033314891497844, 0.061528101864556185, 0.07051004160599812],\n     [0.0, 0.0634183290913442, 0.06834720866562938, 0.06914612071933461, 0.0702263282472859, 0.07657788563767927],\n     [0.0, 0.08520661043231542, 0.08772188110415277, 0.09315472121407559, 0.10080775703280971, 0.10357822211490808],\n     [0.0, 0.08856498626863332, 0.10911471056995574, 0.11729225102681795, 0.12629118906993017, 0.13448641399541536],\n     [0.0, 0.06737206394520054, 0.07237087978074229, 0.09445004662593554, 0.10244796099940952, 0.10666419265073056],\n     [0.0, 0.1772284512996439, 0.18087450661041063, 0.22321979638902212, 0.26995822503187694, 0.3492397094272336],\n     [0.0, 0.10684507307976371, 0.1450485290844591, 0.15468671853000548, 0.1874949397442543, 0.1922686292168594],\n     ...\n   ]\n >}\n```\n\nBefore we compute the statistics from the results, we need to now what are the proportions of each group initially.\n\n```elixir\nseniors_original = Nx.divide(Nx.sum(Nx.equal(y, 1)), Nx.size(y))\nadults_original = Nx.subtract(1, seniors_original)\n{seniors_original, adults_original}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   0.1597892940044403\n >,\n #Nx.Tensor \n   0.8402106761932373\n >}\n```\n\nAs we can see, about 16% of the people are seniors, while 84% are adults. Now, let's calculate the average number of people within the same age group among the nearest neighbors. We should drop the first column since we're not interested in considering the point itself when calculating its neighborhood.\n\n```elixir\nknn = indices[[.., 1..-1//1]]\nage_groups = Nx.take(y, knn)\n\nseniors =\n  Nx.divide(\n    Nx.sum(Nx.multiply(Nx.equal(age_groups, 1), Nx.new_axis(Nx.equal(y, 1), 1))),\n    Nx.sum(Nx.equal(y, 1))\n  )\n  |> Nx.divide(num_neighbors - 1)\n\nadults =\n  Nx.divide(\n    Nx.sum(Nx.multiply(Nx.equal(age_groups, 0), Nx.new_axis(Nx.equal(y, 0), 1))),\n    Nx.sum(Nx.equal(y, 0))\n  )\n  |> Nx.divide(num_neighbors - 1)\n\n{seniors, adults}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   0.28406593203544617\n >,\n #Nx.Tensor \n   0.8579937219619751\n >}\n```\n\nOur intuition was correct: for seniors, 28% of their nearest neighbors in terms of medical results are also seniors, while for adults, nearly 86% of their nearest neighbors are also adults. This indicates that age should always be considered to ensure proper treatment when assessing medical results and drawing conclusions.","ref":"nearest_neighbors.html#kdtrees-in-action-a-practical-example"},{"type":"extras","title":"Approximated nearest neighbors with LargeVis - Nearest neighbors","doc":"The approaches we have seen so far have been exact: they always find the closest `k` nearest neighbors. However, there are situations where exact computations are too expensive and we want to perform approximate computations.\n\nIn the previous example, KDTree was a perfect choice for computing K-Nearest Neighbors. However, when the number of features grows, the efficiency of KDTree decreases significantly, so it's best suited for data with a smaller number of features (<15). For data with many samples but also many features, approximated versions of nearest neighbors are often preferred. They may not give exact solutions, but the deviation is usually minor and accepted as a tradeoff for faster computation.\n\nLargeVis is an approximated nearest neighbors algorithm that uses the concept of a Random Projection Forest. A Random Projection Forest consists of multiple trees. Each tree is built using a divide-and-conquer approach. The entire dataset is used to start the construction, and at each node, the data is projected onto a random hyperplane and split as follows: points with a projection smaller than or equal to the median go into the left subtree, while those with a projection greater than the median go into the right subtree. This process is then recursively applied to the left and right subtrees.\n\nThis structure allows us to quickly create an approximated nearest neighbors graph, which can then be used to find the nearest neighbors efficiently. To better understand this, below is a figure illustrating the process of creating a tree in a random projection forest. We need to note that in Scholar implementation in each step points are splitted in half, but on the figure below this condition is not satisifed. However, the general idea of splitting points is the same.\n\n| ![Random Projection Forest](files/rpKNN.png)                                                                                       |\n| :--------------------------------------------------------------------------------------------------------------------------------: |\n| Figure 3: A process of creating a tree in Random Projection Forest (Source: http://www.math.umassd.edu/~dyan/Slides_rpForests.pdf) |\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nOn the next figure, we have a comparison of how space is partitioned using KDTree and Random Projection Forest. This time splitting is the same as the output of Scholar implementation of Random Projection Forest.\n\n![](files/rbf_and_kdtree.png)\nFigure 4: Left: A spatial partitioning of $\\mathbf{R}^2$ inducedby a KDTree with three levels. The dots are data points; the cross marks a query point $q$ for KNN. Right: Partitioning induced by an Random Projection Tree.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nHere's a brief explanation of LargeVis. First, we compute an approximated nearest neighbors graph using a Random Projection Forest. Then, we refine the graph by examining the neighbors of each point's neighbors for a fixed number of iterations. Now, let's use LargeVis in practice.\n\nIn this example, we will use a dataset on online news popularity. This dataset summarizes a diverse set of features about articles published by Mashable over a two-year period. To read more about this dataset, click [here](https://archive.ics.uci.edu/dataset/332/online+news+popularity). Usually, the dataset is used to predict the number of shares on social networks, but we'll use it differently. KNN and Approximated Nearest Neighbors can be valuable in data mining, particularly in business scenarios. Imagine you're an editor at an online news agency, and you want to create a series of articles that boost your reach. One strategy could be to examine groups of the most popular articles and analyze them to identify specific patterns. We can use Approximated Nearest Neighbors to do this. For each news article, we'll calculate its nearest neighbors and then sort these groups by the total number of shares they receive on social networks.\n\nThe first step is to load the dataset from the Internet archive.\n\n```elixir\ndigits_url =\n  \"https://archive.ics.uci.edu/static/public/332/online+news+popularity.zip\"\n\nraw_data =\n  Req.get!(digits_url).body\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n<<80, 75, 3, 4, 10, 0, 0, 0, 0, 0, 76, 147, 191, 70, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21, 0, 28,\n  0, 79, 110, 108, 105, 110, 101, 78, 101, 119, 115, 80, 111, 112, 117, 108, 97, 114, 105, 116, 121,\n  ...>>\n```\n\nNext, we load the data into a dataframe.\n\n```elixir\n{:ok, [_, {_, data}]} = :zip.extract(raw_data, [:memory])\ndata = String.replace(data, \", \", \",\")\ndf_data = DF.load_csv!(data, delimiter: \",\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```text\n#Explorer.DataFrame \n```\n\nNext, we split the data into two parts: the target variable (shares, or $y$) and the rest of the data ($x$). We separate the shares from $x$ to avoid introducing bias. We also remove URLs because tensors do not support non-numerical data types, and we exclude the timedelta feature, which is an auxiliary feature used in creating the dataset. Finally, we apply standard scaling to $x$.\n\n```elixir\nx = df_data |> DF.discard([\"url\", \"timedelta\", \"shares\"]) |> Nx.stack(axis: 1)\ny = Nx.stack(df_data[[\"shares\"]], axis: 1)\n\nx = Scholar.Preprocessing.StandardScaler.fit_transform(x)\n{x, y}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [-0.16777552811051025, -0.16587029232809064, -0.1678798688132422, -0.16787677252413957, -0.16787847173106735, -0.1678491604112801, -0.16786756848647255, -0.16787677252406877, -0.167885976561665, -0.16784289830350915, -0.16783995637368385, -0.167885976561665, -0.16787677252406877, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.16332077391393496, -0.16332077391393496, -0.16332077391393496, -0.16787677252406877, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.167885976561665, -0.16788137149445206, -0.16788249486817522, -0.16788560835713134, -0.167885596778704, -0.16788560727226615, -0.16788117557784715, -0.16788512461768915, -0.16788555628597568, -0.1678858504789582, -0.16787889653274482, -0.16788385255298896, -0.16788249157833882, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [\n     [593],\n     [711],\n     [1500],\n     [1200],\n     [505],\n     [855],\n     [556],\n     [891],\n     [3600],\n     [710],\n     [2200],\n     [1900],\n     [823],\n     [10000],\n     [761],\n     [1600],\n     [13600],\n     [3100],\n     [5700],\n     [17100],\n     [2800],\n     [598],\n     [445],\n     [1500],\n     [852],\n     [783],\n     [1500],\n     [1800],\n     [462],\n     [425],\n     [3200],\n     [3900],\n     [480],\n     [573],\n     [3600],\n     [343],\n     [7700],\n     [2600],\n     [504],\n     [2100],\n     [1100],\n     [401],\n     [4200],\n     [1200],\n     [1500],\n     [575],\n     [507],\n     [1200],\n     ...\n   ]\n >}\n```\n\nNow, let's move on to computing the approximated nearest neighbors. Notice that this dataset has significantly more features than the previous one (58 compared to 7).\n\n```elixir\nnum_neighbors = 6\n{indices, _dist} = Scholar.Neighbors.LargeVis.fit(x, num_neighbors: num_neighbors)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{#Nx.Tensor \n   [\n     [0, 65, 21, 69, 2, 25],\n     [1, 66, 28, 61, 9, 32],\n     [2, 25, 69, 21, 57, 65],\n     [3, 56, 33, 63, 53, 72],\n     [4, 62, 43, 40, 60, 44],\n     [5, 39, 35, 30, 18, 14],\n     [6, 51, 74, 7, 64, 44],\n     [7, 74, 64, 6, 44, 60],\n     [8, ...],\n     ...\n   ]\n >,\n #Nx.Tensor \n   [\n     [0.0, 0.0018657036251906384, 0.0036749279899371094, 0.005279428978234045, 0.006727966127223153, 0.006837100863281978],\n     [0.0, 6.756593401603524e-5, 1.2678390136218736e-4, 1.6633686250093157e-4, 2.2233078504659554e-4, 3.514499996064942e-4],\n     [0.0, 5.173768744997381e-4, 0.001809295719831838, 0.003793044425862508, 0.006135524442190683, 0.006203436890693119],\n     [0.0, 5.252092785567814e-4, 6.090372028619397e-4, 6.319163895009465e-4, 7.10570116086926e-4, 8.120833196596075e-4],\n     [0.0, 5.067837852972306e-5, 9.174324571151536e-5, 3.706506657349229e-4, 4.427973160256605e-4, 4.6200572899281997e-4],\n     [0.0, 0.03510952981931922, 0.03666978946624632, 0.038386790160971074, 0.05585061734567715, 0.05899400609953473],\n     [0.0, 1.4980733767922953e-4, 2.6223120133330976e-4, 2.703271990114324e-4, 4.1551509313155296e-4, 5.712522600732687e-4],\n     [0.0, 2.0471372823266868e-5, 1.485758818753876e-4, 2.703271990114324e-4, 3.053468283568113e-4, 3.2301368793209305e-4],\n     ...\n   ]\n >}\n```\n\nFinally, we compute the nearest neighbors and sort them in descending order based on the sum of shares.\n\n```elixir\nshares_of_neighbors = Nx.take(y, indices) |> Nx.squeeze(axes: [-1]) |> Nx.sum(axes: [-1])\nord = Nx.argsort(shares_of_neighbors, direction: :desc)\nNx.take(indices, ord)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Nx.Tensor \n  [\n    [8055, 5407, 9365, 6100, 8717, 8217],\n    [8065, 8717, 9365, 5407, 6100, 8055],\n    [5407, 9365, 8055, 8717, 8065, 6100],\n    [8717, 8065, 9365, 6100, 5407, 8055],\n    [6100, 8717, 9365, 5407, 8065, 8055],\n    [9365, 8065, 5407, 8717, 6100, 8055],\n    [16467, 14218, 11843, 26259, 16268, 16334],\n    [16334, 16268, 16467, 15205, 26259, 14218],\n    [16268, 16334, ...],\n    ...\n  ]\n>\n```\n\nNotice that the first six groups are quite similar. The condition that a neighbor of your neighbor is also your neighbor is often satisfied, which is why LargeVis uses this fact to refine the results of Random Projection Forests.","ref":"nearest_neighbors.html#approximated-nearest-neighbors-with-largevis"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,48]}}